/*
 * Frequency-response matrix object
 *
 * Copyright (c) Solexa 2005, 2006
 * Author: Klaus Maisinger
 * This source file is covered by the "Solexa Public Source License"
 * agreement and bound by the terms therein.
 */

/* classes to describe a fully functional nucleotide (FullyFunctional)
 * with a dye attached, and the spectral response matrix (ResponseMatrix)
 * of the instrument (intensity in a colour as a function of incorporated base) 
 */

#if defined(__BORLANDC__)
#include <stdlib.h>
#else
#include <cstdlib>
#endif
#include <cassert>
#include "Matrix.h"
#include "DNACodes.h"
#include "LinAlg.h" 

#define MIN(x,y) ((x) < (y) ? (x) : (y))

//---------------------------------------------------------------------------
/*
 * Simple container to hold the information to describe a fully functional
 * (i.e. a nucleotide with a dye attached, plus the filter used to observe
 * the dye).
 *
 * TODO: Not sure yet what the right format is to describe filter and
 * dye.
 * Note: The extra information is not actually used.
 */ 

FullyFunctional::FullyFunctional(unsigned char abase, int adye, int afilter)
{
    base = abase;
    if (adye >= 0)
        dye = adye;
    else
        dye = get_base_index(base);
    if (afilter >= 0)
        filter = afilter;
    else
        filter = dye;
}

FullyFunctional::FullyFunctional(const FullyFunctional &temp)
{
    base = temp.base;
    dye = temp.dye;
    filter = temp.filter;
}

FullyFunctional::~FullyFunctional()
{}

//---------------------------------------------------------------------------
/*
 * Object that holds the spectral response matrix of the instrument.
 * The matrix describes the intensity response in a frequency band
 * (determined by the dye's spectrum and the filter) for a given nucleotide.
 * The matrix can be read from a file or set directly. It can be applied
 * to obtain the intensity in all channel for a given combination of
 * different bases, or its inverse can be applied to recover the base
 * proportions from the observed intensities.
 * A change in power level in one channel corresponds to a rescaling
 * of the corresponding matrix row (the matrix columns corrspond to the 
 * nucleotides and the matrix rows to the frequency channels).
 *
 * TODO:
 * - The dimension of the data may be less than the 
 *   matrix dimension if only a subset of the nucleotides are incorporated.
 *   The matrix ought to be permuted automatically in this case 
 *   (implement this in get_matrix()). Of course, this is of limited value
 *   for real sequencing runs.
 * - Of course, the number of observed channels may be different from
 *   the number of nucleotides incorporated. This cannot be dealt with
 *   by matrix inversion (unless the number of channels is greater than
 *   the number of nucleotides, in which case some channels can simply
 *   be ignored). However, a more sophisticated treatment than matrix
 *   inversion might be required anyway
 *   in the presence of high noise levels.
 */
ResponseMatrix::ResponseMatrix(int n)
{
    int i;
    ndim = n;
    matrix = new float [ ndim*ndim ];
    channels = new FullyFunctional * [ndim];
    for (i = 0; i < ndim; i++)
        channels[i] = NULL;
    power = NULL;
    exposure_times = new float[ndim];
    for (i = 0; i < ndim; i++)
        exposure_times[i] = 0.;
}

void ResponseMatrix::set_matrix(float *mat, int n,
                                FullyFunctional **ffns, float *exposures)
{
    reset_matrix();
    if (n > ndim)
        return;
    int i, j;
    for (i = 0; i < n; i++) {
        for (j = 0; j < n; j++)
            matrix[i*ndim+j] = mat[i*n+j];
        channels[i] = new FullyFunctional(*(ffns[i]));
    }
    if (exposures) {
        for (i = 0; i < n; i++)
            exposure_times[i] = exposures[i];
    }
}

void ResponseMatrix::set_matrix(float *mat, int n,
                                unsigned char *bases, float *exposures)
{
    if (n == 0)
        n = ndim;
    if (n > ndim)
        return;
    FullyFunctional **ch;
    ch = new FullyFunctional * [ n ];
    int i;
    if (bases == NULL)
        for (i = 0; i < n; i++)
            ch[i] = new FullyFunctional(BaseUpperCase[i%4]);
    else
        for (i = 0; i < n; i++)
            ch[i] = new FullyFunctional(bases[i]);
    set_matrix(mat, n, ch, exposures);

    for (i = 0; i < n; i++)
        delete ch[i];
    delete [] ch;
}

/*
 * Update a permutation lookup table <index> so that it holds a mapping
 * from the order of bases given by the <n>-dimenensional array
 * <bases> to the corresponding matrix columns. (This is to work around
 * the fact that the matrix columns may be given in a different order from
 * the actual base order used in the experiment. In other words, the order
 * in which nucleotides are scanned may change in the future.)
 */
void ResponseMatrix::find_reordering(unsigned char *bases, int n, int *index)
{
    assert(n <= ndim);
    int i, j;
    for (i = 0; i < ndim; i++)
        index[i] = i;
    if ((bases == NULL) || (n <= 0))
        return;
    for (i = 0; i < n; i++)
        for (j = 0; j < ndim; j++)
            if (channels[j] && (channels[j]->get_base() == bases[i])) {
                index[i] = j;
                break;
            }
}

/*
 * Return a pointer to a copy the <n>-dimensional subsets of rows 
 * and columns of the matrix. The columns and rows are reordered to 
 * fit the base-order given by the <n>-dimensional array <bases>.
 */
float *ResponseMatrix::get_matrix(int n, unsigned char *bases,
                                  float *exposures)
{
    float *result = NULL;
    if ((n <= 0) || (n > ndim))
        n = ndim;

    int i, j;
    int *index = new int [ndim];
    find_reordering(bases, n, index);
    result = new float [ n*n];
    for (i = 0; i < n; i++)
        for (j = 0; j < n; j++)
            result[i*n+j] = matrix[index[i]*ndim+index[j]];

    // apply power correction
    if (power != NULL)
        for (j = 0; j < n; j++)
            for (i = 0; i < n; i++)
                result[i*n+j] *= power[i];
            
    // adjust exposure times
    if (exposures != NULL) {
        int flag = 1;
        for (i = 0; i < ndim; i++) {
            if ((exposures[i] <= 0.) || (exposure_times[i] <= 0.))
                flag = 0;         
        }
        if (flag) {
            for (i = 0; i < ndim; i++)
                for (j = 0; j < ndim; j++)
                    result[i*ndim+j] *= exposures[i] / exposure_times[index[i]];
        }
    }

    delete [] index;
    return result;
}

/*
 * return the inverse matrix of get_matrix().
 */
float *ResponseMatrix::get_inverse_matrix(int n, unsigned char *bases,
                                          float *exposures)
{
    assert((n > 0) && (n <= ndim));
    float *mat = get_matrix(n, bases, exposures);
    float *inverse = new float [n*n];

    matinvert(mat, inverse, n);
  
    delete [] mat;
    return inverse;
}

/* 
 * Load the matrix from a text file.
 * The file has the following format:
 * - Lines starting with '#' contain comments.
 * - Lines starting with "<" describe a fully functional nucleotide
 * - All other lines contain the rows of the matrix
 */
int ResponseMatrix::load(const char filename[])
{
    char buffer[256];
    FILE *fp = fopen(filename, "r");
    int dim = 0;
    int row = 0;
    if (fp) {
        reset_matrix();
        while (fgets(buffer, sizeof(buffer), fp)) {
            // ignore comments
            if (buffer[0] == '#')
                continue;
            // read column descriptors
            if ((buffer[0] == '>') && (dim < ndim)) {
                unsigned char base;
                sscanf(buffer, "> %c %f", &base, &exposure_times[dim]);
                channels[dim] = new FullyFunctional(base);
                dim ++;
            }
            else {                    // read matrix rows
                if (row < ndim) {
                    int col = 0;
                    char *begin, *end;
                    end = buffer;
                    do {
                        begin = end;
                        matrix[row*ndim+col] = strtod(begin, &end);
                        col++;
                    }
                    while ((col < ndim) && (end != NULL) && (end != begin));
                    row++;
                }
            }
        }
        fclose(fp);
    }
    return MIN(dim, row);
}

/*
 * set a number of scale factors to rescale the powers in different 
 * frequency channels. This can be used to correct for any fluctuations
 * in relative dye brightnesses.
 */
void ResponseMatrix::set_power_correction(float *p, int n)
{
    assert((n <= ndim) && (n >= 0));
    if (power)
        delete [] power;
    power = NULL;
    if (n > 0) {
        int i;
        power = new float [n];
        for (i = 0; i < n; i++)
            power[i] = p[i];
    }
}

void ResponseMatrix::apply(float *input, float *output, int ndata,
                           int n, unsigned char *bases)
{
    if (n == 0)
        n = ndim;
    float *mat = get_matrix(n, bases);
    // multiply input by matrix to obtain output
    int i, j, k;
    for (k = 0; k < ndata; k++) {
        for (i = 0; i < n; i++) {
            output[k*n+i] = 0.;
            for (j = 0; j < n; j++)
                output[k*n+i] += mat[i*n+j] * input[k*n+j];
        }
    }
    delete [] mat;
}

void ResponseMatrix::apply_inverse(float *input, float *output, int ndata,
                                   int n, unsigned char *bases)
{
    if (n == 0)
        n = ndim;
    float *inverse = get_inverse_matrix(n, bases);

    // multiply inverse by input to obtain output
    int i, j, k;
    for (k = 0; k < ndata; k++)
        for (i = 0; i < n; i++) {
            output[k*n+i] = 0.;
            for (j = 0; j < n; j++)
                output[k*n+i] += inverse[i*n+j] * input[k*n+j];
        }
    delete [] inverse;
}

/*
 * set matrix element to zero and clear fully functionals.
 */
void ResponseMatrix::reset_matrix()
{
    int i;
    for (i = 0; i < ndim*ndim; i++)
        matrix[i] = 0.;
    for (i = 0; i < ndim; i++)
        matrix[i] = 1.;
    for (i = 0; i < ndim; i++)
        if (channels[i] != NULL)
            delete channels[i];
}

ResponseMatrix::~ResponseMatrix()
{
    reset_matrix();
    delete [] matrix;
    delete [] channels;
    delete [] exposure_times;
    if (power != NULL)
        delete [] power;
}
