/*
 * Main program for per-cycle intensity correlation calculation.
 * 
 * Copyright (c) Solexa 2005, 2006
 * Author: Klaus Maisinger
 * This source file is covered by the "Solexa Public Source License"
 * agreement and bound by the terms therein.
 */

/*
 * Estimate the relative correlation between signals from clusters in
 * different sequencing cycles
 */
 
#include <vector>
#include <cassert>
#include "DNACodes.h"
#include "DNAProcess.h"
#include "IntensityData.h"
#include "Matrix.h"
#include "Maths.h"
#include "Stats.h"

const int MAX_CLUSTERS = 50000;
const int MIN_CLUSTERS =   500;
const int MAX_CYCLES = 20;
const int PURE_BASES = 12;
const float CHASTITY = 0.6;

/*
 * Attempt to estimate ratio of signal from current cycle to some
 * other cycle, for each base separately.
 */
float *estimate_phasing(std::vector<float *>cycledata, int nclusters,
                        int *sequences, int step, int firstcycle,
                        int lastcycle)
{
    int   ndim = lastcycle-firstcycle;
    float *phasing = new float [ 4 * ndim];
    int   cycle;
    float follow[4], nofollow[4];
    int   count[4], nocount[4];

    for (cycle = firstcycle; cycle < lastcycle; cycle++) {
        int i, j;
        for (j = 0; j < 4; j++) {
            follow[j] = nofollow[j] = 0.f;
            count[j] = nocount[j] = 0;
        }
        for (i = 0; i < nclusters; i++) {
            /* skip clusters with zero intensity */
            //assert(cycle*nclusters+i < nclusters*lastcycle);
            //assert((cycle+step)*nclusters+i < nclusters*lastcycle);
            if ((sequences[cycle*nclusters+i] == -1) ||
                (sequences[(cycle+step)*nclusters+i] == -1)) {
                continue;
            }
            assert(cycle < lastcycle);
            for (j = 0; j < 4; j++) {
                /* skip clusters with the current base-call */
                if (sequences[cycle*nclusters+i] == j)
                    continue;
                float value = cycledata[cycle][i*4+j];
                if (sequences[(cycle+step)*nclusters+i] == j) {
                    // previous base hast the same base-call
                    follow[j] += value;
                    count[j] += 1;
                }
                else {
                    // previous base has different base-call
                    nofollow[j] += value;
                    nocount[j] += 1;
                }
            }
        }
        for (j = 0; j < 4; j++) {
            if (count[j] > 0)
                follow[j] /= count[j];
            if (nocount[j] > 0)
                nofollow[j] /= nocount[j];
            /* 
             * subtract general signal background (previous base different)
             * from signal left over from previous (next) cycle
             */
            phasing[j*ndim+cycle-firstcycle] = follow[j] - nofollow[j];
        }
    }
    return phasing;
}

/*
 * apply purity filtering to the data set
 */
void filter_reads(std::vector<float *>cycledata, int *nclusters, int ncycles)
{
    int i, inew, k, j;
    int npurebases = ncycles;
    if (npurebases > PURE_BASES)
        npurebases = PURE_BASES;
    inew = 0;
    std::vector<int> lookup(*nclusters);
    for (i = 0; i < *nclusters; i++) {
        lookup[i] = -1;
        float purity = 1.;
        for (k = 0; k < npurebases; k++) {
            float *cyclesignal = &(cycledata[k][i*4]);
            float max1 = 1.;
            float max2 = 1.;
            int   imax1 = -1;
            int   imax2 = -1;
            // find brightest base
            for (j = 0; j < 4; j++) 
                if (cyclesignal[j] > max1) {
                    max1 = cyclesignal[j]; imax1 = j;
                }  
            // find second brightest base
            for (j = 0; j < 4; j++) 
                if ((j != imax1) && (cyclesignal[j] > max2)) {
                    max2 = cyclesignal[j]; imax2 = j;
                }
            if ((imax1 == -1) || (imax2 == -1)) {
                purity = 0.; break;
            }
            purity = max1/(max1+max2);
            if (purity < CHASTITY)
                break;
            
        }
        if (purity < CHASTITY)
            continue;
        lookup[inew] = i;
        inew++;
    }
    *nclusters = inew;
    for (k = 0; k < ncycles; k++)
        for (i = 0; i < *nclusters; i++)
            memcpy(&(cycledata[k][i*4]), 
                   &(cycledata[k][lookup[i]*4]), 4*sizeof(float));
        
}

/* 
 * Perform some base-calling based on the intensity values.
 * The bases are encoded as number from 0-3 for bases A, C, G, T
 * respectively (-1 for no base-call).
 * This could be replaced by sequences read from a sequence file.
 */ 
void call_bases(std::vector<float *>cycledata, int nclusters,
                int *bases)
{
    int i, j, k;
    int ncycles = (int)cycledata.size();

    for (k = 0; k < ncycles; k++) {
        for (i = 0; i < nclusters; i++) {
            float max = 0.f;
            int idx = -1;
            for (j = 0; j < 4; j++)
                if (cycledata[k][i*4+j] > max) {
                    max = cycledata[k][i*4+j];
                    idx = j;
                }
            bases[k*nclusters+i] = idx;
        }
    }    
}

/*
 * format phasing and prephasing estimates for output
 */
void print_phasing(float *phasing, float *prephasing, int ncycles,
                   int phasing_firstcycle, int phasing_lastcycle,
                   int prephasing_firstcycle, int prephasing_lastcycle)
{
    int i, j;
    float values1[4], values2[4];
    char line[1024];
    for (i = 0; i < ncycles; i++) {
        for (j = 0; j < 4; j++) {
            values1[j] = 0.;
            values2[j] = 0.;
        }
        if ((i >= phasing_firstcycle) && (i < phasing_lastcycle))
            for (j = 0; j < 4; j++) {
                values1[j] = phasing[j*(phasing_lastcycle-phasing_firstcycle)
                                     +(i-phasing_firstcycle)];
            }
        if ((i >= prephasing_firstcycle) && (i < prephasing_lastcycle))
            for (j = 0; j < 4; j++) {
                values2[j] = prephasing[j*(prephasing_lastcycle-prephasing_firstcycle)
                                     +(i-prephasing_firstcycle)];
            }
        snprintf(line, sizeof(line),
                 "%i\t%6.4f %6.4f\t%6.4f %6.4f %6.4f %6.4f\t%6.4f %6.4f %6.4f %6.4f\n",
                 i + 1, average(values1, 4), average(values2, 4),
                 values1[0], values1[1], values1[2], values1[3],
                 values2[0], values2[1], values2[2], values2[3]);
        printf(line);
    } 
}
 
int main(int argc, char *argv[]) 
{
    int  i, j, k;
    char intensity_filename[FILENAME_MAX] = "";
    char matrix_filename[FILENAME_MAX] = "matrix.txt";
    int  matrixflag = 0;
    int  intensityflag = 0;
    int  filterflag = 1;
    float *matrix = NULL;

    for (i = 1; i < argc; i++) {
        if (argv[i][0] == '-') {
            if (argv[i][1] == '-') {
                switch (argv[i][2]) {
                case 'm': sscanf(argv[i], "--matrix=%s", &matrix_filename);
                    matrixflag = 1;
                    break;
                }
            }
        }
        else {
            snprintf(intensity_filename, FILENAME_MAX, "%s", argv[i]);
            intensityflag = 1;
        }
    }
    
    if (!intensityflag) {
        fprintf(stderr, 
                "Usage: PhasingEstimate [--matrix=<filename>] <intensityfile>\n");
        exit(1);
    }
    
    /* read signal from file */
    float *signal = new float [4*MAX_CYCLES*MAX_CLUSTERS];
    int ncol = 4 * MAX_CYCLES;
    int nclusters = MAX_CLUSTERS;
    if (read_signal_array(intensity_filename, signal, &nclusters, &ncol, 4))
        fprintf(stderr, "Error reading signal file %s\n", intensity_filename);
    int ncycles = ncol/4;
    if (nclusters < MIN_CLUSTERS) {
        fprintf(stderr, 
                "Warning: Not enough clusters in %s to estimate phasing.\n",
                intensity_filename);
        delete [] signal;
        return 0;
    }
    
    /* try to read frequency cross-talk matrix */
    if (matrixflag) {
        ResponseMatrix *fmatrix = new ResponseMatrix(4);
        int num = fmatrix->load(matrix_filename);
        if (num) {
            matrix = fmatrix->get_inverse_matrix(4, (unsigned char*)BaseUpperCase);
        }
        else {
            fprintf(stderr, "Could not find matrix file %s", matrix_filename);
        }
        delete fmatrix;
    }
    
    // convert signal from input array and apply matrix correction
    std::vector<float *>cycledata;
    for (k = 0; k < ncycles; k++) {
        float *array = new float [ nclusters * 4];
        for (i = 0; i < nclusters; i++)
            for (j = 0; j < 4; j++)
                array[i*4+j] = signal[i*4*MAX_CYCLES+k*4+j];
        if (matrix) { // apply matrix correction
            float *array2 = new float [nclusters*4];
            rescale_intensities(array, array2, nclusters, 4, matrix);
            cycledata.push_back(array2);
            delete [] array;
        }
        else
            cycledata.push_back(array);
    }    
    delete [] signal;
    
    /* apply some purity filtering to discard mixed reads */
    if (filterflag) {
        filter_reads(cycledata, &nclusters, ncycles);
        fprintf(stderr, "PF %i\n", nclusters);
        if (nclusters < MIN_CLUSTERS) {
            fprintf(stderr, 
                    "Warning: Not enough filtered clusters in %s to estimate phasing.\n",
                    intensity_filename);
            return 0;
        }
    }

    /* normalise cluster intensities such that the brightest base 
     * has a value of 1
     */    
    for (k = 0;  k < (int)cycledata.size(); k++) {
        for (i = 0; i < nclusters; i++) {
            float norm = 1.;
            for (j = 0; j < 4; j++)
                if (cycledata[k][i*4+j] > norm)
                    norm = cycledata[k][i*4+j];
             for (j = 0; j < 4; j++)
                cycledata[k][i*4+j] /= norm;
        }
    }
    
    /* perform some primitive base-calling */
    int *sequences = new int [nclusters*ncycles];
    call_bases(cycledata, nclusters, sequences);
    
    /* estimate phasing per cycle */
    int phasing_step = -1;
    int phasing_lastcycle = MIN(ncycles, ncycles-phasing_step);
    int phasing_firstcycle = MAX(0, -phasing_step);
    float *phasing = estimate_phasing(cycledata, nclusters,
                                      sequences, phasing_step,
                                      phasing_firstcycle, phasing_lastcycle);
    /* estimate prephasing per cycle */
    int prephasing_step = 1;
    int prephasing_lastcycle = MIN(ncycles, ncycles-prephasing_step);
    int prephasing_firstcycle = MAX(0, -prephasing_step);
    //print prephasing_firstcycle, prephasing_lastcycle
    float *prephasing = estimate_phasing(cycledata, nclusters,
                                         sequences, prephasing_step,
                                         prephasing_firstcycle, 
                                         prephasing_lastcycle);
    
    print_phasing(phasing, prephasing, ncycles,
                  phasing_firstcycle, phasing_lastcycle,
                  prephasing_firstcycle, prephasing_lastcycle);
    delete [] phasing; delete [] prephasing;
    delete [] sequences;
    for (k = 0;  k < (int)cycledata.size(); k++)
        delete [] cycledata[k];
    if (matrix)
        delete [] matrix;
    return 0;
}
