/*
 * Copyright (c) Solexa 2005-2007
 * Authors: Klaus Maisinger, Lorenzo Milazzo
 *
 * This source file is covered by the "Solexa Public Source License"
 * agreement and bound by the terms therein.
  */

//
// CVS - $Id: run_base_caller.cpp,v 1.13 2007/04/11 16:20:53 km Exp $
// Main program for the base-caller ("Bustard").


#include <string>
#include <vector>
#include <cassert>
#include <cstdio>

#include "BaseCallConfig.h"
#include "BaseCallParamsXMLReader.h"
#include "PhasingModel.h"
#include "ProcessBaseCall.h"

#include "DNACodes.h"
#include "ImageXMLReader.h"
#include "IntensityData.h"
#include "XMLSAXParser.h"

#include "Logger.h"
#include "Matrix.h"
#include "Tiling.h"
#include "DNAProcess.h"
#include "Tools.h"

/*
 * parse the QC XML file for information on bases, exposure times and
 * deblocked cycles.
 */
void read_xml_info(const char filename[],
                   unsigned char bases[], float *exposure_times, 
                   int ncol, int *is_deblocked)
{
    int i;
    assert(ncol <= N_COLOURS);
    for (i = 0; i < ncol; i++) {
        exposure_times[i] = 0.;
        bases[i] = BaseUpperCase[i];
        *is_deblocked = 0;
    }
        
    FILE *fh = fopen(filename, "r");
    if (!fh)
        return;

    ImageXMLReader *handler = new ImageXMLReader();
    XMLSAXParser *sax_parser = new XMLSAXParser();
    sax_parser->setContentHandler(handler);
    sax_parser->parse(fh);
    fclose(fh);
    
    int ncol2 = handler->getCount();
    if (ncol2 != N_COLOURS)
        fprintf(stderr, "Error: XML file %s has only %i images.\n",
                filename, ncol2);
    assert((ncol2 == N_COLOURS) || (ncol2 == 0));
    for (i = 0; i < ncol2; i++) {
        bases[i] = handler->getBases()[i];
        exposure_times[i] = handler->getExposureTimes()[i];
        *is_deblocked = handler->getDeblocked()[0];
    }
    delete sax_parser;
    delete handler;
}
/*
 * parse the parameter XML file for options
 */
void read_xml_parameters(const char filename[],
                         float *phasing, float *prephasing)
{
    FILE *fh = fopen(filename, "r");
    if (!fh) {
        fprintf(stderr, "Error: Failed to open paramter file %s\n", filename);
        return;
    }

    BaseCallParamsXMLReader *handler = new BaseCallParamsXMLReader();
    XMLSAXParser *sax_parser = new XMLSAXParser();
    sax_parser->setContentHandler(handler);
    sax_parser->parse(fh);
    fclose(fh);
    *phasing    = handler->getPhasing();
    *prephasing = handler->getPrephasing();
    delete sax_parser;
    delete handler;
}

ResponseMatrix *get_response_matrix(std::vector<ResponseMatrix * > response_matrices,
                                    std::vector<int> start_cycles, 
                                    int current_cycle)
{
    int i;
    int min = current_cycle;
    ResponseMatrix *response_matrix = NULL;
    for (i = 0; i < int(start_cycles.size()); i++) {
        if ((current_cycle >= start_cycles[i]) && 
            (response_matrices[i] != NULL) &&
            (start_cycles[i] <= min)) {
            min = start_cycles[i];
            response_matrix = response_matrices[i];
            
        }
    }

    return response_matrix;
}

/*
 * set up the frequency response matrix
 */
std::vector <float * > get_matrices(std::vector<ResponseMatrix * > response_matrices,
                                    std::vector<int> start_cycles,
                                    int ncol, int ncycles,
                                    unsigned char *bases, 
                                    std::vector<float * > exposure,
                                    Logger *logger)
{
    std::vector <float * > matrices;
    int l;
    for (l = 0; l < ncycles; l++) {
        float *matrix = NULL;
        ResponseMatrix *response_matrix = get_response_matrix(response_matrices, 
                                                              start_cycles, l);
        if (response_matrix) {
            response_matrix->set_power_correction();
            {
                char message[250];
                int pos = 0;
                matrix = response_matrix->get_matrix(ncol, bases, exposure[l]);
                pos = sprintf(message, "Matrix: ");
                snprint_matrix(message+pos, 250-pos, matrix, ncol, ncol);
                logger->info(message);
                delete [] matrix;
            }
            matrix = response_matrix->get_inverse_matrix(ncol, bases, exposure[l]);
        }
        else 
            matrix = identity_matrix(ncol);
        matrices.push_back(matrix);
    }
    return matrices;
}                                      

void call(ProcessBaseCall *base_caller, Tiling *intensity_name,
          Tiling *sequence_name,
          int ncycles, int lane, int tile, st_fposition *positions,
          unsigned char *bases, std::vector<float * > exposure,
          int *is_deblocked, 
          std::vector<ResponseMatrix * > response_matrices,
          std::vector<int> start_cycles,
          Logger *logger)
{
    int i;
    char msg[256];
    int nobj, ncol, error;
    snprintf(msg, sizeof(msg), "Reading intensity file %s.",
             intensity_name->getIndexFilename(INTENSITY_FILE, tile).c_str());
    logger->info(msg);
    std::vector<float * > intensities 
        = read_signal_array(intensity_name->getIndexFilename(INTENSITY_FILE,
                                                             tile).c_str(),
                            &nobj, &ncol, &error);
    assert((int(intensities.size()) >= ncycles) ||
	   (int(intensities.size()) == 0));
    snprintf(msg, sizeof(msg), "Reading noise     file %s.",
             intensity_name->getIndexFilename(NOISE_FILE, tile).c_str());
    logger->info(msg);
    std::vector<float * > noise 
        = read_signal_array(intensity_name->getIndexFilename(NOISE_FILE,
                                                             tile).c_str(),
                            &nobj, &ncol, &error);
    assert((int(noise.size()) >= ncycles) ||
	   (int(noise.size()) == 0));
    assert(int(noise.size()) == int(intensities.size()));
    if (ncol == 0)
      ncol = N_COLOURS;
    if (intensities.size() == 0)
        for (i = 0; i < ncycles; i++) {
	    intensities.push_back(0);
	    noise.push_back(0);
        }
        
    for (i = int(intensities.size()) - 1; i >= ncycles; i--) {
        delete [] intensities[i];
        delete [] noise[i];
        intensities.pop_back();
        noise.pop_back();
    }

    std::vector<float * > matrices = get_matrices(response_matrices, start_cycles, 
                                                  ncol, ncycles,
                                                  bases, exposure, logger);

    std::vector<float * > probabilities(ncycles);
    std::vector<float * > matrix_signal(ncycles);
    std::vector<float * > signal(ncycles);
    logger->info("Starting to base-call...");
    unsigned char *sequence = new unsigned char [nobj * ncycles];
    int cycle;
    for (cycle = 0; cycle < ncycles; cycle++) {
        probabilities[cycle] = new float [nobj*ncol];
        matrix_signal[cycle] = new float [nobj*ncol];
        signal[cycle]        = new float [nobj*ncol];
        
        sprintf(msg, "Calling cycle %i...", cycle+1);
        logger->info(msg);
        base_caller->call(intensities, noise, matrices, nobj, ncol, cycle, 
                          bases, is_deblocked, 
                          sequence, probabilities[cycle], 
                          signal[cycle], matrix_signal[cycle]);
    }


    /* save corrected intensities and probabilities to file */
    error = 0;
    if (!error) {
        snprintf(msg, sizeof(msg), "Writing %s.",
                 sequence_name->getIndexFilename(SIGNAL2_FILE, tile).c_str());
        logger->info(msg);
        error = write_signal_array(sequence_name->getIndexFilename(SIGNAL2_FILE,
                                                                   tile).c_str(),
                                   signal, nobj, ncol,
                                   lane, sequence_name->getTile(tile), positions);
    }                                   
    if (!error) {
        snprintf(msg, sizeof(msg), "Writing %s.",
                 sequence_name->getIndexFilename(PROBABILITY_FILE, tile).c_str());
        logger->info(msg);
        error = write_quality_file(sequence_name->getIndexFilename(PROBABILITY_FILE,
                                                           tile).c_str(),
                                   probabilities, nobj, ncol);
    }
    if (!error) {
        snprintf(msg, sizeof(msg), "Writing %s.",
                 sequence_name->getIndexFilename(SEQUENCE_FILE, tile).c_str());
        logger->info(msg);
        error = write_sequence_file(sequence_name->getIndexFilename(SEQUENCE_FILE, tile).c_str(),
                                    sequence, nobj, ncycles,
                                    lane, sequence_name->getTile(tile), positions);
    }
    if (error) {
        fprintf(stderr, "Error: Cannot recover from IO error! Exiting...\n");
        perror("Error: ");
        exit(1);
    }

    delete [] sequence;
    int l;
    for (l = 0; l < ncycles; l++) {
        delete [] intensities[l];
        delete [] noise[l];
        delete [] matrices[l];
        delete [] matrix_signal[l];
        delete [] signal[l];
        delete [] probabilities[l];
    }
}

/*
 * set up and run the base-caller
 */
void run_base_caller(const char path_image_analysis[],
                     const char path_basecaller[], 
                     const char path_intensities[], 
                     const char sample[],
                     const std::vector<std::string> matrices, 
                     std::vector<int> starts_matrices,
                     int ncycles, int ncolours, int ntiles,
                     int channel, int tile, std::string logfile,
                     float phasing_rate, float prephasing_rate, 
                     std::vector<int> starts_phasing)
{
    assert(ncycles > 0);
    assert(ncolours > 0);
    assert(ncolours <= N_COLOURS);

    /* set up logger */
    Logger *logger = new Logger(logfile.c_str());
    logger->add_handler(new LogHandler());
    char msg[256];
    sprintf(msg, "%s%s", software, software_version);
    logger->set_source(msg); 
    sprintf(msg, "> `Bustard` - base-caller");
    logger->info(msg);
    sprintf(msg, "Base-caller path: %s", path_basecaller);
    logger->info(msg);
    sprintf(msg, "Log file: %s", logfile.c_str());
    logger->info(msg);
    
    
    // load frequency response matrix
    std::vector<ResponseMatrix*> response_matrices;
    response_matrices.reserve(matrices.size());

    int i;
    for (i = 0; i < (int)matrices.size(); i++) {
        response_matrices.push_back(new ResponseMatrix(N_COLOURS));
        int num = response_matrices.at(i)->load(matrices.at(i).c_str());
        if (num) {
            sprintf(msg, "Successfully loaded %ix%i matrix from %s", 
                    num, num, matrices.at(i).c_str());
        }
        else {
            sprintf(msg, "Could not find matrix file %s", matrices.at(i).c_str());
            delete response_matrices.at(i);
            response_matrices.at(i) = NULL;
        }
    }
    logger->info(msg);

    sprintf(msg, "Using phasing=%6.4f, prephasing=%6.4f", 
            phasing_rate, prephasing_rate);
    logger->info(msg);

    // initialise objects for creation of filenames
    std::string path = std::string(path_image_analysis) + std::string(sample);
    Tiling *sequencer_name = new Tiling(path.c_str());
    sequencer_name->setChannel(channel, tile, ntiles);

    // initialise objects for creation of filenames
    path = std::string(path_basecaller) + std::string(sample);
    Tiling *base_caller_name = new Tiling(path.c_str());
    base_caller_name->setChannel(channel, tile, ntiles);

    // initialise objects for creation of filenames
    path = std::string(path_intensities) + std::string(sample);
    Tiling *intensity_name = new Tiling(path.c_str());
    intensity_name->setChannel(channel, tile, ntiles);

    ProcessBaseCall *base_caller = new ProcessBaseCall();
    base_caller->setCycleNumber(ncycles);
    PhasingModel phasing(phasing_rate, prephasing_rate, starts_phasing);
    base_caller->setPhasingModel(phasing);

    /* 
     * load intensity files and call base-caller
     */
    {
        unsigned char bases[4] = { 'A', 'C', 'G', 'T' };
        int tile = 0;
        int nmax = default_cluster_allocation;
        int ncol = ncolours;
        int *is_deblocked = new int [ncycles];
        int i;

        /* read cluster positions */
        st_fposition *referencepositions = new st_fposition[nmax];
        read_positions(sequencer_name->getIndexFilename(POSITION_FILE,
                                                        tile).c_str(),
                       referencepositions, nmax); // TODO: set nmax dynamically
        unsigned char *allbases = new unsigned char [N_COLOURS * ncycles];
        float *allexposures = new float [ N_COLOURS * ncycles ];
        int cycle;
        for (cycle = 0; cycle < ncycles; cycle++) {
            for (i = 0; i < N_COLOURS; i++) {
            	allbases[cycle*N_COLOURS+i] = bases[i];
            	allexposures[cycle*N_COLOURS+i] = 0.;
            }
           /* read exposure times and block info from XML file */
            read_xml_info(sequencer_name->getIndexFilename(QC_FILE,
                                                           tile, cycle).c_str(),
                          allbases+cycle*N_COLOURS,
                          allexposures+cycle*N_COLOURS, N_COLOURS,
                          &is_deblocked[cycle]);
        }
        
        sprintf(msg, "Intensity file: %s",   
                intensity_name->getIndexFilename(INTENSITY_FILE,
                                                 tile, cycle).c_str());
        logger->info(msg);
        sprintf(msg, "Noise file:     %s",   
                intensity_name->getIndexFilename(NOISE_FILE,
                                                 tile, cycle).c_str());
       logger->info(msg);
            
        std::vector <float * > exposures;
        for (cycle = 0; cycle < ncycles; cycle++) {
            float *exposure_times = new float [N_COLOURS];

            /* read intensity values */
            for (i = 0; i < ncol; i++) {
                int baseidx = get_base_index(allbases[cycle*N_COLOURS+i], bases, ncol);
                assert((baseidx >= 0) && (baseidx < N_COLOURS)); 
                exposure_times[baseidx] = allexposures[cycle*N_COLOURS+i]; 
            }
            int offset = sprintf(msg, "Exposure:");
            for (i = 0; i < N_COLOURS; i++)
                offset += sprintf(msg+offset, "  %c %6.3f",
                                  BaseUpperCase[i], exposure_times[i]); 
            logger->info(msg);
            exposures.push_back(exposure_times);
        }     
        /* call base-caller */
        call(base_caller, intensity_name, base_caller_name,
             ncycles, channel, tile, referencepositions,
             (unsigned char *)BaseUpperCase, exposures,
             is_deblocked, response_matrices, starts_matrices, logger);

        for (cycle = 0; cycle < ncycles; cycle++)
            delete [] exposures[cycle];
        for (i = 0; i < int(response_matrices.size()); i++)
            if (response_matrices.at(i))
                delete response_matrices.at(i);
        delete [] allexposures;
        delete [] allbases;
        delete [] referencepositions;
        delete [] is_deblocked;
    }

    delete base_caller;
    delete logger;
    delete base_caller_name;
    delete sequencer_name;
    delete intensity_name;
}


int main(int argc, char *argv[]) 
{
    char matrix_filename[FILENAME_MAX] = "matrix.txt";
    char tmp_prefix[FILENAME_MAX] = "tempseq";
    char tmp_prefix2[FILENAME_MAX] = "tempseq";
    char tmp_prefix3[FILENAME_MAX] = "../tempseq";
    char parameter_file[FILENAME_MAX] = "";
    char sample[FILENAME_MAX] = "s";
    float phasing_rate = 0., prephasing_rate = 0.;
    int ncolours = N_COLOURS, ntiles = 1, ncycles = 0;
    int channel = 0, tileoffset = 0;
    int i;
    std::string logfile;
    int restart;
    const char delimiter[] = ",";
    char* token;
    std::vector<int> starts_phasing;
    std::vector<int> starts_matrix;
    std::vector<std::string> matrices;

    
    starts_phasing.push_back(1);

    for (i = 1; i < argc; i++) {
        if (argv[i][0] == '-')
            switch (argv[i][1]) {
            case 't': 
                switch (argv[i][2]) {
                case 'i': 
                    if (sscanf(argv[i], "-tile=%i,%i", 
                               &channel, &tileoffset) != 2)
                        if (sscanf(argv[i], "-tile=%[^_]_%d_%d",
                                   sample, &channel, &tileoffset) == 3) {
                            channel --; tileoffset--;
                        }
                    break;
                } break;
            case 'm': 
                switch (argv[i][2]) {
                case 'a':
                    switch (argv[i][3]) {
                    case 't': 
                        sscanf(argv[i], "-matrix=%s", matrix_filename);
                        matrices.push_back(matrix_filename);
                        starts_matrix.push_back(0);
                        break;
                    } 
                    break;
                case 'u':
                    switch (argv[i][3]) {
                    case 'l':
                        sscanf(argv[i], "-multiple_matrix=%s", matrix_filename);
                        token = strtok(matrix_filename, delimiter);
                        matrices.push_back(token);
                        token = strtok(NULL, delimiter);
                        starts_matrix.push_back((int)strtod(token, NULL)-1);
                        break;
                    } 
                    break;
                }
                break;
            case 'p': 
                if (sscanf(argv[i], "-phasing=%f", &phasing_rate))
                    break;
                if (sscanf(argv[i], "-prephasing=%f", &prephasing_rate))
                    break;
                if ((sscanf(argv[i], "-prefix=%s", tmp_prefix) == 1)
                     && (!strcmp(tmp_prefix2, "tempseq"))) {
                        strcpy(tmp_prefix2, tmp_prefix);
                        break;
                }
                if (sscanf(argv[i], "-prefix2=%s", tmp_prefix2))
                    break; 
                if (sscanf(argv[i], "-prefix3=%s", tmp_prefix3))
                    break; 
                if (sscanf(argv[i], "-params=%s", parameter_file)) {
                    read_xml_parameters(parameter_file,
                                        &phasing_rate, &prephasing_rate);
                }
                break;
            case 's': sscanf(argv[i], "-sample=%s", sample); break;
            case 'n': sscanf(argv[i], "-nc=%i", &ncycles); break;
            case 'l':
            	{
            	    char path[FILENAME_MAX];
                    sscanf(argv[i], "-logfile=%s", path);
                    logfile = path;
            	} break;
            case 'r':
                sscanf(argv[i], "-restart=%d", &restart);
                starts_phasing.push_back(restart);
                break;
            }
    }

    if ((argc < 2)) {
        printf("Usage: run_base_caller [options]\n"
               "Options:\n"
           "  -matrix=<filename>\n     set the path for matrix file (single matrix)\n"
           "  -multiple_matrix=<filename>,<cycle>\n     set the path for matrix file and the cycle at which\n     the matrix is applied (multiple matrices)\n"
           "  -phasing=<x>\n     set phasing rate (between 0.0 and 1.0)\n"
           "  -prephasing=<x>\n     set prephasing rate (between 0.0 and 1.0)\n"
           "  -restart=<cycle>\n     set cycle at which the phasing/prephasing effect is reset (1 indexed)\n"
           "  -params=<filename>\n     read options from parameter XML file\n"
           "  -prefix=<path>\n     set path for image analysis output files\n"
           "  -prefix2=<path>\n     set prefix for output files\n"
           "  -prefix3=<path>\n     set prefix for input intensity files\n"
           "  -sample=<text>\n     set sample id (must not contain underscores)\n"
           "  -tile=<lane>,<tile>\n     set lane/tile (0 indexed)\n"
           "  -tile=<sample>_<lane>_<tile>\n     set sample/lane/tile (1 indexed)\n"
           "  -nc=<n>\n     set number of cycles sequenced\n"
           "  -logfile=<filename>\n     set the path for log file\n");
        exit(0);
    }
    assert((phasing_rate >= 0.) && (phasing_rate < 0.1));
    assert((prephasing_rate >= 0.) && (prephasing_rate < 0.1));

    if (logfile.length() == 0) {
        logfile = tmp_prefix2;
	logfile += software;
        logfile += ".log";
    }
    printf("Logfile %s\n", logfile.c_str());
    run_base_caller(tmp_prefix, tmp_prefix2, tmp_prefix3, sample,
                    matrices, starts_matrix,
                    ncycles, ncolours, ntiles, 
                    channel, tileoffset, logfile,
                    phasing_rate, prephasing_rate, starts_phasing);

    return 0;
}
