//  Time-stamp: <26 Jun 2006 16:57 BST Tiling.cpp km>
/*
 * Set the tiling structure and runfolder names for a data analysis run.
 * This object is used to generate suitable filenames for output files,
 * according to the run-folder specification. Internally, the image
 * analysis does not need to know about these conventions; tiles (positions
 * on the chip) are simply numbered sequentially.
 *
 * Copyright (c) Solexa 2005, 2006
 * Author: Klaus Maisinger
 * This source file is covered by the "Solexa Public Source License"
 * agreement and bound by the terms therein.
 */

#if defined(__BORLANDC__)
#include <stdio.h>
#else
#include <cstdio>
#endif
#include "Tiling.h"

Tiling::Tiling(const char prefix[]) : nchannel(0), tileoffset(0)
{
    path = prefix;
}

Tiling::~Tiling()
{
    if (nchannel > 0) {
        delete [] channelidx; delete [] channelcount;
    }
}

/*
 * use <nchan>  channels with indices given by the <nchan>-dim 
 * array <chanidx>. each channel has <chancount> tiles.
 */
void Tiling::setChannels(int *chanidx, int *chancount, int nchan)
{
    nchannel = nchan;
    channelcount = new int [nchan];
    channelidx = new int [nchan];
    int i;
    for (i = 0; i < nchan; i++) {
        channelcount[i] = chancount[i];
        channelidx[i] = chanidx[i];
    }
}

/*
 * only use one channel with index <chanidx>. there are <ntiles> tiles,
 * counting starts from <chanoffset>
 */
void Tiling::setChannel(int chanidx, int chanoffset, int ntiles)
{
    nchannel = 1;
    channelcount = new int [1];
    channelidx = new int [1];
    channelcount[0] = ntiles+chanoffset;
    channelidx[0] = chanidx;
    tileoffset = chanoffset;
}

int Tiling::getChannel(int idx) 
{ 
    int i = 0, total = 0;
    while (i < nchannel) {
        total += channelcount[i];
        if (idx < total)
            return channelidx[i];
    }
    return 0; 
}

int Tiling::getTile(int idx) { 
    int i = 0, total = 0;
    while (i < nchannel) {
        total += channelcount[i];
        if (idx < total)
            return idx-(total-channelcount[i]) + tileoffset;
    }
    return idx+tileoffset; 
}

int Tiling::getIndex(int channel, int tile) {
    int i = 0, total = 0;
    while ((i < nchannel) && (channelidx[i] < channel))
        total += channelcount[i];
    return tile+total;
}

static std::string typestring[12] = 
    { "pos", "idx", "int", "nse", "hdr", "off", "clu", "qcm", 
      "sig", "seq", "prb", "sig2" };

std::string Tiling::getTileFilename(filename_type type, int channel, 
                                    int tile, int cycle, int colour)
{
    char filename[FILENAME_MAX];
    if (type == QC_FILE) {
        snprintf(filename, FILENAME_MAX,
                 "%s_%01d_%04d_%02d_%s.xml", path.c_str(), channel + 1,
                 tile + 1, cycle + 1, typestring[type].c_str());
        return std::string(filename);
    }
    if (cycle < 0) {
        snprintf(filename, FILENAME_MAX,
                 "%s_%01d_%04d_%s.txt", path.c_str(), channel + 1,
                 tile + 1, typestring[type].c_str());
    }
    else {
        if (colour < 0)
            snprintf(filename, FILENAME_MAX,
                    "%s_%01d_%04d_%02d_%s.txt", path.c_str(),
                    channel + 1, tile + 1, 
                    cycle + 1, typestring[type].c_str());
        else
             snprintf(filename, FILENAME_MAX,
                    "%s_%01d_%04d_%02d_%1d_%s.txt", path.c_str(),
                    channel + 1, tile + 1, 
                    cycle + 1, colour + 1, typestring[type].c_str());
            
    }
    return std::string(filename);
}

std::string Tiling::getIndexFilename(filename_type type, 
                                     int idx, int cycle, int colour)
{
    return getTileFilename(type, getChannel(idx), getTile(idx), cycle, colour);
}

std::string Tiling::getFilePrefix()
{
    std::string s = path;
    return s;
}
