/*
 * Copyright (c) 2006 Solexa
 * This source file is covered by the "Solexa Public Source License"
 * agreement and bound by the terms therein.
 */
#ifndef INCLUDED_ELAND_INNER_H
#define INCLUDED_ELAND_INNER_H


#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fstream>
#include <iostream>
#include <vector>
#include <deque>
#include <string>
#include <cmath>
#include <algorithm>
#include <map>
#include <dirent.h>
#include <pthread.h>
#include <utility>
#include <sys/resource.h>

#include "GlobalUtilities.h"
#include "ELAND_outer.h"


// Identifier to embed in source and binary files
static char* eland_inner_h_id="@(#) $Id: ELAND_inner.h,v 1.3 2006/10/23 15:41:25 tc Exp $";

/*
PROJECT: ELAND (Efficient Local Alignment of Nucleotide Data)
MODULE: ELAND_inner.h
AUTHOR: A. J. Cox

Copyright (c) 2003-2006 Solexa Limited. All rights reserved.

*/


// Stands for Efficient Local Alignment of Nucleotide Data

// Matches oligos to genome data allowing for substitution errors and
// for ambiguity codes

// The basic plan: oligos are divided into 4 contiguous fragments, e.g.
// a 24-mer into 4 fragments of 6 bases - call them ABCD.  
// These 4 fragments can be partitioned into (assuming a 24 base oligo) 
// a 12 base prefix and a 12 base suffix in 6 different ways:
//
// Prefix  Suffix
// A B     C D
// C D     A B
// A C     B D
// B D     A C
// A D     B C
// B C     A D
//
// The idea is to convert the prefix to a number using 2 bits per base encoding
// and use this number as an index into a look up table.
//
// The look up table gives you the suffixes of all oligos that have the same
// prefix as the oligo you are searching for. These suffixes are then compared
// inexactly with the query suffix by doing an exclusive-OR then using the
// result to index the "suffix score table" - another look up table which
// gives the positions of any differing bases.
//
// The key point is that we can catch all possible 2 substitution error 
// matches by checking the 6 combinations above: the worst case scenario is 
// that the two errors occur in different fragments - for one of the six 
// combinations both these fragments will be in the suffix and so will be 
// caught by the inexact matching step.
//
// 'N's in the data (and some 2 base ambiguity codes) can be handled by
// additionally storing an AND mask for each oligo which is applied after
// the exclusive-OR and prior to accessing the suffix score table. 
//
// It makes sense to cover the six possible partitions in 3 passes
// through the genome, making use of the fact that the partitions can be 
// paired so that the prefix of one is the suffix of another.


// The way in which an oligo matches to the genome data is described by
// one instance of MatchPosition 
// one instance of MatchDescriptor
// The match as described by these two quantities may be categorised as
// being in one of seven states:
// 
// 1: NM - No Match for this oligo
// 2: UE - Unique Exact match for this oligo
// 3: U1 - Unique match for this oligo with a single base substitution error
// 4: U2 - Unique match for this oligo, having 2 base substitution errors
// 5: RE - Unique Exact match for this oligo
// 6: R1 - Unique match for this oligo with a single base substitution error
// 7: R2 - Unique match for this oligo, having 2 base substitution errors


// MatchPosition stores a position in the genome (chromosome number + position)
// as a single unsigned int
// Capacity of an unsigned int is split into blocks according to the
// value of the most significant byte
// 0 : no match
// 1 - 239 : reserved for sequence data
// 240 - 255 : repeat
// blockRepeat + x : this oligo is the same as oligo number x in the batch
// ~0 (all bits 1) : match not attempted - oligo failed quality control (too many Ns etc)



// Hash tables work as follows
// i) Using 2 bits per base encoding convert prefix of oligo or genome fragment
// into a number  
// ii) Use this number to index into entryPointer_ - look up table of 
// TablePointers
// iii) Value of corresponding TablePointer gives you the position in hashRem_
// of the first TableEntry for that prefix

typedef uint TablePointer;

template<> bool TableEntry<false>::operator<(const TableEntry<false>& rhs) const
{
  return ((mask<rhs.mask)
	  ||((mask==rhs.mask)
	     &&((suffix.ui<rhs.suffix.ui)
		||((suffix.ui==rhs.suffix.ui)&&(position<rhs.position)))));
} // ~bool operator<(const TableEntry& rhs)

template<> bool TableEntry<false>::operator==(const TableEntry<false>& rhs) const
{
  return ( (suffix.ui==rhs.suffix.ui)
	   &&(mask==rhs.mask)
	   &&(position==rhs.position));
} // ~bool operator==(const TableEntry& rhs)

template<> bool TableEntry<false>::operator!=(const TableEntry<false>& rhs) const
{
  return ( (suffix.ui!=rhs.suffix.ui)
	   ||(mask!=rhs.mask)
	   ||(position!=rhs.position));
} // ~bool operator==(const TableEntry& rhs)

template<> bool TableEntry<true>::operator<(const TableEntry<true>& rhs) const
{
  return ( ((position&splitPrefixMaskHigh)<(rhs.position&splitPrefixMaskHigh))
	   ||(((position&splitPrefixMaskHigh)==(rhs.position&splitPrefixMaskHigh))
	      &&((mask<rhs.mask)
		 ||((mask==rhs.mask)
		    &&((suffix.ui<rhs.suffix.ui)
		       ||((suffix.ui==rhs.suffix.ui)&&(position<rhs.position)))))));
} // ~bool operator<(const TableEntry& rhs)

template<> bool TableEntry<true>::operator==(const TableEntry<true>& rhs) const
{
  return ( (suffix.ui==rhs.suffix.ui)
	   &&(mask==rhs.mask)
	   &&(position==rhs.position));
	   //	   &&((position&splitPrefixMaskHigh)==(rhs.position&splitPrefixMaskHigh)) );
} // ~bool operator==(const TableEntry& rhs)

template<> bool TableEntry<true>::operator!=(const TableEntry<true>& rhs) const
{
  return ( (suffix.ui!=rhs.suffix.ui)
	   ||(mask!=rhs.mask)
	   ||(position!=rhs.position));
	   //	   ||((position&splitPrefixMaskHigh)!=(rhs.position&splitPrefixMaskHigh)) );
} // ~bool operator==(const TableEntry& rhs)


// NB this won't work for n=32
#define hashsize(n) ((uint)1<<(n))
#define hashmask(n) (hashsize(n)-1)



// Oligo format, e.g. for 21 mers, 1 character = 1 bit:
// ui0.............................ui1.............................
// uc0.....uc1.....uc2.....uc3.....uc4.....uc5.....uc6.....uc7.....
// 2.1.1.1.1.1.1.1.1.1.1.9.8.7.6.5.4.3.2.1.0.  <-base numbers 0 20
// 0 9 8 7 6 5 4 3 2 1 0

// QueryGenerator: generates a set of query sequences for an oligo, allowing 
// for:
// i)  reverse complements
// ii) Ns
// iii) 2 base ambiguity codes (TBD)

template< int DIFF> struct ReverseShifter
{
  ReverseShifter() : 
    prefixMask_( (Word)0x3 ),
    suffixMask_( ((Word)0x3)<<(numBitsPerBase*(suffixLength-1)) ) {}


  void operator()( const Oligo& ol, Oligo& rc ) const
  {
    // should never be compiled, always goes to a specialization
    STATIC_ASSERT(1==0);
  } // ~operator()

  const Word prefixMask_;
  const Word suffixMask_;

}; // ~template< int DIFF> struct ReverseShifter 


template<> void ReverseShifter<0>::operator()
( const Oligo& ol, Oligo& rc ) const
{
  // prefix and suffix are same size, don't need to swap bits between elements
  rc.ui[0]>>=((maxBasesPerWord-suffixLength)*numBitsPerBase);
  rc.ui[1]>>=((maxBasesPerWord-prefixLength)*numBitsPerBase);
} // ~template<> void ReverseShifter<0>::operator()


template<> void ReverseShifter<1>::operator()
( const Oligo& ol, Oligo& rc ) const
{
  //  cout << "move pref to suff" << endl;
  // move 1 base from prefix to suffix
  rc.ui[0]<<=numBitsPerBase;
  rc.ui[0]>>=((maxBasesPerWord-suffixLength)*numBitsPerBase);
  rc.ui[1]>>=((maxBasesPerWord-prefixLength)*numBitsPerBase);
  rc.ui[1]^=(ol.ui[1]&prefixMask_);
  //  rc.ui[1]^=prefixMask_;
} // ~template<> void ReverseShifter<1>::operator()

template<> void ReverseShifter<-1>::operator()
( const Oligo& ol, Oligo& rc ) const
{
  //  cout << "move suff to pref \n" ;
  // move 1 base from suffix to prefix
  rc.ui[0]>>=numBitsPerBase;
  rc.ui[0]>>=((maxBasesPerWord-suffixLength)*numBitsPerBase);
  rc.ui[1]>>=((maxBasesPerWord-prefixLength)*numBitsPerBase);
  rc.ui[0]|=(ol.ui[0]&suffixMask_);
  rc.ui[0]^=suffixMask_;
} // ~template<> void ReverseShifter<-1>::operator()


class QueryGenerator
{
public:

  QueryGenerator()
  {}
    
  // generate a set of query sequences from ASCII sequence data
  int operator()( const char* buf, const uint oligoNum, 
		  vector<Oligo>& queryOligo,
		  vector<Oligo>& queryMask,
		  vector<OligoNumber>&  queryOligoNum );

  // convert a single oligo from ASCII to binary. 
  // Fail if the oligo contains Ns or other funny stuff
  void operator()( const char* buf, Oligo& oligo );

  // convert an oligo from ASCII to binary and produce a 
  // mask for the Ns - return number of Ns found
  int encodeOligo( const char* buf, Oligo& oligo, Oligo& mask );

  void reverseOligo( const Oligo& ol, Oligo& rc ) const
  {

    rc.uc[0]=reverseChar[ol.uc[7]]; 
    rc.uc[1]=reverseChar[ol.uc[6]];
    rc.uc[2]=reverseChar[ol.uc[5]];
    rc.uc[3]=reverseChar[ol.uc[4]];
    rc.uc[4]=reverseChar[ol.uc[3]];
    rc.uc[5]=reverseChar[ol.uc[2]];
    rc.uc[6]=reverseChar[ol.uc[1]];
    rc.uc[7]=reverseChar[ol.uc[0]];
    reverseShift_( ol, rc );
  } // ~reverseOligo
  
protected:
  char tempBuf1[OLIGO_LENGTH+1];
  char tempBuf2[OLIGO_LENGTH+2];
  ReverseShifter<prefixLength-suffixLength> reverseShift_;

}; // ~class QueryGenerator : public ExpandedTranslationTable


// convert an oligo from ASCII to binary and produce a 
// mask for the Ns - return number of Ns found
// template< int oligoLength > 
int QueryGenerator::encodeOligo( const char* buf, Oligo& oligo, Oligo& mask )
{
  int numNs(0);
  bool isN;

  for (int i(0);i<prefixLength;++i, ++buf)
  {
    isN=isBlank(*buf);
    //    isN=( (*buf=='n')||(*buf=='N') );
    oligo.ui[1]<<=numBitsPerBase;
    oligo.ui[1]|=(isN?0:whichBase[*buf]);
    mask.ui[1]<<=numBitsPerBase;
    mask.ui[1]|=0x3*isN;
    numNs+=1*isN;
  } // ~for i

  for (int i(0);i<suffixLength;i++, ++buf) // was max bases per word
  {
    isN=isBlank(*buf);
    //    isN=( (*buf=='n')||(*buf=='N') );
    oligo.ui[0]<<=numBitsPerBase;
    oligo.ui[0]|=(isN?0:whichBase[*buf]);
    mask.ui[0]<<=numBitsPerBase;
    mask.ui[0]|=0x3*isN;
    numNs+=1*isN;
  } // ~for i

  return numNs;
} // ~int encodeOligo( const char* buf, Oligo& oligo, Oligo& mask )


// convert a single oligo from ASCII to binary. 
// Fail if the oligo contains Ns or other funny stuff
// template< int oligoLength > 
void QueryGenerator::operator()( const char* buf, Oligo& oligo )
{
  Oligo dummy;
  oligo.ui[0]=0;
  oligo.ui[1]=0;
  if (encodeOligo( buf, oligo, dummy)!=0)
  {
    cerr << "Error: did not expect to find Ns in oligo!" << endl;
  }
} // ~void operator()( const char* buf, Oligo& oligo )


// generate a set of query sequences from ASCII sequence data
// template< int oligoLength > 
int QueryGenerator::operator()
( const char* buf, const uint oligoNum, 
  vector<Oligo>& queryOligo,
  vector<Oligo>& queryMask,
  vector<OligoNumber>&  queryOligoNum )
{
  int numNs(0), tailSize(0), headSize(0);
  //  char tempBuf1[OLIGO_LENGTH+1], tempBuf2[OLIGO_LENGTH+1];
  tempBuf1[OLIGO_LENGTH]='\0';
  tempBuf2[OLIGO_LENGTH]='\0';
  queryOligo.clear();
  queryMask.clear();
  queryOligoNum.clear();

  //  cout << endl << buf << " - initial " << endl;

  for (int i(0);i<OLIGO_LENGTH;i++)
    numNs+=isBlank(buf[i]);
  //    numNs+=((buf[i]=='n')||(buf[i]=='N'));
  if (numNs>4)
  {
    //    cout << "Too many Ns, rejecting" << endl;
    return numNs;
  } // ~if


  //  while ( ((buf[headSize]=='N')||(buf[headSize]=='n'))
  //	  &&(headSize<OLIGO_LENGTH) )
  while ( isBlank(buf[headSize])&&(headSize<OLIGO_LENGTH) )
    headSize++;

  //  while( ((buf[OLIGO_LENGTH-tailSize-1]=='N')
  //	  ||(buf[OLIGO_LENGTH-tailSize-1]=='n'))&&(tailSize<OLIGO_LENGTH))
  while( (isBlank(buf[OLIGO_LENGTH-tailSize-1])&&(tailSize<OLIGO_LENGTH)))
    tailSize++;


  // copy to buffer, shifting any head->tail 

  for ( int i(0);i<OLIGO_LENGTH-headSize;i++)
    tempBuf1[i]=buf[i+headSize];

  for ( int i(OLIGO_LENGTH-headSize);i<OLIGO_LENGTH;i++)
    tempBuf1[i]='N';

  tailSize += headSize; // can now forget about headSize

  int numInternalNs(numNs-tailSize);

  if (numInternalNs>2)
  {
    // cout << "Too many separate Ns, rejecting" << endl;
    return numNs;
  }

  // OK, output oligo as is
  //  cout << tempBuf1 << " - shifted " << endl;
  queryOligo.push_back( Oligo() );
  queryMask.push_back( Oligo() );
  queryOligoNum.push_back(oligoNum);
 
  assert( numNs==encodeOligo(tempBuf1, queryOligo.back(), queryMask.back()));
  //    printWord(queryOligo.back().ui[1],prefixLength_);
  //    printWord(queryOligo.back().ui[0],16);
  //   cout << " - encoded oligo" << endl;
  //    printWord(queryMask.back().ui[1],prefixLength_);
  //    printWord(queryMask.back().ui[0],16);
  //  cout << " - encoded mask" << endl;

  // if >=1 error, copy removing first N, send to output list
  if (numInternalNs>=1)
  {
    int i, firstN, secondN;

    tempBuf2[OLIGO_LENGTH-1]='n';
    //    for (i=0;((i<OLIGO_LENGTH-tailSize)
    //      &&(tempBuf1[i]!='n')
    //      &&(tempBuf1[i]!='N'));i++)
    for (i=0;((i<OLIGO_LENGTH-tailSize)&&(!isBlank(tempBuf1[i])));i++)
    {
      tempBuf2[i]=tempBuf1[i];
    } // ~for i
    assert(i!=OLIGO_LENGTH-tailSize);
    firstN=i;
    i++;

    for (;(i<OLIGO_LENGTH);i++)
    {
      tempBuf2[i-1]=tempBuf1[i];
    } // ~for i

    //     cout << "After removal of first internal N: " << endl;
    //   cout << tempBuf2 << " - modified buffer" << endl;
    // OK, output oligo
    queryOligo.push_back( Oligo() );
    queryMask.push_back( Oligo() );
    queryOligoNum.push_back(oligoNum);
    
    assert( numNs==encodeOligo(tempBuf2, queryOligo.back(), queryMask.back()));
    assert( (queryMask.back().ui[0]&0x3)==0x3 );
    //  queryOligo.back().ui[0] |= nonIncorpFirstN;
    queryOligo.back().ui[0] |= (nonIncorpFirstN&queryMask.back().ui[0]);
    queryOligo.back().ui[1] |= (nonIncorpFirstN&queryMask.back().ui[1]);

    //      printWord(queryOligo.back().ui[1],prefixLength_);
    //   printWord(queryOligo.back().ui[0],16);
    //   cout << " - encoded ol " << endl;
    //     printWord(queryMask.back().ui[1],prefixLength_);
    //    printWord(queryMask.back().ui[0],16);
    //   cout << endl;

    if (numInternalNs==2)
    {
      //      for (i=firstN+1;((i<OLIGO_LENGTH-tailSize)
      //	       &&(tempBuf1[i]!='n')
      //	       &&(tempBuf1[i]!='N'));i++);
      for (i=firstN+1;((i<OLIGO_LENGTH-tailSize)&(!isBlank(tempBuf1[i])));i++);

      assert(i!=OLIGO_LENGTH-tailSize);
      secondN=i;

      for (i=0;i<firstN;i++) tempBuf2[i]=tempBuf1[i];
      for (i=firstN+1;i<secondN;i++) tempBuf2[i-1]=tempBuf1[i];
      for (i=secondN+1;i<OLIGO_LENGTH;i++) tempBuf2[i-2]=tempBuf1[i];
      tempBuf2[OLIGO_LENGTH-2]='n';
      tempBuf2[OLIGO_LENGTH-1]='n';
      //      cout << "After removal of both internal Ns: " << endl;
      //  cout << tempBuf2 << " - buffer" << endl;
      queryOligo.push_back( Oligo() );
      queryMask.push_back( Oligo() );
      queryOligoNum.push_back(oligoNum);
    
      assert( numNs
	      ==encodeOligo(tempBuf2, queryOligo.back(), queryMask.back()));
      assert( (queryMask.back().ui[0]&0x3)==0x3 );
      //      queryOligo.back().ui[0] |= nonIncorpBothNs;
      queryOligo.back().ui[0] |= (nonIncorpBothNs&queryMask.back().ui[0]);
      queryOligo.back().ui[1] |= (nonIncorpBothNs&queryMask.back().ui[1]);


      //         printWord(queryOligo.back().ui[1],prefixLength_);
      //    printWord(queryOligo.back().ui[0],16);
      //   cout << " - encoded " << endl;
      //    printWord(queryMask.back().ui[1],prefixLength_);
      //   printWord(queryMask.back().ui[0],16);
      //  cout << endl;

      if (secondN!=firstN+1)
      {
	for (i=0;i<secondN;i++) tempBuf2[i]=tempBuf1[i];
	for (i=secondN+1;i<OLIGO_LENGTH;i++) tempBuf2[i-1]=tempBuf1[i];
	tempBuf2[OLIGO_LENGTH-1]='n';
	//	 cout << "After removal of second internal N:" << endl;
	//	 cout << tempBuf2 << " - buffer" << endl;
	queryOligo.push_back( Oligo() );
	queryMask.push_back( Oligo() );
	queryOligoNum.push_back(oligoNum);
    
	assert( numNs
		==encodeOligo(tempBuf2, queryOligo.back(), queryMask.back()));
	assert( (queryMask.back().ui[0]&0x3)==0x3 );
	//	queryOligo.back().ui[0] |= nonIncorpSecondN;
	queryOligo.back().ui[0] |= (nonIncorpSecondN&queryMask.back().ui[0]);
	queryOligo.back().ui[1] |= (nonIncorpSecondN&queryMask.back().ui[1]);

	//		printWord(queryOligo.back().ui[1],prefixLength_);
	//		printWord(queryOligo.back().ui[0],16);
	//	 cout << " - encoded" << endl;
	//	printWord(queryMask.back().ui[1],prefixLength_);
	//	printWord(queryMask.back().ui[0],16);
	//	 cout << endl;

      } // ~if
      else
      { 
	//	 cout << "Internal Ns are adjacent" << endl;
      }
   } // ~if
  } // ~if

#ifndef DONT_SEARCH_REVERSE_STRAND
  // add reverse to all oligos in pile
  const int numToDo(queryOligo.size());
  for (int i(0);i<numToDo;++i)
  {
    queryOligo.push_back( Oligo() );
    queryMask.push_back( Oligo() );
    queryOligoNum.push_back(queryOligoNum[i]|isReverseOligo);
    reverseOligo( queryOligo[i], queryOligo.back());

    // check for self-complementary sequences. These are rare but can
    // cause the oligo concerned to be spuriously flagged as a repeat
    if (queryOligo[i]!=queryOligo.back())
    { // not self-complementary, carry on 
      reverseOligo( queryMask[i], queryMask.back());
      //      queryMask.back().ui[0] ^= ((uint)~0); // TBD
      // changed 03.06.03 TC
      // rhs of next line is supposed to contain all ones in its first 
      // 2*prefixLength_ bits - does not work for prefixLength_=16
      //	queryMask.back().ui[1] ^= ~( ((uint)~0)<< (prefixLength_<<1) );
      //      queryMask.back().ui[1] ^= 
      //	( ((uint)~0) >> ((maxBasesPerWord-prefixLength_)<<1) );
      // Changed to match new oligo format TC 05.08.03

      queryMask.back().ui[0] ^= 
	( ((uint)~0) >> ((maxBasesPerWord-suffixLength)<<1) );
      queryMask.back().ui[1] ^= 
	( ((uint)~0) >> ((maxBasesPerWord-prefixLength)<<1) );

    } // ~if
    else
    { // self-complementary
      queryOligo.pop_back();
      queryMask.pop_back();
      queryOligoNum.pop_back();
      //	cout << oligoNum << " is self-complementary" << endl;
    } // ~else

  } // ~for
#endif
  
  return numNs;

} // ~int QueryGenerator::operator()





// class Hasher: scrambles an oligo into a hash value
template<int PASS> class Hasher
{
public:
  Hasher();
  inline void operator()( const Oligo& ol, Oligo& out ) const;

  // These are logical AND masks for the four fragments
  const Word fragMaskA_;
  const Word fragMaskB_;
  const Word fragMaskC_;
  const Word fragMaskD_;

  //  const int hashShift1_; - moved to ConvertOligo
  const int hashShift2_;

  int getLowerFragSizePart1( void ) const;
  Word getLowerFragMaskPart1( void ) const;
  const FragmentErrorType*  getLowerFragScorePart1
  ( const SuffixScoreTable& table ) const;
  const FragmentErrorType*  getUpperFragScorePart1
  ( const SuffixScoreTable& table ) const; 
  int getLengthPart1( void ) const;

  int getLowerFragSizePart2( void ) const;
  Word getLowerFragMaskPart2( void ) const;
  const FragmentErrorType*  getLowerFragScorePart2
  ( const SuffixScoreTable& table ) const;
  const FragmentErrorType*  getUpperFragScorePart2
  ( const SuffixScoreTable& table ) const; 
  int getLengthPart2( void ) const;

private:
}; // ~class Hasher

template<int PASS> Hasher<PASS>::Hasher() :
  fragMaskA_(hashmask(numBitsPerBase*fragLengthA)),
  fragMaskB_(hashmask(numBitsPerBase*fragLengthB)
	     <<(numBitsPerBase*fragLengthA)),
  fragMaskC_(hashmask(numBitsPerBase*fragLengthC)),
  fragMaskD_(hashmask(numBitsPerBase*fragLengthD)
	     <<(numBitsPerBase*fragLengthC)),
  hashShift2_(numBitsPerBase*((OLIGO_LENGTH%fragmentsPerOligo)==2))
{
  //  cerr << fragLengthA << " " << fragLengthB << " " 
  //      << fragLengthC << " " << fragLengthD << endl;

  assert(fragLengthA+fragLengthB+fragLengthC+fragLengthD
	 ==OLIGO_LENGTH);
  assert((fragMaskA_&fragMaskB_)==0);
  assert((fragMaskC_&fragMaskD_)==0);

} // ~template<> Hasher<0>::Hasher( int keyBits, int oligoBases ) :

// Parameters for pass 0: partition1=AB, partition2=CD
template<> int Hasher<0>::getLowerFragSizePart1( void ) const
{ return fragLengthA; }
template<>  Word Hasher<0>::getLowerFragMaskPart1( void ) const
{ return fragMaskA_; }
template<> const FragmentErrorType*  Hasher<0>::getLowerFragScorePart1
( const SuffixScoreTable& table ) const
{ return &table.scoreFragA_[0]; }
template<> const FragmentErrorType*  Hasher<0>::getUpperFragScorePart1
( const SuffixScoreTable& table ) const
{ return &table.scoreFragB_[0]; }
template<>  int Hasher<0>::getLengthPart1( void ) const
{ return fragLengthA+fragLengthB; }

template<> int Hasher<0>::getLowerFragSizePart2( void ) const
{ return fragLengthC; }
template<>  Word Hasher<0>::getLowerFragMaskPart2( void ) const
{ return fragMaskC_; }
template<> const FragmentErrorType*  Hasher<0>::getLowerFragScorePart2
( const SuffixScoreTable& table ) const
{ return &table.scoreFragC_[0]; }
template<> const FragmentErrorType*  Hasher<0>::getUpperFragScorePart2
( const SuffixScoreTable& table ) const
{ return &table.scoreFragD_[0]; }
template<>  int Hasher<0>::getLengthPart2( void ) const
{ return fragLengthC+fragLengthD; }

// Parameters for pass 1: partition1=CB, partition2=AD
template<> int Hasher<1>::getLowerFragSizePart1( void ) const
{ return fragLengthC; }
template<>  Word Hasher<1>::getLowerFragMaskPart1( void ) const
{ return fragMaskC_; }
template<> const FragmentErrorType*  Hasher<1>::getLowerFragScorePart1
( const SuffixScoreTable& table ) const
{ return &table.scoreFragC_[0]; }
template<> const FragmentErrorType*  Hasher<1>::getUpperFragScorePart1
( const SuffixScoreTable& table ) const
{ return &table.scoreFragB_[0]; }
template<>  int Hasher<1>::getLengthPart1( void ) const
{ return fragLengthC+fragLengthB; }

template<> int Hasher<1>::getLowerFragSizePart2( void ) const
{ return fragLengthA; }
template<>  Word Hasher<1>::getLowerFragMaskPart2( void ) const
{ return fragMaskA_; }
template<> const FragmentErrorType*  Hasher<1>::getLowerFragScorePart2
( const SuffixScoreTable& table ) const
{ return &table.scoreFragA_[0]; }
template<> const FragmentErrorType*  Hasher<1>::getUpperFragScorePart2
( const SuffixScoreTable& table ) const
{ return &table.scoreFragD_[0]; }
template<>  int Hasher<1>::getLengthPart2( void ) const
{ return fragLengthA+fragLengthD; }

// Parameters for pass 2: partition1=DB, partition2=CA
template<> int Hasher<2>::getLowerFragSizePart1( void ) const
{ return fragLengthD; }
template<>  Word Hasher<2>::getLowerFragMaskPart1( void ) const
{ return (fragMaskD_>>(numBitsPerBase*fragLengthC)); }
template<> const FragmentErrorType*  Hasher<2>::getLowerFragScorePart1
( const SuffixScoreTable& table ) const
{ return &table.scoreFragD_[0]; }
template<> const FragmentErrorType*  Hasher<2>::getUpperFragScorePart1
( const SuffixScoreTable& table ) const
{ return &table.scoreFragB_[0]; }
template<>  int Hasher<2>::getLengthPart1( void ) const
{ return fragLengthD+fragLengthB; }

// ordering of A and C swapped so that A is in the lower bits of the
// fragment
template<> int Hasher<2>::getLowerFragSizePart2( void ) const
{ return fragLengthA; }
template<>  Word Hasher<2>::getLowerFragMaskPart2( void ) const
{ return fragMaskA_; }
template<> const FragmentErrorType*  Hasher<2>::getLowerFragScorePart2
( const SuffixScoreTable& table ) const
{ return &table.scoreFragA_[0]; }
template<> const FragmentErrorType*  Hasher<2>::getUpperFragScorePart2
( const SuffixScoreTable& table ) const
{ return &table.scoreFragC_[0]; }
template<>  int Hasher<2>::getLengthPart2( void ) const
{ return fragLengthC+fragLengthA; }



// hash0: just convert an oligo into two partitions without scrambling, 
// e.g. for a 24-mer
// LSB    ----------------->    MSBlsb    ----------------->    msb 
// ol[0]                           ol[1] 
// 2 2 2 2 1 1 1 1 1 1 1 1 1 1 
// 3.2.1.0.9.8.7.6.5.4.3.2.1.0.9.8.7.6.5.4.3.2.1.0.xxxxxxxxxxxxxxxx
// ->
// out[0]                          out[1] 
// 2 2 2 2 1 1 1 1 1 1 1 1         1 1 
// 3.2.1.0.9.8.7.6.5.4.3.2.xxxxxxxx1.0.9.8.7.6.5.4.3.2.1.0.xxxxxxxx
// A---------->B---------->        C---------->D---------->

template<> void Hasher<0>::operator()( const Oligo& ol, Oligo& out ) const
{
  //  printWord(ol.ui[1],16);
  //  cout << '-';
  //  printWord(ol.ui[0],16);
  // cout << " -> ";

  out.ui[0]=ol.ui[0];
  out.ui[1]=ol.ui[1];

  //  printWord(out.ui[1],16);
  //  cout << '-';
  //  printWord(out.ui[0],16);
  //  cout << endl;

  //  cout << "done for this oligo" << endl << endl;

} // ~void Hasher::operator()


// hash1: convert an oligo into two partitions and scramble as follows, 
// e.g. for a 24-mer
// LSB    ----------------->    MSBlsb    ----------------->    msb 
// ol[0]                           ol[1] 
// 2 2 2 2 1 1 1 1 1 1 1 1 1 1 
// 3.2.1.0.9.8.7.6.5.4.3.2.1.0.9.8.7.6.5.4.3.2.1.0.xxxxxxxxxxxxxxxx
// ->
// out[0]                          out[1]
// 2 2 2 2 1 1 1 1                 1 1 1 1 1 1
// 3.2.1.0.9.8.1.0.9.8.7.6.xxxxxxxx7.6.5.4.3.2.5.4.3.2.1.0.xxxxxxxx
// A---------->C---------->        B---------->D---------->


template<> void Hasher<1>::operator()( const Oligo& ol, Oligo& out ) const
{

  //  printWord(ol.ui[1],16);
  //  cout << '-';
  //  printWord(ol.ui[0],16);
  //  cout << " -> ";

  out.ui[0]=ol.ui[0]&fragMaskB_; // out.ui[0] = 0B
  out.ui[0]|=(ol.ui[1]&fragMaskC_);// out.ui[0] = CB
  out.ui[1]=ol.ui[1]&fragMaskD_; // out.ui[1] = 0D
  //  out.ui[1]|=(ol.ui[1]&fragMaskA_); // out.ui[1] = AD - oops! AC 6.6.3
  out.ui[1]|=(ol.ui[0]&fragMaskA_); // out.ui[1] = AD

  //  printWord(out.ui[1],16);
  //  cout << '.';
  //  printWord(out.ui[0],16);
  //  cout << endl;

} // ~void Hasher::hash1( const Oligo& ol, Oligo& out ) const

// hash2: convert an oligo into two partitions and scramble as follows, 
// e.g. for a 24-mer
// LSB    ----------------->    MSBlsb    ----------------->    msb 
// ol[0]                           ol[1] 
// 2 2 2 2 1 1 1 1 1 1 1 1 1 1 
// 3.2.1.0.9.8.7.6.5.4.3.2.1.0.9.8.7.6.5.4.3.2.1.0.xxxxxxxxxxxxxxxx
// ->
// out[0]                          out[1] 
// 2 2 2 2 1 1                     1 1         1 1 1 1 1 1
// 3.2.1.0.9.8.5.4.3.2.1.0.xxxxxxxx1.0.9.8.7.6.7.6.5.4.3.2.xxxxxxxx
// A---------->D---------->        C---------->B---------->

template<> void Hasher<2>::operator()( const Oligo& ol, Oligo& out ) const
{
  //  printWord(ol.ui[1],16);
  //  cout << '-';
  //  printWord(ol.ui[0],16);
  //  cout << " -> ";

  out.ui[0]=ol.ui[0]&fragMaskB_; // out.ui[0] = 0B
  out.ui[0]<<=hashShift2_; // shift B if necessary so D fits next to it
  out.ui[0]|=ol.ui[1]>>(numBitsPerBase*fragLengthC); // out.ui[0]=DB

  // ordering of A and C swapped in second fragment TC 03.07.03
  //  out.ui[1]&=fragMaskC_; // out.ui[1]=C0
  // out.ui[1]|=((ol.ui[0]&fragMaskA_)<<(numBitsPerBase*fragLengthC));
  out.ui[1]=ol.ui[1]&fragMaskC_;
  out.ui[1]<<=(numBitsPerBase*fragLengthA);
  out.ui[1]|=(ol.ui[0]&fragMaskA_);

  //  printWord(out.ui[1],16);
  //  cout << '-';
  //  printWord(out.ui[0],16);
  //  cout << endl;

} // ~void Hasher::hash2( const Oligo& ol, Oligo& out ) const


// end of Hasher stuff




// class RepeatTable: reads in an ASCII table of repeats and filters them out 
// of the incoming oligos.

class RepeatTable
{
public:

  // struct IsLess: returns true if the unmasked bits of a are < b
  // allows binary searching for oligos containing Ns
  struct IsLess
  {
    IsLess( const Oligo& mask ) : 
      mask0_(~mask.ui[0]), mask1_(~mask.ui[1]) {}
    bool operator()( const Oligo& a, const Oligo& b )
    {
      return( ((a.ui[1]&mask1_)<(b.ui[1]&mask1_))
      	      ||( ((a.ui[1]&mask1_)==(b.ui[1]&mask1_))
		  &&((a.ui[0]&mask0_)<(b.ui[0]&mask0_))) );
    } // ~operator()
    const Word mask0_;
    const Word mask1_;
  }; // ~struct isLess

  RepeatTable( const char* repeatFileName );
  ~RepeatTable() 
  {
    cerr << "deleting repeat table" << endl;
  } // ~dtor

  bool check( const Oligo& oligo, const Oligo& mask )
  {
    return (!(binary_search(oligos_.begin(), 
			    oligos_.end(), 
			    oligo,
			    IsLess(mask) )));
  } // ~check

  void checkOligos( OligoSource& oligos, MatchTable& results );

private:
  vector<Oligo> oligos_;
}; // ~class RepeatTable


RepeatTable::RepeatTable( const char* repeatFileName )
{
  OligoSourceRaw repeats(repeatFileName);
  Oligo last;
  uint numRepeats(0), numEntries(0);
  QueryGenerator makeOligos;
  const char* pRepeat;

  cerr << "creating repeat table" << endl;
  last.ui[1]=~0;


  oligos_.push_back( Oligo() );
  while (1)
  {
    ++numEntries;
    if ((pRepeat=repeats.getNextOligo())==NULL) break;
    makeOligos( pRepeat, oligos_.back() );
    if (oligos_.back()!=last)
    {
      ++numRepeats;
      if (numEntries!=1) assert((last<oligos_.back())||(last==oligos_.back()));
      last=oligos_.back();
      oligos_.push_back( Oligo() );
    } // ~if
    //   else cout << " REPEATED";
    //   cout << endl;
  } // ~while
  
  oligos_.pop_back();

  sort(oligos_.begin(), oligos_.end());
  cerr << "storing " << numRepeats << " repeats out of " 
       << numEntries << " entries" << endl;
  assert(numRepeats==oligos_.size());
} // ~void RepeatTable:RepeatTable( const char* repeatFileName )

// return false if a repeat, true if OK to hash
void RepeatTable::checkOligos( OligoSource& oligos, MatchTable& results )
{
  const char* pOligo;
  Oligo oligo, mask;
  uint numRepeats(0), numEntries(0);
  QueryGenerator makeOligos;
  cerr << "checking oligos for repeats" << endl;
  while (1)
  {
    if ((pOligo=oligos.getNextOligo())==NULL) break;
    ++numEntries;

    // bug fix: oligo & mask need to be zeroed before each encoding - TC 18.7.3
    // (can get away with this in most cases because oligos are not reused)
    mask.ui[0]=0;
    mask.ui[1]=0;
    oligo.ui[0]=0;
    oligo.ui[1]=0;

    // changed to allow oligos with Ns to be checked - TC 11.07.03
    makeOligos.encodeOligo( pOligo, oligo, mask );
    if (check( oligo, mask )==false)
    {
      ++numRepeats;
      results.resize(numEntries+1);

      //      results.matchPosition_[numEntries]=repeatMasked;
      results.setRepeatMasked__(numEntries);

    } // ~if
  } // ~while

  oligos.rewind();
  cerr << "found " << numRepeats << " repeats in " << numEntries << " oligos" << endl;
  results.resize(numEntries+1);

} // ~void RepeatTable::isRepeat



// struct HashTableDataStore
// This is a wrapper for the data used by PartitionHashTable. Idea is
// this persists between passes so saves unnecessary allocation/dellocation
// of memory
struct HashTableDataStore
{
  vector<TablePointer> entryPointer_;
  vector<TableEntry<useSplitPrefix> > hashRem_; 
};



struct PHTHelperData
{
  PHTHelperData( HashTableDataStore& data, MatchTable& results ) : 
    data_(data), results_(results) {}
    
  HashTableDataStore& data_;

  TablePointer* pCount_;
  
  int lowerFragSize_;
  
  Word lowerFragMask_;

  const FragmentErrorType* lowerFragScore_;

  const FragmentErrorType* upperFragScore_;

  MatchTable& results_;

  map< Word, uint > maskMap_;
  vector<Word> maskTable_;

  //   next two quantities only used if useSplitPrefix==true
  Word splitPrefixMask_; 
  int splitPrefixShift_;
};

template<class Child, bool useSplitPrefix> struct PHTHelperBase : 
public PHTHelperData
{
  PHTHelperBase( HashTableDataStore& data, MatchTable& results ) : 
    PHTHelperData( data, results )
  {} 
  const Child& asChild( void ) const 
  { return static_cast<const Child&>(*this); }
  bool wantMatch( const FragmentErrorType errorLow, 
		  const FragmentErrorType errorHigh, 
		  const Word mask  ) const
  { return (asChild()).wantMatch(errorLow, errorHigh, mask); }
  Word getInfoN( const Word suff, const Word mask ) const 
  { return (asChild()).getInfoN(suff, mask); }
  inline void countKey( const Word key, 
			const Word prefixMask,
			const Word suffixMask);
  
  inline void hashEntry
  ( Word key, const Word keyMask,
    const Word entry, const Word entryMask,
    OligoNumber oligoNum );

  inline void check
  ( Word prefix, const Word suff, 
    const MatchPosition sequencePos);
  inline void check__
  ( const vector<pair<Oligo, MatchPosition> >& h );
  //  ( const uint a, const uint b, const Word suff, 
  //  const MatchPosition sequencePos);
  //  ( const uint a, const uint b, const Word suff, 
  //  const MatchPosition sequencePos);
};

// Specialization for "standard" mode (prefix not split)
template<class Child> struct PHTHelperBase<Child, false> : 
public PHTHelperData
{
  PHTHelperBase( HashTableDataStore& data, MatchTable& results ) : 
    PHTHelperData( data, results )
  {} 
  const Child& asChild( void ) const
  { return static_cast<const Child&>(*this); }
  bool wantMatch( const FragmentErrorType errorLow, 
		  const FragmentErrorType errorHigh, const Word mask ) const
  { return (asChild()).wantMatch(errorLow, errorHigh, mask); }
  Word getInfoN( const Word suff, const Word mask ) const 
  { return (asChild()).getInfoN(suff, mask); }

  inline void countKey( const Word key, 
			const Word prefixMask,
			const Word suffixMask)
  {
    if (prefixMask==0)
    {
      ++maskMap_[suffixMask];
      ++pCount_[key];
    } // ~if
  } // ~void countKey( const Word key, const Word mask )

  inline void hashEntry
  ( Word key, const Word keyMask,
    const Word entry, const Word entryMask,
    OligoNumber oligoNum )
  {
    if (keyMask==0)
    {
      data_.hashRem_[pCount_[key]].suffix.ui=entry;
      //      data_.hashRem_[pCount_[key]].mask.ui=entryMask;
      data_.hashRem_[pCount_[key]].mask=maskMap_[entryMask];
      data_.hashRem_[pCount_[key]].position=oligoNum;
      //    position_[pCount_[key]]=oligoNum;
      pCount_[key]++;
    } // ~if
  } // ~void hashEntry

  inline void check
  ( Word prefix, const Word suff, 
    const MatchPosition sequencePos ) 
  {
    register Word thisMatch(0), thisMask;
    register FragmentErrorType errorLow, errorHigh;
    for (uint i(pCount_[prefix]);
         i!=pCount_[prefix+1];
         ++i)
    {
#ifdef DEBUG
      printWord(data_.hashRem_[i].suffix.ui, maxBasesPerWord );
      cout << " - ";
      printWord(maskTable_[data_.hashRem_[i].mask],maxBasesPerWord); 
      cout << endl;
#endif

      thisMatch = (suff^data_.hashRem_[i].suffix.ui);
      thisMask = maskTable_[data_.hashRem_[i].mask];
      //      thisMatch &= (~data_.hashRem_[i].mask.ui);
      thisMatch &= (~thisMask);
      //      printWord(thisMatch,10); cout << endl;
     
      //      errorLow=lowerFragScore_[thisMatch&lowerFragMask_];
      if ((errorLow=lowerFragScore_[thisMatch&lowerFragMask_])
          <moreThanTwoErrors__)
      {
        thisMatch>>=(numBitsPerBase*lowerFragSize_); // eliminate temporary
        if ((errorHigh=upperFragScore_[thisMatch])
            <moreThanTwoErrors__)
        {

          //   numErrors=(errorLow>>(2*errorBits))+(errorHigh>>(2*errorBits));
#ifdef DEBUG
          numErrors=errorLow;
          numErrors&=(~(errorMask1|errorMask2));
          numErrors+=errorHigh;
          numErrors>>=(2*errorBits);
          cout << (data_.hashRem_[i].position&(~isReverseOligo)) << " "
               << ((data_.hashRem_[i].position&isReverseOligo)?'R':'F') << " "
               << numErrors << " " << sequencePos-blockSize << " ";
          printWord(prefix,10);  cout << " ";
          printWord(suff, 10);
#endif
          //            cout << "ck " << sequencePos-16777216 << " " << (errorLow&errorPosMask1__) << " "
          //   << ((errorLow&errorPosMask2__)>>8) << " "
          //   << (errorHigh&errorPosMask1__) << " "
          //  << ((errorHigh&errorPosMask2__)>>8) << endl;


          if (wantMatch(errorLow, errorHigh, thisMask))
          {
            //              cout << "wtd " << sequencePos-16777216 << " " << (errorLow&errorPosMask1__) << " "
            //    << ((errorLow&errorPosMask2__)>>8) << " "
            //  << (errorHigh&errorPosMask1__) << " "
            //  << ((errorHigh&errorPosMask2__)>>8) << endl;


            results_.addMatch( data_.hashRem_[i], thisMask, sequencePos, 
                               errorLow, errorHigh ); 
          } // ~if

        } // ~if
      } // ~if
    } // ~for   
    
  }  // ~check


  inline void check__
  ( const vector<pair<Oligo, MatchPosition> >& h )
  {

    register Word thisMatch(0), thisMask;

#ifdef DEBUG
    cout << sequencePos-blockSize << " - position " << endl;
    printWord( prefix, maxBasesPerWord ); cout << " - prefix" << endl;
    printWord( suff, maxBasesPerWord); cout << " - suffix" << endl;
#endif

    vector<pair<uint, uint> >p(h.size());
    for (uint j(0);j<h.size();++j)
    {
      p[j].first=pCount_[h[j].first.ui[0]];
      p[j].second=pCount_[h[j].first.ui[0]+1];
    } // ~for j

    register FragmentErrorType errorLow, errorHigh;

    for (uint j(0);j<p.size();++j)
    {
      for (uint i(p[j].first);
           i!=p[j].second;
           ++i)
      {
        //    for (uint i(pCount_[prefix]);
        //   i!=pCount_[prefix+1];
        //   ++i)

        //      cout << "XX " << b-a << endl;
#ifdef DEBUG
        printWord(data_.hashRem_[i].suffix.ui, maxBasesPerWord );
        cout << " - ";
        printWord(maskTable_[data_.hashRem_[i].mask],maxBasesPerWord); 
        cout << endl;
#endif

        thisMatch = (h[j].first.ui[1]^data_.hashRem_[i].suffix.ui);

        thisMask = maskTable_[data_.hashRem_[i].mask];
        //      thisMatch &= (~data_.hashRem_[i].mask.ui);
        thisMatch &= (~thisMask);
        //      printWord(thisMatch,10); cout << endl;
     
        //      errorLow=lowerFragScore_[thisMatch&lowerFragMask_];
        if ((errorLow=lowerFragScore_[thisMatch&lowerFragMask_])
            <moreThanTwoErrors__)
        {
          thisMatch>>=(numBitsPerBase*lowerFragSize_); // eliminate temporary
          if ((errorHigh=upperFragScore_[thisMatch])
              <moreThanTwoErrors__)
          {

            // numErrors=(errorLow>>(2*errorBits))+(errorHigh>>(2*errorBits));
#ifdef DEBUG
            numErrors=errorLow;
            numErrors&=(~(errorMask1|errorMask2));
            numErrors+=errorHigh;
            numErrors>>=(2*errorBits);
            cout << (data_.hashRem_[i].position&(~isReverseOligo)) << " "
                 <<((data_.hashRem_[i].position&isReverseOligo)?'R':'F') << " "
                 << numErrors << " " << sequencePos-blockSize << " ";
            printWord(prefix,10);  cout << " ";
            printWord(suff, 10);
#endif
            //            cout << "ck " << sequencePos-16777216 << " " << (errorLow&errorPosMask1__) << " "
            //   << ((errorLow&errorPosMask2__)>>8) << " "
            //   << (errorHigh&errorPosMask1__) << " "
            //  << ((errorHigh&errorPosMask2__)>>8) << endl;


            if (wantMatch(errorLow, errorHigh, thisMask))
            {
              //              cout << "wtd " << sequencePos-16777216 << " " << (errorLow&errorPosMask1__) << " "
              //    << ((errorLow&errorPosMask2__)>>8) << " "
              //  << (errorHigh&errorPosMask1__) << " "
              //  << ((errorHigh&errorPosMask2__)>>8) << endl;


              results_.addMatch( data_.hashRem_[i], thisMask, h[j].second, 
			       errorLow, errorHigh ); 
            } // ~if

          } // ~if
        } // ~if
      } // ~for	i
    } // ~for j
    
  }  // ~check__

}; // ~template<class Child> struct PHTHelperBase<Child, false>


// Specialization for split prefix mode
template<class Child> struct PHTHelperBase<Child, true> : 
public PHTHelperData
{
  PHTHelperBase( HashTableDataStore& data, MatchTable& results ) : 
    PHTHelperData( data, results )
  {} 
  const Child& asChild( void ) const 
  { return static_cast<const Child&>(*this); }
  bool wantMatch( const FragmentErrorType errorLow,
		  const FragmentErrorType errorHigh,
		  const Word mask ) 
  { return (asChild()).wantMatch( errorLow, errorHigh, mask); }
  Word getInfoN( const Word suff, const Word mask ) const 
  { return (asChild()).getInfoN(suff, mask); }
  inline void countKey( const Word key, 
			const Word prefixMask,
			const Word suffixMask )
  {
    if (prefixMask==0)
    {
      ++maskMap_[suffixMask];
      ++pCount_[key>>splitPrefixShift_];
    } // ~if
    //    pCount_[key>>splitPrefixShift_]+=(mask==0);
  } // ~void countKey( const Word key, const Word mask )

  inline void hashEntry
  ( Word key, const Word keyMask,
    const Word entry, const Word entryMask,
    OligoNumber oligoNum )
  {

    if (keyMask==0)
    {
      oligoNum|=((key&splitPrefixMask_)<<MAX_HASH_BITS);
      key >>= splitPrefixShift_;
      data_.hashRem_[pCount_[key]].suffix.ui=entry;
      //      data_.hashRem_[pCount_[key]].mask.ui=entryMask;
      data_.hashRem_[pCount_[key]].mask=maskMap_[entryMask];
      data_.hashRem_[pCount_[key]].position=oligoNum;
      pCount_[key]++;
    } // ~if
  } // ~void hashEntry

  inline void check__
  ( const vector<pair<Oligo, MatchPosition> >& h )
  {}


  inline void check
  ( Word prefix, const Word suff, 
    const MatchPosition sequencePos )
  {
    //    printWord( prefix, 12 ); cout << "****" << endl;
    register Word thisMatch(0), thisMask;
    FragmentErrorType errorLow, errorHigh;
    OligoNumber thisSplitPrefix((prefix&splitPrefixMask_)<<MAX_HASH_BITS);
    prefix>>=splitPrefixShift_;
    uint i(pCount_[prefix]);
    
    for (;i!=pCount_[prefix+1];++i)
    {
      if ((data_.hashRem_[i].position&splitPrefixMaskHigh)
	    ==thisSplitPrefix) break;
    } // ~for

    for (;i!=pCount_[prefix+1];++i)
    {
      if ((data_.hashRem_[i].position&splitPrefixMaskHigh)!=thisSplitPrefix) 
	break;
      
      thisMatch = ((suff^data_.hashRem_[i].suffix.ui));
      //      thisMatch &= (~data_.hashRem_[i].mask.ui);
      thisMask = maskTable_[data_.hashRem_[i].mask];
      thisMatch &= (~thisMask);


      if ((errorLow=lowerFragScore_[thisMatch&lowerFragMask_])
	  <moreThanTwoErrors__)
      {
	thisMatch>>=(numBitsPerBase*lowerFragSize_); // eliminate temporary
	if ((errorHigh=upperFragScore_[thisMatch])
	    <moreThanTwoErrors__)
	{
	  //	  numErrors=errorLow;
	  //  numErrors&=(~(errorMask1|errorMask2));
	  //	  numErrors+=errorHigh;
	  //  numErrors>>=(2*errorBits);

	  if (wantMatch(errorLow, errorHigh, thisMask))
	    results_.addMatch( data_.hashRem_[i], thisMask, sequencePos, 
			       errorLow, errorHigh ); 
	} // ~if errorHigh
      } // ~if errorLow
    } // ~for
  } // ~check

}; // ~template<class Child> struct PHTHelperBase<Child, true>

template<int PASS, bool isFwd> struct PHTHelper : 
  public PHTHelperBase<PHTHelper<PASS,isFwd>,useSplitPrefix>
{
  PHTHelper( HashTableDataStore& data, MatchTable& results ) : 
    PHTHelperBase<PHTHelper<PASS,isFwd>,useSplitPrefix>( data, results )
  {} 
  bool wantMatch( const FragmentErrorType errorLow, 
		  const FragmentErrorType errorHigh,
		  const Word mask ) const;
  Word getInfoN( const Word suff, const Word mask ) const;
}; // ~template<int PASS, bool isFwd> struct PHTHelper


// Functionality specific to the 6 partitions
template<> bool PHTHelper<0, true>::wantMatch
( const FragmentErrorType errorLow, 
  const FragmentErrorType errorHigh, 
  const Word mask ) const
{ 
  // pass 0 in forward direction: want any match with 2 differences 
  // or less - only way this doesn't occur is if there are two 
  // errors in one segment and at least one in the other

  //  return (!( ((errorLow>=twoErrors)&&(errorHigh>=oneError))
  //     ||((errorHigh>=twoErrors)&&(errorLow>=oneError))));

  return (!( ((errorLow>oneError__)&&(errorHigh>noErrors__))
	     ||((errorHigh>oneError__)&&(errorLow>noErrors__))));



  //  return (numErrors<=2); 
} // ~template<> bool PHTHelper<0, true>::wantMatch
template<> Word PHTHelper<0, true>::getInfoN
( const Word suff, const Word mask ) const
{ return noNonIncorps; }

template<> bool PHTHelper<0, false>::wantMatch
( const FragmentErrorType errorLow,
  const FragmentErrorType errorHigh,
  const Word mask ) const
{ 
  // pass 0 in reverse direction: want any match with 1 or 2 errors,
  // only want zero error matches if mask is non zero

  //  return ( ((errorLow!=0)||(errorHigh!=0)||(mask!=0))
  //   &&( !( ((errorLow>=twoErrors)&&(errorHigh>=oneError))
  //	  ||((errorHigh>=twoErrors)&&(errorLow>=oneError)))));

  return ( ((errorLow!=0)||(errorHigh!=0)||(mask!=0))
	   &&( !( ((errorLow>oneError__)&&(errorHigh>noErrors__))
		  ||((errorHigh>oneError__)&&(errorLow>noErrors__)))));

  //  return ( ((numErrors==1)||(numErrors==2))
  //	   ||((numErrors==0)&(mask!=0)) ); 
}
template<> Word PHTHelper<0, false>::getInfoN
( const Word suff, const Word mask ) const
{ return ( ((mask&0x3)!=0x3) ? noNonIncorps : (suff&0x3) ); }

template<> bool PHTHelper<1, true>::wantMatch
( const FragmentErrorType errorLow,
  const FragmentErrorType errorHigh,
  const Word mask ) const
{ 
  // pass 1 in forward direction: want any match with 2 errors or less, 
  // provided it has errors or Ns in 2 distinct segments (if all errors/Ns
  // are in a single segment it will be matched on pass 0.

  //  return (     (  ((errorLow!=0)||((mask&lowerFragMask_)!=0))
  //		  && ((errorHigh!=0)||((mask&(~lowerFragMask_))!=0)))
  //	   && ( !( ((errorLow>=twoErrors)&&(errorHigh>=oneError))
  //		   ||((errorHigh>=twoErrors)&&(errorLow>=oneError)))));

  return (     (  ((errorLow!=0)||((mask&lowerFragMask_)!=0))
		  && ((errorHigh!=0)||((mask&(~lowerFragMask_))!=0)))
	   && ( !( ((errorLow>oneError__)&&(errorHigh>noErrors__))
		   ||((errorHigh>oneError__)&&(errorLow>noErrors__)))));

} // ~template<> bool PHTHelper<1, true>::wantMatch


template<> Word PHTHelper<1, true>::getInfoN
( const Word suff, const Word mask ) const
{ return noNonIncorps; }

template<> bool PHTHelper<1, false>::wantMatch
( const FragmentErrorType errorLow,
  const FragmentErrorType errorHigh,
  const Word mask ) const
{ 
  // pass 1 in reverse direction: want any match with 2 errors or less, 
  // provided it has errors or Ns in 2 distinct segments (if all errors/Ns
  // are in a single segment it will be matched on pass 0.

  //  return (     (  ((errorLow!=0)||((mask&lowerFragMask_)!=0))
  //		  && ((errorHigh!=0)||((mask&(~lowerFragMask_))!=0)))
  //   && ( !( ((errorLow>=twoErrors)&&(errorHigh>=oneError))
  //	   ||((errorHigh>=twoErrors)&&(errorLow>=oneError)))));

  return (     (  ((errorLow!=0)||((mask&lowerFragMask_)!=0))
		  && ((errorHigh!=0)||((mask&(~lowerFragMask_))!=0)))
	   && ( !( ((errorLow>oneError__)&&(errorHigh>noErrors__))
		   ||((errorHigh>oneError__)&&(errorLow>noErrors__)))));

} // ~template<> bool PHTHelper<1, false>::wantMatch


template<> Word PHTHelper<1, false>::getInfoN
( const Word suff, const Word mask ) const
{ return ( ((mask&0x3)!=0x3) ? noNonIncorps : (suff&0x3) ); }

template<> bool PHTHelper<2, true>::wantMatch
( const FragmentErrorType errorLow,
  const FragmentErrorType errorHigh,
  const Word mask ) const
{ 
  // pass 2 in forward direction: want any match with 2 errors or less, 
  // provided it has errors or Ns in 2 distinct segments (if all errors/Ns
  // are in a single segment it will be matched on pass 0.

  //  return (     (  ((errorLow!=0)||((mask&lowerFragMask_)!=0))
  //	  && ((errorHigh!=0)||((mask&(~lowerFragMask_))!=0)))
  //   && ( !( ((errorLow>=twoErrors)&&(errorHigh>=oneError))
  //	   ||((errorHigh>=twoErrors)&&(errorLow>=oneError)))));

  return (     (  ((errorLow!=0)||((mask&lowerFragMask_)!=0))
		  && ((errorHigh!=0)||((mask&(~lowerFragMask_))!=0)))
	   && ( !( ((errorLow>oneError__)&&(errorHigh>noErrors__))
		   ||((errorHigh>oneError__)&&(errorLow>noErrors__)))));

} // ~template<> bool PHTHelper<2, true>::wantMatch

template<> Word PHTHelper<2, true>::getInfoN
( const Word suff, const Word mask ) const
{ return noNonIncorps; }

template<> bool PHTHelper<2, false>::wantMatch
( const FragmentErrorType errorLow,
  const FragmentErrorType errorHigh,
  const Word mask ) const
{ 
  // pass 2 in reverse direction: want any match with 2 errors or less, 
  // provided it has errors or Ns in 2 distinct segments (if all errors/Ns
  // are in a single segment it will be matched on pass 0.

  //  return (     (  ((errorLow!=0)||((mask&lowerFragMask_)!=0))
  //	  && ((errorHigh!=0)||((mask&(~lowerFragMask_))!=0)))
  //   && ( !( ((errorLow>=twoErrors)&&(errorHigh>=oneError))
  //	   ||((errorHigh>=twoErrors)&&(errorLow>=oneError)))));

  return (     (  ((errorLow!=0)||((mask&lowerFragMask_)!=0))
		  && ((errorHigh!=0)||((mask&(~lowerFragMask_))!=0)))
	   && ( !( ((errorLow>oneError__)&&(errorHigh>noErrors__))
		   ||((errorHigh>oneError__)&&(errorLow>noErrors__)))));

} // ~template<> bool PHTHelper<2, false>::wantMatch

template<> Word PHTHelper<2, false>::getInfoN
( const Word suff, const Word mask ) const
{ return ( ((mask&0x3)!=0x3) ? noNonIncorps : (suff&0x3) ); }



template< int i > class OligoHashTable;
// class PartitionHashTable
// This stores all the oligo information for one of the six partitions
// The partitions are searched two at a time, thus three runs through the
// genome are required in total
template< bool useSplitPrefix, int PASS, bool isFwd > class PartitionHashTable
{
  friend class OligoHashTable<0>;
  friend class OligoHashTable<1>;
  friend class OligoHashTable<2>;
public:
  PartitionHashTable( HashTableDataStore& data, MatchTable& results ) :
    data_(data), sps_(data, results), results_(results)
  {} // ~ctor

  void setTable( int prefixLength, 
		 int lowerFragSize, 
		 Word lowerFragMask,
		 const FragmentErrorType* lowerFragScore,
		 const FragmentErrorType* upperFragScore );

  void makePointerArray( void );
  //  void removeRepeatedEntries( vector<MatchPosition>& matchPosition_ );__
  void removeRepeatedEntries( MatchTable& results );

  // private:
  // data storage for hash table
  int numBits_;

  HashTableDataStore& data_;
  // Moved to HashTableDataStore
  //  vector<TablePointer> entryPointer_;
  //  vector<TableEntry<useSplitPrefix> > hashRem_; 
  //  vector<OligoNumber> position_; 

  PHTHelper<PASS, isFwd> sps_;
  //  TablePointer* pCount_;
  //  int lowerFragSize_;
  //  Word lowerFragMask_;
  //  const FragmentErrorType* lowerFragScore_;
  //  const FragmentErrorType* upperFragScore_;

  MatchTable& results_;

}; // ~class PartitionHashTable

template<bool useSplitPrefix, int PASS, bool isFwd> 
void PartitionHashTable<useSplitPrefix,PASS,isFwd>::setTable
( int prefixLength, 
  int lowerFragSize, 
  Word lowerFragMask,
  const FragmentErrorType* lowerFragScore,
  const FragmentErrorType* upperFragScore )
{
  bool needToClear( data_.entryPointer_.size() != 0 );
  if ((2*prefixLength)>MAX_HASH_BITS)
  {
    numBits_= MAX_HASH_BITS;
    sps_.splitPrefixShift_= (2*prefixLength)-MAX_HASH_BITS;
    cerr << "partition hash table will run in split prefix mode, split is " 
	 << numBits_ << "/" << sps_.splitPrefixShift_ << endl;
    sps_.splitPrefixMask_=hashmask(sps_.splitPrefixShift_);
  } // ~if
  else
  {
    numBits_=2*prefixLength; 
    // shift and mask still need to be explicitly set to zero because this
    // branch can still be executed when useSplitPrefix is true: if
    // oligo length is 26 and max bits is 25, then on pass 2 there is one
    // 25 bit table and 1 24 bit table TC 27.02.04
    sps_.splitPrefixShift_=0;
    sps_.splitPrefixMask_=0;
  } // ~else
  data_.entryPointer_.resize( 2 + (1<<numBits_) );
  if (needToClear)
    memset(&data_.entryPointer_[0], '\0', 
	   sizeof(TablePointer)*data_.entryPointer_.size());

  sps_.lowerFragSize_ = lowerFragSize;
  sps_.lowerFragMask_ = lowerFragMask;
  sps_.lowerFragScore_ = lowerFragScore;
  sps_.upperFragScore_ = upperFragScore;
  sps_.pCount_ = &data_.entryPointer_[2];

  cerr << "Setting partition hash table size to " 
       << numBits_ << " bits" << endl;
} // ~void setTable( int numBits )


template< bool useSplitPrefix, int PASS, bool isFwd > 
void PartitionHashTable<useSplitPrefix,PASS,isFwd>::makePointerArray( void )
{
  TablePointer last(0);

  bool needToClear(data_.hashRem_.size()!=0);

  for ( vector<TablePointer>::iterator i(data_.entryPointer_.begin()+2);
	i!=data_.entryPointer_.end(); i++)
    { *i +=last; last=*i; }
  cerr << "will place " << data_.entryPointer_.back() << " entries in table"<< endl;
  data_.hashRem_.resize( data_.entryPointer_.back());
  //  position_.resize( hashRem_.size() );

  if (needToClear)
  {
    memset(&data_.hashRem_[0], '\0', sizeof(TableEntry<useSplitPrefix>)*data_.hashRem_.size());
  } // ~if
  
  sps_.pCount_--; // now pCount_ == data_.entryPointer_ + 1

  //cout<<"Number of different masks found: " << sps_.maskMap_.size() << endl;
  assert(sps_.maskMap_.size()<65535);
  sps_.maskTable_.reserve(sps_.maskMap_.size());
  uint j(0);
  for( map<Word, uint>::iterator i(sps_.maskMap_.begin());
       i!=sps_.maskMap_.end();
       ++i, ++j)
  {
    //    printWord(i->first, 16);
    //    cout << '\t' << i->second << endl;
    i->second=j;
    sps_.maskTable_.push_back(i->first);
  }


} // ~void PartitionHashTable::makePointerArray( void )

template< bool useSplitPrefix, int PASS, bool isFwd > 
void PartitionHashTable<useSplitPrefix,PASS,isFwd>::removeRepeatedEntries
                                                         //( vector<MatchPosition>& matchPosition_ )
( MatchTable& results )
{
  int numSaved(0);

  sps_.pCount_--; // now sps_.pCount_ == data_.entryPointer_

  TableEntry<useSplitPrefix> lastEntry;
  //  Word lastEntry;

  uint k,l(0);
  OligoNumber existingOligo, newOligo;

  // remove repeats from table

  // Sort entries for each hash value into order then eliminate any repeats.
  // This bit is based on code in IndexGenome.cpp

  for ( int i(0) ; i < (1<<numBits_) ; i++ )
  {

    if (sps_.pCount_[i+1]-sps_.pCount_[i]>1)
    {
      //      cout << "### " <<  (pCount_[i+1]-pCount_[i]) << endl;
      sort( data_.hashRem_.begin()+sps_.pCount_[i], data_.hashRem_.begin() + sps_.pCount_[i+1] );
      //   mirrorSort( hashRem_.begin(), 
      //	  hashRem_.begin()+pCount_[i], 
      //	  hashRem_.begin() + pCount_[i+1],
      //	  swapper );
      
    } // ~if

    k = sps_.pCount_[i];
    sps_.pCount_[i] = l;
    //    cout << i << ": " << l << " from " << k << endl;
    //    lastEntry = hashRem_[k].suffix.ui;
    //   lastEntry ^= ~0; // ensures false on first iteration
    lastEntry = data_.hashRem_[k];
    lastEntry.suffix.ui ^= ~0; // ensures false on first iteration

    for ( ; k!= sps_.pCount_[i+1] ; k++ )
    {
      if (data_.hashRem_[k].mask==lastEntry.mask) numSaved++;
      if (data_.hashRem_[k] != lastEntry)
      { // no repeat, copy as normal 
	//	entry_[l]=entry_[k];
	data_.hashRem_[l]=data_.hashRem_[k];
	//	position_[l]=position_[k];
	//	cout << i << ": "<< hashRem_[l].suffix.ui << " to "<< l<< " from "<< k<< endl; 
	++l;
      } // ~if
      else
      { // repeat 
	--l; // shift back to first repeated entry
	//	type_[l]=repeatInBatch;
	//	entry_[l]=entry_[k];
	//	entry_[l].position_ = repeatFlag + repeatInGenome + 1;
	do
	{
	  // flag current entry as being same as that entry
	  //	  oligos_.flagRepeat( oligoNum_[l]&(~reverseComplementFlag), 
	  //	      oligoNum_[k]&(~reverseComplementFlag)); 
	  //cout<<((hashRem_[k].position&splitPrefixMaskLow)&(~isReverseOligo))
	    //	       << ((hashRem_[k].position&isReverseOligo)?'R':'F')
	  //   << " is the same as "
	  // << ((hashRem_[l].position&splitPrefixMaskLow)&(~isReverseOligo))
	  //   << ((hashRem_[l].position&isReverseOligo)?'R':'F') 
	  //   << endl;
	  existingOligo 
	    = (data_.hashRem_[l].position&splitPrefixMaskLow)&(~isReverseOligo);
	  newOligo 
	    = (data_.hashRem_[k].position&splitPrefixMaskLow)&(~isReverseOligo);
          results.setSameAs__( newOligo, existingOligo );
#ifdef ZZZZ
	  if (  (matchPosition_[newOligo]<blockRepeat)
		&& (matchPosition_[existingOligo]!=blockRepeat+newOligo) )
	  {
	    matchPosition_[newOligo] = blockRepeat+existingOligo;
	  } // ~if
#endif
	    //	  matchPosition_[position_[k]&(~isReverseOligo)]
	    //   =blockRepeat+(position_[l]&(~isReverseOligo));

	  k++;
	} // ~do
	while ( (data_.hashRem_[k]==lastEntry)&&(k!=sps_.pCount_[i+1]) );
	k--; // need to look at next entry again
	l++; // and shift destination pointer
      } // ~else
      lastEntry = data_.hashRem_[k];
    } // ~if
  } // ~for i
  cerr << sps_.pCount_[1<<numBits_] << " entries reduced to " << l << " entries"
       << endl;

  //  numOligos_ = pCount_[1<<numBits_];
  sps_.pCount_[1<<numBits_] = l;

  data_.hashRem_.resize(l);
  //  position_.resize(l);

} // ~void PartitionHashTable::removeRepeatedEntries( void )


// OligoHashTable: manage build and scan of the two PartitionHashTables on
// each of the three passes
// pass 0:
// - search for partitions 0 = ABCD and 1 = CDAB
// pass 1: 
// - search for partitions 2 = ACBD and 3 = BDAC
// pass 2: 
// - search for partitions 4 = ADCB and 5 = CBAD
template< int PASS > class OligoHashTable
{
public: 
  OligoHashTable( int oligoLength, 
		  HashTableDataStore& htds1,
		  HashTableDataStore& htds2, 
		  const SuffixScoreTable& score, MatchTable& results );

  // buildTable: read oligos from source and generate hash tables
  // return false if no oligos were hashed
  bool buildTable( OligoSource& oligos );

  // hashThisOligo: returns true if an oligo should be placed in the hash 
  // table on this pass
  bool hashThisOligo( OligoNumber oligoNum, const vector<Oligo>& queryMask );

  // scan: scan hash table across a chromosome file stored in the 
  // 2-bits-per-base format generated by squashGenome
  MatchPosition scan( FileReader& file, const MatchPosition currentBlock );

  // checkOligo: look in the two hash tables for matches against a position
  // in the genome
  inline void checkOligo
  ( const Oligo& ol, const MatchPosition sequencePos );

  void checkOligo
    ( const vector<Oligo>& v, const MatchPosition starts );



  // checkManyOligos: fast scan across multiple consecutive positions in the
  // genome
  MatchPosition checkManyOligos
  ( Oligo& ol, const Word* pWord, 
    const Word* pLastWord, MatchPosition seqNum );

  int getNumOligos( void ) const { return numOligos_; }

  // shiftOligo: shift a base off of thisWord and onto ol
  void shiftOligo( Oligo& ol, Word& thisWord ) const
  {
    register Word thisCarry;

    ol.ui[1]<<=numBitsPerBase;
    thisCarry=(ol.ui[0]>>((suffixLength-1)*numBitsPerBase));
    ol.ui[1]|=thisCarry;
    ol.ui[0]<<=numBitsPerBase;
    thisCarry=(thisWord>>((maxBasesPerWord-1)*numBitsPerBase));
    ol.ui[0]|=thisCarry;
    thisWord<<=numBitsPerBase;
    ol.ui[1]&=(~prefixMask);
    ol.ui[0]&=(~suffixMask);
  } // ~shiftOligo

private:

  // numOligos_: on pass 0, buildTable sets this to the number of oligos 
  // found in the source
  int numOligos_;

  const int oligoLength_;
  const int prefixLength_;

  const Word prefixMask_;

  // data storage for match scoring table

  // Score of match between suffixes A and B is suffixScore_[ A XOR B ]
  const SuffixScoreTable& score_;

  // data storage for match information for each oligo

  MatchTable& results_;
  //  vector<MatchPosition> matchPosition_;
  //  vector<MatchDescriptor> matchType_;

  // data storage for oligos
  PartitionHashTable<useSplitPrefix, PASS, true> table1_;
  PartitionHashTable<useSplitPrefix, PASS, false> table2_;

  // hasher - templatized
  Hasher<PASS> hash_;

}; // ~class OligoHashTable

template< int PASS > OligoHashTable<PASS>::OligoHashTable
( int oligoLength, 
  //  PartitionHashTable<useSplitPrefix>& table1, 
  //  PartitionHashTable<useSplitPrefix>& table2, 
  HashTableDataStore& htds1,
  HashTableDataStore& htds2,
  const SuffixScoreTable& score, MatchTable& results ) :
  numOligos_(-1),
  oligoLength_(oligoLength),
  prefixLength_(oligoLength-16),
  // TC 04.06.03 - changed to work for oligoLength of 32
  //  prefixMask_((1<<(2*prefixLength_))-1),
  // NB new version probably doesn't work for oligoLength of 16!
  prefixMask_( ((Word)~0) 
	       >> (numBitsPerBase*(maxBasesPerWord-prefixLength_)) ),
  //  partitionNum_(0),
  score_(score),
  results_(results),
  table1_(htds1, results),
  table2_(htds2, results),
  hash_()
{
  //  assert(oligoLength_>=16);
  //  assert(oligoLength_<=24);

  cerr << "made hash table, pointer table sizes = "
       << table1_.data_.entryPointer_.size() << ", "
       << table2_.data_.entryPointer_.size() << endl;

} // ~OligoHashTable::OligoHashTable



template< int PASS > bool OligoHashTable<PASS>::hashThisOligo
( OligoNumber oligoNum, const vector<Oligo>& queryMask )
{

  // if mask is empty then nothing to hash (because of QC failure)
  // so obviously return false
  if (queryMask.size()==0) 
  {
    results_.setQualityFailed__(oligoNum);
    return false;
  } // ~if
  else 
  { // ask MatchTable if it is interested in the results
    return results_.isInterested__( oligoNum, 
                                    PASS,    
                                    ((queryMask[0].ui[0]!=0)
                                     ||(queryMask[0].ui[1]!=0)));
  } // ~else
} // ~OligoHashTable<PASS>::hashThisOligo

#ifdef ZZZZZ
// Template specialization on pass zero - always hash unless oligo has
// failed quality control or been repeat masked.
template<> bool OligoHashTable<0>::hashThisOligo
( OligoNumber oligoNum, const vector<Oligo>& queryMask )
{
  // if repeat masking not used, results_ not set up first time called on 
  // first pass: hash unless queryMask is empty 
  if (results_.size()==0)
  {
    return ((queryMask.size()==0)?false:true);
  } // ~if
  else
  {
    //    if (results_.matchPosition_[oligoNum]==repeatMasked)
    if (results_.isRepeatMasked__(oligoNum))
      return false;
    // if oligo has failed quality control, QueryGenerator does not generate
    // any queries so queryMask is empty
    if (queryMask.size()==0) 
    {
      results_.setQualityFailed__(oligoNum);
      //      results_.matchPosition_[oligoNum]=qualityFailed;
      return false;
    } // ~if
    return true;
  }
} // ~hashThisOligo( OligoNumber oligoNum, const vector<Oligo>& queryMask )

// General template definition called on pass 1 and 2
template< int PASS > bool OligoHashTable<PASS>::hashThisOligo
( OligoNumber oligoNum, const vector<Oligo>& queryMask )
{
  //  if ((results_.matchPosition_[oligoNum]==repeatMasked)
  //     ||(results_.matchPosition_[oligoNum]==qualityFailed))
  if (results_.isRepeatMasked__(oligoNum)
      ||results_.hasQualityFailed__(oligoNum))
    return false;

  // if oligo has failed quality control, QueryGenerator does not generate
  // any queries so queryMask is empty
  if (queryMask.size()==0) return false;
  // if we are on the first pass all oligos must be hashed
  //  if (partitionNum_==0) return true;
  
  if ((queryMask[0].ui[0]==0)&&(queryMask[0].ui[1]==0))
  {
    // if the oligo contains no Ns or ambiguity codes, all exact or
    // single error matches are caught on the first pass, so only
    // hash if two error matches are still of interest
    //    if ( (results_.matchPosition_[oligoNum]==noMatch)
    //	 ||( ((results_.matchType_[oligoNum].errorType&0x3)==2)
    //     &&(results_.matchType_[oligoNum].r[2]<=maxNumBestMatches))) 
    //   return true;
    //   else
    //   return false;

    // Changed TC 22.08.03
    // If no match yet found, clearly still interested in matches!
    // If best match thus far is unique, still interested in further matches
    if (results_.matchPosition_[oligoNum]==noMatch)
      return true;
    else
    {
      uint numErrors(results_.matchType_[oligoNum].errorType&0x3);
      if (results_.matchType_[oligoNum].r[numErrors]>1)
	return false;
      else
	return true;
    } // ~else	

  } // ~if
  else
  { 
    // if the oligo contains Ns or ambiguity codes, then exact or single 
    // error matches may still be found for this oligo, but we are only 
    // interested if it is not a repeat
    if (  ((results_.matchType_[oligoNum].errorType&0x3)==0)
	  &&(results_.matchType_[oligoNum].r[0]>maxNumBestMatches))
      return false;
    else 
      return true;
  } // ~else
  return false;
} // ~hashThisOligo( OligoNumber oligoNum, const vector<Oligo>& queryMask )
#endif



template< int PASS > bool OligoHashTable<PASS>::buildTable
( OligoSource& oligos )
{
  table1_.setTable( hash_.getLengthPart1(), 
		    hash_.getLowerFragSizePart2(), 
		    hash_.getLowerFragMaskPart2(),
		    hash_.getLowerFragScorePart2(score_),
		    hash_.getUpperFragScorePart2(score_) );

  table2_.setTable( hash_.getLengthPart2(), 
		    hash_.getLowerFragSizePart1(), 
		    hash_.getLowerFragMaskPart1(),
		    hash_.getLowerFragScorePart1(score_),
		    hash_.getUpperFragScorePart1(score_) );

  //  numOligos_=0;

  const char* pOligo;

  Oligo res, mask;

  //  table1_.pCount_ = &table1_.entryPointer_[2]; - now done in setTable
  //  table2_.pCount_ = &table2_.entryPointer_[2]; - now done in setTable

  OligoNumber oligosFound(0), oligosToHash(0);

  oligos.rewind();

  QueryGenerator makeOligosFromASCII;

  vector<Oligo> queryOligo, queryMask;
  vector<OligoNumber> queryOligoNum;

  cerr << "counting oligos" << endl;

  // count occurrences
  while (1)
  {

    if ( (pOligo=oligos.getNextOligo()) == NULL ) break;
    //    ett.translate( pOligo, ol, rc );
    ++oligosFound;

    makeOligosFromASCII
      ( pOligo, 0, queryOligo, queryMask, queryOligoNum );

    //  printWord(ol.ui[1],16);cout <<'-';printWord(ol.ui[0],16); cout << endl;
    
    // if more than 3 exact matches already found, don't search for any more
    //    if ( ( partitionNum_==0)
    // || ( (results_.matchPosition_[oligosFound]<blockRepeat)
    //      && ( ((matchType_[oligosFound].errorType&0x3)!=0)
    //	   || (matchType_[oligosFound].r[0]<=3) ) ) )

    if (hashThisOligo( oligosFound, queryMask )==true)
    {

      ++oligosToHash;
      for ( uint i(0) ; i<queryOligo.size(); i++)
      {
	hash_(queryMask[i],mask);
	hash_(queryOligo[i],res);

	table1_.sps_.countKey( res.ui[0], mask.ui[0], mask.ui[1] );
	table2_.sps_.countKey( res.ui[1], mask.ui[1], mask.ui[0] );

      } // ~for
    } // ~if

  } // ~while   

  if (numOligos_==-1)
  {
    numOligos_=oligosFound;
    assert(numOligos_<maxNumOligos);
  } // ~if
  else
  {
    if (((int)oligosFound)!=numOligos_)
    {
      cerr << "Error: expecting to read " << numOligos_ 
	   << " oligos, but found " << oligosFound << endl;
      exit (1);
    } // ~if
  } // ~else

  cerr << "read " << numOligos_ << " oligos from source" << endl;
  cerr << "Will hash " << oligosToHash << " oligos" << endl;

  if (oligosToHash==0)
  {
    cerr << "No oligos to hash, returning" << endl;

    // if results table has not been built, do so - this ensure alignment 
    // output always produced, even if all sequences are QC failures 
    if (results_.size()==0)
    {
      results_.resize(numOligos_+1);
    } // ~if
    return false;
  } // ~if

  if (PASS==0)
  {
    //    if (results_.matchPosition_.size()==0)
    if (results_.size()==0)
    {
      results_.resize(numOligos_+1);
    } // ~if
    else
    {
      assert(((int)results_.size())==numOligos_+1);
    } // ~else
  } // ~if

  cerr << "converting count values to pointer arrays" << endl;

  table1_.makePointerArray();
  table2_.makePointerArray();

  // rewind and fill in tables

  oligos.rewind();

  //  table1_.pCount_--; // now pCount_ == entryPointer_ + 1
  //  table2_.pCount_--; // now pCount2_ == entryPointer2_ + 1

  for ( int j(1); j <= numOligos_ ; ++j )
  {
    if ( (pOligo=oligos.getNextOligo()) == NULL )
    {
      cerr << "Error: expecting " << numOligos_ << " oligos, only found "
	   << j << endl;
      exit (1);
    } // ~if


    makeOligosFromASCII
      ( pOligo, j, queryOligo, queryMask, queryOligoNum );


    if (hashThisOligo( j, queryMask) ==true )
    {
    
      for ( uint i(0) ; i<queryOligo.size(); i++)
      {
	assert(queryOligo.size()!=0);

	hash_(queryMask[i],mask);
	hash_(queryOligo[i],res);

	table1_.sps_.hashEntry
	  ( res.ui[0], mask.ui[0], res.ui[1], 
	    mask.ui[1], queryOligoNum[i] );
	table2_.sps_.hashEntry
	  ( res.ui[1], mask.ui[1], res.ui[0], 
	    mask.ui[0], queryOligoNum[i] );

      } // ~for i
    } // ~if
  } // ~for j  



  //  table1_.pCount_--; // now pCount_ == entryPointer_
  //  table2_.pCount_--; // now pCount2_ == entryPointer2_

  //  exit(0);

  //  table1_.removeRepeatedEntries( results_.matchPosition_ );
  //  table2_.removeRepeatedEntries( results_.matchPosition_ );
  table1_.removeRepeatedEntries( results_ );
  table2_.removeRepeatedEntries( results_ );

  cerr << "successful table build" << endl;
  return true;
} // ~OligoHashTable::buildTable( ifstream& oligoFile )

template< int PASS > void OligoHashTable<PASS>::checkOligo
( const Oligo& ol, const MatchPosition sequencePos )
{
  //  cout << "co: "<< sequencePos << endl;
  //  exit (-22);
  //  assert(1==0);
  Oligo res;
  hash_(ol, res);
 
  table1_.sps_.check
    ( res.ui[0], res.ui[1], sequencePos );
  //    ( res.ui[0], res.ui[1], sequencePos, partitionNum_ );
  table2_.sps_.check
    ( res.ui[1], res.ui[0], sequencePos );

} // ~OligoHashTable::checkOligo

bool lessThanF( const pair<Oligo, MatchPosition>& lhs, 
                const pair<Oligo, MatchPosition>& rhs )
{
  return (lhs.first.ui[0]<rhs.first.ui[0]);
}

//bool lessThanR( const pair<Oligo, MatchPosition>& lhs, 
//                const pair<Oligo, MatchPosition>& rhs )
//{
//  return (lhs.first.ui[1]<rhs.first.ui[1]);
//}


//  struct S
//  {
//    S() {}
//    S(uint a, uint b, Word suff, MatchPosition pos): a_(a), b_(b), pos_(pos), suff_(suff) {}
//    uint a_;
//    uint b_;
//    Word suff_;
//    MatchPosition pos_;
//  }; // ~S


template< int PASS > void OligoHashTable<PASS>::checkOligo
( const vector<Oligo>& v, const MatchPosition starts )
{

  //  cout << "Co: " << v.size() << "!!" << endl;
  //  vector<Oligo> h(v.size());
  static vector<pair<Oligo, MatchPosition> > h;
  //  static vector<S> p;
  h.resize(v.size());
  //  p.reserve(v.size());
  //  p.resize(0);

  for (uint i(0); i<v.size();i++)
  {
    hash_(v[i],h[i].first);
    h[i].second=starts+i;
  }
  //    cout << i << " " << h.back().second << " ";
  //   printWord(h.back().first.ui[0],16); cout << " ";
  //   printWord(h.back().first.ui[1],16); cout << endl;

  //  uint a, b;
  sort(h.begin(),h.end(),lessThanF);

  //  for (uint i(0); i<h.size();i++)
  //  {
    //    a=table1_.sps_.pCount_[h[i].first.ui[0]];
    //   b=table1_.sps_.pCount_[h[i].first.ui[0]+1];
    //    if (a!=b) p.push_back(S(a,b,h[i].first.ui[1],h[i].second));
    // if (a!=b) table1_.sps_.check__(a,b,h[i].first.ui[1],h[i].second);

  table1_.sps_.check__( h );

    //  }

  //  cout << p.size() << " " << h.size() << " "<< (((double)p.size())/h.size()) << endl;
  //  for (vector<S>::iterator i(p.begin());i!=p.end();i++)
  //  {
  //  table1_.sps_.check__
  //   ( i->a_, i->b_, i->suff_, i->pos_ );
  //  }
  //  p.resize(0);

  Word swap;

  for (uint i(0);i<h.size();i++)
  {
    swap=h[i].first.ui[0];
    h[i].first.ui[0]=h[i].first.ui[1];
    h[i].first.ui[1]=swap;
  }

  sort(h.begin(),h.end(),lessThanF);

  //  sort(h.begin(),h.end(),lessThanR);

  table2_.sps_.check__(h);
  //  for (uint i(0); i<h.size();i++)
  //  {
  //   a=table2_.sps_.pCount_[h[i].first.ui[1]];
  //   b=table2_.sps_.pCount_[h[i].first.ui[1]+1];
  //  if (a!=b) table2_.sps_.check__(a,b,h[i].first.ui[0],h[i].second);
    //    if (a!=b) p.push_back(S(a,b,h[i].first.ui[0],h[i].second));
  //  }
  //  for (vector<S>::iterator i(p.begin());i!=p.end();i++)
  //  {
  //  table2_.sps_.check__
  //      ( i->a_, i->b_, i->suff_, i->pos_ );
  //
    //  }


  //  for (uint i(0); i<h.size();i++)
  //  {
  //   table1_.sps_.check
  //     ( h[i].first.ui[0], h[i].first.ui[1], h[i].second );
  //  }


  //  for (uint i(0); i<h.size();i++)
  //  {
  //   table2_.sps_.check
  //      ( h[i].first.ui[1], h[i].first.ui[0], h[i].second );
  //  }

  //  Oligo res;
  //  hash_(ol, res);
 

  //  table2_.sps_.check
  //  ( res.ui[1], res.ui[0], sequencePos );

} // ~OligoHashTable::checkOligo


#ifdef PREFETCH 
struct PrefetchStruct
{
  PrefetchStruct( const Word* pWord, const Word* pLastWord ) :
    pWord_(pWord), pLastWord_(pLastWord) {}
  const Word* pWord_;
  const Word* pLastWord_;
};

void* prefetch( void* pv )
{
  PrefetchStruct* pf((PrefetchStruct*)pv);
  const Word* pWord(pf->pWord_), *pLastWord(pf->pLastWord_);
  Word pFetch;
  for (;pWord<pLastWord;pWord+=512)
  {
    pFetch = *pWord;
  } // ~for
  //  delete pf;
  return NULL;
} // ~prefetch
#endif



template< int PASS > MatchPosition OligoHashTable<PASS>::checkManyOligos
( Oligo& ol, const Word* pWord, const Word* pLastWord, MatchPosition seqNum )
{
  // ZZZ  const uint queryCacheSize(8*131072);
  //  cout << "cmo" << ((int)(pLastWord-pWord))<< endl;
  //  register uint hashVal;
  register Word thisWord; //, thisCarry;

#ifdef PREFETCH
  PrefetchStruct pf(pWord, pLastWord);
  pthread_t prefetchThread;
  pthread_create( &prefetchThread, NULL, prefetch, (void*) &pf );
  pthread_detach( prefetchThread ); // fire and forget
#endif    

  seqNum -=(oligoLength_-2);


  // ZZZ static vector<Oligo> v(queryCacheSize);
  // ZZZ v.resize(0); 
  //  vector<MatchPosition> n;
  // ZZZ MatchPosition starts=seqNum;

  for ( ; pWord != pLastWord ; ++pWord )
  {
    thisWord = *pWord;
    for (int i(0);i<16;i++, seqNum++)
    {
      
      shiftOligo( ol, thisWord );
#ifdef XXX
      ol.ui[1]<<=numBitsPerBase;
      thisCarry = (ol.ui[0]>>30);
      ol.ui[1] |= thisCarry;
      ol.ui[0]<<=numBitsPerBase;
      thisCarry = (thisWord>>30);
      ol.ui[0]|=thisCarry;
      thisWord<<=numBitsPerBase;
      ol.ui[1]&=prefixMask_;
#endif
      // ZZZ      v.push_back(ol);
      //      n.push_back(seqNum);
        //        n.resize(0);
      
      checkOligo( ol, seqNum );
    } // ~for i
    // ZZZ    if (v.size()>=(queryCacheSize))
    // ZZZ {
    // ZZZ  checkOligo( v, starts);
    // ZZZ  v.resize(0);
    // ZZZ  starts=seqNum;
    // ZZZ   }
  } // ~for pWord
  // ZZZ  checkOligo( v, starts);

  seqNum +=(oligoLength_-2);
#ifdef PREFETCH
  pthread_join( prefetchThread, NULL );
#endif
  return seqNum;


} // ~int OligoHashTable::checkManyOligos


//#define DEBUG_SCAN
// Have to put definition of scan after checkOligo because
// checkOligo is called by it

template< int PASS > MatchPosition OligoHashTable<PASS>::scan
( FileReader& file, const MatchPosition currentBlock )
{
  MatchPosition seqNum(0), seqLast;
  register Word thisWord;
  const ValidRegion *pValid(file.getFirstValid()), 
    *pLastValid(file.getLastValid());

  const Word *pWord, *pStart(file.getSeqStart());

  Oligo ol;
#ifdef DEBUG_SCAN
  cout << pLastValid-pValid+1 << " valid regions " << endl;
#endif

  assert((currentBlock&0xF)==0); 

  for ( ; pValid!=pLastValid; ++pValid )
  {
#ifdef DEBUG_SCAN
    cerr << "got valid region " << pValid->start << " " 
	 << pValid->finish << endl;
#endif
    if ((((int)pValid->finish)-((int)pValid->start))+1<oligoLength_) 
    {
#ifdef DEBUG
      cout << "region too short, ignoring" << endl;
#endif
      continue;
    }

    seqNum = pValid->start;
    pWord=pStart+(seqNum>>4);
    seqLast = pValid->finish + currentBlock;
    //    seqNum += currentBlock;
    //   seqLast = pValid->finish + currentBlock;


    // if necessary, shift first word so that first required base
    // takes up the most significant bits
    if ((seqNum&0xF)!=0)
    {
      thisWord=*(pWord++);
      thisWord <<= ((seqNum&0xF)<<1);
 
    } // ~if seqNum

    // fill up ol[0] and ol[1]
    for ( ; seqNum < pValid->start + oligoLength_-1 ; ++seqNum )
    {
      //      cout << "start " << seqNum << endl;
      if ((seqNum&0xF)==0)
      {
      	thisWord=*(pWord++);
      	//	printWord(thisWord,16); cout << endl;
      } // ~if seqNum

#ifdef DEBUG_SCAN
      printWord( ol.ui[1], 16 ); cout << "-";
      printWord( ol.ui[0], 16 ); cout << endl;
#endif


      shiftOligo( ol, thisWord );


    } // ~for seqNum
    //    ol[0] &= prefixMask_;

    seqNum += currentBlock;

    // now scan until the next whole Word
    for ( ; ((seqNum <= seqLast)&&((seqNum&0xF)!=0)) ; ++seqNum )
    {    
      //   cout << "aligning " << seqNum << endl; 
      shiftOligo( ol, thisWord );
#ifdef DEBUG_SCAN
      printWord( ol.ui[1], 16 ); cout << "-";
      printWord( ol.ui[0], 16 ); cout << " - about to check" << endl;
#endif

	// cerr << "scan: " << ol << " " << ol[0] << " " << ol[1] 
	// << " " << seqNum << endl;
	//	checkOligo(ol, seqNum );
	checkOligo(ol, seqNum-oligoLength_+2 );

    } // ~for seqNum

    if (seqNum>=seqLast) 
    {
#ifdef DEBUG_SCAN
      cout << "end of region" << endl;
#endif

      //      cerr << "B: " 
      //   << pValid->start << " " 
      //   << pValid->finish << " " 
      //   << pValid->finish-pValid->start+1 << " " 
      //   << currentBlock << " " 
      //   << seqNum << " " 
      //   << seqLast << " " 
      //   << endl;

      // Bug fix: need to scan one more base if at a word boundary,
      // therefore do not exit loop iteration yet TC - 27.11.03
      //      continue;
      if ((seqNum&0xF)!=0) continue;

    }
    //    assert((seqNum&0xF)==0);
    

    const Word* pLastWord(pStart+((1+pValid->finish)>>4));
    // finish = 33 -> pLastWord = pStart_ + 2
    // finish = 32 -> pLastWord = pStart_ + 2
    // finish = 31 -> pLastWord = pStart_ + 2
    // finish = 30 -> pLastWord = pStart_ + 1
  
#ifdef DEBUG_SCAN
    cout << pLastWord-pWord << endl;
#endif

    // now scan word by word
    //    cout << "cmo " << seqNum << " " << pLastWord-pWord << endl;
    seqNum=checkManyOligos( ol, pWord, pLastWord, seqNum );

    const int basesInLast((1+pValid->finish)&0xF);
    // finish = 33 -> basesInLast = 2
    // finish = 32 -> basesInLast = 1
    // finish = 31 -> basesInLast = 0
    // finish = 30 -> basesInLast = 15


    // if required, do the incomplete word at the end
    if (basesInLast!=0)
    {
      thisWord = *pLastWord;

      for (int i(0);i<basesInLast;i++, seqNum++)
      {
	shiftOligo( ol, thisWord );

	//	cerr << "scanend: " << ol << " " << ol[0] << " " << ol[1] << " " << seqNum << endl;
      //   cout << "finito " << seqNum << endl;
	//        checkOligo(ol, seqNum );
        checkOligo(ol, seqNum-oligoLength_+2 );

      } // ~for i

    } // ~if basesInLast
  
  } // ~for pValid

  return (((seqNum>>blockShift)+1)<<blockShift);

} // ~FileReader::scan( OligoHashTable& hashTable )

// code to scan multiple chromosomes using threads
#if (NUM_THREADS>1)

template <int PASS> void* scan( void* pv );
//void* scan( void* pv );

// ThreadInfo: contains data specific to an individual thread, plus
// static functions & data to maintain a count of the total number of threads

template<int PASS> struct ThreadInfo
{
  ThreadInfo( OligoHashTable<PASS>* pHash, 
	      FileReader* pFile,
	      MatchPosition currentBlock, const char* name ) :
    pHash_(pHash), pFile_(pFile), 
    currentBlock_(currentBlock), name_(name) {}
  ~ThreadInfo()
  {
    delete pFile_;
  } // ~dtor

  OligoHashTable<PASS>* pHash_;
  FileReader* pFile_;
  MatchPosition currentBlock_;
  const char* name_;

  static int threadCount_;
  static pthread_mutex_t threadCountLock_;
  static pthread_cond_t threadFreeSignal_;

  // start thread, waiting for others to complete if necessary
  static void startThread( ThreadInfo* pt )
  {
    pthread_t p;
    pthread_mutex_lock(&threadCountLock_);
    while (threadCount_==NUM_THREADS)
    {
      cout << "Waiting for thread to become free..." << endl;
      pthread_cond_wait(&threadFreeSignal_, &threadCountLock_);
    } // ~while
    cout << "OK, got a free thread..." << endl;
    ++threadCount_;
    pthread_mutex_unlock(&threadCountLock_);
    pthread_create(&p, NULL, scan<PASS>,  (void*) pt);
    pthread_detach(p);

  } // ~startThread

  // decrement thread count, and signal to say a thread is free
  static void freeThread( void )
  {
    cout << "Freeing thread..."  << endl;
    pthread_mutex_lock(&threadCountLock_);
    --threadCount_;

    pthread_cond_signal(&threadFreeSignal_);

    pthread_mutex_unlock(&threadCountLock_);

  } // ~freeThread

  // wait for all outstanding threads to finish
  static void waitForEnd( void )
  {
    pthread_mutex_lock(&threadCountLock_);
    while (threadCount_>0)
    {
      cout << "Waiting for " << threadCount_
	   << " threads to become free..." << endl;
      pthread_cond_wait(&threadFreeSignal_, &threadCountLock_);
    } // ~while
    cout << "All threads done, OK to finish" << endl;
    pthread_mutex_unlock(&threadCountLock_);
  } // ~waitForEnd

}; // ~struct ThreadInfo

template<int PASS> int ThreadInfo<PASS>::threadCount_=0;
template<int PASS> 
pthread_mutex_t ThreadInfo<PASS>::threadCountLock_
=PTHREAD_MUTEX_INITIALIZER;
template<int PASS>
pthread_cond_t ThreadInfo<PASS>::threadFreeSignal_
=PTHREAD_COND_INITIALIZER;

template <int PASS> void* scan( void* pv )
{
  Timer t;
  ThreadInfo<PASS>* pt((ThreadInfo<PASS>*)pv);
  cerr << "Starting thread to scan " << pt->name_ << ": " << t << endl;
  //  currentBlock = hashTable.scan(thisFile, currentBlock);
  (pt->pHash_)->scan(*(pt->pFile_), pt->currentBlock_ );
  cerr << "Finished thread to scan " << pt->name_ << ": " << t << endl;
  ThreadInfo<PASS>::freeThread();
  delete pt; 
  return NULL;
} // ~scan




// scanAll: complete a single pass through all the chromosomes
template< int PASS > void  scanAll
( 
 OligoSource* pOligos,
 const string& directoryName,
 const vector<string>& chromNames,
 vector<MatchPosition>& blockStarts, 
 OligoHashTable<PASS>& hashTable,
 // PartitionHashTable& part1, 
 // PartitionHashTable& part2, 
 // const SuffixScoreTable& scoreTable, 
 // MatchTable& results, 
 Timer& timer
 // const int i 
)
{
  // reset block counter at start of each pass
  MatchPosition currentBlock(blockSize);

  string fullChromName;

  ThreadInfo<PASS>* pThread;


  if (PASS==0) blockStarts.push_back(currentBlock);
  else assert(blockStarts[1]==currentBlock);
		 

  cerr << "About to build hash tables for pass " 
       << PASS << ": " << timer << endl;

  if (hashTable.buildTable( *pOligos )==false)
  {
    cerr << "No oligos to hash, returning" << endl;
    return;
  }

  cerr << "Built hash tables: " << timer << endl;

  for (int j(1);j<chromNames.size();j++)
  {
    //    if (PASS==0) blockStarts.push_back(currentBlock);

    fullChromName = directoryName+chromNames[j];
    cerr << "Scanning file " << fullChromName 
	 << ": " << timer << endl;

    pThread = new ThreadInfo<PASS>
      (&hashTable, NULL, currentBlock, chromNames[j].c_str());

    cerr << "Starting block: " << (currentBlock>>blockShift) << endl;

    pThread->pFile_=new FileReader(fullChromName.c_str());

    cerr << "Last valid base in file: " 
	 << pThread->pFile_->getLastValidBase() << endl;

    currentBlock+=(((pThread->pFile_->getLastValidBase()>>blockShift)+1)
		   <<blockShift);
    if (PASS==0) blockStarts.push_back(currentBlock);
    else (assert(blockStarts[j+1]==currentBlock));
		   
    ThreadInfo<PASS>::startThread(pThread);

    cerr << "Finishing block: " << (currentBlock>>blockShift) << endl;

  } // ~for j

  ThreadInfo<PASS>::waitForEnd();
    //  if (PASS==0) blockStarts.push_back(currentBlock);
  assert(blockStarts.size()==chromNames.size()+1);


} // ~scanAll

// single threaded scanAll
#else

// scanAll: complete a single pass through all the chromosomes
template< int PASS > void  scanAll
( 
 OligoSource* pOligos,
 const string& directoryName,
 const vector<string>& chromNames,
 vector<MatchPosition>& blockStarts, 
 OligoHashTable<PASS>& hashTable,
 // PartitionHashTable& part1, 
 // PartitionHashTable& part2, 
 // const SuffixScoreTable& scoreTable, 
 // MatchTable& results, 
 Timer& timer
 // const int i 
)
{
  // reset block counter at start of each pass
  MatchPosition currentBlock(blockSize);

  string fullChromName;

  //  OligoHashTable hashTable(oligoLength, numBits, part1, part2, scoreTable, results); 
    // set base to extract
  //  hashTable.setPass( i );

  cerr << "About to build hash tables for pass " 
       << PASS << ": " << timer << endl;

  if (hashTable.buildTable( *pOligos )==false)
  {
    cerr << "No oligos to hash, returning" << endl;
    return;
  }
  //  hashTable.buildTable( *pOligos );

  cerr << "Built hash tables: " << timer << endl;

  for (uint j(1);j<chromNames.size();j++)
  {
    if (PASS==0) blockStarts.push_back(currentBlock);


    fullChromName = directoryName+chromNames[j];
    cerr << "Scanning file " << fullChromName 
	 << ": " << timer << endl;

    cerr << "Starting block: " << (currentBlock>>24) << endl;
    FileReader thisFile(fullChromName.c_str());

    currentBlock = hashTable.scan(thisFile, currentBlock);
    cerr << "Finishing block: " << (currentBlock>>24) << endl;

    cerr << "... done " << timer << endl;
  } // ~for j
  if (PASS==0) blockStarts.push_back(currentBlock);
  assert(blockStarts.size()==chromNames.size()+1);


} // ~scanAll
// end of single/multi-threaded conditional code
#endif

#endif // #ifndef INCLUDED_ELAND_H
