#include <fcntl.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fstream>
#include <iostream>
#include <vector>
#include <map>
#include <cmath>
#include <algorithm>
#include <dirent.h>
#include <pthread.h>
#include "GlobalUtilities.h"
#include <errno.h>


// Copyright (c) 2007 Solexa
// This source file is covered by the "Solexa Public Source License"
// agreement and bound by the terms therein.

// PROJECT: ELAND (Efficient Local Alignment of Nucleotide Databases)
// MODULE: Unsquash.cpp
// AUTHOR: A. J. Cox

// Description: grab sequence fragments from a directory of sequence files
// that have been 'squashed' by SquashGenome.cpp into the 2-bits-per-base
// format used by ELAND.
// For convenience, text files in ELAND output format are used as input.


using namespace std;

typedef unsigned char uchar;

// bufSize: maximum size of an input line
const uint bufSize(512);
// batchSize: this many requests will be stored then sorted and fulfilled
// in order of file name and position
const int batchSize(262144);

// maxReadLengthELAND: longest read length allowable by ELAND
const uint maxReadLengthELAND(32);

// StringIndex: assigns an index number to each different string seen.
// This index number can be used to retrieve the corresponding name. 
// map<std::string,int> will
// do pretty much the same job but I want to avoid having to instantiate a 
// std::string object for each line of input parsed. Std::strings are just
// used as convenient stores for const char* strings.
struct StringIndex
{
  StringIndex() 
  { 
    // Need to make sure the vector is not reallocated as all the pointers
    // then become invalid. Assuming each entry refers to a different file
    // is the most conservative assumption.
    names_.reserve(batchSize); 
  } // ~ctor
  struct LessThanString
  {
    bool operator()( const char* a, const char* b ) const
    {
      return (strcmp(a, b)<0);
    } // ~bool lessThanGenomeFrag
  };

  vector<string> names_;
  map<const char*, int, LessThanString > index_;
  
  int getIndex(const char* name)
  {
    if (index_.find(name)==index_.end())
    { // add name to index if not already seen
      names_.push_back(string(name));
      index_[names_.back().c_str()]=names_.size()-1;
    } // ~if
    return index_[name];
  } // ~get 
}; // ~struct StringIndex


struct SeqRequest
{
  SeqRequest( int requestNum, int fileIndex, int filePos, char strand ) :
    requestNum_(requestNum),
    fileIndex_(fileIndex),
    filePos_(filePos),
    strand_(strand)
  {}
  int requestNum_;
  int fileIndex_;
  int filePos_;
  char strand_;
}; // ~struct SeqRequest

bool lessThanRequest (const SeqRequest& a, const SeqRequest& b)
{
  return ((a.fileIndex_<b.fileIndex_)
	  ||((a.fileIndex_==b.fileIndex_)&&(a.filePos_<b.filePos_)));
} // ~bool lessThanRequest;



// SquashFile: provide interface to a 'squashed' chromosome file. Assumes:
// First line might be a comment starting with '>', rest of lines are sequence.
// All sequence lines contain same number of bases, except possibly the last.
// NB Numbering of bases in file starts at 1.
class SquashFile
{
public:
  SquashFile( const char* name );
  ~SquashFile() { munmap( pMap, fileSize ); }

  uint getNextBase( uint i)
  {
    return ( (pWord[i>>4])>>(2*((i&0xF)^0xF)) ) & 0x3; 
  } // ~getNextBase

  char operator[]( int pos )
  {
    if (pos<0) return 'N';
    //    assert(pos<=numChars);
    //  --pos;
    //  pos+=(pos/lineLength);
    // return pStart[pos];
    return(baseNames[getNextBase((uint)pos)]);
  } // ~operator[]

  int getNumChars( void ) const { return numChars; }
private:
  void *pMap;
  const char* pStart;
  const char* pEnd;
  const Word* pWord;
  size_t fileSize;
  int lineLength;
  int numChars;
}; // ~class SquashFile

SquashFile::SquashFile( const char* name )
{
  int fd;
  // memory map sequence file
  fd = open(name, O_RDONLY, S_IRUSR );
  if (fd==-1)
  {
    cerr << "Error: could not open file " 
         << name << endl;
    exit (1);
  } // ~if
  fileSize = lseek(fd, 0, 2);

  pMap=mmap(0, fileSize, PROT_READ, MAP_SHARED, fd, 0);
  if (pMap==NULL)
  {
    cerr << "Error: could not memory map file " 
         << name << endl;
    exit (1);
  } // ~if
  close (fd);

  pStart=(char*)pMap;
  pEnd=pStart+fileSize;

  if (*pStart=='>') 
  {
    while (*pStart!='\n') ++pStart;
    ++pStart;
  } // ~if
  pWord = (const Word*) pStart;

} // ~ctor

void getFragments
( vector<SeqRequest>& requests, 
  vector<char*> frags,
  const StringIndex& files,
  const string& squashDirName,
  const int readLength,
  const int reverseStrandStartOffset,
  const int numRequests)
{
  if (!requests.empty())
  {
    SquashFile* pSquash(NULL);
    string squashFileName;
    int lastFile=requests[0].fileIndex_+1; // ensure false on first go
    sort( requests.begin(), requests.end(), lessThanRequest );

    for ( vector<SeqRequest>::iterator i(requests.begin());
	  i!=requests.end();++i)
    {
      if (i->fileIndex_!=lastFile)
      {
	squashFileName=squashDirName
	  +(string)"/"
	  +files.names_[i->fileIndex_]
	  +(string)".2bpb";
	delete pSquash;
	pSquash=new SquashFile(squashFileName.c_str());
      }
      lastFile=i->fileIndex_;

    //    cout << i->requestNum_ << " " << i->fileIndex_ 
    //	 << " " << i->filePos_ << " " << i->strand_ << " " << lastFile<< endl;

      frags[i->requestNum_][readLength]=0;
      if (i->strand_=='F')
      {
	for (int j(0);j<readLength;j++)
	{
	  frags[i->requestNum_][j]=(*pSquash)[i->filePos_+j-1];
	}
      } // ~if
      else
      {
	for (int j(0);j<readLength;j++)
	{
	  frags[i->requestNum_][readLength-1-j]
	    =reverseCharASCII[(*pSquash)
			      [i->filePos_+j-reverseStrandStartOffset-1]];
	} // ~for
      } // ~else
    } // ~for i

    delete pSquash;
  } // ~if

  for (int i(0);i<numRequests;i++)
    printf("%s\n", frags[i]);

} // ~getFragments

int main( int numArgs, const char** args )
{
  

  if ((numArgs!=3)&&(numArgs!=4))
  {
    cerr << "Usage: cat elandOutput.txt | " << args[0] 
	 << " squashedFileDir readLength [readLengthUsedForELAND]" << endl;
    exit (1);
  }

  string squashDir(args[1]);
  uint readLength(atoi(args[2]));
  assert(readLength>0);

  int readLengthELAND;
  if (numArgs==4)
  {
    readLengthELAND=atoi(args[3]);
  } // ~if
  else
  { // make a sensible guess
    readLengthELAND=(readLength>maxReadLengthELAND)
      ?maxReadLengthELAND:readLength;
  } // ~else

  // In the event of a reverse complement match need to grab bases that
  // come before the read in the forward strand
  const int reverseStrandStartOffset(readLength-readLengthELAND);

  // set up request buffer
  vector<SeqRequest> requests;
  requests.reserve(batchSize);
  // set up fragment buffer
  vector<char*> frags(batchSize);
  for (vector<char*>::iterator i(frags.begin());i!=frags.end();i++)
  {
    *i=new char[readLength+1];
  } // ~for

  StringIndex files;

  char buf[bufSize];
  char chromName[bufSize];
  uint pos;
  char strand;

  int requestNum(0);
  int bn(0);

  while( fgets(buf, bufSize, stdin )!=NULL)
  {

    if (sscanf(buf, "%*s\t%*s\t%*s\t%*s\t%*s\t%*s\t%s\t%d\t%c", 
	       chromName, &pos, &strand )!=3)
    {
      frags[requestNum][0]='\0';
    } // ~if
    else
    {
      requests.push_back
	(SeqRequest(requestNum,
		    files.getIndex(chromName),
		    pos,
		    strand));

    } // ~else
    if (++requestNum==batchSize)
    {
      getFragments(requests, 
		   frags, 
		   files, 
		   squashDir, 
		   readLength, 
		   reverseStrandStartOffset,
		   requestNum);
      requests.clear();
      requestNum=0;
    } // ~if
  } // ~while
  getFragments(requests, 
	       frags, 
	       files, 
	       squashDir, 
	       readLength, 
	       reverseStrandStartOffset,
	       requestNum);

  // deallocate memory
  for (vector<char*>::iterator i(frags.begin());i!=frags.end();i++)
  {
    delete [] *i;
  } // ~for

} // ~main

