/*
 * Class to handle XML output.
 * It provides functions for writing out floats, ints and strings and
 * automatically keeps track of the indentation level.
 *
 * Copyright (c) Solexa 2005, 2006
 * Author: Klaus Maisinger
 * This source file is covered by the "Solexa Public Source License"
 * agreement and bound by the terms therein.
 */

#if defined(__BORLANDC__)
#include <string.h>
#else
#include <cstring>
#endif
#include <cassert>
#include <string>
#include <cstdio>
#include "ImageXMLHandler.h"
#include "XMLSAXParser.h"

void TestXMLHandler::startDocument()
{
    printf("START DOCUMENT\n");
}

void TestXMLHandler::endDocument()
{
    printf("END DOCUMENT\n");
}

void TestXMLHandler::startElement(const char name[])
{
    printf("START ELEMENT %s\n", name);
}

void TestXMLHandler::endElement(const char name[])
{
    printf("END ELEMENT %s\n", name);
}

void TestXMLHandler::characters(const char ch[], int start, int length)
{
    printf("CHARACTERS %s\n", ch);
}

ImageXMLWriter::ImageXMLWriter() 
    : XMLFileWriter(), indentCurrent(0), firstLine(0) 
{
    setFileHandle(0);
    ignorableWhitespace("\n", 0, 1);
}

ImageXMLWriter::ImageXMLWriter(const char filename[], int indentLevel) 
    : XMLFileWriter(), firstLine(0)
{
    setFileHandle(0);
    setFilename(filename);
    setIndentLevel(indentLevel);
    ignorableWhitespace("\n", 0, 1);
}

ImageXMLWriter::~ImageXMLWriter()
{
    flush();
    if (fp)
        fclose(fp);
}

void ImageXMLWriter::setFilename(const char filename[])
{
    flush();
    if (fp)
        fclose(fp);
    FILE *fh = fopen(filename, "w");
    assert(fh != 0);
    setFileHandle(fh);
    firstLine = 1;
}

/*
 * set up the padding in front of a tag depending on the indentation level
 */
void ImageXMLWriter::adjustWhitespace()
{
    int i;
    std::string buffer = "\n";
    if (firstLine) {
        buffer = ""; firstLine = 0;
    }
    for (i = 0; i < indentCurrent; i++)
        buffer += "  ";
    ignorableWhitespace(buffer.c_str(), 0, strlen(buffer.c_str()));
}

void ImageXMLWriter::writeElementString(const char name[], const char element[])
{
    adjustWhitespace();
    XMLFileWriter::startElement(name);
    characters(element, 0, strlen(element));
    XMLFileWriter::endElement(name);
}

void ImageXMLWriter::writeElementFloat(const char name[], float element)
{
    char buffer[20];
    sprintf(buffer, "%g", element);
    writeElementString(name, buffer);
}

void ImageXMLWriter::writeElementInt(const char name[], int element)
{
    char buffer[20];
    sprintf(buffer, "%i", element);
    writeElementString(name, buffer);
}

void ImageXMLWriter::writeElementChar(const char name[], char element)
{
    char buffer[3];
    sprintf(buffer, "%c", element);
    writeElementString(name, buffer);
}

void ImageXMLWriter::startElement(const char name[])
{
    adjustWhitespace();
    XMLFileWriter::startElement(name);
    indentCurrent += 1;
}

void ImageXMLWriter::endElement(const char name[])
{
    indentCurrent -= 1;
    adjustWhitespace();
    XMLFileWriter::endElement(name);
}

void ImageXMLWriter::startDocument()
{
    firstLine = 1;
}

void ImageXMLWriter::endDocument()
{
    characters("\n", 0 , strlen("\n"));
}

#if defined(XMLTEST)
int main(int argc, char *argv[])
{
    ImageXMLWriter *handler;
    handler = new ImageXMLWriter("test.xml");
    handler->startDocument();
    handler->startElement("Channel");
    handler->writeElementInt("ChannelNumber", 2);
    handler->startElement("Tile");
    handler->writeElementInt("TileNumber", 1);
    handler->startElement("Image");
    handler->writeElementInt("Colour", 0);
    handler->writeElementFloat("Background", 34.7);
    handler->writeElementFloat("Noise", 3.6);
    handler->endElement("Image");
    handler->endElement("Tile");
    handler->endElement("Channel");
    handler->endDocument();
    delete handler;
    
    handler = new ImageXMLWriter("test2.xml", 2);
    handler->startDocument();
    handler->startElement("Channel");
    handler->writeElementInt("ChannelNumber", 2);
    handler->startElement("Tile");
    handler->writeElementInt("TileNumber", 1);
    handler->endElement("Tile");
    handler->endElement("Channel");
    handler->endDocument();
    delete handler; 
    
    TestXMLHandler *handler2 = new TestXMLHandler();
    XMLSAXParser *saxparser = new XMLSAXParser();
    FILE *fh = fopen("test.xml", "r");
    saxparser->parse(fh, handler2);
    fclose(fh);  
}    
#endif
