/*
 * Routines to handle cluster intensities in different frequency channels
 * and cycles.
 * 
 * Copyright (c) Solexa 2005, 2006
 * Author: Klaus Maisinger
 * This source file is covered by the "Solexa Public Source License"
 * agreement and bound by the terms therein.
 */

#if defined(__BORLANDC__)
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#else
#include <cstdio>
#include <cstdlib>
#include <cstring>
#endif
#include <cassert>
#include <iostream>
#include <fstream>
#include <sstream>
#include "IntensityData.h"
#include "Stats.h"

int write_index_positions(const char filename[],
                          int lane, int tile,
                          st_fposition *positions, int nobj)
{
    FILE *fp = fopen(filename, "w");
    if (!fp) {
        fprintf(stderr, "Error: Failed to open file %s\n", filename);
        return 1;
    }
    int code = 0;
    int i;
    char buffer[256];
    for (i = 0; i < nobj; i++) {
        sprintf(buffer, "%i\t%i\t%.0f\t%.0f\n",
                lane + 1, tile + 1, 
                int(round(positions[i].x * 100.)) / 100., 
                int(round(positions[i].y * 100.)) / 100.);
        if (fprintf(fp, buffer) < (int)strlen(buffer)) {
            fprintf(stderr, "Error: Failed to write to %s\n", filename);
            code = 2;
        }
    }
    if (fclose(fp)) {
        fprintf(stderr, "Error: Failed to close file %s\n", filename);
        return 3;
    }
    return code;
}

/*
 * auxiliary function to save an array of intensties to disk
 * Each lines represents one cluster.
 * the first three columns are the tile no and the x- and y-
 * image coordinates of the object.
 * all other columns hold the intensity data for each molecule/colony,
 * followed by the background noise estimates.
 */
int write_signal_array(const char filename[],
                       float *array,
                       int nobj, int nall, int ncol,
                       int lane, int tile,
                       st_fposition *positions, int append_flag)
{
    FILE *fp;
    if (append_flag)
        fp = fopen(filename, "a");
    else
        fp = fopen(filename, "w");
    if (!fp) {
        fprintf(stderr, "Failed to open file %s.\n", filename);
        return 1;
    }

    int code = 0;
    int i, j;
    char *buffer = new char [20+2*nall*4*10];
    for (i = 0; i < nobj; i++) {
        int idx = 0;
        if (positions != NULL) {
            idx = sprintf(buffer, "%i\t%i\t%.0f\t%.0f\t", 
                          lane + 1, tile + 1,
                          int(round(positions[i].x * 100.)) / 100.,
                          int(round(positions[i].y * 100.)) / 100.); 
        }
        for (j = 0; j < nall; j++) {
            idx += sprintf(buffer+idx, "%6.1f", array[i*nall+j]);
            if (j != nall-1) {
                if (j%ncol == ncol-1)
                    idx += sprintf(buffer+idx, "\t");
                else
                    idx += sprintf(buffer+idx, " ");
            }
        }
        sprintf(buffer+idx, "\n");
        if (fprintf(fp, buffer) < (int)strlen(buffer)) {
            fprintf(stderr, "Failed to write to %s.\n", filename);
            code = 2;
        }
    }
    delete [] buffer;
    if (fclose(fp)) {
        fprintf(stderr, "Failed to close %s.\n", filename);
        return 3;
    }
    return code;
}

int write_signal_array(const char filename[], std::vector<float * > signal,
                       int nobj, int ncol,
                       int lane, int tile, st_fposition *positions)
{
    FILE *fp = fopen(filename, "w");
    if (!fp) {
        fprintf(stderr, "Error: Failed to open file %s for writing.\n", filename);
        return 1;
    }

    int code = 0;
    int i, j, k;
    int idx, count;
    int ncycles = int(signal.size());
    char *buffer = new char [(22+ncycles*ncol*11)];
    assert(buffer != 0);

    for (i = 0; i < nobj; i++) {
        idx = 0;
        if (positions != NULL) {
            idx = snprintf(buffer, 20, "%i\t%i\t%.0f\t%.0f\t", 
                           lane + 1, tile + 1,
                           int(round(positions[i].x * 100.)) / 100.,
                           int(round(positions[i].y * 100.)) / 100.); 
        }
        for (j = 0; j < ncycles; j++) {
            for (k = 0; k < ncol; k++) {
	        count = snprintf(buffer+idx, 10, "%6.1f", signal[j][i*ncol+k]);
		if (count >= 10)
		    count = 9;
		idx += count;
	        if (k != ncol-1)
                    idx += sprintf(buffer+idx, " ");
                else if (j != ncycles-1)
                    idx += sprintf(buffer+idx, "\t");
		else
                    idx += sprintf(buffer+idx,"\n");
		  
            }
        }
        if (fprintf(fp, buffer) < (int)strlen(buffer)) {
            fprintf(stderr, "Error: Failed to write to %s.\n", filename);
            code = 2;
        }
    }

    delete [] buffer;
    if (fclose(fp)) {
        fprintf(stderr, "Error: Failed to close %s.\n", filename);
        return 3;
    }
    return code;
}

/*
 * read intensity data from a full text file, as described in the
 * run-folder spec.
 */
std::vector<float * > read_signal_array(const char *filename, 
                                        int *nobj, int *ncol, int *error,
                                        int offset)
{
    std::vector<float * > intensities;
    std::vector<std::string> lines;
    lines.reserve(100000);
    std::string buffer;
    buffer.reserve(10000);
    *error = 0;
    *ncol = 0;
    *nobj = 0;

    FILE *fp = fopen(filename, "r");
    if (fp) {
        const int buffersize = 8192;
        char *s = new char [buffersize];
        while (fgets(s, buffersize, fp)) {
            assert(strlen(s) < buffersize-1); // if this fails need to recompile
                                              // with larger buffer size
            if ((strlen(s) == 0) || (s[0] == '#'))
                continue;
            lines.push_back(std::string(s));
        }
        fclose(fp);
        delete [] s;
    }
    else {
        std::cerr << "Error: Failed to open "<<filename << std::endl;
        *error = 1;
        return intensities;
    }

    /*
     * count number of objects, cycles and columns
     */
    *nobj = int(lines.size());
    int ncycles = 0;
    if (*nobj > 0) {
        std::stringstream ss(lines[0]);
        while (getline(ss, buffer, '\t')) {
            ncycles++;
            if (ncycles > offset) {
                float dummy;
                int n = sscanf(buffer.c_str(),
                                  "%f %f %f %f", 
                                  &dummy, &dummy, &dummy, &dummy);
                if (*ncol == 0)
                    *ncol = n;
                if (*ncol != n) {
                    std::cerr << "Error: " << filename << ", line " << 0 <<
                                 ": Wrong number of intensities" << std::endl;
                    *error = 2;
                }
                
             }
        }
    }
    ncycles -= 4;
    
    /* 
     * allocate memory
     */
    int i, j, k;
    for (j = 0; j < ncycles; j++) {
        float *array = new float [*nobj * *ncol];
        for (i = 0; i < *nobj * *ncol; i++)
             array[i] = 0.;
        intensities.push_back(array);
    }
    
    for (i = 0; i < *nobj; i++) {
#if 0   // slow C++ version
        std::stringstream ss(lines[i]);
        int cycle = 0;
        while (getline(ss, buffer, '\t')) {
            j = cycle - offset;
            if (j >= 0) {
                float dummy[4];
                int n = sscanf(buffer.c_str(), "%f %f %f %f", 
                               &(dummy[0]), &(dummy[1]),
                               &(dummy[2]), &(dummy[3]));
                for (k = 0; k < n; k++)
                    intensities[j][i * *ncol + k] = dummy[k];
                if (*ncol != n) {
                    std::cerr << "Error: " << filename << ", line " << i <<
                                 ": Wrong number of intensities" << std::endl;
                    *error = 2;
                }
            }
            cycle++;
            
            if (j >= ncycles) {
                std::cerr << "Error: " << filename << ", line " << i << 
                             ": Wrong number of cycles" << std::endl;
                *error = 3;
            }
        }
#else
        j = -offset;
        int ncolmax = ncycles * *ncol;
        int cycle;
        const char *buffer = lines[i].c_str();
        char *begin, *end;
        float value;
        end = (char *)buffer;
        do {
            begin = end;
            value = (float)strtod(begin, &end);
            if (j >= 0) {
                cycle = j / *ncol;
                k     = j % *ncol;
                intensities[cycle][i * *ncol + k] = value;
            }
            j++;
        } while ((j < ncolmax) && (end != NULL) && (end != begin));
        if (j != ncolmax) {
            fprintf(stderr, 
                    "Error: %s, line %i: File corrupted.\n",
                    filename, i);
            fprintf(stderr,
                    "Error: %s: Expecting %i data columns, found %i\n",
                    filename, ncolmax+offset, j);
            *error = 2;
        }
#endif
    }
    
    return intensities;
}

/*
 * auxiliary function to read intensity data; compare previous function
 */
int read_signal_array(const char filename[],
                      float *signal, int *nobj, int *ncol, int skipcols)
{
    char buffer[4095];
    FILE *fp = fopen(filename, "r");
    int ncolmax = *ncol;
    int nobjmax = *nobj;
    int ncolfirst = 0;
    int lineno = 0;
    int errorflag = 0;
    *nobj = 0;
    *ncol = 0;
    if (fp) {
        while (fgets(buffer, sizeof(buffer), fp)) {
            lineno ++;
            // ignore comments
            if (buffer[0] == '#')
                continue;
            // read molecule intensities
            if (*nobj < nobjmax) {
                int j = 0;
                char *begin, *end;
                float value;
                end = buffer;
                do {
                    begin = end;
                    value = (float)strtod(begin, &end);
                    if (j >= skipcols)
                        signal[(*nobj)*ncolmax+j-skipcols] = value;
                    j++;
                } while ((j < ncolmax+skipcols) && (end != NULL) && (end != begin));
                *ncol = j - skipcols;
                if (ncolfirst == 0) {
                    ncolfirst = *ncol;
                }
                if (*ncol != ncolfirst) {
                    fprintf(stderr, 
                            "Error: %s, line %i: File corrupted.\n",
                            filename, lineno);
                    fprintf(stderr,
                            "Error: %s: Expecting %i data columns, found %i\n",
                            filename, ncolfirst, *ncol);
                    errorflag = 1;
                }
                for (j = *ncol; j < ncolmax; j++)
                    signal[(*nobj)*ncolmax+j] = 0.;
                (*nobj) ++;
            }
        }
        *ncol = ncolfirst;
        fclose(fp);
    }
    else {
        fprintf(stderr, "Error: Failed to open %s.\n", filename);
    }
    return errorflag;
}

