/*
 * Old-stye journal file parsing routines.
 * 
 * (c) Copyright Solexa 2005, 2006
 * Author: Klaus Maisinger
 * This source file is covered by the "Solexa Public Source License"
 * agreement and bound by the terms therein.
 */
/*
 * Journal files provide a list of image filenames to the image processing
 * software, and the nucleotide the image corresponds to. This will soon
 * be replace by an XML file.
 */
 
#include <cassert>
#include "JournalParser.h"

/*
 * Journal files consist of ASCII lines starting with one of the
 * following two commands:
 * 
 * Load "<filename>"
 * Sequence "", <ncycles, <ncolours>: #0-[ACGT], ..."
 */
int parse_sequence_command(const char buffer[],
                           char output_filename[FILENAME_MAX],
                           int *ncycle, int *ncolour,
                           int **stackidx_list, unsigned char **base_list,
                           int *call_base_flag)
{
    char command[250];

    sscanf(buffer, "%s", command);
    if (! strcmp(command, "SeqIntensities"))
        *call_base_flag = 0;
    else if (! strcmp(command, "Sequence"))
        *call_base_flag = 1;
    else 
        return 1;

    {
        // identify filename
        char *f1 = strchr((char *)buffer, '\"')+1;
        char *f2 = strchr(f1, '\"');
        assert(f1 != NULL);
        assert(f2 != NULL);
        if (f2 == NULL) return 2;
        *f2 = '\0';
        snprintf(output_filename, FILENAME_MAX,
                "%s", f1);
        *f2 = '\"';
        // parse cycles and colours
        if (sscanf(f2+1, ", %d, %d", ncycle, ncolour) != 2)
            return 3;

        *base_list = new unsigned char [(*ncycle)*(*ncolour)];
        *stackidx_list = new int [(*ncycle)*(*ncolour)];
    
        // parse stacks and bases
        char *f3 = strchr(f2+1, ':');
        if (f3 == NULL) return 4;
        f3 += 1;
        int i;
        for (i = 0; i < (*ncycle)*(*ncolour); i++) {
            sscanf(f3+1, "#%d-%c", &((*stackidx_list)[i]), &((*base_list)[i]));
            f3 = strchr(f3+1, ',');
            if (f3 == NULL) return 5;
        }
    }
    return 0;
}


int parse_filename_command(const char *buffer, char *fn, 
                           int len)
{
    std::string filename;
    char command[250];
    sscanf(buffer, "%249s", command);
    char *f1 = strchr((char *)buffer, '\"')+1;
    if (! strcmp(command, "Load"))
        filename = f1;
    else 
        return 1;
    // remove trailing "
    std::string::size_type mark = filename.find('\"');
    if (mark != std::string::npos)
        filename.erase(mark);
    snprintf(fn, len, "%s", filename.c_str());
    return 0;
}

int parse_journal(const char *journal_filename,
                  char output_filename[FILENAME_MAX], 
                  int *ncycle, int *ncolour, int *npos,
                  ImageList ***image_list, unsigned char **base_list,
                  int *call_base_flag)
{
    int *stackidx_list;
    FILE *fp = fopen(journal_filename, "r");
    if (!fp)
        return -1;

    char line[2500], command[250];
    while (fgets(line, sizeof(line), fp)) {
        if (line[0] && line[0] != '#') {
            if (sscanf(line, "%s", command) == 1) {
                //printf("%s\n", command);
                if ((! strcmp(command, "Sequence")) 
                    || (! strcmp(command, "SeqIntensities"))) {
                    parse_sequence_command(line, output_filename,
                                           ncycle, ncolour,
                                           &stackidx_list, base_list,
                                           call_base_flag);
#if 0
                    printf("%s %i %i %i\n", output_filename, 
                           *ncycle, *ncolour, *call_base_flag);
                    int i, j;
                    for (i = 0; i < (*ncycle); i++) {
                        printf("%3i ", i);
                        for (j = 0; j < (*ncolour); j++)
                            printf("%c", (*base_list)[i*(*ncolour)+j]);
                        printf("\n");
                    }
#endif
                }
            }
        }
    }
    rewind(fp);
    delete [] stackidx_list;

    *image_list = new ImageList * [(*ncolour)*(*ncycle)];
    int num = 0;
    while (fgets(line, sizeof(line), fp)) {
        if (line[0] && line[0] != '#') {
            if (sscanf(line, "%s", command) == 1) {
                if (! strcmp(command, "Load")) {
                    char filename[FILENAME_MAX];
                    parse_filename_command(line, filename, sizeof(filename));
                    //printf("%s\n", filename);
                    (*image_list)[num] = new ImageList;
#if 1
                    (*image_list)[num]->addFile(filename);
#else
                    if ((*image_list)[num]->load(filename)) {
                        printf("ERROR: Cannot open %s\n", filename);
                        exit(0);
                    }
#endif
                    num++;
                }
            }
        }
    }
    fclose(fp);
#if 1
    if (num < (*ncolour)*(*ncycle)) {
        printf("Not enough images loaded!\n");
        return 1;
    }
#endif

    return 0;
}
