/*
 * Class for image alignment and intensity extraction.
 *
 * Time-stamp: <26 Jan 2007 17:18 GMT Process.cpp km>
 *
 * Performs automated alignment of iteratively passed images.
 * The individual images are processed elsewhere. This class simply
 * combines images taken in different colours, 
 * remaps coordinates suitably (image registration),
 * identifies common objects and extracts intensity informtion for multiple 
 * images. The output is written to disk.
 * 
 * Copyright (c) Solexa 2003, 2004, 2005, 2006
 * Author: Klaus Maisinger
 * This source file is covered by the "Solexa Public Source License"
 * agreement and bound by the terms therein.
 */

#if defined(__BORLANDC__)
#include <ctype.h>
#else
#include <cctype>
#endif
#include <cmath>
#include <cstdio>
#include <cstring>
#include <cstdlib>
#include <cassert>
#include "FFT.h"
#include "DNACodes.h"
#include "Process.h"
#include "IOProcess.h"
#include "Maths.h"
#include "Tools.h"
#include "ImageUtil.h"
#include "QCImage.h"
#include "Positions.h"
#include "IntensityData.h"
#include "ProcessRemapping.h"

//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
/*
 * Class that accumulates data from the camera and performs 
 * image registration and intensity extraction.
 */

DataProcessor::DataProcessor()
{
    initialise();
}

DataProcessor::~DataProcessor()
{
    clear_memory();
}

void DataProcessor::clear_memory()
{
    if (default_transforms != NULL)   delete [] default_transforms;
    if (bases != NULL) delete [] bases;

    if (offset_tracker != NULL) delete offset_tracker;
}

void DataProcessor::setParameters(int cmax, int x, int y, float f, float d)
{
    clear_memory();
    initialise();

    imagemax = 1;
    colourmax = cmax;
    nx = x; ny = y;
    fwhm = f;
    dist = d;
    max_object_number = 2*colourmax*N_MOL;

    default_transforms   = new AlignmentTransform [colourmax];
  
    bases = new unsigned char[colourmax];

    offset_tracker = new OffsetTracker(imagemax, colourmax);
}

void DataProcessor::initialise()
{
    //Initialise all variables
    imagemax = 0;
    colourmax = 4;
    max_stage_offset = 100.;
    max_colour_offset = 5.;
    max_object_number = 0;

    default_transforms   = NULL;
    bases        = NULL;

    offset_tracker = NULL;
    aligner = NULL;
}

void DataProcessor::setDefaultTransforms(AlignmentTransform *transforms,
                                         int nc)
{
    assert (nc == colourmax);
    if (default_transforms != NULL)
        delete [] default_transforms;
    default_transforms = transforms;
}

#if 0
/*
 * Convert cameradata to image and perform data analysis on image
 * to identify molecules.
 * This function is no longer actively tested. Its idea is to enable
 * direct analysis of data provided in memory, as opposed through
 * files.
 */
void DataProcessor::processCamera(unsigned char *cameradata,
                                  int tile, int colour, int cycle)
{
    unsigned short *pixels = new unsigned short [nx*ny];
    Image *image = new Image();

    convert_image_uchar2ushort(cameradata, pixels, nx, ny, 1);
    delete[] cameradata;
    image->setImage(pixels, nx, ny);
    image->set_fwhm(fwhm);
    image->set_distance(dist);

    image->analyse();
    image->data->setSequencingInfo(BaseUpperCase[colour]);
#if 0 // TODO: reactivate
    analyseImage(image, tile, colour, cycle);
#endif
}
#endif

/*
 * This routine calculates the coordinate transforms needed
 * to align images back to the first cycle reference.
 *
 * WARNING: The caller MUST NOT deallocate any image data until
 * all 4 colours in a tile have been completed.
 */
void DataProcessor::analyseImage(Image *image,
                                 st_fposition *referencepositions,
                                 int referencecount, 
                                 int tile, int colour, int cycle)
{
    assert(image);
    assert(image->nx == nx);
    assert(image->ny == ny);
    assert(image->getData());
    assert(image->getData()->status & IMAGEDATA_STATE_ANALYSED);
    assert(image->getData()->filter_image != NULL);
    assert(image->getData()->noise_image  != NULL);
    assert(image->get_objects()->getSize() < 100000);
    image->getData()->lock(); // make sure that image data
                              // cannot be deallocated

    image->getData()->analyseEstimateFWHM();

    /* retrieve reference positions */
    if (colour == 0) {
        aligner = new ProcessAlign(referencepositions, referencecount, 
                                   nx, ny, fwhm);
        aligner->setMaximumOffsets((int)max_stage_offset);
    }

    /*
     * image realignment
     */
    assert(default_transforms[colour].nx == nx);
    assert(default_transforms[colour].ny == ny);
    image->data->transform     = new AlignmentTransform;
    image->data->usedTransform = new AlignmentTransform;
    image->data->transform->copy(default_transforms[colour]);

    if (cycle > 0) {
        /* 
         * alignment on subregions of the image
         * to get an idea of image distortions 
         */
        aligner->setRegionParameters(2, 2);
        aligner->alignRegionPixelPositions(image->get_objects()->getData(),
                                           image->get_objects()->getSize(),
                                           image->data->transform);
    }
    image->data->usedTransform->copy(*(image->data->transform));

    /* "base calling" parameters */
    bases[colour]      = image->data->sequencing->base;

    //  for each position, free temporary data
    if (colour == colourmax-1) {
        delete aligner;
    }
}


int DataProcessor::validateOffsets(std::vector<Image * >images,
                                   int tile, int cycle)
{
    assert(colourmax == int(images.size()));
    int i;
    int is_deblocked = images[0]->data->sequencing->is_deblocked;
    AlignmentTransform *position_transforms 
             = new AlignmentTransform [images.size()];
    int *object_counts = new int [ images.size() ]; 
    for (i = 0; i < colourmax; i++) {
        position_transforms[i].copy(*(images[i]->data->usedTransform));
        object_counts[i] = images[i]->get_objects()->getSize();
    }

    /* 
     * QC of image offsets 
     */

    /* QC for image realignment values */
    int valid_flag = 0;
    int valid_offsets = 0;

    valid_flag = validate_object_counts(object_counts, colourmax, 200);
    for (i = 0; i < colourmax; i++)
        if ((fabs(position_transforms[i].offsetx) > max_stage_offset) ||
            (fabs(position_transforms[i].offsety) > max_stage_offset))
            valid_flag &= 0;

    if (valid_flag)
        valid_offsets = validate_offsets(position_transforms,
                                         colourmax, max_colour_offset,
                                         default_transforms);
    
    if ((valid_offsets == colourmax) && (cycle > 0) && (!(is_deblocked))
        && (valid_flag))
        offset_tracker->add_offsets(position_transforms, colourmax, tile);

    delete [] position_transforms;
    delete [] object_counts;

    return (valid_flag && valid_offsets);
}


void DataProcessor::buildReference(std::vector<Image * >images,
                                   st_fposition *oldreference, int oldcount,
                                   st_fposition **newreference, int *newcount)
{
    int i;
    int *object_counts = new int [colourmax];
    st_fnoisepixel **object_lists = new st_fnoisepixel * [colourmax];
    for (i = 0; i < colourmax; i++) {
        object_counts[i] = images[i]->get_objects()->getSize();
        object_lists[i] = new st_fnoisepixel [object_counts[i]];
        memcpy(object_lists[i], images[i]->get_objects()->getData(),
               object_counts[i] * sizeof(st_fnoisepixel));
    }

    for (i = 0; i < colourmax; i++)
        images[i]->getData()->usedTransform->transformReverse(object_lists[i], 
                                                              object_counts[i]);

    *newcount = 0;
    RemappingLookup *remapper = new RemappingLookup(nx, ny);
    remapper->setReference(oldreference, oldcount);
    remapper->setMaxNewPositions(max_object_number - oldcount);
    for (i = 0; i < colourmax; i++) {
        remapper->addNewPositions(object_lists[i], object_counts[i], dist);
    }
    remapper->findNewPositions(dist);
    remapper->addNewPositionsToReference();
    *newcount = remapper->getNumberNewPositions();
    *newreference = new st_fposition [*newcount];
    memcpy(*newreference, remapper->getNewPositions(), 
           *newcount * sizeof(st_fposition));
    for (i = 0; i < colourmax; i++)
        delete [] object_lists[i];
    delete [] object_counts;
    delete [] object_lists;
    delete remapper;
}

void DataProcessor::analyseTile(std::vector<Image * >images,
                                int tile, int cycle, int *flag,
                                st_fposition *referencepositions,
                                int referencecount,
                                float *intensities, float *noise,
                                int *referenceindices)
{
    int i;
    *flag = validateOffsets(images, tile, cycle);
    if (*flag) 
         analyseTile(images, referencepositions, referencecount,
                     intensities, noise, referenceindices);
    else {
        for (i = 0; i < colourmax * referencecount; i++) {
            intensities[i] = 0.;
            noise[i] = 0.;
            referenceindices[i] = -1;
        }
    }
    for (i = 0; i < colourmax; i++)
        images[i]->getData()->unLock();
}

/*
 * perform the remapping of all colours and generate remapped intensities.
 */
void DataProcessor::analyseTile(std::vector<Image * >images,
                                st_fposition *referencepositions,
                                int referencecount,
                                float *intensities, float *noise,
                                int *referenceindices)
{
    assert(int(images.size()) == colourmax);
    int i;
    int *object_counts = new int [colourmax];
    st_fnoisepixel **object_lists = new st_fnoisepixel * [colourmax];

    for (i = 0; i < colourmax; i++) {
        object_counts[i] = images[i]->get_objects()->getSize();
        object_lists[i] = new st_fnoisepixel [object_counts[i]];
        memcpy(object_lists[i], images[i]->get_objects()->getData(),
               object_counts[i] * sizeof(st_fnoisepixel));
    }

    /*
     * remap molecules for the current position and produce array
     * with molecule intensities in all colours
     */
    int *indices = new int [referencecount * colourmax];
    for (i = 0; i < referencecount*colourmax; i++)
        indices[i] = -1;

    for (i = 0; i < colourmax; i++)
        images[i]->getData()->usedTransform->transformReverse(object_lists[i], 
                                                              object_counts[i]);
                
    RemappingLookup *remapper = new RemappingLookup(nx, ny);
    remapper->setReference(referencepositions, referencecount);
    for (i = 0; i < colourmax; i++) {
        remapper->remap(object_lists[i], object_counts[i], dist);
        memcpy(indices + i * referencecount,
               remapper->getIndex(), referencecount*sizeof(int));
    }    
    delete remapper;

    for (i = 0; i < colourmax; i++) {
        int col = get_base_index(bases[i], bases, colourmax);
        remap_colour(referencepositions, indices + i * referencecount,
                     referencecount,
                     object_lists[i], object_counts[i],
                     intensities, noise, colourmax, col);
    }

    /*
     * go back to images to read out values at the reference positions
     * for missing objects
     */
    for (i = 0; i < colourmax; i++) {
        int col = get_base_index(bases[i], bases, colourmax);
        complete_intensities(referencepositions, 
                             indices + i * referencecount, referencecount, 
                             colourmax, col, 
                             images[i]->getData()->filter_image,
                             images[i]->getData()->noise_image, nx, ny,
                             intensities, noise, fwhm, 0.51,
                             images[i]->getData()->usedTransform);
    }
    
    if (referenceindices != NULL)
        memcpy(referenceindices, indices,
               referencecount * colourmax * sizeof(int));
    delete [] indices;

    for (i = 0; i < colourmax; i++)
        delete [] object_lists[i];
    delete [] object_counts;
    delete [] object_lists;
}


void DataProcessor::displayOffsets()
{
    if (offset_tracker)
        offset_tracker->display(default_transforms);
}
