/*
 * Determine the background noise dispersion of an image.
 *
 * Copyright (c) Solexa 2003-2005, 2006
 * Author: Klaus Maisinger
 * This source file is covered by the "Solexa Public Source License"
 * agreement and bound by the terms therein.
 */

#if defined(__BORLANDC__)
#include <math.h>
#else
#include <cmath>
#endif
#include <cstdio>
#include <cassert>
#include "Maths.h"
#include "Tools.h"
#include "ImageUtil.h"
#include "Optimise.h"
#include "ProcessNoiseEstimate.h"

//---------------------------------------------------------------------------
// image noise estimation
//---------------------------------------------------------------------------

/*
 * determine the maximum value of a histogram and discard values less than
 * lofactor*maximum on the left side and less than hifactor*maximum on the 
 * right side of the maximum (allowance is given for 1-sigma errors on the
 * number counts). return the last valid indices.
 */
static void clip_histogram(int *hist, int n, float lofactor, float hifactor,
                           int *maxindex, int *loindex, int *hiindex)
{
    float threshold;
    int histmax = 0;
    int i;

    // find maximum
    *maxindex = 0;
    for (i = 0; i < n; i++)
        if (hist[i] > histmax) {
            histmax = hist[i];
            *maxindex = i;
        }

    // find first and last index
    *loindex = *maxindex;
    threshold = (histmax-sqrt((float)histmax))*lofactor;
    while ((*loindex-1 >= 0) 
            && (hist[*loindex-1]+sqrt((float)hist[*loindex-1]) >= threshold))
        (*loindex)--;
    *hiindex = *maxindex;
    threshold = (histmax-sqrt((float)histmax))*hifactor;
    while ((*hiindex+1 < n) 
            && (hist[*hiindex+1]+sqrt((float)hist[*hiindex+1]) >= threshold))
        (*hiindex)++;
}

#if 0
/*
 * given a histogram <original> with <on> bins of width <owidth> and
 * a leftmost bin boundary at <omin>, rebin the histogram to a new
 * histogram <hist> with <n> bins of with <width> and leftmost bin
 * boundary <min>. The first bin is [<min>, <min>+<width>).
 */
static void rebin_histogram(int *original, int on, float omin, float owidth,
                            int *hist, int n, float min, float width)
{
    int i;
    for (i = 0; i < n; i++)
        hist[i] = 0;

    int imin = (int)floor((min-omin)/owidth);
    int imax = (int)ceil((min+(n+1)*width-omin)/owidth);
    imin = ((imin < 0) ? 0 : ((imin > on) ? on : imin));
    imax = ((imax < 0) ? 0 : ((imax > on) ? on : imax));
    for (i = imin; i < imax; i++) {
        float arg = omin + (i+0.5) * owidth;
        int index = (int)floor((arg-min)/width);
        if ((index < 0) || (index >= n))
           continue;
        hist[index] += original[i];
    }
}
#endif

/*
 * Auxiliary function to get a percentile from a distribution
 * given by its histogram.
 * <percentile> is quoted as a number between 0. (smallest element) 
 * and 1. (largest element).
 * The histogram <hist> contains <n> equally spaced bins.
 * The bin width is <width>, and the left border of the leftmost bin
 * is given by <minarg>.
 */  
static float get_histogram_percentile(float percentile,
                                      int *hist, int n, float minarg, float width,
                                      float minval = 0., float maxval = 0.) 
{
    int i;
    if (minval == maxval) {
        minval = minarg;
        maxval = minarg + n * width;
    }

    // add up all values in histogram
    long allsum = 0, target;
    for (i = 0; i < n; i++)
        if ((minarg+i*width >= minval) && (minarg+i*width < maxval))
            allsum += hist[i];
    target = (long)(percentile * allsum);

    if (allsum <= 1) {
        printf("Not enough counts in target interval!\n");
        return (maxval+minval)/2.f;
    }
    long sum = 0;
    i = 0;
    while ((i < n) && (sum + hist[i] < target)) {
        if ((minarg+i*width >= minval) && (minarg+i*width < maxval)) 
            sum += hist[i];
        i++;
    }
    float x = minarg + width * i;
    if (i < n) {
        assert(hist[i] > 0);
        x += (width*(target-sum))/hist[i];
    }
#if 0
    printf("Perc %f %6li %6li %6li %5i %5i %f %f %f %f %f\n",
           percentile, target, sum, allsum, i, n, x, minarg, width, minval, maxval);
#endif
    return x;
}

/*
 * Determine the noise dispersion by fitting a Gaussian around the peak
 * of a histogram of pixel values. An initial estimate is obtained from
 * rank statistics.
 * 
 * The whole procedure is a bit of a mess, but it changes so frequently
 * that there is not much point in cleaning it up until it has 
 * stabilised.
 */
void determine_noise_hist_stats(float *image, int n,
                                float *average, float *dispersion)
{
    int i, j;
    /* Specify number of bins per sigma */
    const int nbins = 7;
    /* Avoid small dispersions (less than one camera count). 
     * They are most likely due to blank/saturated/corrupted images. */
    const float minimum_dispersion = 0.9f;
    float filter_av, filter_sdev, rank_av, rank_sdev, back_av, back_sdev;

    /* determine standard deviation and average of original image as initial
     * estimates
     */
    get_array_stats(image, n, &filter_av, &filter_sdev);

    if (filter_sdev < minimum_dispersion) {
        *average = filter_av;
        *dispersion = minimum_dispersion;
        return;
    }

    /*
     * Create a histogram with a very fine initial binning.
     * The binning at this stage is mostly a performance
     * hack for the quick calculation of rank statistics.
     * 
     * The following binning is chosen:
     * Bin by a factor of 10 more finely than any interval; an additional
     * factor of 5. to take into account that the initial standard deviation
     * is going to overestimate the true noise dispersion.
     * Round down to powers of 10. A bin width of less than 0.1 is 
     * not useful as camera data only comes in steps of 1 unit.
     */
    float global_width, global_minarg;
    int n_global_bins;
    global_width = pow(10.f, floor(log10(filter_sdev/nbins/50.f)));
    if (global_width < 0.1f)
        global_width = 0.1f;
    global_minarg = (floor((filter_av-2.5f*filter_sdev)/global_width)-0.5f) 
                    * global_width;
    n_global_bins = (int)ceil(5.f*filter_sdev/global_width);

#if defined(DEBUG)
    printf("Global Av %g Sdev %g Minval %g Cell %g n %i\n",
           filter_av, filter_sdev,
           global_minarg, global_width, n_global_bins);
#endif

    /* perform binning to obtain inital histogram */
    int *global_hist = new int [n_global_bins];
    bin(image, n, global_hist, n_global_bins, global_minarg, global_width);
#if defined(DEBUG) && 0
    long sum = 0.;
    for (i = 0; i < n_global_bins; i++)
        printf("%7.2f %5i %6i\n", global_minarg+(i+0.5)*global_width,
               global_hist[i], sum+= global_hist[i]); 
#endif
    int iter2 = 0;
    rank_av = filter_av;
    rank_sdev = filter_sdev;

    /* Now determine background and noise from rank statistics.
     * Assume background is given by median and noise by the 68th percentile 
     * of the pixel distribution.
     * Iterate and discard extreme values beyond 3 sigma to converge on an estimate.
     */
    float old_av, old_sdev;
    do {
        float minval = rank_av - 3.f*rank_sdev;
        float maxval = rank_av + 3.f*rank_sdev;
        old_av = rank_av;
        old_sdev = rank_sdev;
        rank_av = get_histogram_percentile(0.5f, global_hist, n_global_bins,
                                           global_minarg, global_width,
                                           minval, maxval);
        rank_sdev = (get_histogram_percentile(0.84f, global_hist, n_global_bins,
                                              global_minarg, global_width,
                                              minval, maxval)
                    - get_histogram_percentile(0.16f, global_hist, n_global_bins,
                                               global_minarg, global_width,
                                               minval, maxval))/2.f;
#if defined(DEBUG)
        printf("Rank Av %f Sdev %f\n", rank_av, rank_sdev); 
#endif                                     
        iter2++;
    } while ((iter2 < 10) && (fabs((rank_sdev-old_sdev)/old_sdev) > 0.01f));                                  

    back_av = rank_av;
    back_sdev = rank_sdev;
    if (rank_sdev < minimum_dispersion) {
        back_av = filter_av;
        back_sdev = filter_sdev;
    }

    const float y1sigma = exp(-0.5);
    int ntotbins = 4*nbins+1;
    int *hist = new int [ntotbins];
    float sigma = back_sdev;
    float centre = back_av;
    float maxval = 0.;

    /* this loop is currently not iterated.
     * TODO: Remove it.
     */
    int iter = 0;
    do {
        int lowidx, higidx, maxidx;
        float width, minarg;

        /*
         * Determine the pixel histogram for the subsequent fit.
         * Discard wings of the histogram since they are potentially not 
         * noise-dominated.
         */
        width = sigma / nbins;
        minarg = centre - 2.*sigma - 0.5*width;
        bin(image, n, hist, ntotbins, minarg, width);
        // discard low values in histogram since they are potentially not
        // noise-dominated
        clip_histogram(hist, ntotbins, y1sigma, y1sigma,
                       &maxidx, &lowidx, &higidx);

        centre = minarg+(maxidx+0.5)*width;
        sigma = ((higidx-lowidx)/2.+0.5)*width;
        maxval = hist[maxidx]/width;

#if defined(DEBUG)
        printf("  bin-width %g minarg %g centre %g sigma %g\n",
               width, minarg, centre, sigma);
        printf("  hist: \n");
        for (i = nbins; i < ntotbins-nbins; i++)
            printf("%7.2f %7.2f\n", minarg+width/2+i*width, hist[i]/width);
        printf("  low hi max: %i %i %i %7.2f %7.2f %7.2f\n",
               lowidx, higidx, maxidx,
               minarg+width/2+lowidx*width, minarg+width/2+higidx*width,
               minarg+width/2+maxidx*width);
#endif // defined(DEBUG)
        if (higidx - lowidx < 5) {
#if defined(DEBUG)
            fprintf(stderr,
                    "Not enough histogram bins for fitting (%i %i).\n",
                    lowidx, higidx);
#endif // defined(DEBUG)
            goto noise_hist_terminate;
        }

        assert(width > 0.);
        float twosigma = (higidx - lowidx)*width;
        int isig = int(ceil(twosigma/2./width));

        // in certain conditions in high-density simulations the histogram
        // of the filtered image has a significant shoulder left of the peak.
        // try to cut out this shoulder, but make sure that enough bins
        //  are left over.
        //clip_histogram(hist, ntotbins, 0.8, y1sigma,
        //	       &maxidx, &lowidx, &higidx);
        //if ((lowidx == maxidx) && (lowidx > 0))
        //    lowidx -= 1;
#if defined(DEBUG)
        printf("  low hi max: %i %i %i\n", lowidx, higidx, maxidx);
#endif

        /*
         * Lots of different (fairly ad-hoc) optimisations for corner cases
         * to avoid being contaminated by the negative shoulder of the
         * histogram.
         */
        {
            /*
             * first check for local maxima to the right of the 
             * current maximum. make sure that there are at least
             * two points to the right of and above a local minimum,
             * or two points below and to the left of a local maximum
             */  
            int ioff = 0;
            for (i = 1; i <= higidx-maxidx; i++)
                if (hist[maxidx + i] * 0.9 > hist[maxidx+i-1]) {
                    int count = 0;
                    for (j = maxidx+1; j < maxidx+i; j++)
                        if (hist[j] < 0.9*hist[maxidx+i])
                            count++;
                    if (count >= 2)
                    ioff = i;
                    count = 0;
                    for (j = maxidx+i; j <= higidx; j++)
                        if (hist[j]*0.9 > hist[maxidx+i-1]) 
                            count ++;
                    if (count >= 2)
                        ioff = i;
                }
            if (ioff > 0) {
#if defined(DEBUG)
                printf("  Update: local max, low hi max: %i %i %i\n",
                       lowidx, higidx, maxidx);
#endif
                maxidx += ioff;
            }
            /*
             * now check that the peak isn't too broad; otherwise
             * use outermost point of the peak that is still within 0.8
             * of the global maximum
             */
            ioff = 0;
            for (i = isig; i <= higidx-maxidx; i++)
                if (hist[maxidx + i] > 0.9*hist[maxidx])
                    ioff = i;

            if (ioff > 0) {
                maxidx = maxidx + ioff;
                lowidx = maxidx-1;
                if (hist[lowidx] > hist[maxidx])
                    maxidx = lowidx;

#if defined(DEBUG)
                printf("  Update: broad peak, low hi max: %i %i %i\n",
                       lowidx, higidx, maxidx);
#endif
            }
            ioff = 0;
            /*
             * now cut down on the left shoulder. In fact, the further
             * the shoulder extends, the more cut off.
             */
            assert(lowidx >= 0);
            for (i = isig; i <= maxidx-lowidx; i++)
                if (hist[maxidx - i] > 0.8*hist[maxidx])
                    ioff = i;
            if (ioff > 0) {
                lowidx = maxidx - isig;
                if (ioff > 2*isig)
                    ioff = 2*isig;
                if (ioff > isig)
                    lowidx += ioff-isig;
#if defined(DEBUG)
                printf("  Update: shoulder reduced, low hi max: %i %i %i\n",
                       lowidx, higidx, maxidx);
#endif
            }
            assert(lowidx >= 0);
            /*
             * make sure we capture the slope on the right properly
             */
            while ((hist[higidx] > y1sigma * hist[maxidx])
                   && (higidx + 1 < ntotbins))
                higidx++;
            /*
             * make sure that the left shoulder is no larger than the right
             */
            if (higidx - maxidx < maxidx - lowidx)
                lowidx = maxidx - (higidx-maxidx);
        }

        while ((higidx-lowidx+1 < 5) && (lowidx-1 > 0))
            lowidx -= 1;

        assert(lowidx >= 0);
        assert(higidx < ntotbins);
        assert((maxidx >= lowidx) && (maxidx <= higidx));
        /*
         * now fit a 1d-Gaussian to the histogram by using Levenberg-Marquardt
         * optimisation
         */
        int ndata = higidx-lowidx+1;
        float *x = new float [ndata];
        float *y = new float [ndata];
        float *sig = new float [ndata];
#if defined(DEBUG)
        printf("  low hi max: %i %i %i ndata %i bins %i\n",
               lowidx, higidx, maxidx, ndata, nbins); 
#endif
        for (i = 0; i < ndata; i++) {
            int idx = i + lowidx;
            x[i] = minarg + (idx+0.5f) * width;
            y[i] = (float)hist[idx]/width; // normalise by bin width
            sig[i] = sqrt(y[i]); // Poisson standard deviation
            sig[i] = MAX(sig[i], 1.f);
            sig[i] = 1;
        }
        // initial estimates
        int na = 3;
        float *a = new float [na];
        int *ia = new int [na];
        for (i = 0; i < na; i++)
            ia[i]  = 1;
        if (maxidx == lowidx)
            ia[1] = 0;
        a[0] = y[maxidx-lowidx];
        a[1] = x[maxidx-lowidx];
        a[2] = ndata * width / 2.3f;


#if defined(DEBUG)
        printf("  hist: \n");
        for (i = 0; i < ndata; i++)
            printf(" %7.2f %7.2f %7.2f \n", x[i], y[i], sig[i]);
        printf("  Input:  f(x)=a*exp(-(x-b)*(x-b)/(2.*s*s)), a=%f, b=%f, s=%f\n",
               a[0], a[1], a[2]);
#endif
        float chisq = 0.;
        LM_1d(x, y, sig, ndata, a, ia, na, &chisq, LM_gauss_1d_der);
#if defined(DEBUG)
        printf("  chisquared: %f\n", chisq);
        printf("  Fitted: f(x)=a*exp(-(x-b)*(x-b)/(2.*s*s)), a=%f, b=%f, s=%f\n",
               a[0], a[1], a[2]);
#endif

        // catch failure of fitting procedure and reset to input values
        if ((a[2] < 1.5*sigma) && (a[2] < 1.5* rank_sdev)
            && (a[2] > minimum_dispersion) && (a[2] > rank_sdev/3.)
            && (fabs(a[1]-centre) < back_sdev)
            && (fabs((a[0]-y[maxidx-lowidx])/y[maxidx-lowidx]) < .25)) {
            back_av = a[1];
            back_sdev = a[2];
            maxval = a[0];
        }
        else {
#if defined(DEBUG)
            printf("Fit (%f %f %f) too far from input (%f %f %f); resetting (%f %f)\n",
                   a[0], a[1], a[2], y[maxidx-lowidx], x[maxidx-lowidx], sigma,
                   back_av, back_sdev);
#endif
        }
#if defined(DEBUG)
        printf("  Fitted: f(x)=a*exp(-(x-b)*(x-b)/(2.*s*s)), a=%f, b=%f, s=%f\n",
               maxval, back_av, back_sdev);
#endif

        delete [] x; delete [] y; delete [] sig; delete [] a; delete [] ia;

        iter ++;

    } while (iter < 1);

  noise_hist_terminate:
    delete [] hist;
    delete [] global_hist;

    /* a very small std dev is most likely due to image corruption: */
    if (fabs(back_sdev) < minimum_dispersion) {
        back_av = rank_av;
        back_sdev = rank_sdev;
    }
    *average = back_av;
    *dispersion = back_sdev;
}

/*
 * determine the image threshold to use for the binarisation of the image
 */
void determine_region_noise(float *image, int nx, int ny,
                            int x, int y, int sizex, int sizey,
                            float *background, float *noise)
{
#if defined(DEBUG)
    printf("x y: %i %i\n", x, y);
#endif

    // copy relevant image region to single array
    float *subimage = new float [sizex * sizey ];
    copy_image_region(image, nx, ny, subimage, sizex, sizey, x, y);

    determine_noise_hist_stats(subimage, sizex*sizey, background, noise);

#if defined(DEBUG) 
    printf("final background/noise (%4i %4i) %7.2f %7.2f\n",
           x, y, *background, *noise);
#endif

    delete [] subimage;
}

/*
 * estimate the local noise dispersion of the image.
 * the image is split into a number of tiles that need to be large enough
 * to allow statistically significant estimates of image properties,
 * but small enough to catch local variations in background and noise.
 */
void estimate_noise_image(float *image, int nx, int ny, 
                          float *noise, float *background)
{
    int sizex = get_local_image_size(nx, LOCAL_SIZE);
    int sizey = get_local_image_size(ny, LOCAL_SIZE);
    int ntx = nx / sizex;
    int nty = ny / sizey;
    int i, j, x, y, nsx, nsy;

    /* split the image into a set of tiles */
    for (i = 0; i < ntx; i++)
        for (j = 0; j < nty; j++) {
            x = i * sizex;
            y = j * sizey;
            nsx = sizex;
            nsy = sizey;
            /* add extra pixels if not all pixels have been covered
             * by the last tile */
            if ((i == ntx-1) && (nx - x > nsx))
                nsx = nx - x;
            if ((j == nty-1) && (ny - y > nsy))
                nsy = ny - y;
            float noise_val, background_val;

            /* estimate noise dispersion in local image tile */
            determine_region_noise(image, nx, ny, x, y, nsx, nsy,
                                   &background_val, &noise_val);
            /* copy local noise dispersion to noise image */
            set_image_region_value(noise, nx, ny, x, y, nsx, nsy, noise_val);
      
            /* copy local noise dispersion to background image */
            if (background)
                set_image_region_value(background, nx, ny, x, y, nsx, nsy,
                                       background_val);
        
        }
}

//---------------------------------------------------------------------------
// image noise estimate (obsolete implementations)
//---------------------------------------------------------------------------

/*
 * given an array of data pixels, try to determine the average and dispersion
 * of any Gaussian noise component in the data. The algorithm assumes
 * that the noise component is the dominant image component and uses
 * an iterative k-sigma clipping algorithm to determine the standard
 * deviation of the background fluctuations.
 */
void determine_noise_stats(float *image, int n,
                           float *average, float *dispersion)
{
    float back_av, back_sdev, old_av, old_sdev;
    float value, val_min, val_max;
    int   num, iter, maxiter;
    float factor, tolerance;

    // determine standard deviation and average of original image
    get_array_stats(image, n, &back_av, &back_sdev);

#if defined(DEBUG)
    printf("  filter av: %f sdev: %f\n", back_av, back_sdev);
#endif

    /*
     * find average of background by only considering points within 1 sigma;
     * this is to exploit the fact that these points are heavily dominated by
     * the background noise.
     * this part of the algorithm converges trivially.
     */
    factor = 1.f;
    tolerance = 0.005f;
    maxiter = 8;
    iter = 0;
    int i;

    do {
        old_av = back_av;
        back_av = 0.;
        num = 0;
        val_min = old_av - factor*back_sdev;
        val_max = old_av + factor*back_sdev;
        for (i = 0; i < n; i++) {
            value = image[i];
            if ((value > val_min) && (value < val_max)) {
                back_av += value;
                num++;
            }
        }
        if (num == 0)
            break;
        back_av /= num;
        iter++;
#if defined(DEBUG)
        printf("  background iter %2i: average %f %f\n", iter, back_av,
               fabs((back_av-old_av)/back_av));
#endif
    } while ((iter < maxiter) && (fabs((back_av-old_av)/back_av) > tolerance));

    /* now determine the standard deviation; remove points beyond factor*sigma
     * to reduce bias introduced by molecules.
     * the iteration seems to be fairly stable for sigmas approaching 2.
     * Take 1.7 sigma. This corresponds to 0.7688 of the true standard deviation,
     * a function value of 0.2357 of the peak value of the normal distribution
     * and about 91% (0.9108) of the integrated area.
     */
    factor = 1.7f;
    float integrated = 0.7688f;    // obtained by numerical integration
    float diff = 10., old_diff;
    tolerance = 0.01f;
    maxiter = 10;
    iter = 0;
#if 0
    static int count = 0;
    printf("count %i\n", count);
    count ++;
#endif

    float *imageb = new float [n];
    for (i = 0; i < n; i++) {
        value = fabs(image[i]-back_av);
        imageb[i] = value * value;
    }

    do {
        old_sdev = back_sdev;
        old_diff = diff;
        back_sdev = 0.;
        num = 0;
#if 0
        val_min = back_av - factor*old_sdev;
        val_max = back_av + factor*old_sdev;
        for (i = 0; i < n; i++) {
            value = image[i];
            if ((value > val_min) && (value < val_max)) {
                back_sdev += value*value;
                num++;
            }
        }
        back_sdev /= num;
        back_sdev = sqrt(fabs(back_sdev - back_av*back_av)) / integrated;
#else
        val_max = factor*factor*old_sdev*old_sdev;
        for (i = 0; i < n; i++) {
            if (imageb[i] < val_max) {
                back_sdev += imageb[i];
                num++;
            }
        }
        if (num == 0)
            break;
        back_sdev /= num;
        back_sdev = sqrt(fabs(back_sdev)) / integrated;
#endif

#if defined(DEBUG)
        printf("  background iter %2i: sdev    %f %f\n", iter, back_sdev, diff);
#endif

        iter++;
        diff = fabs((back_sdev-old_sdev)/back_sdev);
        // if algorithm does not converge, back out
        if (diff > old_diff) {
            back_sdev = old_sdev;
            break;
        }
    } while ((iter < maxiter) && (diff > tolerance));

    delete [] imageb;

    *average = back_av;
    *dispersion = back_sdev;
}
