//  Time-stamp: <26 Jun 2006 16:48 BST ProcessObjectDetection.cpp km>
/*
 * Routines to detect objects on a thresholded image
 *
 * Copyright (c) Solexa 2003-2005, 2006
 * Author: Klaus Maisinger
 * This source file is covered by the "Solexa Public Source License"
 * agreement and bound by the terms therein.
 */

#if defined(__BORLANDC__)
#include <math.h>
#include <string.h>
#else
#include <cmath>
#include <cstring>
#endif
#include <cstdio>
#include <cassert>
#include "Maths.h"
#include "Tools.h"
#include "ProcessObjectDetection.h"

#define BINARY_DEFAULT_VALUE -1000.

//---------------------------------------------------------------------------
// pattern recognition
//---------------------------------------------------------------------------

/*
 * Given an initial pixel (x0, y0) on the boundary of a connected area,
 * return an array of pixels delineating the boundary of the area (or exclusion set)
 * such that each consecutive  pixel is just one coordinate change away from
 * the previous (i.e. no diagonal steps) and the order of the pixels corresponds
 * to a clockwise walk.
 * Pixel values must be above threshold for pixels to belong to the area.
 */
void find_pixel_boundary(float *spatial, int nsx, int nsy,
                         int x0, int y0, float threshold,
                         st_pixel *boundary, int nmax, int *size)
{
    /* Directions are labelled as follows:
     *
     *  y            0
     *  ^            |
     *  |        3 --.-- 1
     *  .--> x       |
     *               2
     */
    const int up = 0, right = 1, down = 2, left = 3, none = -1;
    int direction0, new_direction, return_direction;
    int x, y, direction;
    int num;

    *size = 0;
    if (spatial[x0*nsy+y0] <= threshold)
        return;

    /* Identify an initial direction. Since the pixels on the boundary are
     * supposed to describe a closed loop and directions matter, this has 
     * to be the direction from the previous pixel on the periodic boundary,
     * i.e. the last pixel in the whole list. So we need to pick a suitable pixel.
     * Since the parent routine is most likely scanning along the x and y axis,
     * a direction opposite to these is most likely to contain undetected pixels.
     * Hence, as an initial guess go left (i.e. last pixel is to the right)
     * instead of up or right.
     * Furthermore, make sure that
     *   1. no pixel is to the left of direction from the previous pixel
     *      (otherwise we might walk the boundary counter-clockwise)
     *   2. there is a previous pixel in the first place
     */
#if 0

    direction0 = left;
    if ((x0+1 < nsx) && (spatial[(x0+1)*nsy+(y0)] <= threshold)) {
        direction0 = down;
        if ((y0+1 < nsy) && (spatial[(x0)*nsy+(y0+1)] <= threshold)) {
            direction0 = right;
            if ((x0-1 >= 0) && (spatial[(x0-1)*nsy+(y0)] <= threshold)) {
                direction0 = up;
            }
        }
    }
  
#else
    direction0 = left;
    /* check if there is actually a previous pixel from the initial direction */
    if ((x0+1 < nsx) && (spatial[(x0+1)*nsy+(y0)] > threshold)) {
        /* there is a pixel, now make sure that there is no pixel to the left
         * of the previous pixel */
        if ((y0-1 >= 0) && (spatial[(x0)*nsy+(y0-1)] > threshold)) {
            direction0 = up;
            if ((x0-1 >= 0) && (spatial[(x0-1)*nsy+(y0)] > threshold)) {
                direction0 = right;
                if ((y0+1 < nsy) && (spatial[(x0)*nsy+(y0+1)] > threshold)) {
                    fprintf(stderr, "find_pixel_boundary(): Point not on boundary!");
                    return;
                }
            }
        }
    }
    else {
        /* there is no previous pixel, so look for another direction */ 
        if ((y0+1 < nsy) && (spatial[(x0)*nsy+(y0+1)] > threshold))
            direction0 = down;
        else if ((x0-1 >= 0) && (spatial[(x0-1)*nsy+(y0)] > threshold))
            direction0 = right;
        else if ((y0-1 >= 0) && (spatial[(x0)*nsy+(y0-1)] > threshold))
            direction0 = up;
    }
#endif
  

    /* now walk along the boundary (clockwise) */
    x = x0; y = y0;
    direction = direction0;
    num = 0;
    do {
        /* add pixel to list */
        boundary[num].x = (unsigned short)x; boundary[num].y = (unsigned short)y;
        boundary[num].value = 1.;

        return_direction = (direction+2) % 4; // where we came from
        /* Determine a new direction to walk in.
         * Always try to turn left if possible. */
        direction = return_direction;
        new_direction = none;
        do {
            direction = (direction+1) % 4; /* turn right */
            switch (direction) {
            case up:
                if (((y+1) < nsy) && (spatial[(x)*nsy+(y+1)] > threshold)) {
                    y++; new_direction = direction;
                } break;
            case right:
                if (((x+1) < nsx) && (spatial[(x+1)*nsy+y] > threshold)) {
                    x++; new_direction = direction;
                } break;
            case down:
                if (((y-1) >= 0) && (spatial[(x)*nsy+(y-1)] > threshold)) {
                    y--; new_direction = direction;
                } break;
            case left:
                if (((x-1) >= 0) && (spatial[(x-1)*nsy+y] > threshold)) {
                    x--; new_direction = direction;
                } break;
            }
        } while ((direction != return_direction) && (new_direction == none));
        num++;
    } while ((new_direction != none) && (num < nmax)
             && ((x != x0) || (y != y0) || (direction0 != direction)));
    if (num == nmax)
        fprintf(stderr, "WARNING: Object boundary too long (%i)!\n", num);
    *size = num;
    return;
}

// find the maximum in an array of pixels
void find_pixel_maximum(st_pixel *pixels, int n_pixels, st_pixel *max_pixel)
{
    int i;
    max_pixel->x = 0; max_pixel->y = 0; max_pixel->value = 0;
    for (i = 0; i < n_pixels; i++)
        if (pixels[i].value > max_pixel->value)
            memcpy(max_pixel, &pixels[i], sizeof(st_pixel));
}

// find the smallest square containing all pixels
void find_pixel_extent(st_pixel *pixels, int n_pixels,
                       unsigned short *xmin, unsigned short *xmax,
                       unsigned short *ymin, unsigned short *ymax)
{
    assert(n_pixels > 0);
    *xmin = pixels[0].x; *xmax = pixels[0].x;
    *ymin = pixels[0].y; *ymax = pixels[0].y;
    int i;
    for (i = 1; i < n_pixels; i++) {
        if (pixels[i].x < *xmin) *xmin = pixels[i].x;
        if (pixels[i].x > *xmax) *xmax = pixels[i].x;
        if (pixels[i].y < *ymin) *ymin = pixels[i].y;
        if (pixels[i].y > *ymax) *ymax = pixels[i].y;
    }
}

void list_pixels(st_pixel *pixels, int n_pixels)
{
    int i;
    for (i = 0; i < n_pixels; i++)
        printf("%2d: %2d %2d %7.1f\n", i, pixels[i].x, pixels[i].y, pixels[i].value);
}

/*
 * Given a <boundary> of <nb> pixel delineating a compact region
 * on the binarised <nsx>*<nsy>-image <spatial>, return an array containing
 * all <na> pixels lying within this <area>. <xmin>, <xmax>, <ymin>, <ymax>
 * indicate the corner points of the smallest rectangle the area fits in.
 */
void find_pixel_area(st_pixel *boundary, int nb,
                     unsigned short xmin, unsigned short xmax,
                     unsigned short ymin, unsigned short ymax,
                     float *spatial, int nsx, int nsy,
                     st_pixel *area, int *na)
{
    const unsigned short BOUNDARY_UPPER = (1<<1);
    const unsigned short BOUNDARY_LOWER = (1<<2);
    int i, x, y;
    int nx = xmax-xmin+1;
    int ny = ymax-ymin+1;
    unsigned short *bd_array = new unsigned short [nx*ny];

    for (i = 0; i < nx*ny; i++)
        bd_array[i] = 0;

    /* create 2-d array with points on upper and lower boundary
     * marking the area */
    int prev_x, next_x, prev_y, next_y;
    x = boundary[nb-1].x; y = boundary[nb-1].y;
    for (i = 0; i < nb; i++) {
        prev_x = x; prev_y = y;
        x = boundary[i].x; y = boundary[i].y;
        next_x = boundary[(i+1)%nb].x; next_y = boundary[(i+1)%nb].y;

        // mark points on upper boundary (walk is clockwise, so left to right)
        if ((y >= prev_y) && (y >= next_y) && (next_x >= prev_x))
            bd_array[(x-xmin)*ny+(y-ymin)] |= BOUNDARY_UPPER;
        // mark points on lower boundary (walk is clockwise, so right to left)
        if ((y <= prev_y) && (y <= next_y) && (next_x <= prev_x))
            bd_array[(x-xmin)*ny+(y-ymin)] |= BOUNDARY_LOWER;
    }

    /* for each x-value, now scan pixels along y-direction
     * between lower and upper boundary */
    int inarea;
    int idx = 0;
    for (x = xmin; x <= xmax; x++) {
        unsigned short *bd_line = bd_array+(x-xmin)*ny;

        inarea = 0;
        for (y = ymin; y <= ymax; y++) {
            if (bd_line[y-ymin] & BOUNDARY_LOWER) // entering area
                inarea = 1;
            if (inarea) {  // found a pixel
                area[idx].x = x; area[idx].y = y; area[idx].value = spatial[x*nsy+y];
                idx++;
            }
            if (bd_line[y-ymin] & BOUNDARY_UPPER) // leaving area
                inarea = 0;
        }
    }
    delete [] bd_array;

    *na = idx;
    return;
}

void fill_pixel_area(st_pixel *area, int na, float *spatial,
                     int nsx, int nsy, float value)
{
    int i, x, y;
    for (i = 0; i < na; i++) {
        x = area[i].x; y = area[i].y;
        spatial[x*nsy+y] = value;
    }
}

void free_object(st_object *obj)
{
    if (obj->nb > 0)
        delete [] obj->boundary;
    if (obj->na > 0)
        delete [] obj->area;
    obj->xmin = obj->xmax = 0;
    obj->ymin = obj->ymax = 0;
    obj->nb = 0;
    obj->na = 0;
}

/*
 * copy source object to target and shift all coordinate references
 * by (x0, y0)
 */
void shift_object(st_object *src, st_object *target,
                  unsigned short x0, unsigned short y0)
{
    int i;
    target->na = src->na;
    target->nb = src->nb;
    target->xmin = src->xmin+x0;
    target->xmax = src->xmax+x0;
    target->ymin = src->ymin+y0;
    target->ymax = src->ymax+y0;
    target->boundary = src->boundary;
    target->area = src->area;
    for (i = 0; i < target->nb; i++) {
        target->boundary[i].x += x0;
        target->boundary[i].y += y0;
    }
    for (i = 0; i < target->na; i++) {
        target->area[i].x += x0;
        target->area[i].y += y0;
    }
    target->max_pixel = src->max_pixel;
    target->max_pixel.x += x0;
    target->max_pixel.y += y0;
}

/*
 * debugging aid: write out image area as a table of pixel values
 */
void display_object(st_object *a)
{
    int i;
    int nx = a->xmax - a->xmin + 1;
    int ny = a->ymax - a->ymin + 1;
    float *sp = new float [ nx*ny ];

    for (i = 0; i < nx*ny; i++)
    sp[i] = 0;
    for (i = 0; i < a->na; i++) {
        int x = a->area[i].x - a->xmin;
        int y = a->area[i].y - a->ymin;
        sp[x*ny+y] = a->area[i].value;
    }
    display_float_image_region(sp, nx, ny, 0, 0, nx, ny, a->xmin, a->ymin);
    delete [] sp;
}

void set_object_boundary(st_object *obj, st_pixel *boundary, int nb)
{
    assert(nb > 0);
    obj->nb = nb;
    obj->boundary = new st_pixel [nb];
    memcpy(obj->boundary, boundary, nb*sizeof(st_pixel));
}

void set_object_area(st_object *obj, st_pixel *area, int na)
{
    assert(na > 0);
    obj->na = na;
    obj->area = new st_pixel [na];
    memcpy(obj->area, area, na*sizeof(st_pixel));
}

/*
 * apply the object recognition on a thresholded image and identify
 * new objects above a threshold. 
 * the image <spatial> has dimensions <ns>*<ns> and will be set
 * to zero after exit.
 */
void identify_objects(float *spatial, int nsx, int nsy,
                      st_object *objs, int nmax, int *nobjs,
                      unsigned short min_mol_size, float threshold)
{
    int nb, na;
    int nbmax = 4*(nsx+nsy);
    int namax = nsx * nsy;
    st_pixel *boundary = new st_pixel [ nbmax ];
    st_pixel *area = new st_pixel [ namax ];

    int x, y;
    for (x = 0; x < nsx; x++)
        for (y = 0; y < nsy; y++) {
#if 0
            if (x == 168 && (y == 994 || y == 995)) {
                printf("identify %i %i %f %f\n", x, y, spatial[x*nsy+y], threshold);
            }
#endif 
            if (spatial[x*nsy+y] > threshold) {
                if (*nobjs >= nmax) {
                    fprintf(stderr, "WARNING: Too many objects!");
                    return;
                }
                st_object *obj = &objs[*nobjs];
                find_pixel_boundary(spatial, nsx, nsy, x, y, threshold,
                                    boundary, nbmax, &nb);
                set_object_boundary(obj, boundary, nb);
                find_pixel_extent(obj->boundary, obj->nb,
                          &obj->xmin, &obj->xmax, &obj->ymin, &obj->ymax);
                find_pixel_area(obj->boundary, obj->nb, obj->xmin, obj->xmax,
                                obj->ymin, obj->ymax, spatial, nsx, nsy,
                                area, &na);
                if (nb == nbmax) {
                    printf("Large Object: %i %i %i %i %i %i\n", x, y, nb, nbmax, na, namax); 
#if 0
                    for (i = 0; i < nb; i++)
                        printf("%i %i %i %i %i\n", x, y, i, boundary[i].x, boundary[i].y);
#endif
                }
#if 0
                if (x == 168 && (y == 994 || y == 995)) {
                    printf("yikes %i %i %i <= %i %i <= %i\n", x, y, nb, nbmax, na, namax);
                    int flag = 0;
                    for (i = 0; i < na; i++)
                        if (x == area[i].x && y == area[i].y)
                            flag = 1;
                    if (!flag)
                    printf("D... ! First pixel missed.\n");
                } 
                if ((na <= 0)) {
                    int i;
                    printf("YIKES  %i %i %i %i %i %i %i %i\n",
                           na, nb, x, y, obj->xmin, obj->xmax,
                           obj->ymin, obj->ymax);
                    for (i = 0; i < nb; i++)
                        printf("%i %i %f\n",
                                obj->boundary[i].x, obj->boundary[i].y,
                                obj->boundary[i].value);
                }
#endif       
                set_object_area(obj, area, na);
                find_pixel_maximum(obj->area, obj->na, &obj->max_pixel);
                // set all image pixels within the area to zero to avoid counting
                // them again
                fill_pixel_area(obj->area, obj->na, spatial, nsx, nsy, BINARY_DEFAULT_VALUE);
                if ((nb == nbmax) || (na == namax))
                    spatial[x*nsy+y] = BINARY_DEFAULT_VALUE;
#if 0
                    printf("Area %i: boundary %i area %i\n", *nobjs, area.nb, area.na);
                    printf("  max: %i %i %i\n", area.max_pixel.x, area.max_pixel.y,
                           area.max_pixel.value);
#endif
                // discard objects with too few pixels
                if (obj->na >= min_mol_size)
                    (*nobjs)++;
                else
                    free_object(obj);
            }
        }
    delete [] area;
    delete [] boundary;
}

//---------------------------------------------------------------------------
// deblending
//---------------------------------------------------------------------------

void find_local_maxima(st_pixel *maxima, int nm_max, int *nmax,
                       float *sp, int nsx, int nsy,
                       float threshold, float min_dist, float fudge)
{
    int size = (int)floor(min_dist);
    int xmin, xmax, ymin, ymax, x2, y2;
    int flag;
    float value, fudge_value;
    st_pixel global_max;

    *nmax = 0;
    global_max.x = 0;
    global_max.y = 0;
    global_max.value = 0.;
    int x, y;
    for (x = 0;  x < nsx; x++)
        for (y = 0;  y < nsy; y++) {
            value = sp[x*nsy+y];
            fudge_value = fudge * value;
            if (value > threshold) {
                if (value > global_max.value) {
                    global_max.x = x; global_max.y = y;
                    global_max.value = value;
                }
                flag = 1;
                xmin = MAX(0, x-size); xmax = MIN(nsx-1, x+size);
                ymin = MAX(0, y-size); ymax = MIN(nsy-1, y+size);
                x2 = xmin;
                while ((flag) && (x2 <= xmax)) {
                    for (y2 = ymin; y2 <= ymax; y2++) {
                        if (((float)((x2-x)*(x2-x)+(y2-y)*(y2-y)) < min_dist*min_dist)
                            && ((x != x2) || (y != y2))
                            && (sp[x2*nsy+y2] > fudge_value))
                            flag = 0;
                    }
                    x2++;
                }

#if defined(SIMULATION)
                // if only four points contributed to this area, check for saddle
                // points. possible configurations:
                //   xoo  oox
                //   oxo  oxo
                //   oox  xoo
                if ((flag) && (min_dist < sqrt(2.))) {
                    if ((x-1 >= 0) && (y-1 >= 0) && (x+1 < nsx) && (y+1 < nsy)) {
                        if ((sp[(x-1)*nsy+(y-1)] > fudge_value)
                            && (sp[(x+1)*nsy+(y+1)] > fudge_value))
                            flag = 0;
                        else if ((sp[(x-1)*nsy+(y+1)] > fudge_value)
                                 && (sp[(x+1)*nsy+(y-1)] > fudge_value))
                            flag = 0;
                    }
                }
#endif

                // found local maximum
                if ((flag == 1) && (*nmax < nm_max)) {
                    maxima[(*nmax)].x = x; maxima[(*nmax)].y = y;
                    maxima[(*nmax)].value = value;
                    (*nmax)++;
                }
            }
        }

    // now make sure the global maximum is included in the list
    // (depending on the fudge factor, this may not be the case)
    int i;
    flag = 0;
    for (i = 0; i < *nmax; i++) {
        if ((maxima[i].x == global_max.x) && (maxima[i].y == global_max.y))
            flag = 1;
    }
    if ((!flag) && (*nmax < nm_max)) {
        maxima[(*nmax)].x = global_max.x;
        maxima[(*nmax)].y = global_max.y;
        maxima[(*nmax)].value = global_max.value;
        (*nmax)++;
    }
}


/*
 * given a list of local <maxima> and an image region <sp> of
 * dimension <nsx>*<nsy>, create a set of areas around each maximum
 * by assigning each pixel to the maximum that is closest to it
 */
void split_local_maxima(st_object *objs, int nmax, int *nobjs,
                        float *sp, int nsx, int nsy,
                        float threshold, 
                        unsigned short min_mol_size, 
                        st_pixel *maxima, int n_maxima)
{
    if (n_maxima < 2)
        return;
    // for each pixel, find nearest maximum;
    // the index of the nearest maximum is stored in the 
    // array <basins>. also keep track of how many pixels were
    // assigned to each maximum using the array <counts>
    int i, j, k;
    int *basins = new int [nsx*nsy];
    int *counts = new int [n_maxima];
    int *index = new int [n_maxima];
    for (k = 0; k < n_maxima; k++) {
        counts[k] = 0; index[k] = 0;
    }
    for (i = 0; i < nsx; i++)
        for (j = 0; j < nsy; j++) {
            basins[i*nsy+j] = 0;
            if (sp[i*nsy+j] > threshold) {
                int mindist2 = nsx*nsx+nsy*nsy;
                for (k = 0; k < n_maxima; k++) {
                    int xdist = i - maxima[k].x;
                    int ydist = j - maxima[k].y;
                    int dist2 = xdist*xdist+ydist*ydist;
                    if (dist2 < mindist2) {
                        mindist2 = dist2;
                        basins[i*nsy+j] = k+1;
                    }
                }
                counts[basins[i*nsy+j]-1] ++;
            }
        }

    // now allocate arrays to hold all pixels for the maxima.
    // make sure that the maximal number of objects is not exceeded.
    int k2 = 0, na, nnew = 0;
    for (k = 0; k < n_maxima; k++) {
        na = counts[k];
        if (na >= min_mol_size) {
            index[k] = k2+1;
            if (*nobjs+k2 < nmax) {
                st_pixel *area = new st_pixel[na];
                set_object_area(&objs[*nobjs+k2], area, na);
                delete [] area;
                objs[*nobjs+k2].nb = 0;
            }
            k2++;
        }
    }
    nnew = MIN(k2, nmax-(*nobjs));
  
    // fill the arrays with all pixels assinged to a given maximum.
    // also assign maximum and a minimal rectangle around the pixels.
    for (k = 0; k < n_maxima; k++)
        counts[k] = 0;
    for (i = 0; i < nsx; i++)
        for (j = 0; j < nsy; j++) {
            k = basins[i*nsy+j]-1; assert(k < n_maxima);
            if (k < 0)
                continue;
            k2 = index[k] - 1;
            if ((k2 >= nnew) || (k2 < 0))
                continue;
            objs[*nobjs+k2].area[counts[k]].x = i;
            objs[*nobjs+k2].area[counts[k]].y = j;
            objs[*nobjs+k2].area[counts[k]].value = sp[i*nsy+j];
            counts[k] ++;
        }
    for (i = *nobjs; i < *nobjs + nnew; i++) {
        find_pixel_extent(objs[i].area, objs[i].na,
                          &objs[i].xmin, &objs[i].xmax, 
                          &objs[i].ymin, &objs[i].ymax);
        find_pixel_maximum(objs[i].area, objs[i].na, &objs[i].max_pixel);
    }
    *nobjs += nnew;
    delete [] index;
    delete [] counts;
    delete [] basins;
}

/*
 * given an image area <sp> of dimension (<nsx>, <nsy>), find all local maxima
 * in this area and add a list of corresponding objects to the list of <objs>.
 * the image <sp> is assumed to contain at most <npixels> pixels with values 
 * above <threshold>
 */
void find_local_areas(st_object *objs, int nmax, int *nobjs,
                      float *sp, int nsx, int nsy,
                      float threshold, float min_dist,
                      unsigned short min_mol_size, int npixels, int *n_maxima)
{
    TRACE("find_local_areas(%p, %i, %i, %p, %i, %i, %f, %f, %u, %i, %i)\n",
          (void *)objs, nmax, *nobjs, (void *)sp, nsx, nsy, threshold, min_dist, 
          min_mol_size, npixels, *n_maxima);

    st_pixel *maxima = new st_pixel[npixels];
    find_local_maxima(maxima, npixels, n_maxima,
                      sp, nsx, nsy, threshold, min_dist, 1.);
    // now identify new areas
    split_local_maxima(objs, nmax, nobjs, sp, nsx, nsy,
                       threshold, min_mol_size, maxima, *n_maxima);
    delete [] maxima;

    TRACE("~find_local_areas(%p, %i, %i, %p, %i, %i, %f, %f, %u, %i, %i)\n",
          (void* )objs, nmax, *nobjs, (void *)sp, nsx, nsy, threshold, min_dist, 
          min_mol_size, npixels, *n_maxima);
}

/*
 * given the object <original>, try to split area according to some criterion
 * (e.g. number of local minima). If there is more than one new area,
 * the new areas are added to the list <objs> up to a maximum of <nmax>
 * entries
 */ 
void split_object(st_object *original, st_object *objs,
                  int nmax, int *nobjs, int min_mol_size, int max_mol_size,
                  float threshold, float min_dist)
{
    int nsx = original->xmax - original->xmin + 1;
    int nsy = original->ymax - original->ymin + 1;
    unsigned short x0 = original->xmin;
    unsigned short y0 = original->ymin;
    float *sp = new float [nsx*nsy];
    int x, y, i;
    int numold;

    for (i = 0; i < nsx*nsy; i++)
        sp[i] = threshold;

    // copy original area into 2-d array
    for (i = 0; i < original->na; i++) {
        if (original->area[i].value > threshold) {
            x = original->area[i].x-x0;
            y = original->area[i].y-y0;
            sp[x*nsy+y] = original->area[i].value;
        }
#if 0
        printf("original %i: %d %d %d thresh %f\n", i,
               original->area[i].x, y = original->area[i].y,
               original->area[i].value, threshold);
#endif
    }

    // find local maxima
    int n_maxima;
    numold = *nobjs;
#if ! defined(SIMULATION) && 0 
    identify_objects(sp, nsx, nsy, objs, nmax, nobjs, 
                     min_mol_size, original->max_pixel.value/2);
    n_maxima = *nobjs - numold;
    if (n_maxima == 1) {
        // if, using a new threshold, there is exactly one area surviving
        free_object(&objs[numold]);
        *nobjs = numold;
    }
#else
    find_local_areas(objs, nmax, nobjs, sp, nsx, nsy,
                     threshold, min_dist, min_mol_size, original->na,
                     &n_maxima);
#endif
    delete [] sp;

    if (n_maxima > 1) {
        // if the original area was split into at least two new ones
        // overwrite original by at least one of the new areas, shifting the values
        // back to the original coordinates
#if defined(DEBUG) && 0
        printf("Area split!\n");
#endif
        free_object(original);
        shift_object(&objs[numold], original, x0, y0);
        original->is_split = 1;
        // now do the same for the rest of the areas
        for (i = numold+1; i < *nobjs; i++) {
            shift_object(&objs[i], &objs[i-1], x0, y0);
            objs[i-1].is_split = 1;
        }

#if defined(DEBUG) && 0
        int j;
        for (j = 0; j < original->na; j++)
            printf("Split %3d: %3d: %d %d %d\n", 0, j,
                   original->area[j].x, original->area[j].y, original->area[j].value);
        for (i = numold; i < *nobjs-1; i++)
            for (j = 0; j < objs[i].na; j++)
                printf("Split %3d: %3d: %d %d %d\n", i+1, j,
                       objs[i].area[j].x, objs[i].area[j].y, objs[i].area[j].value);
#endif

        (*nobjs) --;
    }
    else {
#if defined(DEBUG) && 0
        printf("Did not split.\n");
#endif
    }

    return;
}

/*
 *  check for local maxima and split up objects if required
 */
void deblend_objects(st_object *objs, int nmax, int *nobjs, float fwhm)
{
    // split up objects of a size above <max_obj_area>
    int max_obj_area = (int)ceil(2.*fwhm*fwhm);
    // discard objects that are too large and most likely artifacts
    // based on some heuristic values
    int max_feature_size = (int)ceil(8.*fwhm);
    int max_feature_area = (int)ceil(16.*fwhm*fwhm);
    int nobjs2 = *nobjs;
    int i;
    for (i = 0; i < nobjs2; i++) {
        // if object exceeds a maximum size
        objs[i].is_split = 0;
        if (objs[i].na > max_obj_area) {
            // if object is too large, discard
            // TODO: actually discard the object rather than
            // just not splitting it up
            if ((objs[i].na < max_feature_area)
                && (objs[i].xmax - objs[i].xmin + 1 < max_feature_size)
                && (objs[i].ymax - objs[i].ymin + 1 < max_feature_size))
                split_object(&objs[i], objs, nmax, nobjs,
                             MIN_MOL_SIZE, max_obj_area, 
                             BINARY_DEFAULT_VALUE, fwhm/2.f);
        }
    }
#if defined(DEBUG) && 0
    printf("number of objects: %4i, before deblending: %4i, new: %4i\n",
           *nobjs, nobjs2, *nobjs-nobjs2); 
    for (i = 0; i < *nobjs; i++)
        printf("%4i: %d %d area size %d %d\n", i, objs[i].max_pixel.x,
               objs[i].max_pixel.y, objs[i].na, objs[i].max_pixel.value);
#endif
}


#if defined(DEBUG)
static void print_vector(float *v, int n, const char *comment = NULL)
{
    int i;
    if (comment != NULL)
        printf(comment);
    printf(" (");
    for (i = 0; i < n-1; i++)
        printf("%f,", v[i]);
    printf("%f)\n", v[n-1]);
}
#endif


/*
 * given a a set of 3 points (xi[i], yi[i]), find a parabola that
 * interpolates the points and determine the maximum of the parabola.
 * this function is used to interpolate the maximum of the correlation
 * function between two images to non-integer lags.
 */
float parabola_extremum(float xi[], float yi[])
{
    float numerator = 0.;
    float denominator = 0.;
    int i;
    for (i = 0; i < 3; i++) {
        int i2 = (i+1) % 3;
        int i3 = (i+2) % 3;
        numerator += yi[i] * (xi[i2]*xi[i2]-xi[i3]*xi[i3]);
        denominator += yi[i] * (xi[i2]-xi[i3]);
    }
    if (fabs(denominator) > 1.e-14)
        return numerator/(2.f*denominator);
    return xi[1];
}

/*
 * Estimate the position of the peak of the object intensity
 * by performing a 3-point parabolic interpolation through 
 * maximum pixel value along the x- and y-axis.
 */
void interpolate_object(float *array, int nx, int ny, st_object *object)
{
    int x = object->max_pixel.x;
    int y = object->max_pixel.y;
    object->position.x = (float)x; object->position.y = (float)y;
    float xv[3], yv[3];
    int i;
    if ((x > 0) && (x < nx-1)) {
        for (i = 0; i < 3; i++) {
            xv[i] = x+i-1;
            yv[i] = array[(x+i-1)*ny+y];
        }
        object->position.x = parabola_extremum(xv, yv);
    }
    if ((y > 0) && (y < ny-1)) {
        for (i = 0; i < 3; i++) {
            xv[i] = y+i-1;
            yv[i] = array[x*ny+(y+i-1)];
        }
        object->position.y = parabola_extremum(xv, yv);
    }
}

void interpolate_objects(float *array, int nx, int ny,
                         st_object *objects, int nobj)
{
    int i;
    for (i = 0; i < nobj; i++)
        interpolate_object(array, nx, ny, &(objects[i]));
}

void detect_objects(float *filter_image, float *image, int nx, int ny, 
                    float *noise, float factor, float fwhm, float min_dist,
                    st_object *objects, int nmax, int *nobj)
{
    float *spatial = new float [nx*ny];

    // threshold image
    threshold_array(spatial, filter_image, noise, factor, nx*ny, BINARY_DEFAULT_VALUE);
   
    // object identification
    identify_objects(spatial, nx, ny, objects, nmax, nobj,
                     MIN_MOL_SIZE, BINARY_DEFAULT_VALUE);

    // deblending
    deblend_objects(objects, nmax, nobj, fwhm);

    // try to obtain better estimate of position of maximum by interpolating 
    // peak
    interpolate_objects(filter_image, nx, ny, objects, *nobj);

    delete [] spatial;
}

