/*
 * QC module for tile remapping
 *
 * Time-stamp: <26 Jun 2006 16:50 BST QCRemapping.cpp km>
 * Copyright (c) Solexa 2005
 * Author: Klaus Maisinger
 * This source file is covered by the "Solexa Public Source License"
 * agreement and bound by the terms therein.
 */

#if defined(__BORLANDC__)
#include <stdio.h>
#include <math.h>
#else
#include <cstdio>
#include <cmath>
#endif
#include <cassert>
#include "QCRemapping.h"

int validate_object_counts(int *counts, int ncolour, int nmin)
{
    int i;
    int valid_offsets = 0;
    for (i = 0; i < ncolour; i++)
    if (counts[i] > nmin)
        valid_offsets |= (1<<i);
    return valid_offsets;
}


#if 1
int validate_offsets(AlignmentTransform *position_transforms,
                     int colourmax, float max_offset,
                     AlignmentTransform *default_transforms,
                     int valid_counts)
{
    int i;
    AlignmentRecord *transforms = new AlignmentRecord [colourmax];
    for (i = 0; i < colourmax; i++) {
        transforms[i].offsetx = position_transforms[i].offsetx 
                       - default_transforms[i].offsetx; 
        transforms[i].offsety = position_transforms[i].offsety 
                       - default_transforms[i].offsety; 
    }
    int flag = validateAlignment(transforms, colourmax,
                                 max_offset, valid_counts);
    int count = 0;                                
    for (i = 0; i < colourmax; i++)
        if (flag & (1<<i))
            count ++;
    if (count > 0) {
        /*
         * extrapolate the outlier offsets from those offsets that 
         * agree with each other and use the default offsets as a
         * reference.
         */
        float xsum = 0., xthsum = 0., ysum = 0., ythsum = 0.;
        for (i = 0; i < colourmax; i++)
            if (flag & (1<<i)) {
                xsum += position_transforms[i].offsetx;
                ysum += position_transforms[i].offsety;
                xthsum += default_transforms[i].offsetx;
                ythsum += default_transforms[i].offsety;
            }
        for (i = 0; i < colourmax; i++)
            if (!(flag & (1<<i))) {
                position_transforms[i].offsetx
                    = (xsum-xthsum)/count + default_transforms[i].offsetx;
                position_transforms[i].offsety 
                    = (ysum-ythsum)/count + default_transforms[i].offsety;
            }
    }
    delete [] transforms;    
    return count;
}
#else
/*
 * QC routine to check that the x- and y-offsets obtained from
 * the image alignment make sense.
 * RETURNS
 *    the number of colours whose offsets are consistent
 */
int validate_offsets(AlignmentTransform *position_transforms,
                     int colourmax, float max_offset,
                     AlignmentTransform *default_transforms,
                     int valid_counts)
{
    if (colourmax == 1)
        return 1;

    int *flags = new int [colourmax];
    int count = 0;
    int targetcount = 0;
    int i, j;
    
    /*
     * number of colours with significant object densities
     */
    for (i = 0; i < colourmax; i++)
        if (valid_counts & (1<<i))
            targetcount ++;
    

    /* for each colour, find all the colours that have similar offsets 
     * within a mximum distance of max_offset
     */
    for (i = 0; i < colourmax; i++)
        flags[i] = 0;
    for (i = 0; i < colourmax; i++) 
        for (j = i+1; j < colourmax; j++)
            if ((fabs((position_transforms[i].offsetx 
                       - default_transforms[i].offsetx) - 
                      (position_transforms[j].offsetx 
                       - default_transforms[j].offsetx)) < max_offset) &&
                (fabs((position_transforms[i].offsety 
                       - default_transforms[i].offsety) - 
                      (position_transforms[j].offsety 
                       - default_transforms[j].offsety)) < max_offset)) {
                flags[i] |= (1<<i) | (1<<j);
                flags[j] |= (1<<i) | (1<<j);
            }
    /* if there is more than one set of mutually incompatible offsets, 
     * no sensible realignment can be made
     */
    int problem_flag = 0;
    for (i = 0; i < colourmax; i++) { 
        if (flags[i] != 0)
            for (j = i+1; j < colourmax; j++)
                if ((flags[j] != 0) && (flags[j] & flags[i] == 0))
                    problem_flag = 1;
    }

#if 0  
    printf("flags : %i %i %i %i %f\n", 
           flags[0], flags[1], flags[2], flags[3], max_offset);
#endif

    /*
     * if no two colours have the same offsets, no realignment can be made.
     * otherwise extrapolate the other offsets from those offsets that 
     * agree with each other.
     */
    if (! problem_flag) {
        for (i = 0; i < colourmax; i++)
            if (flags[i])
                count ++;
        if (count < 2) {
            if (targetcount < 2) {
                /*
                 * if only one colour has significant counts, use that colour
                 * as a reference.
                 */
                for (i = 0; i < colourmax; i++) {
                    if (valid_counts & (1<<i))
                        flags[i] = (1<<i);
                    else
                        flags[i] = 0;
                }
            }
            else 
                problem_flag = 1;
        }
        if (! problem_flag) {
            /*
             * extrapolate the outlier offsets from those offsets that 
             * agree with each other and use the default offsets as a
             * reference.
             */
            float xsum = 0., xthsum = 0., ysum = 0., ythsum = 0.;
            for (i = 0; i < colourmax; i++)
                if (flags[i]) {
                    xsum += position_transforms[i].offsetx;
                    ysum += position_transforms[i].offsety;
                    xthsum += default_transforms[i].offsetx;
                    ythsum += default_transforms[i].offsety;
                }
            for (i = 0; i < colourmax; i++)
                if (flags[i] == 0) {
                    position_transforms[i].offsetx
                        = (xsum-xthsum)/count + default_transforms[i].offsetx;
                          position_transforms[i].offsety 
                        = (ysum-ythsum)/count + default_transforms[i].offsety;
                }
        }
    }
    delete [] flags;
  
    return count;
}
#endif
                 
/* ----------------------------------------------------------------------- */
/*
 * this objects records all offsets obtained from the remapping and
 * uses them to calculate the average offset and its standard deviation.
 */
OffsetTracker::OffsetTracker(int nt, int nc)
{
    int i;
    ntiles = nt; ncolours = nc;
    xsum = new float [ntiles * ncolours];
    ysum = new float [ntiles * ncolours];
    xsum2 = new float [ntiles * ncolours];
    ysum2 = new float [ntiles * ncolours];
    count = new int [ntiles];
    for (i = 0; i < ntiles*ncolours; i++) {
        xsum[i] = 0.; ysum[i] = 0.; xsum2[i] = 0.; ysum2[i] = 0.;
    }
    for (i = 0; i < ntiles; i++)
        count[i] = 0;
}

OffsetTracker::~OffsetTracker()
{
    delete [] xsum; delete [] ysum; delete [] xsum2; delete [] ysum2;
    delete [] count;
}

void OffsetTracker::add_offsets(AlignmentTransform *transform, 
                                int nc, int tile)
{
    assert((tile >= 0) && (tile < ntiles));
    assert(nc == ncolours);
    int i;
    // subtract mean offset
    float xmean = 0., ymean = 0.;
    for (i = 0; i < nc; i++) {
        xmean += transform[i].offsetx; ymean += transform[i].offsety;
    }
 
    for (i = 0; i < nc; i++) {
        float xc = transform[i].offsetx - xmean/nc;
        float yc = transform[i].offsety - ymean/nc;
        xsum[tile*ncolours+i] += xc;     ysum[tile*ncolours+i] += yc;
        xsum2[tile*ncolours+i] += xc*xc; ysum2[tile*ncolours+i] += yc*yc;
    }
    count[tile] ++;
}

int OffsetTracker::get_count(int tile)
{
    int imin = tile; int imax = tile+1;
    if (tile < 0) { 
        imin = 0; imax = ntiles;
    }
    int totalcount = 0;
    int i;
    for (i = imin; i < imax; i++)
        totalcount += count[i];
    return totalcount;
}

/*
 * return the averaged offsets stored.
 * if <tile> is less than 0, all tiles are summed up.
 * if sdev and ysdev are not NULL, return standard deviations.
 */
void OffsetTracker::get_average_offsets(float *x, float *y, int tile, int nc,
                                        float *xsdev, float *ysdev)
{
    assert(nc <= ncolours);
    assert(tile < ntiles);
    if (nc < 0)
        nc = ncolours;

    int imin = tile; int imax = tile+1;
    if (tile < 0) { 
        imin = 0; imax = ntiles;
    }

    int totalcount = get_count(tile);
    int i, j;

    // average over tiles
    for (i = 0; i < nc; i++) {
        x[i] = 0.; y[i] = 0.;
        for (j = imin; j < imax; j++) {
            x[i] += xsum[j*ncolours+i]; y[i] += ysum[j*ncolours+i];
        }
        if (totalcount > 0) {
            x[i] /= totalcount; y[i] /= totalcount;
        }
    }

    // if required, calculate standard devation
    if ((xsdev == NULL) || (ysdev == NULL))
        return;
      
    for (i = 0; i < nc; i++) {
        xsdev[i] = 0.; ysdev[i] = 0.;
        for (j = imin; j < imax; j++) {
            xsdev[i] += xsum2[j*ncolours+i]; ysdev[i] += ysum2[j*ncolours+i];
        }
        if (totalcount > 0) {
            if (x[i]*x[i] < xsdev[i]/totalcount)
                xsdev[i] = sqrt(xsdev[i]/totalcount - x[i]*x[i]);
            else
                xsdev[i] = -sqrt(fabs(xsdev[i]/totalcount - x[i]*x[i]));

            if (y[i]*y[i] < ysdev[i]/totalcount)
                ysdev[i] = sqrt(ysdev[i]/totalcount - y[i]*y[i]);
            else
                ysdev[i] = -sqrt(fabs(ysdev[i]/totalcount - y[i]*y[i]));
        }
    }
}

void OffsetTracker::save(const char filename[], int index)
{
    int j;
    float *xa = new float [ncolours]; float *ya = new float [ncolours];
    float *xs = new float [ncolours]; float *ys = new float [ncolours];

    get_average_offsets(xa, ya, index, ncolours, xs, ys);
    FILE *fp = fopen(filename, "w");
    if (fp) {
        fprintf(fp, "# Best estimate for offsets, generated by the image analysis.\n");
        fprintf(fp, "# Format: <xoffset> <yoffset> <x std dev> <y std dev>\n");
        fprintf(fp, "# %i cycles counted\n", get_count(index));
        for (j = 0; j < ncolours; j++)
            fprintf(fp, "%7.2f %7.2f %7.2f %7.2f\n", 
                    xa[j]-xa[0], ya[j]-ya[0], xs[j], ys[j]);
        fclose(fp);
    }

    delete [] xa; delete [] ya; delete [] xs; delete [] ys;
}

void OffsetTracker::display(AlignmentTransform *default_transforms)
{
    int i, j;
    float *xa = new float [ncolours]; float *ya = new float [ncolours];
    float *xs = new float [ncolours]; float *ys = new float [ncolours];
    printf("       Best offsets:        Default offsets:   Difference:   "
           "Standard deviation:\n");
    for (i = (ntiles > 1 ? -1 : 0); i < ntiles; i++) {
        get_average_offsets(xa, ya, i, ncolours, xs, ys);
    
        for (j = 0; j < ncolours; j++)
            printf("%4i %2i %7.2f %7.2f  %7.2f %7.2f  %7.2f %7.2f  %7.2f %7.2f\n",
                   i, j, xa[j]-xa[0], ya[j]-ya[0],
                   default_transforms[j].offsetx-default_transforms[0].offsetx, 
                   default_transforms[j].offsety-default_transforms[0].offsety, 
                   xa[j]-xa[0] - (default_transforms[j].offsetx -
                                  default_transforms[0].offsetx), 
                   ya[j]-ya[0] - (default_transforms[j].offsety -
                                  default_transforms[0].offsety),
                   xs[j], ys[j]);
    }
    delete [] xa; delete [] ya; delete [] xs; delete [] ys;
}
