/*
 * Auxiliary routines for sorting and selecting arrays.
 *
 * Copyright (c) Solexa 2005, 2006
 * Author: Klaus Maisinger
 * This source file is covered by the "Solexa Public Source License"
 * agreement and bound by the terms therein.
 */

/*
 * TODO: Probably better replace by routines from the standard library.
 */
#if defined(__BORLANDC__)
#include <string.h>
#else
#include <cstring>
#endif
#include <cassert>
#include <cstdlib>
#include "Sorting.h"

/*
 * sort the <n>-dim <array> using straight insertion.
 * The <array> is changed in-place.
 */ 
static void insertionsort(float *array, long n)
{
    long i, j;
    for (i = 1; i < n; i++) {
        float element = array[i];
        j = i-1;
        while ((j >= 0) && (array[j] > element)) {
            array[j+1] = array[j];
            j--;
        }
        array[j+1] = element;   
    } 
}

static void insertionsort_index(float *array, long *index, long n)
{
    long i, j;
    for (i = 1; i < n; i++) {
        float element  = array[i];
        long  element2 = index[i];
        j = i-1;
        while ((j >= 0) && (array[j] > element)) {
            array[j+1] = array[j];
            index[j+1] = index[j];
            j--;
        }
        array[j+1] = element;
        index[j+1] = element2;
    } 
}

#define SWAP(x, y) { float tmp = (x); (x) = (y); (y) = tmp; }
#define LSWAP(x, y) { long tmp = (x); (x) = (y); (y) = tmp; }

/*
 * Auxiliary function to perform the reordering of elements around
 * a pivot (partitioning) element within an <array> partition 
 * given by the indices [<left>, <right>]. The pivot elements is
 * chosen as the median of the left, right and centre elements
 * of the partition.
 * The array must be at least of dimension <right>+1.
 * 
 * Returns the index of the pivot element.
 */
static long partition_exchange(float *array, long left, long right)
{
    long centre = (right+left)/2;  /* centre point of interval */
    /* the pivot (partitioning) element is kept at position <left>+1.
     * make sure that array[left] <= pivot element <= array[right]
     * by choosing the partitioning element as the median of
     * (array[left], array[centre], array[right])
     */
    SWAP(array[centre], array[left+1])
    if (array[left] > array[right])
        SWAP(array[left], array[right])
    if (array[left+1] > array[right])
        SWAP(array[left+1], array[right])
    if (array[left] > array[left+1])
        SWAP(array[left], array[left+1])
    float pivot = array[left+1];
    /* now sweep partition from top and bottom and swap
     * any elements that are out of place
     */
    long itop = right;
    long ibot = left+1;
    do {
        do ibot++; while (array[ibot] < pivot);
        do itop--; while (array[itop] > pivot);
        SWAP(array[itop], array[ibot])
    } while (itop >= ibot);
    SWAP(array[itop], array[ibot]) /* undo last swap */

    /* finally insert pivot element */
    long ipivot = itop;
    array[left+1] = array[ipivot];
    array[ipivot] = pivot;
    return ipivot;
}

/*
 * see partition exchange, but swap second index array at the same time as the
 * original array.
 */
static long partition_exchange_index(float *array, long *index,
                                     long left, long right)
{
    long centre = (right+left)/2;  /* centre point of interval */
    /* the pivot (partitioning) element is kept at position <left>+1.
     * make sure that array[left] <= pivot element <= array[right]
     * by choosing the partitioning element as the median of
     * (array[left], array[centre], array[right])
     */
    SWAP(array[centre], array[left+1]);
    LSWAP(index[centre], index[left+1]);
    if (array[left] > array[right]) {
        SWAP(array[left], array[right]);
        LSWAP(index[left], index[right]);
    }
    if (array[left+1] > array[right]) {
        SWAP(array[left+1], array[right]);
        LSWAP(index[left+1], index[right]);
    }
    if (array[left] > array[left+1]) {
        SWAP(array[left], array[left+1]);
        LSWAP(index[left], index[left+1]);
    }
    float pivot =       array[left+1];
    long  pivot_index = index[left+1];
    /* now sweep partition from top and bottom and swap
     * any elements that are out of place
     */
    long itop = right;
    long ibot = left+1;
    do {
        do ibot++; while (array[ibot] < pivot);
        do itop--; while (array[itop] > pivot);
        SWAP(array[itop], array[ibot]);
        LSWAP(index[itop], index[ibot]);
    } while (itop >= ibot);
    SWAP(array[itop], array[ibot]); /* undo last swap */
    LSWAP(index[itop], index[ibot]); 

    /* finally insert pivot element */
    long ipivot = itop;
    array[left+1] = array[ipivot];
    array[ipivot] = pivot;
    index[left+1] = index[ipivot];
    index[ipivot] = pivot_index;
    return ipivot;
}

/*
 * Returns the <idx>th smallest element from the <n>-dim. array
 * <original>.
 * If <idx> is negative, return the median element.
 * 
 * The implementation follows the method described by NR, which
 * is based on the QuickSort algorithm.
 * The array is repeatedly partioned until the <idx>th smallest element
 * is found.
 */
float select(float *original, long n, int idx)
{
    if (idx < 0)
        idx = n/2;
    /* create temporary copy of input array */
    float *array = new float [n];
    memcpy(array, original, n*sizeof(float));

    /* select the first partition: use the whole array */
    long left  = 0;
    long right = n-1; 

    /* while the active partition contains more than two elements, 
     * keep on partitioning.
     */
    while (right-left > 1) {
        long ipivot = partition_exchange(array, left, right);

        /* select new partition */
        if (ipivot >= idx) right = ipivot-1;
        if (ipivot <= idx) left  = ipivot;
    }
    
    /* active partition contains two elements or less: sort them 
     * if necessary
     */
    if ((right - left == 1) && (array[left] > array[right]))
        SWAP(array[left], array[right])
    float result = array[idx];
    delete [] array;
    return result;
}


struct st_partition {
    long left, right;
};

/*
 * QuickSort implementation following the method described in NR.
 * The <n>-dim. <array> is sorted in-place.
 * If the partition contains 7 elements or less, then use insertion sort
 * instead of quicksort.
 *
 * Instead of using recursive function calls, partitionings are
 * stored in a stack array. While this is the implementation 
 * method used in NR, it is not clear that there is any performance
 * benefit from managing a separate stack. In the present implementation,
 * at least the stack size is variable - however, one probably ought to
 * derive a lower bound for the maximum stack size than the one used here,
 * 1+n/2.
 */
void quicksort(float *array, long n)
{
    st_partition *stack = new st_partition [1+n/2];
    /* select the first partition: use the whole array */
    long istack = 1;
    stack[0].left  = 0;
    stack[0].right = n-1;
    
    /* loop while partitions are remaining on the stack */ 
    while (istack > 0) {
        long left  = stack[istack-1].left;
        long right = stack[istack-1].right;
        if (right - left < 7) { 
            /* for short partitions, use insertion sort */
            insertionsort(&array[left], right-left+1);
            /* remove partition from stack */
            istack--;
        }
        else {
            /* for large partitions, iterate quicksort */
            
            /* choose pivot element and exchange other elements */
            long ipivot = partition_exchange(array, left, right);
            
            /* modify stack to hold the two new partitions 
             * and remove the old one. 
             * process the smaller partition first (to reduce stack usage),
             * so put it on the top of the stack. 
             */
            assert(istack < n/2);
            if (ipivot - left >= right - ipivot) {
                stack[istack-1].left  = left;
                stack[istack-1].right = ipivot - 1;
                stack[istack].left    = ipivot + 1;
                stack[istack].right   = right;
            }
            else {
                stack[istack-1].left  = ipivot + 1;
                stack[istack-1].right = right;
                stack[istack].left    = left;
                stack[istack].right   = ipivot - 1;
            }
            istack++;
        }
    }
    
    delete [] stack;
}
    
void quicksort_index(float *array, long *index, long n)
{
    st_partition *stack = new st_partition [1+n/2];
    /* select the first partition: use the whole array */
    long istack = 1;
    stack[0].left  = 0;
    stack[0].right = n-1;

    /* loop while partitions are remaining on the stack */ 
    while (istack > 0) {
        long left  = stack[istack-1].left;
        long right = stack[istack-1].right;
        if (right - left < 7) { 
            /* for short partitions, use insertion sort */
            insertionsort_index(&array[left], &index[left], right-left+1);
            /* remove partition from stack */
            istack--;
        }
        else {
            /* for large partitions, iterate quicksort */
            
            /* choose pivot element and exchange other elements */
            long ipivot = partition_exchange_index(array, index, left, right);
            
            /* modify stack to hold the two new partitions 
             * and remove the old one. 
             * process the smaller partition first (to reduce stack usage),
             * so put it on the top of the stack. 
             */
            assert(istack < n/2);
            if (ipivot - left >= right - ipivot) {
                stack[istack-1].left  = left;
                stack[istack-1].right = ipivot - 1;
                stack[istack].left    = ipivot + 1;
                stack[istack].right   = right;
            }
            else {
                stack[istack-1].left  = ipivot + 1;
                stack[istack-1].right = right;
                stack[istack].left    = left;
                stack[istack].right   = ipivot - 1;
            }
            istack++;
        }
    }
    
    delete [] stack;
}

#undef SWAP
#undef LSWAP

#if 0
#include <stdio.h>

void print_vector(float *a, long n)
{
    long i;
    printf("  [ ");
    for (i = 0; i < n; i++)
        printf("%g ", a[i]);
    printf("]\n"); 
}

int main(int argc, char *argv[])
{
    float a[8] = { 3., 5., 1., 6., 7., 4., 2., 8. };
    float a2[8];
    float a3[8];
    int i;
    for (i = 0; i < 7; i++)
        printf("%i %f\n", i, select(a, 7, i));
    for (i = 0; i < 8; i++)
        printf("%i %f\n", i, select(a, 8, i));
    for (i = 0; i < 8; i++) {
        a2[i] = a[i];
        a3[i] = a[i];
    }
    quicksort(a, 8);
    print_vector(a, 8);
    long aidx[8];
    for (i = 0; i < 8; i++)
        aidx[i] = i;

    quicksort_index(a2, aidx, 8);
    for (i = 0; i < 8; i++)
        a2[i] = a3[aidx[i]];
    for (i = 0; i < 8; i++)
        if (a[i] != a2[i])
            printf("Index error %i %f %f\n", i, a[i], a2[i]);
    
    for (i = 0; i < 8; i++) {
        a2[i] = a3[i];
    }
    for (i = 0; i < 8; i++)
        aidx[i] = i;
    insertionsort_index(a2, aidx, 8);
    for (i = 0; i < 8; i++)
        if (a[i] != a3[aidx[i]])
            printf("Ins Index error %i %f %f\n", i, a[i], a3[aidx[i]]);
    
    float b[1000];
    float b2[1000];
    float b3[1000];
    long bidx[1000];
    srand(0);
    for (i = 0; i < 1000; i++)
        b[i] = ((float)rand())/RAND_MAX;   
    for (i = 0; i < 1000; i++) {
        b2[i] = b[i];
        b3[i] = b[i];
    }
    quicksort(b, 1000);
    print_vector(b, 1000);
        for (i = 1; i < 1000; i++)
        if (b[i] < b[i-1]) printf("Error %i %f < %f\n", i, b[i], b[i-1]);  
    for (i = 0; i < 1000; i++)
        bidx[i] = i;
    quicksort_index(b2, bidx, 1000);
    for (i = 0; i < 8; i++)
        if (b[i] != b3[bidx[i]])
            printf("Index error %i %f %f\n", i, b[i], b3[bidx[i]]);
    
    return 0;
}
#endif
