/*
 * Main program for image analysis module in the Solexa data analysis
 * pipeline.  
 * Copyright (c) 2005, 2006 Solexa 
 * Author: Klaus Maisinger
 * This source file is covered by the "Solexa Public Source License"
 * agreement and bound by the terms therein.
 */

#include <cstdio>
#include <cassert>
#include <string>
#include "Images.h"
#include "Process.h"
#include "Logger.h"
#include "QCWriter.h"
#include "JournalParser.h"
#include "ImageProcess.h"
#include "IntensityData.h"

#define MIN(x,y) ((x) > (y) ? (y) : (x))

struct analysis_params {
    analysis_params() : threshold(image_default_threshold),
                        fwhm(3.),
                        distance(1.9) {};
    float threshold;
    float fwhm;
    float distance;
};


int snprintf_offsets(char s[], int len, char text[], 
                     std::vector<AlignmentTransform *> transforms)
{
    int i;
    int msg_idx = 0;
    msg_idx = sprintf(s, "%-25s", text); 
    for (i = 0; i < int(transforms.size()); i++)
        msg_idx += snprintf(s+msg_idx, len-msg_idx, "%6.2f %6.2f  ", 
                           transforms[i]->offsetx,
                           transforms[i]->offsety);
    return msg_idx;
}

void run_analysis(ImageList **image_list, unsigned char *base_list,
                  int ncycles, int ncolours, int channel, int tile,
                  analysis_params params,
                  const char path_qc_output[],
                  const char path_image_analysis[],
                  const char sample[], std::string logfile,
                  const char offset_filename[],
                  std::string loadsuffix,
                  int *is_deblocked,
                  int firstcycle)
{
    assert(ncycles > 0);
    assert(ncolours > 0);
    assert(ncolours <= 4);
    assert((params.threshold >= 3.));
    assert((params.fwhm > 1.) && (params.fwhm < 10.));
    assert((params.distance > 1.) && (params.distance < 10.));

    int i, j, k;
    int ntiles, nx, ny;
    i = 0;
    while ((i < ncolours) && (image_list[i]->load()))
        i++; // failed to load image
    if (i == ncolours) {
        fprintf(stderr, "Failed to load any first-cycle image; exiting...\n");
        exit(1);
    }
         
    ntiles = image_list[i]->size();
    nx = image_list[i]->at(0)->nx;
    ny = image_list[i]->at(0)->ny;

    /* set up logging object */
    Logger *logger = new Logger(logfile.c_str());
    logger->add_handler(new LogHandler());
    char msg[256];
    sprintf(msg, "%s%s", software, software_version);
    logger->set_source(msg); 
    logger->info(path_image_analysis);
    logger->info(logfile.c_str());

    sprintf(msg, "Cycles colours tiles %i %i %i", ncycles, ncolours, ntiles);
    logger->info(msg);
    sprintf(msg, "Channel %i Tile %i Num tiles %i",
            channel+1, tile+1, ntiles);
    logger->info(msg);
    sprintf(msg, "nx ny thresh fwhm dist %i %i %5.1f %5.1f %5.1f",
            nx, ny, params.threshold, params.fwhm, params.distance);
    logger->info(msg);

    // initialise objects for creation of filenames
    std::string path = std::string(path_qc_output) + std::string(sample);
    Tiling *qc_name = new Tiling(path.c_str());
    qc_name->setChannel(channel, tile, ntiles);

    // initialise objects for creation of filenames
    path = std::string(path_image_analysis) + std::string(sample);
    Tiling *intensity_name = new Tiling(path.c_str());
    intensity_name->setChannel(channel, tile, ntiles);
    
    DataProcessor *remapper = new DataProcessor();
    remapper->setParameters(ncolours, nx, ny,
                            params.fwhm, params.distance);
    //remapper->setFirstImage(0, firstcycle);
    remapper->setMaxAlignmentOffsets(256., 2.);
    AlignmentTransform *default_transforms = new AlignmentTransform [ncolours];
    for (i = 0; i < ncolours; i++) {
        default_transforms[i].nx = nx;
        default_transforms[i].ny = ny;
    }
    if (loadAlignmentTransforms(offset_filename, default_transforms, ncolours)
        < 0)
        sprintf(msg, "Could not open offset file %s.", offset_filename);
    else
        sprintf(msg, "Successfully read offset file %s", offset_filename);
    logger->info(msg);
    remapper->setDefaultTransforms(default_transforms, ncolours);

    // set up the QC XML output
    QCWriter *qc_handler;
    qc_handler = new QCWriter();
    qc_handler->setIndentLevel(0);

    int cycle;
    std::vector<Image * > images;
    std::vector<AlignmentTransform * > transforms;
    std::vector<float *> intensities;
    std::vector<float *> noisevalues;
    st_fposition *referencepositions = 0;
    int nobj = 0;
    k = 0;
    if (loadsuffix.size() > 0) {
        int nobj2, ncol;
        std::string filename = qc_name->getIndexFilename(POSITION_FILE, k);
        sprintf(msg, "Reading %s", filename.c_str());
        logger->info(msg);
        int status = read_positions(filename.c_str(),
                                    &referencepositions, &nobj);
        if (status) {
            fprintf(stderr, "Error: Failed to read position information.\n");
        }
        
        filename = intensity_name->getIndexFilename(INTENSITY_FILE, k) 
                   + loadsuffix;
        intensities = read_signal_array(filename.c_str(),
                                        &nobj2, &ncol, &status);
        if (status == 0) {
            assert(nobj == nobj2);
            assert(ncol == ncolours);
        } 
        filename = intensity_name->getIndexFilename(NOISE_FILE, k) 
                   + loadsuffix;
        noisevalues = read_signal_array(filename.c_str(),
                                        &nobj2, &ncol, &status);
        if (status == 0) {
            assert(nobj == nobj2);
            assert(ncol == ncolours);
        } 
        firstcycle  = intensities.size();
        sprintf(msg, "Starting from cycle %i\n", firstcycle+1); 
    }
    for (cycle = firstcycle; cycle < ncycles; cycle++) {
        i = cycle;
        // load image files 
        for (j = 0; j < ncolours; j++) {
            if (image_list[i*ncolours+j]->load()) {
                fprintf(stderr, "Warning: Image load failed. Using empty image instead.\n");
                image_list[i*ncolours+j]->createEmptyImages(nx, ny, ntiles);
            }
            assert((int)image_list[i*ncolours+j]->size() == ntiles);
        }
        for (k = 0; k < ntiles; k++) {
            images.clear();
            for (j = 0; j < ncolours; j++) {
                // single-image analysis
                Image *image = image_list[i*ncolours+j]->at(k);
                image->set_threshold(image_default_threshold);
                if (cycle == 0)
                    image->set_threshold(params.threshold);
                image->set_fwhm(params.fwhm);
                image->set_distance(params.distance);
                image->analyse();
                if ((! is_deblocked[cycle]) && (cycle == 0))
                    image->data->analyseObjectStats();
                image->data->setSequencingInfo(base_list[j], is_deblocked[cycle]);
                // image alignment
                remapper->analyseImage(image, referencepositions, nobj,
                                       k, j, cycle);
                images.push_back(image);
                sprintf(msg, "Image %2d %2i %i: %c #obj %5d offset %6.2f %6.2f %7.4f %7.4f exp %3.1f %3.1f",
                        cycle+1, k+1, j+1, 
                        image->data->sequencing->base,
                        image->get_objects()->getSize(), 
                        image->data->transform->offsetx,
                        image->data->transform->offsety,
                        image->data->transform->scalex,
                        image->data->transform->scaley,
                        image->properties->ExposureTime,
                        image->data->estimated_fwhm);
                logger->info(msg);
            }
            int   valid_flag;
            
            transforms.clear();
            for (j = 0; j < ncolours; j++)
                transforms.push_back(images[j]->getData()->transform);
            snprintf_offsets(msg, sizeof(msg), "Original offsets:", transforms);
            logger->info(msg);
         
            /*
             * In first cycle, generate the reference positions for all
             * later cycles and write them to a text file.
             */
            if (cycle == 0) {
                st_fposition *newreference;
                int newcount;
                remapper->buildReference(images, 
                                         referencepositions, nobj,
                                         &newreference, &newcount);
                referencepositions = newreference;
                nobj               = newcount;
                write_positions(qc_name->getIndexFilename(POSITION_FILE,
                                                          k).c_str(),
                                referencepositions, nobj);
                write_index_positions(qc_name->getIndexFilename(INDEX_FILE,
                                                                k).c_str(),
                                      qc_name->getChannel(k), qc_name->getTile(k),
                                      referencepositions, nobj);                
            }

            float *signal = new float [ncolours*nobj];
            float *noise  = new float [ncolours*nobj];
            int *indices  = new int [ncolours*nobj];
            remapper->analyseTile(images, k, cycle, &valid_flag,
                                  referencepositions, nobj,
                                  signal, noise, indices);
            intensities.push_back(signal);
            noisevalues.push_back(noise);

            char s[25];
            snprintf(s, sizeof(s), "Validated offsets (%i):", valid_flag);
            snprintf_offsets(msg, sizeof(msg), s, transforms);
            logger->info(msg);

            /* write output */
            if (cycle == 0) {
                for (j = 0; j < ncolours; j++) {
                    images[j]->getData()->usedTransform->transformReverse(
                                images[j]->get_objects()->getData(),
                                images[j]->get_objects()->getSize());
                    
                    images[j]->data->writeObjectStats(
                                qc_name->getIndexFilename(CLUSTER_FILE, k,
                                                          cycle, j).c_str(),
                                images[j]->get_objects()->getData(),
                                indices+j*nobj, nobj);
                }
            }
            delete [] indices;
            if ((cycle < 3)  || (ntiles > 1)) {
                write_signal_array(qc_name->getIndexFilename(INTENSITY_FILE, 
                                                             k, cycle).c_str(),
                                   signal, nobj, ncolours, ncolours);
                write_signal_array(qc_name->getIndexFilename(NOISE_FILE, 
                                                             k, cycle).c_str(),
                                   noise, nobj, ncolours, ncolours);
            }
            qc_handler->openFile(qc_name->getIndexFilename(QC_FILE, k, cycle).c_str(),
                                 software, software_version);
            for (j = 0; j < ncolours; j++)
                qc_handler->writeImage(images[j], 
                                       qc_name->getChannel(k)+1,
                                       qc_name->getTile(k)+1, cycle+1, j+1);
            qc_handler->closeFile();

            /* clear images */            
            for (j = 0; j < ncolours; j++) {
                Image *image = image_list[i*ncolours+j]->at(k);
                image->clear();
            }
        }
    }
    logger->info("Writing output files...");
    if (ntiles == 1) {
        write_signal_array(intensity_name->getIndexFilename(NOISE_FILE,
                                                            0).c_str(),
                           noisevalues, nobj, ncolours,
                           intensity_name->getChannel(0),
                           intensity_name->getTile(0), referencepositions);
        write_signal_array(intensity_name->getIndexFilename(INTENSITY_FILE,
                                                            0).c_str(),
                           intensities, nobj, ncolours,
                           intensity_name->getChannel(0),
                           intensity_name->getTile(0), referencepositions);
    }
    for (cycle = 0; cycle < ncycles; cycle++) {
        delete [] intensities[cycle];
        delete [] noisevalues[cycle];
    }
    delete [] referencepositions;
    // write out combined intensity and signal files
    remapper->displayOffsets();

    logger->info("Run terminated successfully.");
    delete remapper;
    delete logger;
    delete qc_handler;
    delete qc_name;
    delete intensity_name;
}

int main(int argc, char *argv[]) 
{
    char offset_filename[FILENAME_MAX] = "default_offsets.txt";
    char output_filename[FILENAME_MAX];
    char tmp_prefix[FILENAME_MAX] = "tempseq";
    char tmp_prefix2[FILENAME_MAX] = "tempseq";
    char sample[FILENAME_MAX] = "s";
    int ncolours = 4, ntiles = 1, ncycles = 0;
    int channel = 0, tileoffset = 0;
    int firstcycle = 0;
    ImageList **image_list;
    unsigned char *base_list;
    int call_base_flag;

    char *journalname = NULL;
    int max_shift;
    int nx = N_PIX, ny = N_PIX;
    int i;
    char *deblockedstring = NULL;
    std::string logfile;
    std::string loadsuffix = "";
    analysis_params params;
  
    for (i = 1; i < argc; i++) {
        if (argv[i][0] == '-')
            switch (argv[i][1]) {
            case 'f': sscanf(argv[i], "-fwhm=%f", &(params.fwhm)); break;
            case 't': 
                switch (argv[i][2]) {
                case 'h': sscanf(argv[i], "-threshold=%f", &(params.threshold)); 
                          break;
                case 'i': 
                    if (sscanf(argv[i], "-tile=%i,%i", 
                               &channel, &tileoffset) != 2)
                        if (sscanf(argv[i], "-tile=%[^_]_%d_%d",
                                   sample, &channel, &tileoffset) == 3) {
                            channel --; tileoffset--;
                        }
                    break;
                } break;
            case 'm': 
                switch (argv[i][2]) {
                case 'a':
                    switch (argv[i][3]) {
                    case 'x': sscanf(argv[i], "-maxshift=%i", &max_shift); break;
                    } 
                    break;
                }
                break;
            case 'o': sscanf(argv[i], "-offsets=%s", offset_filename); break;
            case 'd': 
                switch (argv[i][2]) {
                case 'i': sscanf(argv[i], "-dist=%f", &(params.distance)); break;
                case 'e': deblockedstring = new char [ strlen(argv[i]) + 1];
                          sscanf(argv[i], "-deblocked=%s", deblockedstring);
                          break;
                } break;
            case 'p': 
                if ((sscanf(argv[i], "-prefix=%s", tmp_prefix) == 1)
                     && (!strcmp(tmp_prefix2, "tempseq")))
                     strcpy(tmp_prefix2, tmp_prefix);
                else 
                     sscanf(argv[i], "-prefix2=%s", tmp_prefix2); break;
            case 's': sscanf(argv[i], "-sample=%s", sample); break;
            case 'n':
                switch (argv[i][2]) {
                case 'x': sscanf(argv[i], "-nx=%i", &nx); break;
                case 'y': sscanf(argv[i], "-ny=%i", &ny); break;
                case 'c': sscanf(argv[i], "-nc=%i", &firstcycle); break;
                } break;
            case 'l': 
                switch (argv[i][2]) {
                case 'o':
                    switch (argv[i][3]) {
                    case 'g':
                        {
                            char path[FILENAME_MAX];
                            sscanf(argv[i], "-logfile=%s", path);
                            logfile = path;
                        } break;
                    case 'a':
                        {
                            char name[FILENAME_MAX];
                            sscanf(argv[i], "-loadsuffix=%s", name);
                            loadsuffix = name;
                        } break;
                    } break;
                } break;
            }   
        else {
            journalname = argv[i];
        }
    }

    if ((argc < 2) || (journalname == NULL)) {
        printf("Usage: run_image_analysis [options] <journal filename>\n"
              "Options:\n"
           "  -fwhm=<x>          set instrumental FWHM\n"
           "  -threshold=<x>     set 'Gauss factor'\n"
           "  -maxshift=<n>      set maximum shift for alignment (in pixels)\n"
           "  -dist=<x>          set distance parameter for colocalisation\n"
           "  -prefix=<filename> set prefix for internal filenames\n"
           "  -logfile=<filename> set the path to a log file\n"
           "  -sample=<text>      set sample id (must not contain underscores)\n"
           "  -tile=<lane>,<tile> set lane/tile (0 indexed)\n"
           "  -tile=<sample>_<lane>_<tile> set sample/lane/tile (1 indexed)\n");
        exit(0);
    }

    if (parse_journal(journalname, output_filename, &ncycles, &ncolours, &ntiles,
                     &image_list, &base_list, &call_base_flag) < 0) {
        fprintf(stderr, "Error: Cannot open journal file %s\n", journalname);
        exit(1);
    }
    ncycles += firstcycle;
    int *is_deblocked = new int [ncycles];
    for (i = 0; i < ncycles; i++)
        is_deblocked[i] = 0;
    if (deblockedstring) {
        for (i = 0; i < MIN(ncycles, (int)strlen(deblockedstring)); i++)
            if (toupper(deblockedstring[i]) == 'D')
                is_deblocked[i] = 1;
        delete [] deblockedstring;
    }

    if (logfile.length() == 0) {
        logfile = tmp_prefix;
        logfile += software;
	logfile += ".log";
    }
    run_analysis(image_list, base_list, ncycles, ncolours, channel, tileoffset, 
                 params, 
                 tmp_prefix, tmp_prefix2, sample, logfile,
                 offset_filename, loadsuffix, is_deblocked, firstcycle);

    delete [] base_list;
    for (i = 0; i < ncolours * (ncycles-firstcycle); i++)
        delete image_list[i];
    delete [] image_list;
    delete [] is_deblocked;
    return 0;
}
