#!/usr/bin/env perl

# PROJECT: GERALD
# MODULE:  GERALD.pl
# AUTHOR:  A. J. Cox
#
# Copyright (c) 2005-2006 Solexa
# This source file is covered by the "Solexa Public Source License"
# agreement and bound by the terms therein.

# GERALD.pl ... stands for
# Generation of Recursive Analyses Linked by Dependency

# Description: 
# Generate a makefile to auto-generate files of sequence analysis data
# The makefile contains details of dependencies between the various
# analyses
# Also: Create a directory for the analysis files [optional]
# Also: Run the analysis by calling the makefile [optional]

# Usage: [cat configFile.txt |] ./GERALD.pl 
# [--EXPT_DIR value] [--OUT_DIR value] [--OTHER_VAR value] [--RUN]

# Variable/value pairs to go in the makefile may be
# i) stored as environment variable prior to run, e.g.
# setenv TOOLS_DIR /home/me/src/cvs/Tools
# Meant only for specifying things like file locations, not
# experiment-specific parameters
# ii) placed in configFile.txt, e.g.
# READ_LENGTH 25
# BAD_TILES b11a b12c b14c
# or ii) specified as command line arguments, e.g.
# --READ_LENGTH 25 --BAD_TILES b11a b12c b14c
# NB #1: These are listed in increasing order of precedence
# NB #2: preceding "--" needed on command line but not in config file
#
# Following parameters can be specified on command line *only*:
# --FORCE will make OUT_DIR if it doesn't exist and overwrite
# OUT_DIR/Makefile if it does exist, otherwise either of these two
# events will cause an error.
# --RUN on the command line will autorun "make all" in OUT_DIR/Makefile 
# if it is successfully generated

# Mandatory parameters
# EXPT_DIR: Will try to expand this in a sensible way:
# i)  full path
# ii) experiment number only - expands to full path of experiment directory
# - works if only one analysis directory for that experiment
# iii) experiment number / regular expression
# - use to specify a single analysis directory, eg 1661/*root.2

# OUT_DIR: directory in which to put makefile and output. 
# i) If set to EXPT_DIR and --FORCE also set 
# - make a new subfolder of EXPT_DIR and put the output there
# ii) If set to EXPT_DIR/regular expression - must specify a single
# analysis directory already present. If --FORCE set, will overwrite 
# existing Makefile in that directory 
# iii) Else assumed to be a directory name. If --FORCE set, will create
# directory if not already present. If directory is already present, will 
# overwrite any existing Makefiles therein.

# Optional parameters
# READ_LENGTH: Will try to auto-deduce this from EXPT_DIR

use warnings;
use strict;
use File::Copy;
use File::Spec;
use File::Basename;
use XML::Simple;
use Data::Dumper;

my $version = "0.1.9";
my $id;
{
    my $cvs_id='@(#) $Id: GERALD.pl,v 1.68.2.2 2007/06/13 11:08:49 km Exp $';
    my $cvs_version='$Revision: 1.68.2.2 $';
    # Make CVS tag into a displayable string
    $id=$cvs_id;
    $id=~s/\$//g;
} #scope

my $notDef="nOtDeFiNeD";

#my $exptDirRoot="/mnt/sol-snap001/";
my $makeCommand="make";
my $folderRoot="GERALD";

# Try to get user name from environment
my $logName=$ENV{"LOGNAME"};
$logName="unknown" unless defined($logName);


my $tmpName="/tmp/tmp-GERALD-$>-$$.txt";
my $tmpName2="/tmp/tmp-GERALD-$>-$$.txt";

END
{
    if (-e $tmpName)
    {
        print STDERR "Removing temporary file $tmpName\n";
        unlink ($tmpName);
    } # if
} # END

my $numLanes=8;
my $analysisKey="ANALYSIS";
my $laneTag='@LANE@';


# Create XML object - moved to just before XML is read in 
# so variable defs can be incorporated
#my $xml = new XML::Simple(suppressempty=>'');

# Deleted - now read in from external XML
#my $macroDefs = $xml->XMLin("$macroXML");


# $multiLane{"param"} specifies lanes for which param has a lane-specific
# definition 
my %multiLane;
# $multiParam{"param"}{"lane"} specifies definition of param for lane

my %multiParams;

# default parameters
my $thisDir=File::Spec->rel2abs(dirname($0));
my $timeNow=scalar(localtime);



# Deleted - now read from external XML
#my $paramDefs=$xml->XMLin("$paramXML");
#my %params=%$paramDefs;


my @s;
my @s2;
my $lastKey;
my $thisVal;
#my $runMake=0;
#my $forceWrite=0;



#open (XML, ">./GERALD.xml")
#    or die "Failed to open file $params{OUT_DIR}/config.xml";
#print XML "<GeraldConfigParameters>\n";
#print XML XMLout(\%params,NoAttr=>1,RootName=>'DefaultRunParameters');
#print XML XMLout($macroDefs,NoAttr=>1,RootName=>'MacroDefinitions');
#print XML "</GeraldConfigParameters>\n";
#close (XML);

# Create XML 
my $xml = new XML::Simple
    ( suppressempty=>'',
      XMLDecl=>1,
      Variables => { id=>$id, 
		     timeNow=>$timeNow,
		     thisDir=>$thisDir,
		     logName=>$logName } );


my $configHash=$xml->XMLin();

#print Dumper($test),"\n";

my $macroDefs=\%{$configHash->{MacroDefinitions}};
my %params=%{$configHash->{DefaultRunParameters}};



# Look for config params in environment, overwrite defaults if found
# This feature never used as far as I know, can be uncommented if found useful
#foreach( keys %params )
#{
#    if (defined($ENV{"$_"}))
#    { 
#        $params{"$_"}=$ENV{"$_"};
#    } #if
#} #foreach

# Read command line parameters
unless (@ARGV>0)
{
    print STDERR "Usage: $0 [config.txt|-] [--RUN] [--FORCE] [--EXPT_DIR dir] [--OUT_DIR dir]\n";
    print STDERR "Allowable config params and their defaults\n";
    for (sort keys %params)
    {
        print STDERR "$_ => $params{$_}\n";
    } # for
    die;
} # unless



#print Dumper($macroDefs),"\n";
#print Dumper(%params),"\n";


# read config file
unless ($ARGV[0]=~/^\-\-/)
{
    if ($ARGV[0] eq '-')
    {
        open (IN, "<&=STDIN")
            or die "Problem aliasing STDIN: $!";
        print STDERR "Will read config params from STDIN\n";
    } #if
    elsif ($ARGV[0]!~/.xml$/)
    {
        open (IN, "$ARGV[0]")
            or die "Problem opening config file $ARGV[0]: $!";
        print STDERR "Will read config params from $ARGV[0]\n";
 
	shift(@ARGV);

	open (OUT, ">$tmpName")
	    or die "Unable to create temporary file $tmpName: $!";

	while (<IN>)
	{
	    print OUT $_;
	    chomp;
	    if ((/^\#/)||(/^\s*$/))
	    {
		print STDERR "$_\n" unless (/^\s*$/);
		next;
	    } #if
	    @s=split;
	    $lastKey=shift(@s);
	    if (defined($params{$lastKey}))
	    {
		$params{$lastKey}="@s";
	    } # if
	    else
	    {
		my @ss=split(':',$lastKey);
		if ((scalar(@ss)!=2)
		    ||(!defined($params{$ss[1]}))
		    ||($ss[0]!~/^[1-8]+$/))
		{
		    die "Error: Invalid key $lastKey specified in file";
		} # if
		$lastKey=$ss[1];
		@ss=split(//,$ss[0]);
		$multiLane{$lastKey}="" unless (defined($multiLane{$lastKey}));
		foreach my $lane (@ss)
		{
		    die 
			"Error: Parameter $lastKey already defined for lane $lane\n"
			if ($multiLane{$lastKey}=~/$lane/);
		    $multiLane{$lastKey}.=$lane;
		    $multiParams{$lastKey}{"s_$lane"}="@s";
		} # foreach
	    } # else

#        die "Invalid key $lastKey specified in file"
#            unless (defined($params{$lastKey}));
#        $params{$lastKey}="@s";
	} #while

	close (IN);
	close (OUT);

   } #elsif text config file
   else
   { # it's an XML file, hooray
       my $exptHash=$xml->XMLin($ARGV[0]);
        print STDERR "Will read config params from $ARGV[0]\n";
 
	shift(@ARGV);


#print Dumper($test),"\n";

       %params=%{$exptHash->{ChipWideRunParameters}};
       %multiParams=%{$exptHash->{LaneSpecificRunParameters}};

       # next bit sets up multiParams so knows to expect per-lane definitions
       for my $param (keys %multiParams)
       {
	   for my $paramLane (keys %{$multiParams{$param}})
	   {
	       $paramLane=~s/^s_//;
	       $multiLane{$param}.=$paramLane;
	   } # for multiParam
       } # for param

   } # else

} #unless


$lastKey=$notDef;

# parse command-line options; they will overwrite the parameters
# obtained from the parameter file
while($thisVal=shift(@ARGV))
{
    if ($thisVal=~/^\-\-/)
    {
        $lastKey=$thisVal;
        $lastKey=~s/^\-\-//;
        if (($lastKey eq "RUN")||($lastKey eq "FORCE"))
        {            
            $params{$lastKey}=1;
            next;        
        } #if

        die ("Invalid key $lastKey specified on command line")
            unless (defined($params{$lastKey}));
        $params{$lastKey}="";
    } #if
    else
    {
        die ("did not understand command line argument $thisVal")
            unless ($lastKey ne $notDef);
        $params{$lastKey}.="$thisVal ";
    } #else
} #while

# strip whitespace 
foreach (sort keys %params)
{
    $params{$_}=~s/^\s+//;
    $params{$_}=~s/\s+$//;
} #foreach

# Sort out values for ANALYSIS. These need to be in the multiParams

# so they are correctly expanded when the makefile is built from the
# template

#              $multiLane{$lastKey}.=$lane;
#              $multiParams{$lastKey}{"s_$lane"}="@s";

# After the conversion this will always be "12345678"
$multiLane{$analysisKey}="";

for (my $i=1; $i<=$numLanes; $i++)
{
    $multiLane{$analysisKey}.="$i";
    $multiParams{$analysisKey}{"s_$i"}=$params{$analysisKey}
    unless (defined ($multiParams{$analysisKey}{"s_$i"}));  
} # if


# ANALYSIS values should now be ready for expansion


# Sort out EXPT_DIR
die "Did not specify any value for EXPT_DIR"
    unless ($params{"EXPT_DIR"} ne $notDef);

# convert experiment directory to absolute path if necessary
my $abspath = File::Spec->rel2abs($params{"EXPT_DIR"});
# if path exists, separate out EXPT_DIR, EXPT_DIR_ROOT and OUT_DIR
if (-d $abspath) {
    my @dirs = File::Spec->splitdir($abspath);
    my $idx;
    for ($idx = $#dirs; $idx > 0; $idx--) {
	if ($dirs[$idx] eq "Data") {
	    last;
	}
    }
    if ($dirs[$#dirs] =~/GERALD/) {
	$params{"EXPT_DIR"} = File::Spec->catdir(@dirs[($idx-1) .. ($#dirs-1)]);
	$params{"OUT_DIR}"} = File::Spec->catdir("EXPT_DIR", $dirs[$#dirs]);
    }
    else {
	$params{"EXPT_DIR"} = File::Spec->catdir(@dirs[($idx-1) .. $#dirs]);
    }
    $params{"EXPT_DIR_ROOT"} = "";
    if ($idx > 1) {
	$params{"EXPT_DIR_ROOT"} = File::Spec->catdir(@dirs[0 .. ($idx-2)]);
    }
}


# make a search string for $exptDirRoot
my $exptDirRoot=$params{"EXPT_DIR_ROOT"}."/";

my $exptDirRootSearch=$exptDirRoot;
$exptDirRootSearch=~s/\//\\\//g;

if ((-d $params{"EXPT_DIR"})&&($params{"EXPT_DIR"}=~/^$exptDirRootSearch/))
{ 
    die "EXPT_DIR ".$params{"EXPT_DIR"}." present but not readable"
        unless (-r $params{"EXPT_DIR"});
} #if
else
{
    $params{"EXPT_DIR"} = $params{"EXPT_DIR_ROOT"}."/".$params{"EXPT_DIR"};
}
die "Cannot find directory ".$params{"EXPT_DIR"}
unless (-r $params{"EXPT_DIR"});

print STDERR "Found directory ".$params{"EXPT_DIR"}."\n";

print STDERR "Will read sequence files from ".$params{"EXPT_DIR"}."\n";

# Ascertain naming convention of sequence files

#$params{"TILE_ROOT"}  = "s";         # TODO: this should be lane-specific
#$params{"TILE_REGEX"} = $params{"TILE_ROOT"} .
#    "_[1-8]_[0-9][0-9][0-9][0-9]";

# TILE_ROOT and TILE_REGEX now defined in config params
# Here we just check they are sensibly defined
unless ($params{"TILE_REGEX"}=~/^$params{"TILE_ROOT"}/)
{
    die "Parameter TILE_REGEX (value=$params{TILE_REGEX}) does not begin with parameter TILE_ROOT (value=$params{TILE_ROOT})";
} # unless


# Sort out READ_LENGTH
# Extract number of cycles expected to be in sequence directory
@s=split('/', $params{"EXPT_DIR"});

# Attempt to pull out start and finish cycle from directory path
# Work it out as (last cycle num - first cycle num + 1)
# TODO: pull it out from .params files

my $ss='^.*C([0-9]+)\-([0-9]+).*$';
if ($params{"EXPT_DIR"}=~/$ss/)
{
    $thisVal=$2-$1+1;
    print STDERR "Seems to be $thisVal available cycles to analyse\n";
    if ($params{"READ_LENGTH"} eq $notDef)
    {
        print STDERR 
         ".. and no READ_LENGTH specified, so proceeding using that value\n"; 
        $params{"READ_LENGTH"}=$thisVal;
    }
    else
    {
        if ($params{"READ_LENGTH"}>$thisVal)
        {
            print STDERR "WARNING: you have specified read length of $params{'READ_LENGTH'}\n"; 
            print STDERR "but there only seems to be $thisVal cycles of data available\n";
            print STDERR "Proceeding, but if it screws up don't blame me\n";
        } # if
    } # else
} # if
else
{
    die "No READ_LENGTH specified, and could not guess one from directory path"
        unless ($params{"READ_LENGTH"}!=$notDef);
} #else



print STDERR "Will use READ_LENGTH of ". $params{"READ_LENGTH"}."\n";

# Check everything is defined
foreach (sort keys %params)
{
    die "No definition given for parameter $_"
        unless ($params{$_} ne $notDef); 
    print STDERR "$_ $params{$_}\n";
} # for

#my $line;

# sort out OUT_DIR
my $outRoot;
my $outName;

if ($params{"OUT_DIR"} =~ /^EXPT_DIR/)
{
    if ($params{"OUT_DIR"} eq "EXPT_DIR")
    {
	# no further path info has been given; 
        # auto-generate unique name for GERALD folder
	my ($d, $m, $y)=(localtime)[3,4,5];
	$outRoot=sprintf("%02d-%02d-%04d_%s", 
			 $d, $m+1, $y+1900, $logName);
	my $try=1;
	$outName=$outRoot;
	while (-e File::Spec->catdir($params{"EXPT_DIR"},
				     $folderRoot."_".$outName))
	{
	    $try++;
	    $outName="$outRoot.$try";
	} #while

	$params{"OUT_DIR"} = File::Spec->catfile($params{"EXPT_DIR"},
						 $folderRoot."_".$outName);
	print "Output folder ".$params{"OUT_DIR"}."\n";
    } #if
    else {
	# replace string EXPT_DIR with real path
	@s=split('/', $params{"OUT_DIR"});
	$params{"OUT_DIR"}=$params{"EXPT_DIR"}."/".$s[1];

	print STDERR "Will attempt to use directory ".$params{"OUT_DIR"}."\n";
    }
} #if ($params{"OUT_DIR"} =~ /^EXPT_DIR/)

# exit unless FORCE option has been specified
if (!defined($params{"FORCE"})) {
    exit(0);
}
#########################################################################
# now create output folders and makefile
########################################################################
#print Dumper(\%params);
#print Dumper(\%multiParams);
#print Dumper(\%multiLane);
#die;

# now create output directory
# first untaint $params{"OUT_DIR"}
my $outdir;
if ($params{"OUT_DIR"} =~ /^([\w\-\/\.\,]+)$/) {
    $outdir = $1; # $1 is untainted
} else {
    die "folder ".$params{"OUT_DIR"}." has invalid characters.\n";
}

unless (-d $outdir)
{
    mkdir $outdir 
       || die "Failed to create directory ".$outdir.": $!";
    print STDERR "Created directory ".$outdir."\n";
} #unless
else
{
    die "OUT_DIR ".$outdir." present but not readable"
	unless (-r $outdir);
} #if

my $outFileName = File::Spec->catdir($outdir, "Makefile");
print "want to make $outFileName\n";
print "File $outFileName already exists - will overwrite"
    if ((!defined($params{"FORCE"}))&&(-e $outFileName));

open ( OUT, ">$outFileName")
    || die "Failed to open file $outFileName for writing: $!";

my $modVal;
my $multiVal;
my $laneName;

# Read in makefile and substitute params
# NB can only do one @...@ substitution per line
while($thisVal=<main::DATA>)
{
    if ($thisVal=~/(@\w+@)/)
    {
#        print "found key $1\n";
        $lastKey=$1;
        $lastKey=~s/^@//;
        $lastKey=~s/\@$//;
#        print "- $lastKey\n";
        die "Could not find value for $lastKey"
            unless (defined($params{$lastKey})); 

        # print out chip-wide definition
        if ($lastKey ne $analysisKey)
        {
            $modVal=$thisVal;
            $modVal=~s/\@$lastKey\@/$params{$lastKey}/g;
            print OUT $modVal;
        } # if

        # print any lane-specific definitions
        if (defined($multiLane{$lastKey}))
        {
            if ($lastKey ne $analysisKey)
            {
#                $modVal=$laneTag."%:$thisVal";
                $thisVal=$laneTag."%:$thisVal";
            } # if

            foreach my $lane (split(//,$multiLane{$lastKey}))
            {
                die unless (defined ($multiParams{$lastKey}{"s_$lane"}));
                $laneName=$params{"TILE_ROOT"}."_".${lane};
                if ($lastKey eq $analysisKey)
                {
#                    $modVal=$analysis{$multiParams{$lastKey}{"s_$lane"}};
                    $modVal=$multiParams{$lastKey}{"s_$lane"};
		    $modVal='@'.$modVal.'@';
		    $macroDefs->{LANE}=$laneName;
		    while (1)
		    {
#			print "start:\n$line \n";
			if ($modVal=~/@(\w+)@/)
			{
			    my $key=$1;
			    my $val=$macroDefs->{$key};
			    if (defined($val))
			    {
				$val=~s/\s$//;
#				$val=;
				$key='@'.$key.'@';
				
	#			print "key=$key\n";
				$modVal=~s/$key/$val/g;
			    } # if
			    else
			    {
				die "Could not find a def for macro $1";
			    }
#			    print "end:\n$line \n";

			} # if
			else
			{
			    last;
			} # else
    
		    } # while

                } # if analysisKey ...
                else
                {
                    $modVal=$thisVal;
                    $modVal=~s/\@$lastKey\@/$multiParams{$lastKey}{"s_$lane"}/g;
                }
#                print "ZZ $laneName $thisVal $lastKey\n";
#                print $multiParams{$lastKey}{"s_$lane"}."\n";
#                $modVal=$params{"TILE_ROOT"}.${lane}."%:".$thisVal;
#                $modVal=~s/\@$lastKey\@/$multiParams{$lastKey}{"s_$lane"}/g;
                $modVal=~s/$laneTag/$laneName/g;
#                print "OUT $modVal\n";
                print OUT $modVal;
            } # for
        } # if
    } # if
    else
    {
        print OUT $thisVal;
    } # else
} #while

close (OUT);

print STDERR "Successfully created $outFileName\n";

if (-e $tmpName)
{
    print STDERR "Copying config file $tmpName to "
        .$outdir."/config.txt\n";
    copy ($tmpName, $outdir."/config.txt")
        or die "Failed to copy config file: $!";
} # if

print STDERR "Dumping config parameters into $tmpName to "
        .$outdir."/config.xml\n";

open (XML, ">$outdir/config.xml")
    or die "Failed to open file $outdir/config.xml";
print XML "<RunParameters>\n";
print XML XMLout(\%params,NoAttr=>1,
		 RootName=>'ChipWideRunParameters');
print XML XMLout(\%multiParams,NoAttr=>1,
		 RootName=>'LaneSpecificRunParameters');
print XML "</RunParameters>\n";
close (XML);


print STDERR "Running 'make self_test' on $outFileName\n";

$ENV{'PATH'} = "/usr/local/bin:/bin:/usr/bin"; # untaint PATH variable
if (system("make -f $outFileName self_test")!=0)
{
    print STDERR "Problem running 'make self_test' on $outFileName\n\n"; 
    print STDERR "Investigate and fix errors, then run\n";
    print STDERR "$0 $params{OUT_DIR}/config.txt --FORCE\n";
    print STDERR "to build corrected Makefile in same folder\n\n";
    exit (-1);
} # if

print STDERR "Run 'make self_test' on $outFileName completed with no problems\n";



if (defined($params{"RUN"}))
{
    my $comString="cd ".$outdir."; $makeCommand all";
    print STDERR "Will run $comString\n";
    exec ($comString);
} #if
else
{
    print STDERR "Task complete, exiting\n";
} #else


# And now for the grand finale, the master makefile template
# Anything looking like @THIS@ gets substituted.

# NB In Emacs, may have to edit makefile template below as a Makefile then 
# cut-and-paste in the changes - doesn't do tabs correctly in perl mode
__END__
# File: Makefile 
# Description: file for auto-generation of sequence analysis
# Auto-generated by @SOFTWARE_VERSION@
# Auto-generated on @TIME_STAMP@

# Copyright (c) Solexa Limited 2005
# Author: A. J. Cox


# @(#) $Id: GERALD.pl,v 1.68.2.2 2007/06/13 11:08:49 km Exp $

.SUFFIXES:

# General variable definitions

EXPT_DIR:=@EXPT_DIR@


TOOLS_DIR:=@TOOLS_DIR@
QCAL_TOOLS_DIR:=$(TOOLS_DIR)/../QCalibration
GENOME_DIR:=@GENOME_DIR@
GENOME_FILE:=@GENOME_FILE@
GENOME=${GENOME_DIR}/${GENOME_FILE}
READ_LENGTH:=@READ_LENGTH@

CONTAM_DIR:=@CONTAM_DIR@
CONTAM_FILE:=@CONTAM_FILE@
SEQUENCE_FORMAT:=@SEQUENCE_FORMAT@
QUALITY_FORMAT:=@QUALITY_FORMAT@

# If running under Cygwin, need to add a .exe suffix onto
# names of compiled executables

ifdef COMSPEC
EXE:=.exe
endif


# Parameters for called programs...

# ... phageAlign

USE_BASES_CRITERION:=@USE_BASES@
USE_BASES= -useBases ${USE_BASES_CRITERION}
PURE_BASES= -pureBases @PURE_BASES@

PA:=${TOOLS_DIR}/phageAlign${EXE}

PA_PARAMS=$(GENOME) $(READ_LENGTH) $(READ_LENGTH) $(USE_BASES)
PA_PARAMS_MONO=-tag $(GENOME) $(READ_LENGTH) $(READ_LENGTH) $(USE_BASES)


# ... score.pl

SC:=${TOOLS_DIR}/score.pl
SC_PARAMS=
#SC_PARAMS=$(GENOME)

# ... buildCoverage.pl

BC:=${TOOLS_DIR}/buildCoverage.pl
# Second BC_PARAMS is score threshold below which to ignore alignments
# when building coverage. Leave it out to include them all
BC_PARAMS=$(GENOME) 200


# ... plotCoverage.pl

PC:=${TOOLS_DIR}/plotCoverage.pl

# ... plotErrorRate.pl

PE:=${TOOLS_DIR}/plotErrorRate.pl

# ... quahog

QH:=${TOOLS_DIR}/quahog${EXE}
QH_PARAMS= ${USE_BASES} ${PURE_BASES}
# ... qualityFilter.pl

QF:=${TOOLS_DIR}/qualityFilter.pl
FILTER_CRITERION:=@QF_PARAMS@
QF_PARAMS= -stamp ${FILTER_CRITERION}

# ... create_error_thumbnails.pl

CE:=${TOOLS_DIR}/create_error_thumbnails.pl
CE_PARAMS=error
CE_PARAMS_SMT=all
CE_PARAMS_PSH=hist
CE_PARAMS_PEC=perfect
CE_PARAMS_PIN=info

# ... convertToFasta.pl

TF:=${TOOLS_DIR}/convertToFasta.pl
TF_PARAMS=${USE_BASES}

# ... convertFromELAND.pl

FI:=${TOOLS_DIR}/convertFromELAND.pl

# ... Signal_Means.pl
SM:=${TOOLS_DIR}/Signal_Means.pl
SM_PARAMS=$(QF) ${FILTER_CRITERION} ${USE_BASES_CRITERION} $(READ_LENGTH) ${EXPT_DIR}

# ... plotIntensity_for_IVC.pl
PI:=${TOOLS_DIR}/plotIntensity_for_IVC.pl

# ... create_IVC_thumbnail.pl
CI:=${TOOLS_DIR}/create_IVC_thumbnail.pl

# ... Signal_Means_Tile.pl
SMT:=${TOOLS_DIR}/Signal_Means_Tile.pl
SMT_PARAMS=$(QF) ${FILTER_CRITERION} ${USE_BASES_CRITERION} $(READ_LENGTH)

# ... plotIntensity_for_IVC.pl
PIT:=${TOOLS_DIR}/plotIntensity_tiles.pl

# ... Signal_Means_Monotemplates.pl
SMM:=${TOOLS_DIR}/Signal_Means_Monotemplates.pl
SMM_PARAMS=$(QF) ${FILTER_CRITERION} ${USE_BASES_CRITERION} $(READ_LENGTH)

# ... runReport.pl
RR:=${TOOLS_DIR}/runReport.pl
RR_PARAMS=${CURDIR} @EMAIL_SERVER@ \
@EMAIL_DOMAIN@ \
@WEB_DIR_ROOT@ \
@EMAIL_LIST@ 

# ... plotScoreHist.pl
PSH:=${TOOLS_DIR}/plotScoreHist.pl

# ... plot_Error_Curves.pl
PEC:=${TOOLS_DIR}/plot_Error_Curves.pl

# ... plotInformation.pl
PIN:=${TOOLS_DIR}/plotInformation.pl

# ... plotInformation.pl
BTT:=${TOOLS_DIR}/byTileThumbnailView.pl

# ... plotIntensity_monotemplates.pl
PIM:=${TOOLS_DIR}/plotIntensity_monotemplates.pl

# ... create_IVC_monotemplates_thumbnail.pl 
CMT:=${TOOLS_DIR}/create_IVC_monotemplates_thumbnail.pl 

# ... jerboa.pl
JER:=${TOOLS_DIR}/jerboa.pl

# ... reportQuality.pl
RQ:=${TOOLS_DIR}/reportQuality.pl

# ... convertQuality.pl
CQ:=${TOOLS_DIR}/convertQuality.pl
CQ_PARAMS= --useBases ${USE_BASES_CRITERION} $(QUALITY_FORMAT) ${READ_LENGTH}
 
# ... extractQualityParams.py
EQP:=${QCAL_TOOLS_DIR}/extractQualityParams.py
EQP_PARAMS=--sig-suffix=${SIG_FILE_SUFFIX} --dir=${EXPT_DIR} --USE_BASES=${USE_BASES_CRITERION}
 
# ... QualityCalibration
QLC:=${QCAL_TOOLS_DIR}/QualityCalibration${EXE}
 
# ... QualityApply
QLA:=${QCAL_TOOLS_DIR}/QualityApply${EXE}
QLA_PARAMS:=${QUALITY_FORMAT}

# ... filterQualityValues.pl
FQV:=${TOOLS_DIR}/filterQualityValues.pl
FQV_PARAMS=${USE_BASES_CRITERION} ${QUALITY_FORMAT} ${READ_LENGTH}

# ... buildSeq.pl
BS:=${TOOLS_DIR}/buildSeq.pl

# ... unsquash

UNSQ:=${TOOLS_DIR}/unsquash${EXE}
UNSQ_PARAMS=${ELAND_GENOME} ${READ_LENGTH}


# ...

TOOLS:=$(PA) $(SC) $(BC) $(PC) $(RA) $(PE) $(HM) $(XL) $(PCL) $(QH) \
$(QF) $(CE) $(TF) $(FI) $(SM) $(PI) $(CI) $(SMT) $(PIT) $(SMM) $(RR) $(PSH) \
$(PEC) $(PIN) $(BTT) $(PIM) $(CMT) $(JER) $(RQ) $(EQP) $(QLC) $(QLA) $(FQV) \
$(BS) $(UNSQ)



# Following are lists of directoreis files and executables  
# that will be checked by make self_test
# These lists may be added to depending on what analysis is asked for
DIRS_TO_CHECK=$(EXPT_DIR) $(TOOLS_DIR) $(QCAL_TOOLS_DIR)
FILES_TO_CHECK=$(TOOLS) 
EXECS_TO_CHECK:=$(TOOLS)

#
# Find lists of files that are inputs to processing  
#

# Eventually this list will be obtained from an XML file

SIG_FILE_SUFFIX_DEPHASED:=@SIG_FILE_SUFFIX_DEPHASED@
SIG_FILE_SUFFIX_NOT_DEPHASED:=@SIG_FILE_SUFFIX_NOT_DEPHASED@

SEQ_FILE_SUFFIX:=@SEQ_FILE_SUFFIX@
SIG_FILE_SUFFIX:=$(SIG_FILE_SUFFIX_DEPHASED)
PRB_FILE_SUFFIX:=@PRB_FILE_SUFFIX@
QHG_FILE_SUFFIX:=@QHG_FILE_SUFFIX@

TILE:=@TILE_REGEX@

# Get list of sequence files
ALL_SEQ_FILES:= $(wildcard $(EXPT_DIR)/$(TILE)$(SEQ_FILE_SUFFIX))

# Get list of sig files, if blank try non-phase corrected suffix instead
ALL_SIG_FILES:= $(wildcard $(EXPT_DIR)/$(TILE)$(SIG_FILE_SUFFIX))
ifeq ($(strip $(ALL_SIG_FILES)),)
WARNING_MESSAGES+=Info: no intensity files satisfying \
	$(EXPT_DIR)/$(TILE)$(SIG_FILE_SUFFIX_DEPHASED) found, trying \
	$(EXPT_DIR)/$(TILE)$(SIG_FILE_SUFFIX_NOT_DEPHASED) instead
SIG_FILE_SUFFIX:=$(SIG_FILE_SUFFIX_NOT_DEPHASED)
ALL_SIG_FILES:= $(wildcard $(EXPT_DIR)/$(TILE)$(SIG_FILE_SUFFIX))
endif

ALL_PRB_FILES:= $(wildcard $(EXPT_DIR)/$(TILE)$(PRB_FILE_SUFFIX))

ALL_QHG_FILES:= $(wildcard $(EXPT_DIR)/$(TILE)$(QHG_FILE_SUFFIX))

QHG_DIR:=$(EXPT_DIR)
ifeq ($(strip $(ALL_QHG_FILES)),)
QHG_DIR:=.
endif
SM_PARAMS+= $(QHG_DIR)


ALL_TILES:=$(patsubst $(EXPT_DIR)/%$(SEQ_FILE_SUFFIX),%,$(ALL_SEQ_FILES))
ALL_SIG_TILES:=$(patsubst $(EXPT_DIR)/%$(SIG_FILE_SUFFIX),%,$(ALL_SIG_FILES))
ALL_PRB_TILES:=$(patsubst $(EXPT_DIR)/%$(PRB_FILE_SUFFIX),%,$(ALL_PRB_FILES))

# This will be blank if the two tiles lists are the same
DIFF_TILES_SIG:=$(filter-out $(ALL_TILES), $(ALL_SIG_TILES))\
$(filter-out $(ALL_SIG_TILES), $(ALL_TILES))

# This will be blank if the two tiles lists are the same ...
DIFF_TILES_PRB:=$(filter-out $(ALL_TILES), $(ALL_PRB_TILES))\
$(filter-out $(ALL_PRB_TILES), $(ALL_TILES))


BAD_TILES:=@BAD_TILES@
# Any tiles listed in BAD_TILES get completely ignored
TILES:=$(filter-out $(BAD_TILES), $(ALL_TILES) )


ALL_LANE_NUMS:=1 2 3 4 5 6 7 8
ALL_LANES:=$(patsubst %,@TILE_ROOT@_%,$(ALL_LANE_NUMS))
LANES:=$(foreach w, $(ALL_LANES), $(findstring $w, $(TILES)))
LANE_TARGETS:=$(patsubst %,%,$(LANES))


#
# All types of files that can be generated are listed here
#

ALIGN_SUFFIX:=_align.txt
QHG_SUFFIX:=_qhg.txt
QALIGN_SUFFIX:=_qalign.txt
REALIGN_SUFFIX:=_realign.txt
SCORE_SUFFIX:=_score.txt
RESCORE_SUFFIX:=_rescore.txt
RESCORE_PNG_SUFFIX:=_rescore.png
COV_SUFFIX:=_cov.txt
COV_PNG_SUFFIX:=_cov.png
SM_SUFFIX:=_Signal_Means.txt
SMT_SUFFIX:=_all.txt
SMT_PNG_SUFFIX:=_all.png
MONO_SUFFIX:=_monotemplate.txt
PSH_PNG_SUFFIX:=_hist.png
PEC_PNG_SUFFIX:=_errors.png
PIN_PNG_SUFFIX:=_info.png
CALIGN_SUFFIX:=_calign.txt
CDIFF_SUFFIX:=_cdiff.txt
CPREDIFF_SUFFIX:=_cprediff.txt
CREDIFF_SUFFIX:=_crediff.txt
PREALIGN_SUFFIX:=_prealign.txt
CREALIGN_SUFFIX:=_crealign.txt
QREP_SUFFIX:=_qreport.txt
QCALREP_SUFFIX:=_qcalreport.txt
QVAL_SUFFIX:=_qval.txt
QTABLE_SUFFIX:=_qtable.txt
QCAL_SUFFIX:=_qcal.txt
QRAW_SUFFIX:=_qraw.txt
SEQUENCE_SUFFIX:=_sequence.txt
FRAG_SUFFIX:=_frag.txt

OUTPUT_FILE_TYPES+=$(ALIGN_SUFFIX)
OUTPUT_FILE_TYPES+=$(QHG_SUFFIX)
OUTPUT_FILE_TYPES+=$(QALIGN_SUFFIX)
OUTPUT_FILE_TYPES+=$(REALIGN_SUFFIX)
OUTPUT_FILE_TYPES+=$(SCORE_SUFFIX)
OUTPUT_FILE_TYPES+=$(RESCORE_SUFFIX)
OUTPUT_FILE_TYPES+=$(RESCORE_PNG_SUFFIX)
OUTPUT_FILE_TYPES+=$(COV_SUFFIX)
OUTPUT_FILE_TYPES+=$(COV_PNG_SUFFIX)
OUTPUT_FILE_TYPES+=$(SM_SUFFIX)
OUTPUT_FILE_TYPES+=$(SMT_SUFFIX)
OUTPUT_FILE_TYPES+=$(SMT_PNG_SUFFIX)
OUTPUT_FILE_TYPES+=$(MONO_SUFFIX)
OUTPUT_FILE_TYPES+=$(PSH_PNG_SUFFIX)
OUTPUT_FILE_TYPES+=$(PIN_PNG_SUFFIX)
OUTPUT_FILE_TYPES+=$(CALIGN_SUFFIX)
OUTPUT_FILE_TYPES+=$(CDIFF_SUFFIX)
OUTPUT_FILE_TYPES+=$(CPREDIFF_SUFFIX)
OUTPUT_FILE_TYPES+=$(CREDIFF_SUFFIX)
OUTPUT_FILE_TYPES+=$(PREALIGN_SUFFIX)
OUTPUT_FILE_TYPES+=$(CREALIGN_SUFFIX)
OUTPUT_FILE_TYPES+=$(QREP_SUFFIX)


# Chip-wide HTML views
ERROR_THUMB:=Error.htm
SIGNAL_MEANS:=Signal_Means.txt
PLOT_INTENSITY:=plotIntensity_for_IVC
IVC_THUMB:=IVC.htm
SMT_THUMB:=All.htm
PSH_THUMB:=Hist.htm
PEC_THUMB:=Perfect.htm
PIN_THUMB:=Info.htm
BTT_THUMB:=Tile.htm
JER_THUMB:=Summary.htm
RUN_REPORT:=finished.txt

OUTPUT_FILE_TYPES+=$(ERROR_THUMB)
OUTPUT_FILE_TYPES+=$(SIGNAL_MEANS)
OUTPUT_FILE_TYPES+=$(PLOT_INTENSITY)
OUTPUT_FILE_TYPES+=$(IVC_THUMB)
OUTPUT_FILE_TYPES+=$(SMT_THUMB)
OUTPUT_FILE_TYPES+=$(PSH_THUMB)
OUTPUT_FILE_TYPES+=$(PEC_THUMB)
OUTPUT_FILE_TYPES+=$(PIN_THUMB)
OUTPUT_FILE_TYPES+=$(BTT_THUMB)
OUTPUT_FILE_TYPES+=$(JER_THUMB)
OUTPUT_FILE_TYPES+=$(RUN_REPORT)

# Definitions for ELAND
ELAND_RESULT:=eland_result.txt
ELAND_QUERY:=eland_query.txt
ELAND_DIR:=$(TOOLS_DIR)/../Eland/
#ELAND_EXEC:=eland_$(READ_LENGTH)
ELAND_EXEC:=eland
#ELAND_PATH:=$(ELAND_DIR)/$(ELAND_EXEC)
ELAND_GENOME:=@ELAND_GENOME@
ELAND_REPEAT:=@ELAND_REPEAT@
ELAND_MULTIPLE_INSTANCES:=@ELAND_MULTIPLE_INSTANCES@

ifeq ($(ELAND_MULTIPLE_INSTANCES),2)
BREAK_DEPENDENCY_LANE_NUMS:=4
endif
ifeq ($(ELAND_MULTIPLE_INSTANCES),4)
BREAK_DEPENDENCY_LANE_NUMS:=2 4 6
endif
ifeq ($(ELAND_MULTIPLE_INSTANCES),8)
BREAK_DEPENDENCY_LANE_NUMS:=1 2 3 4 5 6 7
endif

BREAK_DEPENDENCY:=$(patsubst %,@TILE_ROOT@_%,$(BREAK_DEPENDENCY_LANE_NUMS))


MAX_ELAND_READ_LENGTH:=32
ALLOWED_ELAND_READ_LENGTHS:=15 16 17 18 19 20 21 22 23 24 25 26 27 28 29 30 31 32


vpath %$(QHG_SUFFIX) $(EXPT_DIR)

# Don't delete any intermediate files
#.SECONDARY:
.PRECIOUS: %$(SMT_SUFFIX) %$(CALIGN_SUFFIX) %$(CREALIGN_SUFFIX) %$(CREDIFF_SUFFIX) %$(CPREDIFF_SUFFIX) %$(RESCORE_SUFFIX) %$(SCORE_SUFFIX) %$(PREALIGN_SUFFIX) %$(REALIGN_SUFFIX) s_%_check_use_bases %$(QTABLE_SUFFIX)

all: self_test warn error $(RUN_REPORT) $(JER_THUMB) $(LANE_TARGETS)
	echo Gerald analysis finished successfully.; @POST_RUN_COMMAND@ 
warn:	
ifneq ($(strip $(WARNING_MESSAGES)),)
	$(warning Warning(s) reported: $(WARNING_MESSAGES))
endif



ifeq ($(strip $(ALL_SEQ_FILES)),)
error_no_seq:	
	$(error Error: no sequence files satisfying $(TILE)$(SEQ_FILE_SUFFIX) were found in EXPT_DIR=$(EXPT_DIR))
ERROR_TARGETS+=error_no_seq
endif

# .. die if lists of seq and sig files are not the same
ifneq ($(strip $(DIFF_TILES_SIG)),)
error_seq_sig_diff:	
	$(error Error: $(TILE)$(SEQ_FILE_SUFFIX) and $(TILE)$(SIG_FILE_SUFFIX)\
	 file lists in EXPT_DIR=$(EXPT_DIR) differ!\
	 Differences=$(DIFF_TILES_SIG))
ERROR_TARGETS+=error_seq_sig_diff
endif

# .. die if lists of seq and prb files are not the same
ifneq ($(strip $(DIFF_TILES_PRB)),)
error_seq_prb_diff:	
	$(error Error: $(TILE)$(SEQ_FILE_SUFFIX) and $(TILE)$(PRB_FILE_SUFFIX)\
	 file lists in EXPT_DIR=$(EXPT_DIR) differ!\
	 Differences=$(DIFF_TILES_PRB))
ERROR_TARGETS+=error_seq_prb_diff
endif

error: $(ERROR_TARGETS)	


### Analysis specification for each lane follows ###

@ANALYSIS@

### End of lane-specific analysis ###


# If the above necessitates some eland analysis, then define
# all the targets necessary to run it
ifdef LAST_ELAND_RESULT
DIRS_TO_CHECK+=$(ELAND_DIR)
FILES_TO_CHECK+=$(ELAND_REPEAT)


# Build Fasta format query from sequence files
#$(ELAND_QUERY): $(patsubst %,$(EXPT_DIR)/%,$(ELAND_PREREQS))
#	cd $(EXPT_DIR); $(TF) $(TF_PARAMS) $(ELAND_PREREQS) > $(CURDIR)/$@

# Run ELAND to generate results
#$(ELAND_RESULT): $(ELAND_QUERY) $(ELAND_EXEC)
#	@echo running eland on  $^
#	$(ELAND_PATH) $(ELAND_QUERY) $(ELAND_GENOME) $(ELAND_RESULT)\
#	$(ELAND_REPEAT)

endif # ELAND_PREREQS

# If the above necessitates some monotemplate analysis, then define
# all the targets necessary to run it
ifdef MONO_PREREQS

# Definitions for monotemplate analysis
MONO_RESULT:=Monotemplate.txt

$(MONO_RESULT): $(MONO_PREREQS)
	cat $^ > $@

PLOT_INTENSITY_MONO:=plotIntensity_monotemplates

$(PLOT_INTENSITY_MONO): $(MONO_RESULT)
	$(PIM) $(MONO_RESULT) $(CURDIR)
	echo dummy > $@

MONO_THUMB:=Monotemplate.htm

$(MONO_THUMB): $(PLOT_INTENSITY_MONO)
	$(CMT) $(CURDIR) $(MONO_LANES) > $@

JER_PREREQS+=$(MONO_THUMB)

OUTPUT_FILE_TYPES+=$(MONO_RESULT) $(PLOT_INTENSITY_MONO) $(MONO_THUMB)


endif # MONO_PREREQS

ifdef CONTAM_FILE

PA_PARAMS_CONTAM=-tag $(CONTAM_DIR)/$(CONTAM_FILE) $(READ_LENGTH) $(READ_LENGTH) $(USE_BASES)
 
contam: $(addsuffix $(CDIFF_SUFFIX), $(CONTAM_TILES))

# Add contaminant target (TBD - make this lane-by-lane)
all+=contam

# Alignment to contaminant
%$(CALIGN_SUFFIX): $(EXPT_DIR)/%$(SEQ_FILE_SUFFIX) $(PA)
	$(PA) $(PA_PARAMS_CONTAM) $<\
	| perl -lane '{ print unless (/^\#/) }' > $@

# Alignment to data minus alignment to contaminant - negative for contaminants 
%$(CDIFF_SUFFIX): %$(CALIGN_SUFFIX) %$(ALIGN_SUFFIX)
	cat $*$(ALIGN_SUFFIX) | grep -v '^\#' | cut -d" " -f2 \
	| paste - $*$(CALIGN_SUFFIX)  \
	| perl -lane '{ print $$F[0]-$$F[2] }' > $@

# Alignment to contaminant using alignment scores from first pass
%$(CREALIGN_SUFFIX): $(EXPT_DIR)/%$(SEQ_FILE_SUFFIX) $(PA)\
%$(SCORE_SUFFIX)
	$(PA) $(PA_PARAMS_CONTAM) -scoreFile $*$(SCORE_SUFFIX) $<\
	| perl -lane '{ print unless (/^\#/) }' > $@

# Alignment to data minus alignment to contaminant - realigned data
%$(CPREDIFF_SUFFIX): %$(CREALIGN_SUFFIX) %$(PREALIGN_SUFFIX)
	cat $*$(PREALIGN_SUFFIX) | grep -v '^\#' | cut -d" " -f2 \
	| paste - $*$(CREALIGN_SUFFIX)  \
	| perl -lane '{ print $$F[0]-$$F[2] }' > $@
# Alignment to data minus alignment to contaminant - realigned + filtered data
%$(CREDIFF_SUFFIX): %$(CPREDIFF_SUFFIX) $(QHG_DIR)/%$(QHG_SUFFIX)
	cat $< | $(QF) $(QF_PARAMS) $(QHG_DIR)/$*$(QHG_SUFFIX) > $@

endif



# Macro definition

GRAB_LANE_TILES=$(filter $(1)%, $(2))

#
# Generic rules to build files
#

# Make "first pass" alignments

# moved to GERALD.xml to define on per-lane basis, as make 3.78 won't override
# a global definition with a per-lane definition when needed
#%$(ALIGN_SUFFIX): $(EXPT_DIR)/%$(SEQ_FILE_SUFFIX) $(PA)
#	$(PA) $(PA_PARAMS) $<\
#	> $@

%$(QVAL_SUFFIX): %$(ALIGN_SUFFIX) $(EQP) \
$(EXPT_DIR)/%$(SIG_FILE_SUFFIX) $(EXPT_DIR)/%$(PRB_FILE_SUFFIX)
	$(EQP) $(EQP_PARAMS) $<\
	> $@
 
#%$(QTABLE_SUFFIX): %$(QVAL_SUFFIX) $(QLC)
#	$(QLC) $< > $@

#%$(QCAL_SUFFIX): %$(QVAL_SUFFIX) %$(QTABLE_SUFFIX) $(QLC)
#	$(QLA) $(QLA_PARAMS) $*$(QTABLE_SUFFIX) $< | \
#	$(FQV) $(FQV_PARAMS) > $@
 
%$(QRAW_SUFFIX): $(EXPT_DIR)/%$(PRB_FILE_SUFFIX) $(CQ)
	$(CQ) $(CQ_PARAMS) $< > $@




# Make scored alignments from quality value files

%$(QALIGN_SUFFIX): $(EXPT_DIR)/%$(PRB_FILE_SUFFIX) $(PA)
	$(PA) $(PA_PARAMS) $<\
	> $@

# Make QC parameter files
%$(QHG_SUFFIX): $(EXPT_DIR)/%$(SIG_FILE_SUFFIX) $(QH)
	$(QH) $(QH_PARAMS) $<\
	> $@

# Default behaviour to make score files

ifdef CONTAM_FILE
%$(SCORE_SUFFIX): %$(ALIGN_SUFFIX) %$(CDIFF_SUFFIX) $(SC)
	cat $< | $(QF) '($$F[0]>0)' $*$(CDIFF_SUFFIX) | $(SC) $(SC_PARAMS) >$@ 
else
%$(SCORE_SUFFIX): %$(ALIGN_SUFFIX) $(SC)
	cat $< | $(SC) $(SC_PARAMS)  > $@ 
endif

# Realignment on per tile basis

# moved to GERALD.xml to define on per-lane basis, as make 3.78 won't override
# a global definition with a per-lane definition when needed
#%$(PREALIGN_SUFFIX): $(EXPT_DIR)/%$(SEQ_FILE_SUFFIX) $(PA)\
#%$(SCORE_SUFFIX)
#	$(PA) $(PA_PARAMS) -scoreFile $*$(SCORE_SUFFIX) $< > $@

%$(REALIGN_SUFFIX): %$(PREALIGN_SUFFIX) $(QHG_DIR)/%$(QHG_SUFFIX)
	cat $< | $(QF) $(QF_PARAMS) $(QHG_DIR)/$*$(QHG_SUFFIX) > $@

ifdef CONTAM_FILE
%$(RESCORE_SUFFIX): %$(REALIGN_SUFFIX) %$(CREDIFF_SUFFIX) $(SC)
	cat $< | $(QF) '($$F[0]>0)' $*$(CREDIFF_SUFFIX)|$(SC) $(SC_PARAMS) >$@ 
else
%$(RESCORE_SUFFIX): %$(REALIGN_SUFFIX)
	cat $< | $(SC) $(SC_PARAMS) > $@ 
endif

# Make pngs from rescore files
%$(RESCORE_PNG_SUFFIX): %$(RESCORE_SUFFIX)
	$(PE) $<

# Make pngs from align, realign and prb files
%$(PSH_PNG_SUFFIX): %$(RESCORE_SUFFIX)
	$(PSH) $@ $*$(ALIGN_SUFFIX) $*$(SCORE_SUFFIX) $*$(REALIGN_SUFFIX) $*$(RESCORE_SUFFIX) $(EXPT_DIR)/$*$(PRB_FILE_SUFFIX) $(USE_BASES_CRITERION)

# Make png of 0,1,2,3,4 errors from realign file
%$(PEC_PNG_SUFFIX): %$(REALIGN_SUFFIX)
	$(PEC) $< $@ 

# Make pngs of information content from align+realign+prb
%$(PIN_PNG_SUFFIX): %$(RESCORE_SUFFIX)
	$(PIN) $@ $*$(ALIGN_SUFFIX) $*$(REALIGN_SUFFIX) \
	$(EXPT_DIR)/$*$(PRB_FILE_SUFFIX) $(USE_BASES_CRITERION)



# Build lane coverage after second alignment
%_cov.txt: %_rescore
	cat $(call GRAB_LANE_TILES, $*, $(GOOD_REALIGN_FILES))\
	| $(BC) $(BC_PARAMS) > $@

all_cov.txt: $(LANES_BEST_ALIGN) $(LANES_COV)
	cat $(LANES_BEST_ALIGN) | $(BC) $(BC_PARAMS) > $@

%$(SMT_SUFFIX): $(EXPT_DIR)/%$(SIG_FILE_SUFFIX) $(QHG_DIR)/%$(QHG_SUFFIX)
	$(SMT) $(SMT_PARAMS) $^ > $@

%$(SMT_PNG_SUFFIX): %$(SMT_SUFFIX)
	$(PIT) $^ $@


# Make error plots of all rescore files
$(ERROR_THUMB):  $(addsuffix $(RESCORE_PNG_SUFFIX), $(THUMBNAIL_TILES))
	$(CE) $(CE_PARAMS) > $@

# Prerequisites for SIGNAL_MEANS
$(SIGNAL_MEANS): $(SM_PREREQS)
	cat $^ > $@

# Run plotIntensity_for_IVC.pl
$(PLOT_INTENSITY): $(SIGNAL_MEANS) 
	$(PI) $(SIGNAL_MEANS) $(CURDIR)
	echo dummy > $@

# Run create_IVC_thumbnail.pl
$(IVC_THUMB): $(PLOT_INTENSITY)
	$(CI) $(CURDIR) > $@


# Build thumbnail of Signal_Means_Tile.pl output
$(SMT_THUMB) : $(addsuffix $(SMT_PNG_SUFFIX), $(SMT_THUMBNAIL_TILES))
	$(CE) $(CE_PARAMS_SMT) > $@

# Build thumbnail of sequence quality histograms
$(PSH_THUMB) : $(addsuffix $(PSH_PNG_SUFFIX), $(THUMBNAIL_TILES))
	$(CE) $(CE_PARAMS_PSH) > $@

# Build thumbnail of sequence quality histograms
$(PEC_THUMB) : $(addsuffix $(PEC_PNG_SUFFIX), $(THUMBNAIL_TILES))
	$(CE) $(CE_PARAMS_PEC) > $@

# Build thumbnail of information content histograms
$(PIN_THUMB) : $(addsuffix $(PIN_PNG_SUFFIX), $(THUMBNAIL_TILES))
	$(CE) $(CE_PARAMS_PIN) > $@



s_1: $(s_1_TARGET_LIST)
	touch s_1_finished.txt
s_2: $(s_2_TARGET_LIST)
	touch s_2_finished.txt
s_3: $(s_3_TARGET_LIST)
	touch s_3_finished.txt
s_4: $(s_4_TARGET_LIST)
	touch s_4_finished.txt
s_5: $(s_5_TARGET_LIST)
	touch s_5_finished.txt
s_6: $(s_6_TARGET_LIST)
	touch s_6_finished.txt
s_7: $(s_7_TARGET_LIST)
	touch s_7_finished.txt
s_8: $(s_8_TARGET_LIST)
	touch s_8_finished.txt


# Build by_tile thumbnail view of whatever is there
JER_PREREQS+= $(IVC_THUMB) $(SMT_THUMB) $(LANE_TARGETS)
$(JER_THUMB): $(JER_PREREQS)
	$(JER) $(CURDIR) > $@

$(BTT_THUMB) : $(ERROR_THUMB) $(IVC_THUMB) $(SMT_THUMB) $(PSH_THUMB) $(PEC_THUMB) $(PIN_THUMB) $(JER_THUMB)
	$(BTT) $(CURDIR) > $@

$(RUN_REPORT): $(BTT_THUMB)
	$(RR) $(RR_PARAMS)
	@echo dummy > $@



# GERALD will run make self_test on the created make file
# and parse the output to check for problems
s_%_check_use_bases: 
	@echo $(USE_BASES_CRITERION) | perl -lane \
	'{ die "\nERROR: Invalid USE_BASES string: $$_\n\n" unless (/^(odd|all|[NnYn]+)$$/);}'
	@echo $(USE_BASES_CRITERION) $(READ_LENGTH) | \
	perl -lane '{ if ($$F[0]=~/^[NnYn]+/) { ($$b=$$F[0])=~s/[Nn]//g; $$a=length($$b); die "\nERROR: $$a Ys in USE_BASES string $$F[0], but need $$F[1] to match READ_LENGTH\n\n" unless ($$a==$$F[1])}}'


# target for debugging
s_%_self_test: s_%_check_use_bases
	@echo $(DIRS_TO_CHECK) | perl -lane \
	'{for (@F) {die "\nERROR: did not find dir $$_\n\n" unless (-d $$_)}}'
	@echo $(FILES_TO_CHECK) | perl -lane \
	'{for (@F) {die "\nERROR: did not find file $$_\n\n" unless (-e $$_)}}'
	@echo $(EXECS_TO_CHECK) | perl -lane \
	'{for (@F) {die "\nERROR: file $$_ not executable\n\n" unless(-x $$_)}}'
self_test: $(patsubst %,%_self_test,$(LANES))



qhg: $(ALL_QHG_FILES)


qalign: $(LANES:%=%_qalign)

rescore: $(LANES:%=%_rescore)

lra: $(ALL_LANES_RA) summary_lra.txt plot_calls 

tidy:
	find . -empty -exec rm {} \;

clean_qcal:
	find . -name "*_qraw.txt" -or -name "*_qtable.txt" -or -name "*_qcal.txt" -or -name  "*qreport.txt" -or -name  "*qcalreport.txt" | xargs $(RM) 

clean0:
	find . -maxdepth 1 -name "*.png" -or -name "*_coverage" -or -name "*_sequence.txt" | xargs $(RM)
	find . -maxdepth 1 -name "core" $(addprefix -or -name "*, $(addsuffix ", $(OUTPUT_FILE_TYPES)))  | xargs $(RM)
	-$(RM) core core.*

clean: clean_qcal clean0
	@echo Finished cleaning.

clean_intermediate:
	@echo Nothing to be cleaned.

compress:
	gzip *.txt

uncompress:
	gunzip *.txt.gz

# End of generated Makefile
