#!/usr/bin/env perl

# PROJECT: GERALD
# MODULE:  buildSeq.pl
# AUTHOR:  A. J. Cox
#
# Copyright (c) 2006 Solexa
# This source file is covered by the "Solexa Public Source License"
# agreement and bound by the terms therein.

# Build a fastq/fasta/scarf format file from a set of _seq.txt files and 
# a set of quality files

my $version;
my $id;
{
    my $cvs_id='@(#) $Id: buildSeq.pl,v 1.8 2007/02/27 18:31:40 km Exp $';
    my $cvs_version='$Revision: 1.8 $';
    # Make CVS tag into a displayable string
    $version=$cvs_version;
    $version=~s/\$//g;
    $version=~s/Revision//;
    $id=$cvs_id;
#    $id=~s/\$//g;
} #scope


use warnings;
use strict;
use File::Basename;

use FindBin;
use lib "$FindBin::Bin/../";
use Gerald::Common;

my $usage=<<"END";
Usage: $0 
[--useBases useBases] 
[--qualityFilter filterPath filterParam] 
[--contamFilter contamParam] 
[--qhgPath qhgPath]
[--fastq|--fasta|--scarf] 
[--prefix=sequenceNamePrefix]
readLength exptDir qualityFile[s]
END

my %outputModes=
( '--fastq'=>0,
  '--fasta'=>1,
  '--scarf'=>2 );


die $usage unless (@ARGV>=3);

my $useBases=undef;
my $qualityParam=undef;
my $qualityFilter=undef;
my $contamParam=undef;
my $qhgPath=undef;

if ($ARGV[0] eq '--useBases')
{
    shift (@ARGV);
    $useBases=shift(@ARGV); 
} # if

if ($ARGV[0] eq '--qualityFilter')
{
    shift (@ARGV);
    $qualityFilter=shift(@ARGV);
    die "Error: could not find quality filter executable $qualityFilter" 
	unless (-e $qualityFilter);
    die "Error: file $qualityFilter is not executable" 
	unless (-x $qualityFilter);
    $qualityParam=shift(@ARGV); 
} # if

if ($ARGV[0] eq '--contamFilter')
{
    shift (@ARGV);
    $contamParam=shift(@ARGV); 
} # if

if ($ARGV[0] eq '--qhgPath')
{
    shift (@ARGV);
    $qhgPath=shift(@ARGV); 
} # if

my $outputMode=$outputModes{$ARGV[0]};
if (defined($outputMode))
{
    shift (@ARGV);
} # if
else
{
    $outputMode=0; 
} # else

$"="_";
$"=":" if ($outputMode==2);

my $outputPrefix=undef;
if ($ARGV[0]=~/\-\-prefix=(.*)/)
{
    $outputPrefix=$1;
    shift (@ARGV);
} # if


die $usage unless (@ARGV>=3);

my $readLength=shift(@ARGV);

die "Specified read length of $readLength, need positive value"
    unless ($readLength>0);

my $exptDir=shift (@ARGV);

die "Error: could not find experiment directory $exptDir" 
    unless (-e $exptDir);
die "Error: $exptDir is not a directory" 
    unless (-d $exptDir);


my @s1;


unless (defined($outputPrefix))
{
    my $bustardFolder=File::Spec->rel2abs(dirname($exptDir));
    @s1=split('/',$bustardFolder);
#    $outputPrefix=$s1[-3]; # in a just world this would be the runfolder name
#    $outputPrefix=Common::getShortRunFolderName($s1[-3]); 
    $outputPrefix=getShortRunFolderName($s1[-3]); 

#    if (defined ($outputPrefix))
#    {
#	$outputPrefix=~s/^\d{6}_//; # strip date
#	$outputPrefix=~s/^slxa-//; # strip unnecessary machine name
#	$outputPrefix=~s/_0+/_/; # strip leading zeroes
#	$outputPrefix=~s/_[Ff][Cc]\d+$//; # remove any _FCxxxx idiocy at end
#    } # if
#    else 
#    {
#	$outputPrefix='s';
#    } # else
} # unless


my $file;
my $stem;
my $type;
my $pipe;
my $dir;
my $line;
#my @s2;
#my @s3;


my @baseIndex;
#my $readLength=25;
$useBases="NYYYYYYYYYYYYYYYYYYYYYYYYYYY" unless defined($useBases);

my %total;
my %error;
my $i;
my $j;
my $qv;
my $numBases=0;

my $seqFileName;

sub parseUseBases
{
    my ($useBases)=@_;
    if ($useBases eq "all")
    {
        $useBases= "Y" x $readLength;
    } # if
    elsif ($useBases eq "odd")
    {
        $useBases= "YN" x ($readLength);
        $useBases=substr($useBases,0,length($useBases)-1);
    } # elsif
    else
    {
        die "Use bases string $useBases must be odd, all or YNYN..."
            unless ($useBases=~/^[YNyn]+$/);
        $useBases = uc ($useBases);
    } # else

    my @ub=split(//,$useBases);
  #  my @baseIndex;
    for ($i=0;$i<@ub;$i++)
    {
        push (@baseIndex, $i) if ($ub[$i] eq "Y"); 
    } # if

    die "Need $readLength bases of base information, only ", 
    scalar(@baseIndex), " specified in $useBases"
        unless (scalar(@baseIndex)>=$readLength);
    @baseIndex=splice(@baseIndex,0,$readLength);
} # sub parseUseBases

parseUseBases($useBases);

my $finished=0;
my $doQuality=defined($qualityParam);
my $numTiles;
#my $tileList;

my $qualString;
my $seqString;

for $file (@ARGV)
{
    $numTiles++;
    $dir=dirname($file);
    if (!defined($qhgPath)) {
	$qhgPath=$dir;
    }

#    print dirname($file), "\n";
    unless (basename($file)=~/(s_[1-8]_[0-9][0-9][0-9][0-9])_(.*\.txt)$/)
    {
	die "Could not parse tile from file name $file";
    } # unless
    $stem=$1;
    $type=$2;
#    $tileList.=$stem;
#    $tileList.=((($numTiles%8)==0)?"\n":" ");

    unless (-f $file)
    {
	print STDERR basename($0), 
	": Error - could not find file $file\n";
	exit (0);
    } # unless

    $seqFileName="${exptDir}/${stem}_seq.txt";
    
    unless (-f $seqFileName)
    {
	print STDERR basename($0), 
	": Error - could not find file $seqFileName associated with file $file\n";
	exit (0);
    } # unless

    $pipe = "paste ${seqFileName} ${file} |";
    $pipe .= " ${qualityFilter} '${qualityParam}' ${qhgPath}/${stem}_qhg.txt |"
	if ($doQuality); 

    if (defined($contamParam)) 
    {
	if (defined($qualityParam)) 
	{        
	    $pipe .= " qualityFilter.pl '${contamParam}' ${dir}/${stem}_crediff.txt |";
	} # if
	else
	{
	    $pipe .= " qualityFilter.pl '${contamParam}' ${dir}/${stem}_cprediff.txt |";
	} # else
    } # if
#    print STDERR "$pipe \n";
    open (LINE, $pipe) || die "Failed to open $pipe";
    while (<LINE>)
    { 
	chomp;
#	print "$_\n";
	$seqString="";
	@s1=split('\t',$_);
	$qualString=pop(@s1);
	
	for ($i=0;$i<scalar(@baseIndex);$i++)
	{
	    $seqString.=substr($s1[-1],$baseIndex[$i],1);
	}
	$seqString=~tr/\./N/;
	pop (@s1);
	unshift(@s1, $outputPrefix);
	if ($outputMode==0)
	{
	    print "@@{s1}\n$seqString\n+@{s1}\n$qualString\n";
	}
	elsif ($outputMode==1)
	{
	    print ">@{s1}\n$seqString\n";
	}
	elsif ($outputMode==2)
	{
	    print "@{s1}:${seqString}:${qualString}\n";
	}


    } # while  

    close (LINE);

} # for file






