#!/usr/bin/env perl


# Jerboa produces an experimental summary sheet in html format
# The html prints some summary statistics
# The html provides links to other information
#
# Copyright (c) 2006 Solexa
# Author L. J. Davies
# This source file is covered by the "Solexa Public Source License"
# agreement and bound by the terms therein.

# The program is called using directory name

use warnings; 
use strict;
use Cwd;
use XML::Simple;
use File::Basename;
use File::Spec;
use Data::Dumper;

my $version;
my $id;
{
    my $cvs_id='@(#) $Id: jerboa.pl,v 1.31 2007/03/05 17:52:15 km Exp $';
    my $cvs_version='$Revision: 1.31 $';
    # Make CVS tag into a displayable string
    $version=$cvs_version;
    $version=~s/\$//g;
    $version=~s/Revision//;
    $id=$cvs_id;
#    $id=~s/\$//g;
} #scope

sub info
{
    my $message = shift @_;
    print STDERR "Info: ", basename($0), $message, "\n";
}

my $maxNumLanes=8;
my $numChannels=4;

my ($i,$j,$tiles,$file,$sigFile,$length,$count,$PF_count,$PF_score,$runFolder);
my ($lanes,$monotemplates,$cov_graphs,$align,$PF_align,$file_name);
my ($error,$score,$filtered,$oneSig,$twentySig,$number,$template_count,$status);
my (@results,@array,@array2,@template);
my (%lane_type,%tags);

my $machineName;

# Next def makes sure all unknown fields are deonted the same way
my $unknown='unknown';

my $seqFileRegex='s_[1-8]_[0-9][0-9][0-9][0-9]';
my $seqSuffix='_seq.txt';
my $alignSuffix='_align.txt';
my $scoreSuffix='_score.txt';
my $realignSuffix='_realign.txt';
my $rescoreSuffix='_rescore.txt';
my $allSuffix='_all.txt';

# Intensities less than minIntensity get ignored in dropoff calculations
my $minIntensity=5; 
my $runFolderPath="$ARGV[0]/../../../..";
my $bustardPath="$ARGV[0]/..";
my $firecrestPath="$ARGV[0]/../..";

my ($infoContent, $threeSig, $infoContentRaw, $errorRaw);

my $xml = new XML::Simple();
my $xmlData;

my $phasing=undef;
my $prephasing=undef;

my @laneResults;
#my %chipResults;

my $thisStat;

# Read XML file into a hash. Return 1 if successfully completed
# Check for zero padding at end of file
sub readFileXML
{
    my ($fileName)=@_;
    return undef unless (-f $fileName);
    unless(open (FILE, $fileName)) 
    { 
	warn "Failed to open $fileName: $!"; return undef; 
    } 
    my $text="";
    while (<FILE>)
    {
	$text.=$_;
    } # while
    close (FILE);

    my $l=length($text);

    while (($l!=0)&&(substr($text,$l-1,1) eq "\0"))
    {
	$l--;
    } # while

    if ($l==0)
    {
	warn "file $fileName seems to be completely blank";
	return undef;
    } # if   

    $text=substr($text,0,$l) unless ($l==length($text));

    my $data = $xml->XMLin($text);

    return $data;
} # sub readFileXML


# Take a series of n points (xi, yi)
# Convert yi to log(yi)
# Fit a best fit line to obtain average slope in x linear - y log space
# Convert this to a best fit percent loss per cycle

# Formula used for best fit is y ~ ax+b
# a=(sigma(xi yi) - n * mean x * mean y)/(sigma(xi xi) - n * mean x * mean x)  
# b = mean y - a * mean x (b not worked out)

# This is formula 18.34 From Jeffrey 2nd edition ISBN 0-442-30728-4 
# (& 000's of other maths books)

# Version in code uses fact that mean (xi) = sigma (xi) / n


sub logLinearFit
{
    my ($x, $y)=@_;
    die unless (scalar(@{$x})==scalar(@{$y}));
    my ($x_sum, $y_sum, $xx_sum, $xy_sum, $i, $ylog, $a, $b)
    =(0,0,0,0,0,0,0,0);
    my $numDataPoints=0;

    for ($i=0; $i<@{$x}; $i++)
    {
#        print "$i ".$x->[$i]." ".$y->[$i]."\n";
	next unless ($y->[$i]>=$minIntensity);
	++$numDataPoints;

        $ylog=log($y->[$i]);
        $x_sum+=$x->[$i];
        $y_sum+=$ylog;
        $xx_sum+=($x->[$i]*$x->[$i]);
        $xy_sum+=($x->[$i]*$ylog);
    } # for
#    print "$i\n";

    unless ($numDataPoints>=2)
    {
	warn "Not enough valid data points to do fit";
	return 100.0;
    } # unless
    
    # Work out gradient of best fit line
    $a=($xy_sum-($x_sum*$y_sum/$i))/($xx_sum-($x_sum*$x_sum/$i));

    # Convert to a percent loss per cycle
    return (100*(1-exp($a)));
} # sub logLinearFit


### Check the input parameters

if(@ARGV != 1){
    print STDERR "$0 Gerald_dir\n";
    die;
} #  if @ARGV


## Look for experimental number

# Following line enables you to do e.g. "jerboa.pl ."
$ARGV[0]=File::Spec->rel2abs($ARGV[0]);

## Check to see if directory names ends with '/'
if($ARGV[0] !~ m/\/$/){
    $ARGV[0] = $ARGV[0]."/";
}

@array = split('/',$ARGV[0]);
$runFolder = undef;

# Check for valid as-agreed-in-spec-months ago run-folder name
# or eight digit run-folder name (grrr....)
for($i=0;$i<@array;$i++)
{
    next if (($array[$i] !~ m/0000\d{4}/) 
	     && ($array[$i] !~ m/^\d{6}_(.*)_\d{4}$/)) ;
    $runFolder = $array[$i];
    $machineName=$1 if defined ($1);
    last;
} # for

# Last ditch attempt at finding a run folder name
$runFolder=$array[-5] if ((defined($array[-5]))&&(!defined($runFolder)));

# Give up...
$runFolder=$unknown unless (defined($runFolder));

info(": deduced run folder name $runFolder");

unless (defined($machineName))
{
    my $paramsFile="${runFolderPath}/${runFolder}.params";

    print STDERR basename($0), 
    ": looking in $paramsFile for machine name\n";
    if(defined($xmlData=readFileXML($paramsFile)))
    {
#	print Dumper($xmlData);
	if (defined($xmlData->{run}{instrument}))
	{
	    $machineName=$xmlData->{run}{instrument};	
	    print STDERR basename($0), 
	    ": found machine name $machineName\n";
	} # if
	else
	{
	    warn "Failed to find instrument name in $paramsFile\n"
	} # else
    }
} # unless



my $phasingFile="${firecrestPath}/.params";
info(": looking in $phasingFile for phasing information");

if(defined($xmlData=readFileXML($phasingFile)))
{ 

#    $xmlData=$xml->XMLin($phasingFile);
#    print Dumper($xmlData), "\n";;

    if (defined($xmlData->{Run}))
    {
	if (ref($xmlData->{Run}) eq 'ARRAY')
	{
	    for ($i=0;$i<scalar(@{$xmlData->{Run}});$i++)
	    {
		if ($xmlData->{Run}[$i]{Name} eq $array[-2])
		{
		    $phasing=$xmlData->{Run}[$i]{Parameters}[0];
		    $prephasing=$xmlData->{Run}[$i]{Parameters}[1];
		    printf STDERR "%s: found phase=%5.3f prephase=%5.3f\n",
		    basename($0), $phasing, $prephasing;
		    last;
		} # if
	    } # for
	} # if
	elsif (ref($xmlData->{Run}) eq 'HASH')
	{
	    $phasing=$xmlData->{Run}{Parameters}[0]
		if (defined($xmlData->{Run}{Parameters}[0]));
	    $prephasing=$xmlData->{Run}{Parameters}[1]
		if (defined($xmlData->{Run}{Parameters}[1]));
	    printf STDERR "%s: found phase=%5.3f prephase=%5.3f\n",
	    basename($0), $phasing, $prephasing;
	} # elsif
    } # if defined xmlData->Run
} # if
#else
#{
#    warn "Unable to find file $phasingFile";
#} # else

for($lanes=1;$lanes<=$maxNumLanes;$lanes++) {
    $laneResults[$lanes]{phasingapplied} = undef;
    $laneResults[$lanes]{prephasingapplied} = undef;
    $laneResults[$lanes]{phasingmeasured} = undef;
    $laneResults[$lanes]{prephasingmeasured} = undef;
    $laneResults[$lanes]{nonZeroTiles} = 0;
}
for($lanes=1;$lanes<=$maxNumLanes;$lanes++) {
    if ($phasing > 0.) {
	$laneResults[$lanes]{phasingapplied} = $phasing;
    }
    if ($prephasing > 0.) {
	$laneResults[$lanes]{prephasingapplied} = $prephasing;
    }
}

# look for lane-specific phasing values (pipeline version 0.1.7 and later)
for($lanes=1;$lanes<=$maxNumLanes;$lanes++) {
    $phasingFile="${bustardPath}/Phasing/s_".$lanes."_phasing.xml";


    if(defined($xmlData=readFileXML($phasingFile)))
    { 
        #print STDERR Dumper($xmlData), "\n";;
	if (defined($xmlData->{Phasing})) {
	    $laneResults[$lanes]{phasingmeasured} = $xmlData->{Phasing};
	}
	if (defined($xmlData->{Prephasing})) {
	    $laneResults[$lanes]{prephasingmeasured} = $xmlData->{Prephasing};
	}
    } # if
#    else
#    {
#	warn "Unable to find file $phasingFile";
#    } # else
}
for($lanes=1;$lanes<=$maxNumLanes;$lanes++) {
    $phasingFile="${bustardPath}/params".$lanes.".xml";

    if(defined($xmlData=readFileXML($phasingFile)))
    { 
        #print STDERR Dumper($xmlData), "\n";;
	if (defined($xmlData->{Phasing})) {
	    $laneResults[$lanes]{phasingapplied} = $xmlData->{Phasing};
	}
	if (defined($xmlData->{Prephasing})) {
	    $laneResults[$lanes]{prephasingapplied} = $xmlData->{Prephasing};
	}
    } # if
#    else
#    {
#	warn "Unable to find file $phasingFile";
#    } # else
}


$machineName=$unknown unless (defined($machineName));

my $sampleFile="${runFolderPath}/samples.xml";
info(": looking in $sampleFile for sample information");

if(defined($xmlData=readFileXML($sampleFile)))
{
    info(": found sample file $sampleFile, trying to parse");

#    $xmlData=undef;
#    $xmlData=$xml->XMLin($sampleFile);
    for ($i=0;$i<scalar(@{$xmlData->{Lane}});$i++)
    {
    print STDERR basename($0), 
    ": found sample ID ",$xmlData->{Lane}[$i]{SampleID},
    " for lane ", $xmlData->{Lane}[$i]{LaneNumber}, "\n";

#	print $xmlData->{Lane}[$i]{LaneNumber}
#	    , " -- ", $xmlData->{Lane}[$i]{SampleID}, "\n";
	$laneResults[$xmlData->{Lane}[$i]{LaneNumber}]{sample}
	    =$xmlData->{Lane}[$i]{SampleID};
    } # for
} # if
else
{
    warn "Unable to find file $sampleFile";
    for ($lanes=1; $lanes<=$maxNumLanes; $lanes++)
    {
	$laneResults[$lanes]{sample}=$unknown;
    } # for
} # else

# now look for reference, read-length and filter in config.xml file

for ($lanes=1; $lanes<=$maxNumLanes; $lanes++) {
    foreach $thisStat (qw/template type length purity eland/) {
	$laneResults[$lanes]{$thisStat} = "NOT STORED";
    }
}

my $configFile=File::Spec->catdir($ARGV[0], "config.xml");
info(": looking in $configFile for target information");
my %configKeywords = ( "READ_LENGTH" => "length",
		       "ANALYSIS"    => "type",
		       "GENOME_FILE" => "template",
		       "ELAND_GENOME" => "eland",
		       "QF_PARAMS"   => "purity" );
if(defined($xmlData=readFileXML($configFile)))
{
    info(": found config file $configFile, trying to parse");

    if (defined($xmlData->{ChipWideRunParameters})) {
	foreach my $keyword (keys %configKeywords) {
	    if (defined($xmlData->{ChipWideRunParameters}{$keyword})) {
		for ($lanes=1; $lanes<=$maxNumLanes; $lanes++) {
		    $laneResults[$lanes]{$configKeywords{$keyword}}=
			$xmlData->{ChipWideRunParameters}{$keyword};
		} # for
	    } # if
	} # foreach
    } # if
    
    if (defined($xmlData->{LaneSpecificRunParameters})) {
	my %data = %{$xmlData->{LaneSpecificRunParameters}};
	foreach my $keyword (keys %configKeywords) {
	    if (defined($data{$keyword})) {
		foreach my $key (keys %{$data{$keyword}}) {
		    my ($lane) = $key =~ m/[^_]+_(\d+)/;
		    $laneResults[$lane]{$configKeywords{$keyword}} 
		        = $data{$keyword}{$key};
		} # for
	    } # if
	} # foreach
    } # if

    # no output for ANALYSIS none, use ELAND_GENOME as Eland target
    for ($lanes=1; $lanes<=$maxNumLanes; $lanes++) {
	$laneResults[$lanes]{type} = uc($laneResults[$lanes]{type});
	if ($laneResults[$lanes]{type} eq "NONE") {
	    foreach my $value (qw/length template/) {
		$laneResults[$lanes]{$value} = $unknown;
	    }
	} # if 
	if ($laneResults[$lanes]{type} eq "SEQUENCE") {
	    foreach my $value (qw/template/) {
		$laneResults[$lanes]{$value} = $unknown;
	    }
	} # if 
	if ($laneResults[$lanes]{type} eq "ELAND") {
	    $laneResults[$lanes]{template} 
                = basename($laneResults[$lanes]{eland});
	} # if
    } # for
} #if


my $chipFile="${runFolderPath}/chip.xml";
my $chipName=$unknown;
info(": looking in $chipFile for chip information");

if(defined($xmlData=readFileXML($chipFile)))
{
    info(": found chip file $chipFile, trying to parse");

#    $xmlData=undef;
#    $xmlData=$xml->XMLin($chipFile);
    if (defined ($xmlData->{ChipID}))
    {
	print STDERR basename($0), 
	": found chip ID ",$xmlData->{ChipID},"\n";
	$chipName=$xmlData->{ChipID};
    } # if
    else
    {
	warn "Unable to find field ChipID in file $chipFile"; 
    } # else
} # if
#else
#{
#    warn "Unable to find file $chipFile";
#} # else



##### Going to first go through the whole directory and 
##### store the required information
##### Only Use realign files to specify contents as most reliable source
print STDERR basename($0), ": creating tile list from files of form  $ARGV[0]${seqFileRegex}${alignSuffix}\n";

my @seqFiles=glob("${bustardPath}/${seqFileRegex}${seqSuffix}");
my $stem;

if (@seqFiles==0)
{
    warn "No files of form ${seqFileRegex}${seqSuffix} found in ${bustardPath}\n"; 
    exit (0);
} # if

# Ensure sensible (or at least consistent) file ordering
@seqFiles=sort(@seqFiles);

#opendir (DIR,$ARGV[0]) or die "can't open $ARGV[0]\n";

$tiles = 0;


#while (defined($file = readdir (DIR)))
foreach my $seqFile (@seqFiles)
{
    unless ($seqFile=~/.*(${seqFileRegex})${seqSuffix}/)
    {
	warn "Unexpected problem: $file did not match ${seqFileRegex}${alignSuffix}";
	next;
    } # unless
    else
    {
	$stem="$ARGV[0]/$1";
    } # else
    $file = '${stem}${alignSuffix}';


#    next if ($file !~ m/_align/);

    ## Determine Lane and Tile


    @array = split("_",basename($seqFile));

    $results[$tiles]{lane} = $array[1];
    $results[$tiles]{tile} = $array[2];
    
#    $file = $ARGV[0]."$file";
    $file_name = $file;


    ## Now going to set so defaults - mostly for Eland

    $results[$tiles]{purity}   = $laneResults[$results[$tiles]{lane}]{purity};
    $results[$tiles]{template} = $laneResults[$results[$tiles]{lane}]{template};
    $results[$tiles]{length}   = $laneResults[$results[$tiles]{lane}]{length};
    $results[$tiles]{type}     = $laneResults[$results[$tiles]{lane}]{type};
    $results[$tiles]{count} = 0;
    $results[$tiles]{error} = 0;
    if ($results[$tiles] ne "NONE") {
        $lane_type{$results[$tiles]{lane}} = $results[$tiles]{type};
    }
    #### Dealing With Align Files

    open(IN,"${seqFile}") 
	|| warn "can't open ${seqFile}\n";
    $count = 0;

# This loop changed to remove a lot of unnecessary string matches
# Only first character needs to be checked for "commenthood"
# Only need to search for tags if line is a comment line
# Once one is found, no need to try others


    while(<IN>)
    {
	if($_ =~ m/^\#/)
	{
	    next;
	}	
	$count++;
    } # while

    $results[$tiles]{count} = $count;    
    
    close(IN);
    
    ##### Dealing With realign files


# This kind of thing is not robust - if you change order of file opening 
# everything messes up
#    $file = $file_name;
#    $file =~ s/align/realign/;
    
#    if (defined(open(IN,$file))){ 


    if (defined(open(IN,"${stem}${realignSuffix}")))
    { 
	
	$PF_count = 0;
	$PF_score = 0;
	$PF_align = 0;
	

	while(<IN>)
	{

	    # Changed: 
	    # only first character needs to be checked to see whether
	    # it's a comment
	    # only need to look for FILTER_CRITERION if line
	    # is a comment

	    if ($_ =~ m/^\#/)
	    {
		if($_ =~ m/FILTER_CRITERION/){

		    ## Just doing a quick change here to fix bug
		    
		    $_ =~ s/\#FILTER_CRITERION//;

		    # Convert any '&' and '<' characters to XML equivalents
		    $_=~s/\&/&amp;/g;
		    $_=~s/</&lt;/g;
		    
		    $results[$tiles]{purity} = $_;
		} # if
		next;
	    } # if


	    @array = split(' ',$_);
	    
	    $PF_count++;
	    $PF_score += $array[1];
	    
	    if($array[2] == 1)
	    {
		$PF_align++;
	    } # IF

	} # while (<IN>)

	if($PF_count != 0)
	{
	    $PF_score /= $PF_count;
	    $PF_align = $PF_align / $PF_count * 100;
	    $laneResults[$results[$tiles]{lane}]{nonZeroTiles}++;
	} # if
	else 
	{
	    $PF_score = 0;
	    $PF_align = 0;
	} # else

	if ($count != 0)
	{
	    $PF_count = $PF_count / $count * 100;
	} # if
	else 
	{
	    $PF_count = 0;
	} # else
	$results[$tiles]{PF_count} = $PF_count;
	$results[$tiles]{PF_score} = $PF_score;
	$results[$tiles]{PF_align} = $PF_align;

	
	close(IN);

    } # if
    else 
    { 
	$results[$tiles]{PF_count} = 0;
	$results[$tiles]{PF_score} = 0;
	$results[$tiles]{PF_align} = 0;
    } # else	

    #### Dealing with rescore files

#    $file = $file_name;
#    $file =~ s/align/rescore/;
my $infoContent;    

my $infoContentKey='infoContent';
my $errorKey='error';


foreach my $suffix($rescoreSuffix, $scoreSuffix)
{
    $infoContent=undef;

    if (defined(open(IN,"${stem}${suffix}")))
    { 

	while(<IN>)
	{
#	    print $_; die;
	    if ($_ =~ m/ percent error rate/)
	    {
		@array = split(' ',$_);

		$results[$tiles]{$errorKey} = $array[0];
#		last;
	    } # if
	    elsif ($_ =~ /^\~/)
	    { # parse information content line
		$infoContent=$_;
	    } # elsif
	    else
	    {
		last if (defined($infoContent));
	    } # else


	} # while <IN>
	close(IN);

	if (defined($infoContent))
	{
	    chomp($infoContent);

	    $infoContent=~s/^\~//;

	    @array=split('\t',$infoContent);

	    $results[$tiles]{$infoContentKey}=$array[4]/$array[0]
		if ((@array==7)&&($array[0]!=0));
	} # if
	else
	{
	    $results[$tiles]{$infoContentKey}=0;
	} # else

    } # if
    else 
    {
	$results[$tiles]{$errorKey} = 0;
	$results[$tiles]{purity}=0;
    } # else

    $infoContentKey='infoContentRaw';
    $errorKey='errorRaw';

} # foreach suffix

    ### Now dealing with _all.txt files

    
#    $file = $file_name;
#    $file =~ s/align/all/;

$results[$tiles]{oneSig} = 0;
$results[$tiles]{twentySig} = 0;

#    if (defined(open(IN,$file))){ 

my $thisIntensity;
my @cyc2to10num;
my @cyc2to10inten;
my @cyc10to20num;
my @cyc10to20inten;


    if (defined(open(IN,"${stem}${allSuffix}")))
    { 
	@cyc2to10num=();
	@cyc2to10inten=();
	@cyc10to20num=();
	@cyc10to20inten=();
	
	while(<IN>)
	{
	    
	    # Changed: only check if first character is a comment
	    next if ($_ =~ m/^\#/);
	    
	    @array = split('\t',$_);

	    $thisIntensity
		=($array[2]+$array[3]+$array[4]+$array[5])/$numChannels;

	    if (($array[1]>=2)&&($array[1]<=4))
	    {
		$results[$tiles]{sigAvg2to4}+=$thisIntensity;
	    }
	    
	    if (($array[1]>=2)&&($array[1]<=10))
	    {
		push @cyc2to10num, $array[1];
		push @cyc2to10inten, $thisIntensity;
	    }

	    if (($array[1]>=10)&&($array[1]<=20))
	    {
		push @cyc10to20num, $array[1];
		push @cyc10to20inten, $thisIntensity;
	    }


	    if ($array[1] =~/^[123]$/)
	    {
	
		$results[$tiles]{threeSig}+=$thisIntensity;
		
		
		$results[$tiles]{oneSig}+=$thisIntensity
		    if ($array[1] eq '1');

	    } # if 
	    elsif($array[1] eq '20')
	    {

		$results[$tiles]{twentySig}+=$thisIntensity;
#		for($i=0;$i<4;$i++)
#		{
#		    $results[$tiles]{twentySig} += $array[$i+2];
#		}
#		
#		$results[$tiles]{twentySig} /= 4;
		last; # NB assumes cycle 20 data always comes after cycle 1
	    } # elsif 

	} # while

	if($results[$tiles]{oneSig} != 0)
	{
	    $results[$tiles]{twentySig} 
	    =  ($results[$tiles]{twentySig}
		/$results[$tiles]{oneSig}) * 100;
	} # if
	
	if (defined($results[$tiles]{threeSig}))
	{
	    $results[$tiles]{threeSig}/=3;
	} # i
	if (defined($results[$tiles]{sigAvg2to4}))
	{
	    $results[$tiles]{sigAvg2to4}/=3;
	}
	$results[$tiles]{sigLoss2to10}=logLinearFit
	    ( \@cyc2to10num, \@cyc2to10inten );
	$results[$tiles]{sigLoss10to20}=logLinearFit
	    ( \@cyc10to20num, \@cyc10to20inten );



    } # if
    else 
    {
	$results[$tiles]{oneSig} = 0;
	$results[$tiles]{threeSig} = 0;
	$results[$tiles]{twentySig} = 0;
    } # else
    
    $tiles++;
    
} # foreach (@seqFiles)



### Outputting the Lane Information

#print  "</table>\n";



#my @stats = qw/oneSig twentySig error PF_count PF_align PF_score count threeSig infoContent infoContentRaw errorRaw/;

my @stats_raw= qw/count infoContentRaw errorRaw/;

my @stats_filtered= qw/oneSig twentySig error PF_count PF_align PF_score threeSig infoContent sigAvg2to4 sigLoss2to10 sigLoss10to20/;

my @stats=( @stats_raw, @stats_filtered );

my $thisVal;
my $tileCount;
my $numSamples;

## Computing the Lane Results Table

for($lanes=1;$lanes<=$maxNumLanes;$lanes++)
{

    $oneSig = 0;
    $twentySig = 0;
    $error = 0;
    $filtered = 0;
    $score = 0;
    $align = 0;
    $number = 0;
    $tileCount = 0;

    $threeSig=0;
    $infoContent=0;
    $infoContentRaw=0;
    $errorRaw=0;


    if(exists $lane_type{$lanes})
    {

	for ($i=0;$i<$tiles;$i++)
	{

	    next if($results[$i]{lane} != $lanes);

#	    my @stats = qw/oneSig twentySig error PF_count PF_align PF_score count threeSig infoContent infoContentRaw errorRaw/;

	    foreach $thisStat (@stats_raw)
	    {
		$thisVal = 0;
		if (defined($results[$i]{$thisStat})) {
		    $thisVal=$results[$i]{$thisStat};
		}
		$laneResults[$lanes]{$thisStat}{mean}+=$thisVal;
		$laneResults[$lanes]{$thisStat}{sumsq}+=($thisVal*$thisVal);
	    } # foreach @stats

	    if ($results[$i]{PF_count}!=0)
	    {
		foreach $thisStat (@stats_filtered)
		{
		    $thisVal=$results[$i]{$thisStat};
		    $laneResults[$lanes]{$thisStat}{mean}
		    +=$thisVal;
		    $laneResults[$lanes]{$thisStat}{sumsq}
		    +=($thisVal*$thisVal);
		} # foreach @stats
	    } # if
	    $tileCount++;

	} # for tiles

	


	foreach $thisStat (@stats)
	{
	    # Need to average % of purity filtered sequences over all tiles
	    # Other stats should only be averaged over tiles that
	    # contained at least 1 sequence that passed the filter

	    $numSamples=$laneResults[$lanes]{nonZeroTiles};

	    if($thisStat eq 'PF_count')
	    {
		$numSamples=$tileCount;
	    } # if
	    else
	    {
		foreach (@stats_raw)
		{
		    if ($thisStat eq $_)
		    {
			$numSamples=$tileCount; 
#			last;
		    } # if
		} # foreach
	    } # else
#	    print STDERR "$thisStat $numSamples $lanes", "\n";
	    if (defined($laneResults[$lanes]{$thisStat})) {
		$laneResults[$lanes]{$thisStat}{stdev}
		=($laneResults[$lanes]{$thisStat}{sumsq}
		  - (($laneResults[$lanes]{$thisStat}{mean}
		      *$laneResults[$lanes]{$thisStat}{mean})/$numSamples))
		    unless ($numSamples==0);
		$laneResults[$lanes]{$thisStat}{stdev}
		/=($numSamples-1) if ($numSamples>=2);
		$laneResults[$lanes]{$thisStat}{stdev}
		=sqrt($laneResults[$lanes]{$thisStat}{stdev});
		$laneResults[$lanes]{$thisStat}{mean}/=$numSamples
		    unless ($numSamples==0);
	    }
	} # foreach (@stats)


#	$laneResults[$lanes]{runFolder}=$runFolder;
#	$laneResults[$lanes]{machine}=$machineName;
#	$chipResults{runFolder}=$runFolder;
#	$chipResults{machine}=$machineName;
	
    } # if
} #  for lanes

# Now ready to start printing the page

# Print some basic information 


print  "<!--RUN_TIME ".scalar(localtime)." -->\n";
print  "<!--SOFTWARE_VERSION $id-->\n";



print  "<html>\n<body>\n\n";
print  "<a name=\"Top\"><h2><title>$runFolder Summary</title></h2></a>\n";
print  "<h1>Summary Information For Experiment $runFolder on Machine $machineName</h1>\n";

# Print the chip summary. Idea is this contains all parameters 
# known prior to the experiment

print  "<h2><br></br>Chip Summary<br></br></h2>\n";

print  "<table border=\"1\" cellpadding=\"5\">\n";
printf  ("<tr><td>Machine</td><td>%s</td></tr>\n",uc($machineName));
printf  ("<tr><td>Run Folder</td><td>%s</td></tr>\n",
	 $runFolder);
printf  ("<tr><td>Chip ID</td><td>%s</td></tr>\n",
	 $chipName); # TBD
print "</table>\n";




# Print the chip summary. Idea is this contains all parameters 
# known prior to the experiment

print  "<h2><br></br>Lane Parameter Summary<br></br></h2>\n";

print  "<table border=\"1\" cellpadding=\"5\">\n<tr>\n";
print  "<td>Lane</td>\n";
print  "<td>Sample ID</td>\n";
print  "<td>Sample Target</td>\n";
print  "<td>Sample Type</td>\n";
print  "<td>Length</td>\n";
print  "<td>Filter</td>\n";
print  "<td>Tiles</td>\n";
print  "</tr>\n";


for($i=1;$i<=$maxNumLanes;$i++)
{
    print  "<tr>\n";
    print  "<td>$i</td>\n";
    printf  ("<td>%s</td>\n",$laneResults[$i]{sample});
    print  "<td>$laneResults[$i]{template}</td>\n";
    print  "<td>$laneResults[$i]{type}</td>\n";
    print  "<td>$laneResults[$i]{length}</td>\n";
    print  "<td>$laneResults[$i]{purity}</td>\n";
    print  "<td><a href=\"#Lane$i\">Lane $i</a></td>\n";
    print  "</tr>\n";
}

print  "</table>\n";

# Now print the lane summary table

print  "<h2><br></br>Lane Results Summary<br></br></h2>\n";

print  "<table border=\"1\" cellpadding=\"5\">\n<tr>\n\n";
print  "<td>Lane </td>\n";
print  "<td>Clusters </td>\n";
print  "<td>Av 1st Cycle Int </td>\n";
print  "<td>% intensity after 20 cycles </td>\n";
print  "<td>% PF Clusters </td>\n";
print  "<td>% Align (PF) </td>\n";
print  "<td>Av Alignment Score (PF) </td>\n";
print  "<td> % Error Rate (PF) </td>\n";
print  "</tr>\n";



for($lanes=1;$lanes<=$maxNumLanes;$lanes++)
{
    if(exists $lane_type{$lanes})
    {
	print  "<tr>\n";
	print "<td>$lanes</td>\n";

	foreach $thisStat 
	    (qw/count oneSig/)
	{
	    if (defined($laneResults[$lanes]{$thisStat})) {
		printf  ("<td>%d +/- %d</td>\n",
			 $laneResults[$lanes]{$thisStat}{mean},
			 $laneResults[$lanes]{$thisStat}{stdev});
	    }
	    else {
		printf ("<td>${unknown}</td>\n");
	    }
	} # for
	foreach $thisStat 
	    (qw/twentySig PF_count PF_align PF_score error/)
	{
	    if (defined($laneResults[$lanes]{$thisStat})) {
		printf  ("<td>%3.2f +/- %3.2f</td>\n",
			 $laneResults[$lanes]{$thisStat}{mean},
			 $laneResults[$lanes]{$thisStat}{stdev});
	    }
	    else {
		printf ("<td>${unknown}</td>\n");
	    }
	    
	} # for

	print  "</tr>\n";
    } # if
} # for lanes

print  "</table>\n";

# Now print the expanded lane summary table

print  "<h2><br></br>Expanded Lane Summary<br></br></h2>\n";

print  "<table border=\"1\" cellpadding=\"5\">\n<tr>\n\n";

print "<tr><td colspan=\"2\">Lane Info</td>\n";
print "<td colspan=\"2\">Phasing Info</td>\n";
print "<td colspan=\"2\">Raw Data</td>\n";
print "<td colspan=\"7\">Filtered Data</td></tr>\n";


print  "<td>Lane </td>\n";
print  "<td>Clusters </td>\n";
print  "<td>% Phasing </td>\n";
print  "<td>% Prephasing </td>\n";
print  "<td>% Error Rate (raw) </td>\n";
print  "<td> Equiv Perfect Clusters (raw) </td>\n";

print  "<td>% retained </td>\n";

print  "<td>Cycle 2-4 Av Int </td>\n";
print  "<td>Cycle 2-10 Av % Loss </td>\n";
print  "<td>Cycle 10-20 Av % Loss </td>\n";


#print  "<td>Av 1-3 Cycle Int </td>\n";
#print  "<td>% intensity after 20 cycles </td>\n";

print  "<td>% Align (PF) </td>\n";
print  "<td>% Error Rate (PF) </td>\n";
print  "<td> Equiv Perfect Clusters (PF) </td>\n";

print  "</tr>\n";


for($lanes=1;$lanes<=$maxNumLanes;$lanes++)
{

    if(exists $lane_type{$lanes})
    {


	print  "<tr>\n";
	print "<td>$lanes</td>\n";
	printf  ("<td>%d</td>\n",$laneResults[$lanes]{count}{mean});

	
	if (defined ($laneResults[$lanes]{phasingapplied}))
	{
	    printf  ("<td>%6.4f</td>\n",
		     $laneResults[$lanes]{phasingapplied}*100.);
	} # if
	else
	{
	    print "<td>$unknown</td>\n";
	} # else

	if (defined ($laneResults[$lanes]{prephasingapplied}))
	{
	    printf  ("<td>%6.4f</td>\n",
		     $laneResults[$lanes]{prephasingapplied}*100.);
	} # if
	else
	{
	    print "<td>$unknown</td>\n";
	} # else

	printf  ("<td>%3.2f</td>\n",$laneResults[$lanes]{errorRaw}{mean});
	printf  ("<td>%d</td>\n",$laneResults[$lanes]{infoContentRaw}{mean});
	if (defined($laneResults[$lanes]{PF_count})) {
	    printf  ("<td>%3.2f</td>\n",$laneResults[$lanes]{PF_count}{mean});
	}
	else {
	    print "<td>$unknown</td>\n";
	}

	if (defined($laneResults[$lanes]{sigAvg2to4})) {
	    printf  ("<td>%d +/- %d</td>\n",
		     $laneResults[$lanes]{sigAvg2to4}{mean},
		     $laneResults[$lanes]{sigAvg2to4}{stdev});
	} else {
	    print "<td>$unknown</td>\n";
	}
	foreach $thisStat (qw/sigLoss2to10 sigLoss10to20/)
	{
	    if (defined($laneResults[$lanes]{$thisStat})) {
		printf  ("<td>%3.2f +/- %3.2f</td>\n",
			 $laneResults[$lanes]{$thisStat}{mean},
			 $laneResults[$lanes]{$thisStat}{stdev});
	    } else {
		print "<td>$unknown</td>\n";
	    }
	}

#	printf  ("<td>%3.2f</td>\n",$laneResults[$lanes]{threeSig}{mean});
#	printf  ("<td>%3.2f</td>\n",$laneResults[$lanes]{twentySig}{mean});



	foreach $thisStat (qw/PF_align error/)
	{
	    if (defined($laneResults[$lanes]{$thisStat})) {
		printf  ("<td>%3.2f</td>\n",
			 $laneResults[$lanes]{$thisStat}{mean});
	    } else {
		print "<td>$unknown</td>\n";
	    }
	}
	if (defined($laneResults[$lanes]{infoContent})) {
	    printf  ("<td>%d</td>\n",$laneResults[$lanes]{infoContent}{mean});
	} else {
	    print "<td>$unknown</td>\n";
	}

	print  "</tr>\n";
    } # if
} # for lanes

print  "</table>\n";


print  "<b><br></br>IVC Plots</b>\n";
print  "<p> <a href='IVC.htm' target=\"_blank\"> IVC.htm\n </a></p>\n";
print  "<b><br></br>All Intensity Plots</b>\n";
print  "<p> <a href='All.htm' target=\"_blank\"> All.htm\n </a></p>\n";
print  "<b><br></br>Error graphs: </b>\n";
print  "<p> <a href='Error.htm' target=\"_blank\"> Error.htm\n </a></p>\n";
print  "<td><a href=\"#Top\">Back to top</a></td>\n";

## Making the results table

for($lanes=1;$lanes<=$maxNumLanes;$lanes++){

    if(exists $lane_type{$lanes}){

	print  "<a name=\"Lane$lanes\"><h2><br></br>Lane $lanes<br></br></h2></a>\n";

	print  "<table border=\"1\" cellpadding=\"5\">\n<tr>\n";
	print  "<td>Lane </td>\n";
	print  "<td>Tile </td>\n";
	print  "<td>Clusters </td>\n";
	print  "<td>Av 1st Cycle Int </td>\n";
	print  "<td>% intensity after 20 cycles </td>\n";
	print  "<td>% PF Clusters </td>\n";
	print  "<td>% Align (PF) </td>\n";
	print  "<td>Av Alignment Score (PF) </td>\n";
	print  "<td> % Error Rate (PF) </td>\n";
	print  "</tr>\n";


	
	for ($i=0;$i<$tiles;$i++){

	    next if($results[$i]{lane} != $lanes);
	    print  "<tr>\n";
	    print  "<td>$results[$i]{lane}</td>\n";
	    print  "<td>$results[$i]{tile}</td>\n";
	    print  "<td>$results[$i]{count}</td>\n";
	    printf  ("<td>%3.2f</td>\n",$results[$i]{oneSig});
	    printf  ("<td>%3.2f</td>\n",$results[$i]{twentySig});
	    printf  ("<td>%3.2f</td>\n",$results[$i]{PF_count});
	    printf  ("<td>%4.1f</td>\n",$results[$i]{PF_align});
	    printf  ("<td>%4.1f</td>\n",$results[$i]{PF_score});
	    print  "<td>$results[$i]{error}</td>\n";
	    print  "</tr>\n";
	}
	print  "</table>\n";


#print  "<h3>Links to graphical outputs</h3>";

	if($lane_type{$lanes} =~ m/DEFAULT/){

	    $cov_graphs = "s_".$lanes."_cov.png";
	    
	    print  "<b><br></br>Coverage plot: </b>\n";


	    print  "<p> <a href='${cov_graphs}' target=\"_blank\"> $cov_graphs\n </a></p>\n";
	    print  "<td><a href=\"#Top\">Back to top</a></td>\n";
	}

# Print out an extra table if this lane is a monotemplate lane

	if($lane_type{$lanes} =~ m/MONOTEMPLATE/){
#	    $monotemplates  = $ARGV[0]."Monotemplate_".$lanes.".txt";
	    $monotemplates  = "$ARGV[0]/s_${lanes}_monotemplate.txt";

#	    print STDERR $monotemplates;
	    if(! open(TEMPLATES,"$monotemplates")){
		print "<p> No template summary available</p>\n";
	    }
	    else {  ###### if data exsists
		@template=();
		
		print  "<h3>Monotemplate Summary</h3>\n";

		print  "<table border=\"1\" cellpadding=\"5\">\n\n\n";
		print  "<tr>\n";
		print  "<td>Lane </td>\n";
		print  "<td>Template </td>\n";
		print  "<td>Count </td>\n";
		print  "<td>Percent </td>\n";
		print  "<td>True 1st Cycle Intensity </td>\n";
		print  "<td>Av Error Rate </td>\n";
		print  "<td> % Perfect </td>\n";
		print  "</tr>\n";
		
		$template_count = 0;
		my $totalTemplates=0;

		while(<TEMPLATES>){
		    next if ($_ =~ m/\#/);
		    @array = split('\t',$_);

		    $status = 0;
		    ### check if new template
		    if($template_count == 0)
		    {
			$template[$template_count]{lane} = $array[0];
			$template[$template_count]{tag} = $array[2];
			$template[$template_count]{number} = $array[3];
			$totalTemplates+=$array[3];

			if($array[2] =~ m/^A/){
			    $template[$template_count]{intensity} = $array[4];
			}
			elsif($array[2] =~ m/^C/){
			    $template[$template_count]{intensity} = $array[5];
			}
			elsif($array[2] =~ m/^G/){
			    $template[$template_count]{intensity} = $array[6];
			}
			elsif($array[2] =~ m/^T/){
			    $template[$template_count]{intensity} = $array[7];
			}

			$template[$template_count]{error} 
			=  $array[12] + $array[13] + $array[14] 
			    + $array[15] + $array[16];

			$template[$template_count]{count}++;
			$template[$template_count]{perfect} = $array[17];
			$template_count++;
		    } # if
		    else 
		    {
			for($i=0;$i<$template_count;$i++)
			{
			    if($template[$i]{tag} =~ m/$array[2]/)
			    {
				## see this one before
				$template[$i]{error} 
				= $template[$i]{error} + $array[12] 
				    + $array[13] + $array[14] + $array[15] 
				    + $array[16];
				if($template[$i]{perfect} > $array[17])
				{
				    $template[$i]{perfect} = $array[17];
				} # if
				$template[$i]{count}++;
				
				$status++;
				
			    } # if
			} # for i
			if($status == 0)
			{
			    ## new template 
			    $template[$template_count]{lane} = $array[0];
			    $template[$template_count]{tag} = $array[2];
			    $template[$template_count]{number} = $array[3];
			    $totalTemplates+=$array[3];
			    

			    if($array[2] =~ m/^A/)
			    {
				$template[$template_count]{intensity} 
				= $array[4];
			    }
			    elsif($array[2] =~ m/^C/)
			    {
				$template[$template_count]{intensity} 
				= $array[5];
			    }
			    elsif($array[2] =~ m/^G/)
			    {
				$template[$template_count]{intensity} 
				= $array[6];
			    }
			    elsif($array[2] =~ m/^T/)
			    {
				$template[$template_count]{intensity} 
				= $array[7];
			    }

			    $template[$template_count]{error} 
			    = $array[12] + $array[13] + $array[14] 
				+ $array[15] + $array[16];

			    $template[$template_count]{count}++;
			    $template[$template_count]{perfect} = $array[17];
			    $template_count++;
			} # if
		    } # else
		} # while
		
		my $percentTemplate;
		for($i=0;$i<$template_count;$i++)
		{
		    $percentTemplate=0;
		    $percentTemplate
			=100.0*$template[$i]{number}/$totalTemplates
			unless ($totalTemplates==0);

		    print "<tr>\n";
		    print "<td>$template[$i]{lane}</td>\n";
		    print "<td>$template[$i]{tag}</td>\n";
		    print "<td>$template[$i]{number}</td>\n";
		    printf "<td>%5.2f</td>\n", $percentTemplate;
		    
#		    print "<td>$template[$i]{intensity}</td>\n";
		    printf "<td>%d</td>\n", $template[$i]{intensity};
		    $template[$i]{error}  /=  $template[$i]{count}
		    unless ($template[$i]{count}==0);
#		    print "<td>$template[$i]{error}</td>\n";
		    printf "<td>%5.2f</td>\n", $template[$i]{error};


		    print "<td>$template[$i]{perfect}</td>\n";
		    print "</tr>\n";
		}
		
		print  "</table>\n";

	    }
	    
	    print  "<b><br></br>Monotemplate IVC plot: </b>\n";
	    print  "<p> <a href='Monotemplate.htm' target=\"_blank\"> Monotemplate.htm\n </a></p>\n";
	    print  "<td><a href=\"#Top\">Back to top</a></td>\n";
	}
    } # if exists ($lane_type...

} # for lanes

# Final entry in HTML file
print  "</body>\n</html>\n";
