#!/usr/bin/env python

# Copyright (c) 2005, 2006 Solexa
# Author: Klaus Maisinger
# This source file is covered by the "Solexa Public Source License"
# agreement and bound by the terms therein 

"""Small library to serialise Python objects consisting of lists,
mappings, float, ints and strings to XML. This is similar to the Perl
XML::Simple module. The sematics of the <KeyAttr> options have changed
somewhat."""

import re, types, string, sys

def value_to_string(value):
    """Serialisation of simple data types to a string."""
    if type(value) == types.ListType:
        if len(value) > 1:
            print value
            sys.stderr.write("Error: Too many items in list for string "
                             "conversion!\n")
            sys.exit(1)
        return value_to_string(value[0])
    if type(value) == types.IntType:
        return "%i" % value
    if type(value) == types.FloatType:
        return "%f" % value
    if type(value) == types.StringType:
        return value
    

def XMLout(data,
           RootName   = 'opt',
           KeyAttr    = ['name', 'id', 'key'],
           XMLDecl    = 0,
           NoIndent   = 0,
           ContentKey = 'content',
           OutputFile = None,
           NoAttr     = 0,
           NoSort     = 0,
           mergelines = 1  # internal variable
           ):
    """Take a data structure of arbitrary types and convert it to
    XML. If <KeyAttr> is a list, then all elements of this list
    are treated as attributes rather than separate tags. If <KeyAttr>
    is a dictionary, then the values in the dictionary are treated as
    attributes to all XML tags that are given by the
    corresponding keys. If the flag <mergelines> is not set, the
    output is returned 
    as a list of strings (lines), without any line termination
    characters. If <XMLDecl> is a string, then the output starts with
    an optional xml declaration."""
    lines = []
    # optional XML declaration
    if XMLDecl:
        if type(XMLDecl) == types.StringType:
            lines.append(XMLDecl)
        else:
            lines.append("<?xml version=\"1.0\"?>")

    # remove any one-element arrays - this also undoes the ForceArray
    # option
    while (type(data) == types.ListType) and len(data) == 1:
        data = data[0]
    # complex data lives in dictionaries
    if type(data) == types.DictType:
        tags = data.keys()
        fulltags = [ RootName ]
        # assemble list of valid attributes for current tag
        attributes = []
        if not NoAttr:
            if type(KeyAttr) == types.ListType:
                attributes = KeyAttr[:]
            elif type(KeyAttr) == types.DictType:
                if KeyAttr.has_key(RootName):
                    if type(KeyAttr[RootName]) == types.ListType:
                        attributes = KeyAttr[RootName][:]
                    else:
                        attributes = [ KeyAttr[RootName] ]
        # now parse data for those attributes
        if not NoSort:
            attributes.sort()
        for tag in attributes:
            if data.has_key(tag):
                fulltags.append("%s=\"%s\"" % (tag, value_to_string(data[tag])))
                tags.remove(tag)
        if (len(tags) == 0):
            lines.append("<%s />" % " ".join(fulltags))
        elif (len(tags)==1) and (tags[0] == ContentKey):
            lines.append("<%s>%s</%s>" % (" ".join(fulltags),
                                          value_to_string(data[tags[0]]),
                                          RootName))
        else:
            lines.append("<%s>" % " ".join(fulltags))
            # now recursively parse the content of each tag, and
            # take care of indentation
            if not NoSort:
                tags.sort()
            indent = 2
            if NoIndent:
                indent = 0
            for tag in tags:
                lines += map(lambda x: "%s%s" % (" "* indent, x),
                             XMLout(data[tag], RootName = tag,
                                    KeyAttr = KeyAttr, mergelines = 0))
            lines.append("</%s>" % RootName)
    # data are in a list with more than element
    if (type(data) == types.ListType) and len(data) > 1: 
        for element in data:
            lines += XMLout(element, RootName=RootName,
                            KeyAttr = KeyAttr, mergelines = 0)
    # empty list
    elif (type(data) == types.ListType) and (len(data) == 0):
        lines.append("<%s />" % RootName)
    # simple data types: float, string, int
    elif type(data) in [ types.ListType, types.StringType,
                         types.FloatType, types.IntType]:
        lines.append("<%s>%s</%s>" % (RootName, value_to_string(data),
                                      RootName))

    # postprocess output:
    # merge lines to single string, write out to file if
    # <OutputFile> is specified
    if OutputFile or mergelines:
        if not NoIndent:
            lines = map(lambda x: x+"\n", lines)
        line = ''.join(lines)
        if OutputFile:
            if type(OutputFile) == types.StringType:
                filehandle = None
                try:
                    filehandle = open(OutputFile, "w")
                except:
                    sys.stderr.write("Unable to open %s\n" %
                                     OutputFile)
                    filehandle = None
            else:
                filehandle = OutputFile
            if filehandle:
                filehandle.writelines(lines)
                filehandle.close()
        return line
        
    return lines

def string_to_value(element):
    """Try to convert a string to a numeric data type."""
    try:
        value = int(element)
    except:
        try:
            value = float(element)
        except:
            value = element
    return value
    
def parse_tag(line):
    """Given a line of XML text, extract the first tag, comment or
    declaration from this line."""
    start = re.compile("<([^><]*)>")
    m = start.search(line)
    if m:
        fulltag = m.group(1)
        is_terminated = 0
        pos = [ m.start(), m.end()]
        # identify comments and declarations
        if fulltag[0] == "?" and fulltag[-1] == "?":
            return { "type": "declaration",
                     "value": fulltag[1:-1],
                     "pos": pos}
        if fulltag[0:3] == "!--" and fulltag[-2:] == "--":
            return { "type": "comment",
                     "value": fulltag[3:-2],
                     "pos" : pos}
        if fulltag[-1] == "/":
            is_terminated = 1
            fulltag = fulltag[:-1]
        # now parse the contents of the tag and identify atrributes
        parts = []
        quotes = [ "\'", "\""]
        is_quoted = 0
        is_nonwhitespace = 0
        first = 0
        for i in range(len(fulltag)):
            if fulltag[i] != " ":
                is_nonwhitespace = 1
                if fulltag[i] in quotes:
                    is_quoted = not is_quoted
            else:
                if is_quoted:
                    continue
                if is_nonwhitespace:
                    parts.append(fulltag[first:i])
                    first = i+1
                is_nonwhitespace = 0
        if is_nonwhitespace:
            parts.append(fulltag[first:])
        attributes = {}
        for i in range(1, len(parts)):
            try:
                attr, val = parts[i].split("=")
            except:
                sys.stderr.write("Attribute error: %s\n" % parts[i])
                sys.exit(1)
            if (val[0] in quotes) and (val[-1] in quotes):
                val = val[1:-1]
            attributes[attr] = val
        tag = parts[0]

        # now get content of tag
        value = None
        if not is_terminated:
            end = re.compile("</%s>" % tag)
            m = end.search(line)
            value = line[(pos[1]):(m.start())]
            pos[1] = m.end()
        return { "type": "tag",
                 "tag" : parts[0],
                 "value": value,
                 "attributes": attributes,
                 "pos" : pos}
    else:
        return None

def XMLin(line,
          ForceArray = 0,
          KeyAttr = ["id", "name", "key"],
          KeepRoot = 0,
          NoAttr = 0,
          ContentKey = 'content',
          ForceContent = 0):
    """Parse lines of XML produced by the XMLout() function. The
    <KeyAttr> argument is just ignored. If <KeepRoot> is not set,
    discard the top-level tag. This routine does not support mixed content."""
    # if line does not contain XML tags, treat it as a filename
    if (type(line) == types.StringType) and (not line.find("<") >= 0):
        filename = line
        line = open(filename).readlines()
    # convert input to a single string
    if type(line) == types.ListType:
        line = ''.join(line)
    result = {}
    # parse input string, one tag at a time
    current = parse_tag(line)
    while current:
        newresult = {}
        if current["type"] == "tag":
            tag = current["tag"]
            newresult[tag] = {}
            if not NoAttr:
                # read attributes
                for attr in current["attributes"].keys():
                    value = string_to_value(current["attributes"][attr])
                    if ForceArray:
                        value = [ value ]
                    newresult[tag][attr] = value

            if current["value"]:
                if parse_tag(current["value"]):
                    # more tags nested inside the current one: recurse
                    tmp = XMLin(current["value"], ForceArray=ForceArray,
                                KeyAttr=KeyAttr, KeepRoot=1)
                    for key in tmp.keys():
                        newresult[tag][key] = tmp[key]
                else:
                    # tag content is a simple type
                    value = string_to_value(current["value"])
                    if ForceArray:
                        value = [ value ]
                    if (ForceContent or
                        ((not NoAttr) and current["attributes"])):
                        newresult[tag][ContentKey] = value
                    else:
                        newresult[tag] = value

            # add results back into main data structure.
            # if tags are repeated, create a list.
            if result.has_key(tag):
                if type(result[tag]) == types.ListType:
                    if type(newresult[tag]) != types.ListType:
                        result[tag].append(newresult[tag])
                    else:
                        result[tag] += newresult[tag]
                        
                else:
                    result[tag] = [ result[tag], newresult[tag] ]
            else:
                result[tag] = newresult[tag]
                if ForceArray and type(newresult[tag]) != types.ListType:
                    result[tag] = [ result[tag] ]
                    
        line = line[(current["pos"][1]+1):]
        current = parse_tag(line)

    if not KeepRoot:
        tag = result.keys()[0]
        if ((type(result[tag]) == types.ListType)
            and len(result[tag]) == 1):
            return result[tag][0]
        return result[tag]
    return result

if __name__ == "__main__":
    testobject =\
               { "field1": 1.,
                 "field2" : [ { "name": 2 },
                              {"name": 3}],
                 "field3": "test",
                 "field4": [ 1, 2, 3],
                 "field5" : { "name": "test",
                              "second1": [ 1, 2 ],
                              "second2": 2.
                              },
                 "field6" : [ { "name" : 1,
                                "third1" : 3,
                                "third2" : "test3"
                                },
                              { "name" : 2,
                                "third1" : 4,
                                "third2" : "test4"
                                }
                              ],
                 "field7" : "This is a very long string including some\
                 line breaks.\n This is a very long string containing\
                 some line breaks.\n",
                 "field8": { "name": "test", "content": "value" }
                 }

    lines = XMLout(testobject, RootName="testobject",
                   KeyAttr=[])
    print lines
    lines = XMLout(testobject, RootName="testobject",
                   KeyAttr={"field6" : "name"})
    print lines
    lines = XMLout(testobject, RootName="testobject")
    print lines
    print testobject
    print XMLin(lines, ForceArray=0)
    print XMLin(lines, ForceArray=1)
    lines = XMLout(XMLin(lines, ForceArray=1),
                   RootName = "testobject")
    print XMLin(lines, ForceArray=0)
    print lines
    print XMLin(open("test1.xml").readlines())
    print
    print XMLin(open("SBS121.xml").readlines())
