#! /usr/bin/env python

#
# Copyright (c) Solexa 2005-2007
# Author: Klaus Maisinger
#
# This source file is covered by the "Solexa Public Source License"
# agreement and bound by the terms therein.

#
# program name: 'Bustard'
# CVS - $Id: bustard.py,v 1.31.2.1 2007/06/12 11:46:57 km Exp $      
#             
# script to configure the analysis environment and
# generate the Makefiles for:
#  . base-calling
#  . sequencing


import sys, os, os.path, glob, getopt, shutil
import runfolder, common, makefile, XMLSimple, gerald

basecaller_software         = "Bustard"
basecaller_software_version = "1.8.28"
basecaller_executable = "run_base_caller"
phasingpath           = "Phasing"
phasing1_executable   = os.path.join(phasingpath, "phasing_estimate")
phasing2_executable   = os.path.join(phasingpath, "phasing_summary.py")
phasing3_executable   = os.path.join(phasingpath, "select_phasing.py")


class BaseCallParams:
    def __init__(self): 
        self.phasing_rate = 0.
        self.prephasing_rate = 0.
        self.autophasingflag = 1
        self.autophasinglane = 0
        self.automatrixflag = 1
        self.automatrixlane = 0
        self.matrixfile = ""
        self.phasing_restarts = []
        self.options = [ ]

    def parse(self, options):
        """Set values according to command-line options from
        getopt.getopt"""
        for o, a in options:
            if o == "--matrix":
                if (a.lower().find("auto") == 0):
                    self.automatrixflag = 1
                    if (len(a) == 5):
                        try:
                            self.automatrixlane = int(a[4])
                            if self.automatrixlane > 8:
                                self.automatrixlane = 8
                        except:
                            sys.stderr.write("Error: Cannot parse option %s\n" % o)
                elif (a.lower().find("default") == 0):
                    self.automatrixflag = 0
                else:
                    self.automatrixflag = 0
                    self.matrixfile = os.path.abspath(a)
            if o == "--phasing":
                if (a.lower().find("auto") == 0):
                    self.autophasingflag = 1
                    self.phasing_rate = 0.
                    self.prephasing_rate = 0.
                    if (len(a) == 5):
                        try:
                            self.autophasinglane = int(a[4])
                            if self.autophasinglane > 8:
                                self.autophasinglane = 8
                        except:
                            sys.stderr.write("Error: Cannot parse option %s\n" % o)
                else:
                    self.autophasingflag = 0
                    self.phasing_rate = float(a)
        
            if o == "--prephasing":
                self.autophasingflag = 0
                self.prephasing_rate = float(a)
            if o == "--reset-phasing-cycle":
                self.phasing_restarts.append(int(a))

    def print_params(self):
        # Matrix
        if not self.automatrixflag:
            print "Matrix file: %s" % self.matrixfile
        else:
            if self.automatrixlane:
                print ("Matrix:      Using automatic matrix estimation "
                       " (lane %i only)." % self.automatrixlane)
            else:
                print "Matrix:      Using automatic matrix estimation."
        
        # Phasing, Prephasing
        if self.autophasingflag:
            if self.autophasinglane:
                print ("Phasing:     Using automatic phasing estimation" +
                       " (lane %i only).") % (self.autophasinglane)
            else:
                print "Phasing:     Using automatic phasing estimation."
        else:
            print "Phasing: %6.4f  prephasing: %6.4f" \
                  % (self.phasing_rate, self.prephasing_rate)
            if ((self.phasing_rate > 0.05) or
                (self.prephasing_rate > 0.05)):
                sys.stderr.write("Error: phasing/prephasing values "
                                 "suspiciously high! A typo ??\n")
                sys.exit(1)
        if self.phasing_restarts:
            print "Multiple runs on the same cluster (resetting phasing cycles)"
        

def print_tile_counts(tiledic):
    tiles = tiledic.keys()
    tiles.sort()
    for key in tiles:
        print "%10s %3i" % (key, tiledic[key])
    

def getTiles(analysisfolder):
    paramsfile = os.path.join(runfolder.get_data_folder(analysisfolder),
                              ".params")
    if paramsfile:
        params = runfolder.get_params(paramsfile, ["ImageAnalysis", "data"])
        run = runfolder.get_run(params, analysisfolder)
        if run:
            tiles = runfolder.get_tiles_from_tile_selection(run)
            maxcycle = run["Cycles"]['Number']
            tiledic = {}
            for tile in tiles:
                tiledic[tile] = maxcycle
            return tiledic
    print "Could not identify tiles from parameter file."
    return common.count_intensity_files(analysisfolder)


def getReadLength(tiledic):
    return min(tiledic.values())


def getGeraldConfigs(geraldconfigfiles, run_folder, readlength):
    geraldconfigs = []
    for geraldconfigfile in geraldconfigfiles:
        geraldconfig = gerald.GeraldConfig(geraldconfigfile, run_folder)
        geraldconfig.setReadLength(readlength)
        if geraldconfig.modified:
            print "Updating Gerald config.:"
            geraldconfig.write()
        geraldconfigs.append(geraldconfig)
    return geraldconfigs


def processGeraldConfig(geraldflag, geraldconfigs,
                        analysis_folder, sequence_folder):
    for i in range(len(geraldconfigs)):
        geraldconfig = geraldconfigs[i]
        filename = geraldconfig.filename
        if geraldconfig.modified:
            filename = os.path.join(analysis_folder,
                                    "config%i.txt" % (i+1))
            print "Writing config. to %s." % filename
            geraldconfig.writeFile(filename)
        if geraldflag:
            gerald.run_GERALD(runfolder.get_executable(gerald.gerald_executable),
                              filename, sequence_folder)
            

def createBasecallerFolders(sequence_folder, tiles):
    lanes = runfolder.get_lane_samples(tiles)
    try:
        os.makedirs(sequence_folder)
        os.mkdir(os.path.join(sequence_folder, runfolder.sequencesubdir))
        os.mkdir(os.path.join(sequence_folder, runfolder.phasingdir))
    except:
        sys.stderr.write("Encountered error while creating folder %s. "
                         "Perhaps you do not have the permissions?"
                         % sequence_folder)
        sys.exit(1)


def processParamsFile(sequence_folder, tiledic,
                      basecaller_phasing_rate, basecaller_prephasing_rate):
    for lane in runfolder.get_lane_samples(tiledic.keys()).keys():
        paramsfile = os.path.join(sequence_folder, "params%i.xml" % (lane+1))
        params = { "Phasing":    basecaller_phasing_rate,
                   "Prephasing": basecaller_prephasing_rate }
        XMLSimple.XMLout(params, RootName="Parameters", OutputFile = paramsfile)


def usage():
    print "Usage: bustard.py [--tiles=<tile1>[,<tile2>[,...]]] \
[--maxcycle=<number>] [--make] \
[--matrix=<matrix filename>|auto] \
[--phasing=<phasing rate>|auto|auto1|...] [--prephasing=<prephasing rate>] \
[--reset-phasing-cycle=<restart cycle>] \
[--GERALD=<config file>] \
<image analysis directory>"


#
# -- main program --
#
if __name__ == "__main__":
    
    print "Starting the Bustard base-caller, %s version %s"\
          % (basecaller_software,
             basecaller_software_version)

    runfolder.validate_executables([ basecaller_executable,
                                     phasing1_executable ])

    # parsing command-line options
    try:
        (options, args) = getopt.gnu_getopt(sys.argv[1:], ":m",
                                            ["tiles=",
                                             "maxcycle=", "matrix=",
                                             "make",
                                             "phasing=", "prephasing=",
                                             "reset-phasing-cycle=",
                                             "GERALD="])
    except getopt.GetoptError:
        sys.stderr.write("Error: failed to parse command-line options.\n")
        usage()
        sys.exit(1)
        
    if (len(args) != 1):
        usage()
        sys.exit(2)
        
    #print options, args
    
    selected_tiles = []
    maxcycle = -1
    runflag = 0
    makeflag = 0
    geraldflag = 0
    geraldconfigfiles = []
    base_params = BaseCallParams()
    base_params.parse(options)
    for o, a in options:
        if o == "--tiles":
            selected_tiles = a.split(",")
        if o == "--folders":
            selected_folders = a.split(",")
        if o == "--maxcycle":
            maxcycle = int(a)
        if o == "--make":
            makeflag = 1
            runflag = 1
        if o == "--GERALD" or o == "--config":
            path = os.path.abspath(a)
            if not os.path.exists(path):
                sys.stderr.write("Error: Cannot find %s!\n" %
                                 path)
                sys.exit(2)
            geraldconfigfiles.append(path)
            if o == "--GERALD":
                geraldflag = 1


    # Analysis folder
    analysis_folder = os.path.abspath(args[0])
    if os.path.split(analysis_folder)[1] == runfolder.analysissubdir:
        analysis_folder = os.path.split(analysis_folder)[0]
    
    # grouping filenames by position on chip
    tiledic = getTiles(analysis_folder)
    tiledic = common.select_tiles(tiledic, selected_tiles)
    if maxcycle > 0:
        for tile in tiledic.keys():
            tiledic[tile] = min([ tiledic[tile], maxcycle ])
    if (not tiledic) or (len(tiledic.values()) == 0):
        sys.stderr.write("Could not locate image analysis run.\n")
        sys.exit(1)
    print_tile_counts(tiledic)

    # Sequence folder
    sequence_folder =\
                runfolder.get_new_sequence_folder(analysis_folder,
                                                  basecaller_software,
                                                  basecaller_software_version)
    print  "Analysis folder: %s" % analysis_folder
    print  "Sequence folder: %s" % sequence_folder

    # Instrument
    instrument = common.get_instrument(analysis_folder)
    if instrument:
        print "Instrument:  %s" % instrument

    # Matrix
    if not base_params.automatrixflag:
        instrument_folder = os.getenv('INSTRUMENT_DIR')
        base_params.matrixfile = common.get_matrix(base_params.matrixfile,
                                                   instrument_folder,
                                                   analysis_folder)

    base_params.print_params()

    # Gerald configuration file
    readlength = getReadLength(tiledic)
    geraldconfigs = []
    geraldconfigs = getGeraldConfigs(geraldconfigfiles,
                                     runfolder.get_run_folder(analysis_folder),
                                     readlength)

    #--
    if not runflag:
        print ("You can generate the Makefiles by using the \"--make\" "
               "option.")
        sys.exit(0)
    #--

    # Base-caller folders
    tiles = tiledic.keys()
    tiles.sort()
    createBasecallerFolders(sequence_folder, tiles)

    if base_params.matrixfile and not base_params.automatrixflag:
        for lane in runfolder.get_lane_samples(tiledic.keys()).keys():
            matrixfile2 = os.path.join(sequence_folder,
                                       "matrix%i.txt" % (lane+1))
            shutil.copyfile(base_params.matrixfile, matrixfile2)

    if not base_params.autophasingflag:
        processParamsFile(sequence_folder, tiledic,
                          base_params.phasing_rate,
                          base_params.prephasing_rate)

    print "Generating Makefiles and parameter files ..."

    # generating Makefile for the base-caller
    makefile.generate_basecaller_makefile(tiledic, analysis_folder,
                                          sequence_folder, None,
                                          base_params.phasing_rate,
                                          base_params.prephasing_rate,
                                          base_params.phasing_restarts,
                                          base_params.automatrixflag,
                                          base_params.autophasingflag,
                                          base_params.automatrixlane,
                                          base_params.autophasinglane)

    # generating .params file for base-calls
    cyclerange = range(1, max(tiledic.values())+1)
    params = [ base_params.phasing_rate,
               base_params.prephasing_rate ]
    runfolder.write_sequence_params(analysis_folder, sequence_folder,
                                    tiles, cyclerange,
                                    basecaller_software,
                                    basecaller_software_version,
                                    runfolder.get_login(),
                                    params, base_params.matrixfile)

    # processing Gerald configuration file
    processGeraldConfig(geraldflag, geraldconfigs,
                        analysis_folder, sequence_folder)
