#! /usr/bin/env python
#
# Copyright (c) Solexa 2007
# Author: Klaus Maisinger
#
# This source file is covered by the "Solexa Public Source License"
# agreement and bound by the terms therein.

import sys, os, glob, re
import runfolder

"""Common routines between goat_pipeline and bustard."""

def select_tiles(tiledic, selected_tiles =[]):
    """Select tiles from a dictionary whose keys refer to tiles. Only part
    of the key needs to match the selected tiles."""
    if selected_tiles:
        a = re.compile("(%s)" % "|".join(selected_tiles))
        for key in tiledic.keys():
            if not a.match(key):
                del tiledic[key]
    return tiledic

def count_intensity_files(analysis_folder):
    """Count the number of intensity files per tile in the image analysis
    working directory."""
    searchfolder = os.path.join(analysis_folder, runfolder.analysissubdir)
    searchfolder = os.path.join(searchfolder, "*_int.txt")
    filenames = glob.glob(searchfolder)
    if (len(filenames) == 0):
        searchfolder = os.path.join(analysis_folder, runfolder.analysissubdir)
        searchfolder = os.path.join(searchfolder,
                                    runfolder.LaneDirectory.wildcard)
        searchfolder = os.path.join(searchfolder, "*_int.txt")
        filenames = glob.glob(searchfolder)
    filenames.sort(runfolder.FileName.cmp)
    # group filenames by position on chip
    tiledic = group_dictionary(runfolder.FileName.tilestr, filenames)
    # validate file counts
    tiles = tiledic.keys()
    tiles.sort()
    for key in tiles:
        cycles = map(lambda x: int(x.split("_")[-2]), tiledic[key])
        if min(cycles) != 1:
            sys.stderr.write(
                "Warning: first cycle data missing for tile %s."
                "Removing tile from analysis.\n" % key)
            del tiledic[key]
            continue
        if max(cycles) != len(cycles):
            sys.stderr.write(
                "Warning: Data missing for tile %s. "
                "Removing tile from analysis.\n" % key)
            del tiledic[key]
            continue
        tiledic[key] = max(cycles)
    return tiledic

def group_dictionary(function, list):
    """Create a dictionary where each key corresponds to a function
    value and the corresponding values are all list elements with the
    same function value.""" 
    dict = {}
    for item in list:
        id = function(item)
        dict.setdefault(id, []).append(item)
    return dict
        

def group(function, list):
    """Create a list of lists each containing elements that yield the
    same result when the function is applied. This creates groups
    of elements that have the same function value."""
    # sort by key value and return array
    dic = group_dictionary(function, list)
    keys = dic.keys()
    keys.sort()
    return [ dic[key] for key in keys ]


def get_instrument(analysis_folder):
    """Return name of instrument from runfolder parameter file"""
    paramsfile = runfolder.get_run_params_filename(analysis_folder)
    instrument = ""
    if os.path.isfile(paramsfile):
        params = runfolder.get_params(paramsfile)
        try:
            instrument = params["Run"][0]["instrument"]
        except:
            print params
            sys.stderr.write("No instrument id found in parameter file %s\n"
                             % paramsfile)
    else:
        sys.stderr.write("Parameter file %s not found!\n" % paramsfile)
    return instrument

def get_matrix(matrixfilename, instrument_folder, analysis_folder):
    """Return path to matrix file"""
    matrixfile = matrixfilename or \
                 runfolder.find_instrument_matrix_file(analysis_folder,
                                                       instrument_folder,
                                                       get_instrument(analysis_folder))
    if not os.path.exists(matrixfile):
        sys.stderr.write("No matrix file %s found!\n" % matrixfile)
        return ""
    else:
        return matrixfile
