#!/usr/bin/env python

# parse GERALD input file and update readlengths if necessary
#
# Copyright (c) Solexa 2005, 2006
# Author: Klaus Maisinger
# This source file is covered by the "Solexa Public Source License"
# agreement and bound by the terms therein.

import os, sys, string
import XMLSimple

"""Parse Gerald config files and modify read-lengths."""

gerald_executable     = "GERALD.pl"
gerald_config         = "GERALD.xml"
toolspath             = os.getenv("TOOLS_DIR") or "../Gerald"
gerald_executable     = os.path.join(toolspath, gerald_executable)

def get_my_path():
    return os.path.dirname(os.path.abspath(sys.argv[0]))

class GeraldConfig:
    """Read and write Gerald config files."""
    def __init__(self, filename = None, run_folder = None):
        self.data = {}
        self.data["lane"] = [ {} for i in range(8) ]
        self.run_folder = run_folder
        if filename:
            self.read(filename)

    def read(self, filename):
        """Read configuration file from disk."""
        try:
            lines = open(filename).readlines()
        except:
            sys.stderr.write("Error: Failed to read %s\n" % filename)
            return
        for line in lines:
            # skip comments
            idx = line.find('#')
            if idx >=0:
                line = line[:idx]
            # parse key, value pairs
            pairs = line.split()
            if len(pairs) == 0:
                continue
            if len(pairs) == 1:
                pairs.append("")
            value = " ".join(pairs[1:])
            parts2 = pairs[0].split(":")
            if len(parts2) >= 2:
                lanes = map(lambda x: int(x)-1, list(parts2[0]))
                for lane in lanes:
                    self.data["lane"][lane][parts2[1]] = value
            else:
                self.data[pairs[0]] = value

        self.modified = 0
        self.filename = filename

        # copy general tags to lanes for all ANALYSIS targets
        for tag in ["ANALYSIS", "READ_LENGTH", "USE_BASES",
                    "CONTAM_FILE", "CONTAM_DIR"]:
            if self.data.has_key(tag):
                for lane in range(len(self.data["lane"])):
                    if ((not self.data["lane"][lane].has_key(tag))
                        and (self.data["lane"][lane].get("ANALYSIS") !=
                             "none")):
                        self.data["lane"][lane][tag] = self.data[tag]
                del self.data[tag]

        # no ANALYSIS target means ANALYSIS default
        for lane in self.data["lane"]:
            if (not lane.has_key("ANALYSIS")):
                lane["ANALYSIS"] = "none"
        
        # copy general tags to lanes for default, monotemplate ANALYSIS targets
        for tag in ["GENOME_FILE", "GENOME_DIR" ]:
            if self.data.has_key(tag):
                for lane in range(len(self.data["lane"])):
                    if ((not self.data["lane"][lane].has_key(tag))
                        and (self.data["lane"][lane].get("ANALYSIS") in
                             ["monotemplate", "default", "expression"])):
                        self.data["lane"][lane][tag] = self.data[tag]
                del self.data[tag]
        # copy general tags to lanes for eland ANALYSIS targets
        for tag in ["ELAND_FILE", "ELAND_DIR"]:
            if self.data.has_key(tag):
                for lane in range(len(self.data["lane"])):
                    if ((not self.data["lane"][lane].has_key(tag))
                        and (self.data["lane"][lane].get("ANALYSIS") ==
                             "eland")):
                        self.data["lane"][lane][tag] = self.data[tag]
                del self.data[tag]
            

    def setReadLength(self, length):
        """Update Gerald parameters to a new maximum read-length."""
        for lane in self.data["lane"]:
            if lane.get("ANALYSIS").lower() == "none":
                continue
            if not lane.get("USE_BASES"):
                continue
            readlength = length
            if (lane.has_key("READ_LENGTH")):
                readlength = int(lane["READ_LENGTH"])
                if readlength > length:
                    readlength = length
                    self.modified = 1
            else:
                self.modified = 1
            if lane.get("ANALYSIS").lower() in ["monotemplate", "sequence",
                                                "default", "eland"]:
                # find length of shortest tag in monotemplate reference file
                if lane.get("ANALYSIS").lower() == "monotemplate":
                    if lane.has_key("GENOME_FILE"):
                        genomedir = (lane.get("GENOME_DIR")
                                     or os.getenv("GENOME_DIR")
                                     or os.path.join(get_my_path(),
                                                     "../../Genomes"))
                        # try to handle relative paths in GENOME_DIR.
                        # problems is that relative paths are with
                        # respect to potentially not yet existing
                        # Gerald folders
                        if not os.path.isabs(genomedir):
                            parts = genomedir.split(os.sep)
                            if len(parts) > 5:
                                i = 0
                                while ((parts[i] == "..")
                                       and (i < 4)):
                                    i += 1
                                genomedir = os.path.join(self.run_folder,
                                                         os.sep.join(parts[i:]))
                                       
                        filename = os.path.join(genomedir,
                                                lane.get("GENOME_FILE"))
                        taglength = readlength
                        try:
                            taglength = min(map(len,
                                                map(string.rstrip, open(filename).readlines() )) )
                        except:
                            sys.stderr.write("Error: Could not read %s\n" %
                                             filename)
                        if taglength < readlength:
                            self.modified = 1
                            readlength = taglength
                # update USE_BASES string if necessary
                if lane.get("USE_BASES").lower().find("y") >= 0:
                    bases = lane["USE_BASES"]
                    ignorecount = 0
                    for i in range(len(bases)):
                        if bases[i].lower() == "n":
                            ignorecount += 1
                    use_bases = length - ignorecount
                    # subtract last cycle (used for dephasing) if READ_LENGTH
                    # not specified
                    if use_bases > 20 and not (lane.has_key("READ_LENGTH") and
                                               lane["READ_LENGTH"] >= use_bases):
                        use_bases -= 1
                    if readlength > use_bases:
                        readlength = use_bases
                        self.modified = 1
                    # unless explicit read-length specified, restrict
                    # BAC control lanes to 25 cycles
                    #if (lane["ANALYSIS"] == "default" and
                    #    lane.get("GENOME_FILE").find("BAC") >= 0):
                    #    if not lane.has_key("READ_LENGTH"):
                    #        if readlength > 25: 
                    #            self.modified = 1
                    #            readlength = 25
                    # put correct number of "Y"s into USE_BASES string
                    if len(bases) != readlength+ignorecount:
                        #print len(bases), readlength, ignorecount
                        self.modified = 1
                    if len(bases) > readlength+ignorecount:
                        lane["USE_BASES"] = bases[:(readlength+ignorecount)]
                    if len(bases) < readlength+ignorecount:
                        lane["USE_BASES"] = bases + "Y"*(readlength
                                                         +ignorecount
                                                         -len(bases))
                        
                lane["READ_LENGTH"] = "%i" % readlength
                if readlength < 12:
                    lane["PURE_BASES"] = "%i" % readlength

    def write(self, output=sys.stdout):
        """Print out current configuration settings."""
        keys = self.data.keys()
        keys.remove("lane")
        keys.sort()
        for key in keys:
            output.write("%s %s\n" % (key, self.data[key]))
        for lane in range(len(self.data["lane"])):
            for key in self.data["lane"][lane]:
                output.write("%i:%s %s\n" % (lane+1, key,
                                             self.data["lane"][lane][key]))

    def writeFile(self, filename):
        """Write configuration to a file."""
        try:
            fd = open(filename, "w")
            self.write(fd)
            fd.close()
        except:
            sys.stderr.write("Error: Could not write %s\n")



def getGeraldXML():
    data = {}
    filename = os.path.join(toolspath, gerald_config)
    filename = os.path.join(get_my_path(), filename)
    try:
        data = XMLSimple.XMLin(filename)
    except:
        sys.stderr.write("Error: Failed to read %s\n" % filename)
    return data
    
def getQhgInfo():
    data = getGeraldXML()
    pure_bases = 12
    use_bases = "all"
    if data.has_key('DefaultRunParameters'):
        if data['DefaultRunParameters'].has_key('PURE_BASES'):
            pure_bases = int(data['DefaultRunParameters']['PURE_BASES'])
        if data['DefaultRunParameters'].has_key('USE_BASES'):
            use_bases = data['DefaultRunParameters']['USE_BASES']
    return (pure_bases, use_bases)


def run_GERALD(executable, configfile, sequencefolder, norun = 0):
    """Run the GERALD makefile generator for sequence alignment."""
    commandline = "%s %s --EXPT_DIR %s" % (executable,
                                           configfile, sequencefolder)
    print
    if not norun:
        commandline += " --FORCE"
    print commandline
    os.system(commandline)
    #output = os.popen4(commandline)[1].read()
    #print output
