#! /usr/bin/env python

#
# Copyright (c) 2005-2007 Solexa
# Author: Klaus Maisinger, Lorenzo Milazzo
#
# This source file is covered by the "Solexa Public Source License"
# agreement and bound by the terms therein.
#

# program name: 'Goat-Pipeline'
# CVS - $Id: goat_pipeline.py,v 1.62.2.4 2007/06/28 14:44:44 km Exp $      
#             
# script to configure the analysis environment;
# generation of the Makefiles for:
#  . image analysis 
#  . base-calling
#  . sequencing

import os, os.path, sys, getopt, glob, re, shutil
import runfolder, makefile, matrix, XMLSimple, common, bustard, gerald
from manteia     import ManteiaFilePath
from swerdlotron import SwerdlotronFileName

"""Script to start the Solexa Image and Sequence Analysis Pipeline on
a given experimental run-folder produced by a Solexa instrument."""

class ImageParams:
    def __init__(self): 
        self.threshold           =  4.
        self.fwhm                =  2.7
        self.remapping_distance  =  1.5

image_software              = "Firecrest"
image_software_version      = "1.8.28"
image_executable      = "run_image_analysis"
image2_executable     = "analyse_image"
offset_executable     = "parse_offset_scaling.py"
offset2_executable    = "update_offsets.py"
matrixpath            = "Matrix"
matrix1_executable    = os.path.join(matrixpath, "savinio")
matrix3_executable    = os.path.join(matrixpath, "summarise_matrices.py")
matrix4_executable    = os.path.join(matrixpath, "select_matrices.py")


def get_tiff_files(files):
    """Return all tiff files given in a list of files or directories."""
    allnames = []
    for file in files:
        if os.path.isdir(file):
            allnames += glob.glob(os.path.join(file, "*.tif"))
            allnames += glob.glob(os.path.join(file, "*.stk"))
        else:
            allnames += [file]
    allnames = filter(re.compile("\.[(tif)|(stk)]").search, allnames)
    return allnames


def get_bottom_level_folders(filelist):
    """Obtain folder names in which image files reside."""
    folders = []
    for folder in filter(None, map(runfolder.get_bottom_level_folder,
                                   filelist)):
        if folder not in folders:
            folders.append(folder)
    return folders


def get_cycle_params(files, FileNameClass):
    """Given a list of image files, extract number of cycles, colours
    and a string describing the alternation of incorporation/deblock
    cycles."""
    if (len(files) == 0):
        return (0, 0, "")
    firstcycle = FileNameClass.cycle(files[0])
    
    firstdeblock = FileNameClass.is_deblocked(files[0])
    ncolours = 1
    for i in range(1, len(files)):
        if ((FileNameClass.cycle(files[i]) == firstcycle) and
            ((FileNameClass.is_deblocked(files[i]) == firstdeblock))):
            ncolours += 1
        else:
            break
    ncycles = len(files)/ncolours
    
    deblock = ['C'] * ncycles
    for i in range(ncycles):
        if FileNameClass.is_deblocked(files[i*ncolours]):
            deblock[i] = 'D'
    return (ncycles, ncolours, "".join(deblock))
        

def write_image_journal(filename, files, FileNameClass):
    """Write out a journal file for the image analysis. This is used
    to pass on a list of filenames to be analysed to the image
    analysis software. The format is archaic and derived from a legacy
    application. This will be replaced by an XML file."""
    ncycles, ncolours, deblock = get_cycle_params(files, FileNameClass)
    #print ncycles, ncolours
    try:
        fd = open(filename, "w")
        for file in files:
            fd.write("Load \"%s\"\n" % file)
            command = "SeqIntensities"
            sequencecommand = "%s \"%s\", %d, %d: %s" % \
                      (command, "", ncycles, ncolours,
                       ",".join(map(lambda x: "#%d-%c"
                                    % (x, FileNameClass.base(files[x]).upper()),
                                    range(len(files)))))
        fd.write("%s\n" % sequencecommand)
        fd.close()
    except:
        sys.stderr.write("Error: Could not create %s\n" % filename)
        sys.exit(2)


def run_generate_journal(analysis_folder, sampleid, lane, tile,
                         files, FileNameClass):
    """Generate journal file for the image analysis."""
    # generate folder for intermediate files
    tmp_folder = runfolder.get_lane_subfolder(analysis_folder,
                                              runfolder.analysissubdir,
                                              sampleid, lane, tile)
    journalfile = os.path.join(tmp_folder,
                               runfolder.FileName.name(sampleid, "jnl",
                                                       lane, tile))
    write_image_journal(journalfile, files, FileNameClass)


def rename_image_folders(old_analysis_folder, analysis_folder):
    """Rename image folders."""
    try:
        os.rename(old_analysis_folder, analysis_folder)
        for filename in glob.glob(os.path.join(analysis_folder,
                                           "*.old")):
            os.remove(filename)
        for filename in glob.glob(os.path.join(analysis_folder,
                                           "*_int.txt")):
            os.rename(filename, filename+".old")
        for filename in glob.glob(os.path.join(analysis_folder,
                                           "*_nse.txt")):
            os.rename(filename, filename+".old")
        for filename in glob.glob(os.path.join(analysis_folder,
                                           "*finished.txt")):
            os.remove(filename)
        
    except:
        sys.stderr.write("Failed to rename %s to %s!\n"
                         % (old_analysis_folder, analysis_folder))
        sys.exit(1)


def create_image_folders(analysis_folder, tiles):
    lanes = runfolder.get_lane_samples(tiles)
    try:
        os.makedirs(analysis_folder)
        os.mkdir(os.path.join(analysis_folder,
                              runfolder.analysissubdir))
        for lane in lanes.keys():
            folder = runfolder.get_lane_subfolder(analysis_folder,
                                                  runfolder.analysissubdir,
                                                  lanes[lane], lane, 0)
            os.mkdir(folder)
        os.mkdir(os.path.join(analysis_folder, runfolder.offsetdir))
        os.mkdir(os.path.join(analysis_folder, runfolder.matrixdir))
        os.mkdir(os.path.join(analysis_folder, runfolder.focusdir))
    except:
        sys.stderr.write("Encountered error while creating folder %s. "
                         "Perhaps you do not have the permissions?"
                         % analysis_folder)
        sys.exit(1)


def get_cycle_counts_dic(tile_dict, FileNameClass):
    """Count count number of images per cycle per tile"""
    counts = {}
    for tile in tile_dict.keys():
        counts[tile] = get_cycle_counts(tile_dict[tile], FileNameClass)
    return counts
    
    
def get_cycle_counts(files, FileNameClass):
    """Count count number of images per cycle give a list of files"""
    counts = [ [ 0, 0 ] ]
    lastid = 0
    
    lastcycle = FileNameClass.cycle(files[0])
    
    for file in files:
        id = FileNameClass.is_deblocked(file)
        cycle = FileNameClass.cycle(file)
        if ((lastid == 1 and id == 0) or (counts[-1][id] >= 4) or
            (cycle != lastcycle)):
            counts.append([0,0])
        counts[-1][id] += 1
        lastid = id
        lastcycle = cycle

    return counts


def print_cycle_overview(tile_dict, FileNameClass):
    """Print a summary of the number of images found in enzymology and
    deblocking scans for each tile."""
    counts = get_cycle_counts_dic(tile_dict, FileNameClass)
    maxcycle = max(map(len, counts.values()))
    keys = tile_dict.keys()
    keys.sort()
    print "Tile " + " " * max([len(keys[0])-4, 0]) +" images  %s" %\
          (" ".join(map(lambda x: "C%02d" % x, range(1, maxcycle+1))))
    for tile in keys:
        print "%4s  %3d    " % (tile, len(tile_dict[tile])),
        for cycleno, deblockedno in counts[tile]:
            print "%1d-%1d" % (cycleno, deblockedno),
        print

        
def validate_cycles(tiledic, FileNameClass, autocycleflag=1):
    """Runs some validation checks on a list of filenames to make sure
    that the directory structure is not corrupted."""
    validated = 1
    counts = get_cycle_counts_dic(tiledic, FileNameClass)
    maxcycle = 0
    mincycle = -1
    for tile in tiledic.keys():
        deblocked = counts[tile][0][1]
        ncycles = len(counts[tile])
        # make sure all tiles have the same number of images
        if maxcycle == 0:
            maxcycle = ncycles
        if mincycle < 0:
            mincycle = ncycles
        if maxcycle != ncycles and not autocycleflag:
            validated = 0
            sys.stderr.write(
                "Error: Tile %s: "
                "Different number of cycles: %i instead of %i. "
                "The downstream pipeline cannot currently handle this. "
                "You may need to restrict the number of cycles using "
                "the \"--cycles\" option (or use \"--cycles=auto\").\n"
                % (tile, ncycles, maxcycle))
        if mincycle > ncycles:
            mincycle = ncycles
        if maxcycle < ncycles:
            maxcycle = ncycles
            
        firstcyclecount = counts[tile][0][0]
        for i in range(ncycles):
            cyclecount, deblockedcount = counts[tile][i]
            # check that each cycle has the same number of images as
            # in for the first cycle 
            if (cyclecount != firstcyclecount):
                validated = 0
                sys.stderr.write(
                    "Error: Tile %s cycle %2i: "
                    "Wrong number of images. Your image folders "
                    "are not consistent. Expected "
                    "%i images, but found %i.\n"
                    % (tile, i+1, firstcyclecount, cyclecount))
            # deblocked images also need to have the same number of
            # images (apart from the last cycle), or no deblocked
            # images at all. 
            if (deblockedcount not in [0,firstcyclecount]):
                if (i == ncycles-1):
                    sys.stderr.write(
                        "Warning: Tile %s cycle %2i: "
                        "Discarding last %i deblocked images "
                        "to make number of images consistent "
                        "with runfolder spec.\n"
                        % (tile, i+1, deblockedcount))
                    tiledic[tile] = tiledic[tile][:-deblockedcount]
                    counts[tile][i][1] = 0
                else:
                    validated = 0
                    sys.stderr.write(
                        "Error: Tile %s cycle %2i: Wrong number of "
                        "deblocked images. Expected 0 or %i images, "
                        "but found %i images\n"
                        % (tile, i+1, firstcyclecount, deblockedcount))
            if ((deblockedcount != deblocked) and (i != ncycles-1)):
                validate = 0
                sys.stderr.write(
                    "Error: Tile %s cycle %2i: " 
                    "Must have deblocked images in all cycles or none."
                    % (tile, i+1))
    if autocycleflag and maxcycle != mincycle:
        print "Setting uniform read length of %i cycles." % mincycle
        for tile in tiledic.keys():
            tiledic[tile] = filter(lambda x: FileNameClass.cycle(x)
                                   <= mincycle,
                                   tiledic[tile])
            #print tile, tiledic[tile]
            
    return validated


def missing_files_error():
    """error message (image files)"""
    sys.stderr.write("Not enough image files found.\n")
    sys.exit(1)


def parse_cycles(a):
    numbers = []
    for arg in a.split(","):
        arg2 = map(int, arg.split("-"))
        if len(arg2) == 1:
            arg2.append(arg2[0])
        numbers += range(int(arg2[0]), int(arg2[1])+1)
    return numbers


def get_qc_filenames(dic, FileNameClass, datafolder):
    """Generate a list of all used lanes in the experiment, and a
    list of the corresponding filenames for QC files."""
    lanes = []
    for tilestr in dic.keys():
        lane = FileNameClass.lane(tiledic[tilestr][0]) 
        if not lane in lanes:
            lanes.append(lane)

    filenames = []
    folder = os.path.join(datafolder, runfolder.analysissubdir)
    tilestr = FileNameClass.tilestr(tiledic.values()[0][0])
    prefix = tilestr.split("_")[0] 
    for lane in lanes:
        filenames.append(os.path.join(folder, "%s_%i_qcm.xml" %
                                      (prefix, lane+1)))
    return lanes, filenames


def get_read_length(tiledic, FileNameClass):
    length = 1
    tiles = tiledic.keys()
    if len(tiles) > 0:
        length = get_cycle_params(tiledic[tiles[0]], FileNameClass)[0]
    return length


def create_instrument_offset_file(analysis_folder, instrument_folder,
                                  instrument):
    print "Auto-generating offset file..."
    path = runfolder.get_instrument_folder(analysis_folder,
                                           instrument_folder)
    
    if not os.path.exists(path):
        try:
            os.mkdir(path)
        except:
            sys.stderr.write("Error: Failed to create %s\n" % path)
            sys.exit(1)
    path = os.path.join(path, instrument)
    if not os.path.exists(path):
        try:
            os.mkdir(path)
        except:
            sys.stderr.write("Error: Failed to create %s\n" % path)
            sys.exit(1)
    filename = os.path.join(os.path.dirname(os.path.abspath(sys.argv[0])),
                            runfolder.offsetfile)
    try:
        shutil.copyfile(filename,
                        os.path.join(path, runfolder.offsetfile))
    except:
        sys.stderr.write("Error: Failed to copy %s to %s\n"
                         % (filename, os.path.join(path, runfolder.offsetfile)))
        sys.exit(1)
    return os.path.join(path, runfolder.offsetfile)
        

def get_lanes(selected_tiles):
    dic = {}
    for tile in selected_tiles:
        try:
            lane = int(tile.split("_")[1]) - 1
        except:
            lane = -1
        if lane >= 0:
            dic[lane] = 1
    return dic.keys()


def select_lanes(folders, selected_tiles):
    lanes = get_lanes(selected_tiles)
    if lanes:
        folders = filter(lambda x: runfolder.LaneDirectory.lane(x) in
                         lanes or runfolder.LaneDirectory.lane(x) < 0,
                         folders)
    return folders


def run_lawn_analysis(dic, datafolder):
    """Simplified analysis for chips that have a dense mono-template
    instead of clusters."""
    for tile in tiledic.keys():
        outputfile = os.path.join(datafolder, tile + "_out.txt")
        for file in dic[tile]:
            commandline = "%s -v -v %s >> %s" %\
                          (runfolder.get_executable(image2_executable),
                           file, outputfile)
            print commandline
            os.system(commandline)


def mergeCycleRangePair(cycle_range1, cycle_range2, restarts):
    offset = cycle_range1[-1]
    restart = offset + 1
    restarts.append(restart)
    for i in range(len(cycle_range2)):
        cycle_range2[i] = cycle_range2[i] + offset
    cycle_range1 = cycle_range1 + cycle_range2
    return cycle_range1, restarts


def mergeCycleRanges(cycle_ranges, restarts):
    if (len(cycle_ranges) == 1):
        return cycle_ranges[0], restarts
    cycle_range1 = cycle_ranges[0][:]
    for cr in range(1, len(cycle_ranges)):
        cycle_range1, restarts = mergeCycleRangePair(cycle_range1, cycle_ranges[cr], restarts)
    return cycle_range1, restarts


def getCycleRange(tileimagelist):
    """For a given Run-Folder, determine the cycle range."""    
    allfilenames = reduce(lambda x, y: x+y, tileimagelist)    
    folders = map(runfolder.get_bottom_level_folder, allfilenames)    
    cycles = filter(None, map(runfolder.CycleDirectory.cycle, folders))
    
    # remove all multiple references to the same cycle
    unique_cycles = []
    for cycle in cycles:
        if cycle not in unique_cycles:
            unique_cycles.append(cycle)
    unique_cycles.sort()
    
    return unique_cycles


def getFullCycleRange(tiledics, restarts):
    """Given two or more Run-Folders, determine the full cycle range."""
    if (len(tiledics) == 1):
        cyclerange = getCycleRange(tiledics[0].values())
        return cyclerange, restarts
    cycle_ranges = []
    for td in range(len(tiledics)):
        cycle_ranges.append(getCycleRange(tiledics[td].values()))
    fullcyclerange, restarts = mergeCycleRanges(cycle_ranges, restarts)
    return fullcyclerange, restarts


def mergeTileDicPair(tiledic1, tiledic2):
    for k in tiledic1.keys():
        tiledic1[k] = tiledic1[k] + tiledic2[k]
    return tiledic1


def linkRunFolders(tiledics):
    """Given two or more Run-Folders, merge the tile dictionaries."""
    if (len(tiledics) == 1):
        return tiledics[0]
    tiledic1 = tiledics[0].copy()
    for td in range(1, len(tiledics)):
        tiledic1 = mergeTileDicPair(tiledic1, tiledics[td])
    return tiledic1


def getTileDic(run_folder,
               selected_folders, selected_cycles, selected_tiles,
               nodeblocked):
    """For a given Run-Folder, define a tile dictionary."""
    folders = runfolder.find_image_folders(run_folder)
    
    if selected_tiles:
        # this is just a hack to avoid reading unneeded folders
        # tiles get filtered anyway by the select_tiles function
        folders = select_lanes(folders, selected_tiles)
    if selected_cycles:
        folders = filter(lambda x: runfolder.CycleDirectory.cycle(x)
                         in selected_cycles, folders)
    if nodeblocked:
        folders = filter(lambda x: not runfolder.CycleDirectory.is_deblocked(x),
                         folders)
        
    folders = runfolder.select_image_folders(folders, selected_folders)

    filenames = get_tiff_files(folders)
    filenames += map(os.path.abspath,
                     filter(lambda x: x.find(".tif") > 0, args))
    if len(filenames) <= 0:
        missing_files_error()

    # detect filenaming convention
    # (this is used to support legacy instruments)
    FileNameClass = runfolder.FileName
    for nameclass in [ ManteiaFilePath, SwerdlotronFileName,
                       runfolder.RowColumnFileName ]:
        if nameclass.match(filenames[0]):
            print "%s detected." % nameclass.instrument()
            FileNameClass = nameclass
    
    filenames = filter(FileNameClass.match, filenames)

    # exclude filenames with spaces
    spacefilenames = filter(lambda x: x.find(" ") >= 0, filenames)
    if (spacefilenames):
        sys.stderr.write("Warning: Detected spaces in filenames \"%s\"."
                         "This is not supported by the pipeline and "
                         "the files are removed from the analysis.\n"
                         % "\",\"".join(spacefilenames))
    filenames = filter(lambda x: x.find(" ") < 0, filenames)

    # sort by filename (ignoring the full path)
    filenames.sort(FileNameClass.cmp)

    # group filenames by position on chip
    tiledic = common.group_dictionary(FileNameClass.tilestr, filenames)
    
    # remove unselected positions
    tiledic = common.select_tiles(tiledic, selected_tiles)        
    
    # remove positions with too few files
    if (min_len > 0):
        print "Removing tiles with less than %i files" % min_len
    for key in tiledic.keys():
        if len(tiledic[key]) < min_len:
            del tiledic[key]
    if (len(reduce(lambda x, y: x+y, tiledic.values())) <= 0):
        missing_files_error()

    # discard sequences that start with a deblocked cycle,
    # since there is no clue as to what base has been incorporated
    for key in tiledic.keys():
        if FileNameClass.is_deblocked(tiledic[key][0]) != 0:
            del tiledic[key]
    if (len(tiledic.values()) <= 0):
        missing_files_error()

    # count number of enzymology cycles and
    # deblocked cycles per position
    if not validate_cycles(tiledic, FileNameClass, autocycleflag):
        sys.exit(0)

    return (tiledic, FileNameClass)

    
def usage():
        print "Usage: goat_pipeline.py [--tiles=<tile1>[,<tile2>[,...]]] \
[--cycles=<cycle1>[-cycle2][,cycle3][...]] \
[--make] [--deblocked] [--nobasecall] \
[--offsets=<offsets filesname>|auto] \
[--matrix=<matrix filename>|auto] \
[--phasing=<phasing rate>|auto|auto1|...] [--prephasing=<prephasing rate>] \
[--reset-phasing-cycle=<restart cycle>] \
[--GERALD=<config file>] \
<Run-Folder1> [...]"


#
# -- main program --
#
if __name__ == "__main__":

    print "Starting Solexa/Goat-Pipeline, %s version %s"\
          % (image_software, image_software_version)

    runfolder.validate_executables([ image_executable, image2_executable,
                                     matrix1_executable,
                                     bustard.basecaller_executable,
                                     bustard.phasing1_executable ])
    
    #-------------------------------------------------------------------#
    
    # parsing command-line options
    try:
        (options, args) = getopt.gnu_getopt(sys.argv[1:], ":m",
                                            ["sequence=", "tiles=",
                                             "cycles=",
                                             "matrix=",
                                             "offsets=", "threshold=",
                                             "fwhm=", "distance=",
                                             "make",
                                             "phasing=", "prephasing=",
                                             "reset-phasing-cycle=",
                                             "folders=", "lawn",
                                             "deblocked", "nobasecall",
                                             "GERALD=", "config=",
                                             "directory=",
                                             "sequence-directory=",
                                             "INSTRUMENT_DIR="])
    except getopt.GetoptError:
        sys.stderr.write("Error: failed to parse command-line options.\n")
        usage()
        sys.exit(1)
        
    if len(args) < 1:
        sys.stderr.write("Error: you need to specify at least one "
                         "Run-Folder for analysis!\n")
        usage()
        sys.exit(2)
        
    print options, args

    min_len = 0
    refsequence=""
    offsetfile = ""
    old_analysis_folder = ""
    old_sequence_folder = ""
    instrument_folder = os.getenv('INSTRUMENT_DIR')
    selected_tiles = []
    selected_cycles = []
    selected_folders = []
    runflag = 0
    lawnflag = 0
    nodeblocked = 1
    basecallflag = 1
    autooffsetflag = 0
    autocycleflag = 0
    makeflag = 0
    geraldflag = 0
    geraldconfigfiles = []
    image_params = ImageParams()
    base_params = bustard.BaseCallParams()
    base_params.parse(options)
    for o, a in options:
        if o == "-m":
            min_len = int(a)
        if o == "--sequence":
            refsequence = a
        if o == "--tiles":
            selected_tiles = a.split(",")
        if o == "--folders":
            selected_folders = a.split(",")
        if o == "--INSTRUMENT_DIR":
            instrument_folder = os.path.abspath(a)
        if o == "--offsets":
            if (a.lower().find("auto") == 0):
                autooffsetflag = 1
                offsetfile = ""
            else:
                offsetfile = os.path.abspath(a)
        if o == "--threshold":
            image_params.threshold = float(a)
        if o == "--fwhm":
            image_params.fwhm = float(a)
        if o == "--distance":
            image_params.remapping_distance = float(a)
        if o == "--cycles":
            if (a.lower().find("auto") == 0):
                autocycleflag = 1
            else:
                selected_cycles = parse_cycles(a)
        if o == "--make":
            makeflag = 1
            runflag = 1
        if o == "--lawn":
            lawnflag = 1
        if o == "--deblocked":
            nodeblocked = 0
        if o == "--nobasecall":
            basecallflag = 0
        if o == "--GERALD" or o == "--config":
            path = os.path.abspath(a)
            if not os.path.exists(path):
                sys.stderr.write("Error: Cannot find %s!\n" %
                                 path)
                sys.exit(2)
            geraldconfigfiles.append(path)
            if o == "--GERALD":
                geraldflag = 1
        if o == "--directory":
            # choose an existing analysis directory
            old_analysis_folder = os.path.abspath(a)
            if not os.path.exists(old_analysis_folder):
                sys.stderr.write("Cannot find analysis directory %s\n" %
                                 old_analysis_folder)
                sys.exit(2)
            basecallflag = 0
                
    #-------------------------------------------------------------------#
            
    starttime = runfolder.get_date_time()

    FileNameClass = runfolder.FileName

    # Tile dictionaries
    #
    # case `single Run-Folder` (standard case):
    #       the list 'tiledics' contains only one dictionary
    # case `multi Run-Folders` (e.g. paired reads):
    #       the list 'tiledics' contains two or more dictionaries
    tiledics = []
    
    for run_folder in args:
        tiledic, FileNameClass = getTileDic(run_folder,
                                            selected_folders, selected_cycles,
                                            selected_tiles, nodeblocked)
        tiledics.append(tiledic)

    tiledic = linkRunFolders(tiledics) 

    print_cycle_overview(tiledic, FileNameClass)

    # Cycle range; cycle(s) at which the phasing/prephasing effect is reset
    restarts = []
    cyclerange, restarts = getFullCycleRange(tiledics, restarts)
    base_params.phasing_restarts = base_params.phasing_restarts + restarts
    base_params.phasing_restarts.sort()

    # Analysis folder 
    analysis_folder =\
                runfolder.get_new_analysis_folder(os.path.abspath(tiledic.values()[0][0]),
                                                  cyclerange, image_software,
                                                  image_software_version)
    # if some analysis already exists, identify the directories
    if old_analysis_folder:
        print  "Original analysis folder: %s" % old_analysis_folder
    print  "Analysis folder: %s" % analysis_folder

    # Sequence folder
    if basecallflag:
        sequence_folder =\
                runfolder.get_new_sequence_folder(old_analysis_folder
                                                  or analysis_folder,
                                                  bustard.basecaller_software,
                                                  bustard.basecaller_software_version)
        sequence_folder = os.path.join(analysis_folder,
                                       os.path.split(sequence_folder)[1])
        print  "Sequence folder: %s" % sequence_folder

    # Instrument
    instrument = common.get_instrument(analysis_folder)
    if instrument:
        print "Instrument:  %s" % instrument

    # Offset data
    data_offsetfile = os.path.join(runfolder.get_data_folder(analysis_folder),
                                   runfolder.offsetfile)
    instrument_offsetfile = ""
    if not offsetfile:
        offsetfile = runfolder.find_data_offset_file(analysis_folder)
    if not offsetfile:
        instrument_offsetfile = \
                runfolder.find_instrument_offset_file(analysis_folder,
                                                      instrument_folder,
                                                      instrument)
        offsetfile = instrument_offsetfile
    if offsetfile and os.path.exists(offsetfile):    
        print "Offset file: %s" % offsetfile
    else:
        sys.stderr.write("Error: No offset file found!\n")
        if not autooffsetflag:
            sys.stderr.write("Warning: Now turning on the "
                             "\"--offsets=auto\" option!\n")
            autooffsetflag = 1
        instrument_offsetfile = \
               create_instrument_offset_file(analysis_folder,
                                             instrument_folder,
                                             instrument)
        offsetfile = instrument_offsetfile
        if offsetfile and os.path.exists(offsetfile):    
            print "Offset file: %s" % offsetfile       
    if instrument_offsetfile:
        print "Will update offsets file %s after run!" % instrument_offsetfile

    # Matrix
    if not base_params.automatrixflag:
        base_params.matrixfile = \
            common.get_matrix(base_params.matrixfile, instrument_folder, analysis_folder)

    base_params.print_params()
 
    # Gerald configuration file
    readlength = get_read_length(tiledic, FileNameClass)   
    geraldconfigs = []
    geraldconfigs = bustard.getGeraldConfigs(geraldconfigfiles,
                                             runfolder.get_run_folder(analysis_folder),
                                             readlength)

    #--
    if not runflag:
        print ("You can generate the Makefiles by using the \"--make\" "
               "option.")
        sys.exit(0)
    #--

    #-------------------------------------------------------------------#

    # create output folders
    
    tiles = tiledic.keys()
    tiles.sort()
    
    if old_analysis_folder:
        rename_image_folders(old_analysis_folder, analysis_folder)
    else:
        create_image_folders(analysis_folder, tiles)

    if basecallflag:
        if old_sequence_folder:
            try:
                oldfolder = os.path.join(os.path.split(sequence_folder)[0],
                                         os.path.split(old_sequence_folder)[1])
                print oldfolder
                print os.path.split(sequence_folder)[0]
                print os.path.split(old_sequence_folder[1])
                os.rename(oldfolder, sequence_folder)
            except:
                sys.stderr.write("Failed to rename folder %s to %s!\n"
                                 % (old_sequence_folder, sequence_folder))
                sys.exit(0)
                
        else:
            # Base-caller folders
            bustard.createBasecallerFolders(sequence_folder, tiledic.keys())

        # Matrix
        if base_params.matrixfile and not base_params.automatrixflag:
            for lane in runfolder.get_lane_samples(tiledic.keys()).keys():
                matrixfile2 = os.path.join(sequence_folder,
                                           "matrix%i.txt" % (lane+1))
                shutil.copyfile(base_params.matrixfile, matrixfile2)

        # Phasing/prephasing effect        
        if not base_params.autophasingflag:
            bustard.processParamsFile(sequence_folder, tiledic,
                                      base_params.phasing_rate,
                                      base_params.prephasing_rate)

    # run lawn analysis
    if lawnflag:
        run_lawn_analysis(tiledic, analysis_folder)
        sys.exit(0)

    offsetdir = os.path.join(analysis_folder, runfolder.offsetdir)
    offsetsummaryfile = os.path.join(offsetdir, "offsets.txt")
    ncolours = 4
    
    print "Generating journals, Makefiles and parameter files ..."
    
    #-------------------------------------------------------------------#

    # generating journal files
    for tilestr in tiles:
        sample, lane, tile = runfolder.FileName.split_tile(tilestr)
        
        files = tiledic[tilestr]        
        firstcycle = 0
        
        run_generate_journal(analysis_folder, sample, lane, tile,
                             files[ncolours*firstcycle:], FileNameClass)

    #-------------------------------------------------------------------#

    # generating Makefiles    
    makefile.generate_makefile(tiledic, analysis_folder, FileNameClass,
                               offsetfile, offsetsummaryfile,
                               data_offsetfile, instrument_offsetfile,
                               autooffsetflag, "", "", old_analysis_folder,
                               image_params)

    if basecallflag:
        makefile.generate_basecaller_makefile(tiledic, analysis_folder,
                                              sequence_folder, None,
                                              base_params.phasing_rate,
                                              base_params.prephasing_rate,
                                              base_params.phasing_restarts,
                                              base_params.automatrixflag,
                                              base_params.autophasingflag,
                                              base_params.automatrixlane,
                                              base_params.autophasinglane)
        
    #-------------------------------------------------------------------#
    
    # generating .params file for intensity remapping
    params = [ image_params.threshold,
               image_params.fwhm,
               image_params.remapping_distance ]
    imagefolders = get_bottom_level_folders(reduce(lambda x, y: x + y,
                                                   tiledic.values()))
    runfolder.write_analysis_params(analysis_folder, tiles,
                                    cyclerange, imagefolders,
                                    image_software, image_software_version,
                                    runfolder.get_login(), image_params,
                                    starttime,
                                    runfolder.read_default_offsets(offsetfile),
                                    old_analysis_folder)

    #-
    if not basecallflag:
        sys.exit(0)
    #-

    # generating .params file for base-calls
    params = [ base_params.phasing_rate,
               base_params.prephasing_rate ]
    runfolder.write_sequence_params(analysis_folder, sequence_folder,
                                    tiles, cyclerange,
                                    bustard.basecaller_software,
                                    bustard.basecaller_software_version,
                                    runfolder.get_login(),
                                    params, base_params.matrixfile,
                                    old_sequence_folder)

    #-------------------------------------------------------------------#

    # processing Gerald configuration file
    bustard.processGeraldConfig(geraldflag, geraldconfigs,
                                analysis_folder, sequence_folder)
