#!/usr/bin/env python
#
# Copyright (c) 2005, 2006 Solexa
# Author: Klaus Maisinger
# This source file is covered by the "Solexa Public Source License"
# agreement and bound by the terms therein.

import os, os.path, types, sys, shutil
import runfolder, goat_pipeline, bustard, gerald

"""Generate makefiles for the image analysis and base-calling software."""

makefilepath          = "Makefiles"

def get_max_cycle(value, ncolours=4):
    if type(value) == types.IntType:
        return value
    if type(value) == types.ListType:
        return len(value)/ncolours
    raise "Wrong type for value in makefile.get_max_cycle()"

def get_var_tile(tilestr):
    sample, lane, tile = runfolder.FileName.split_tile(tilestr)
    return tilestr.replace(sample, "$(sample%i)" % (lane+1))
                           
def generate_makefile(tiledic, analysis_folder, FileNameClass,
                      offsetfile, offsetsummaryfile,
                      data_offsetfile, instrument_offsetfile,
                      autooffsetflag,
                      sequence_folder, matrixfile,
                      old_analysis_folder, image_params):
    filename = os.path.join(analysis_folder, "Makefile")
    lines = []
    tiles = tiledic.keys()
    tiles.sort()
    lanes = runfolder.get_lane_samples(tiles)

    lines.append("# Makefile auto-generated by makefile.py (version %s)\n# %s\n\n"
                 % (goat_pipeline.image_software_version,
                    runfolder.get_date_time()))
    image_exe = runfolder.get_executable(goat_pipeline.image_executable)
    installation_dir = os.path.split(image_exe)[0]
    lines.append("installationdir:=%s\n" %  installation_dir)
    lines.append("image_exe:=%s\n" %
             runfolder.get_executable(goat_pipeline.image_executable)\
                 .replace(installation_dir, "$(installationdir)"))
    lines.append("offsets_exe:=%s\n" %
             runfolder.get_executable(goat_pipeline.offset_executable)\
                 .replace(installation_dir, "$(installationdir)"))
    lines.append("offsets2_exe:=%s\n" %
             runfolder.get_executable(goat_pipeline.offset2_executable)\
                 .replace(installation_dir, "$(installationdir)"))
    lines.append("matrix1_exe:=%s\n" %
             runfolder.get_executable(goat_pipeline.matrix1_executable)\
                 .replace(installation_dir, "$(installationdir)"))
    #lines.append("matrix2_exe:=%s\n" %
    #         runfolder.get_executable(goat_pipeline.matrix2_executable)\
    #             .replace(installation_dir, "$(installationdir)"))
    lines.append("matrix3_exe:=%s\n" %
             runfolder.get_executable(goat_pipeline.matrix3_executable)\
                 .replace(installation_dir, "$(installationdir)"))
    lines.append("matrix4_exe:=%s\n" %
             runfolder.get_executable(goat_pipeline.matrix4_executable)\
                 .replace(installation_dir, "$(installationdir)"))
    lines.append("PASTE=paste\n")
    lines.append("CP=cp\n")
    lines.append("analysisdir:=%s\n" % analysis_folder)
    for lane in lanes.keys():
        lines.append("sample%i:=%s\n" % (lane+1, lanes[lane]))
    lines.append("analysis_subfolder:=%s\n" %
                 runfolder.analysissubdir)
    lines.append("adir2:=.\n")
    lines.append("lanes:=%s\n" % " ".join(map(lambda x: "$(sample%i)_%i" % (x+1, x+1),
                                              lanes.keys())))
    lines.append("lanefile:=lanes.txt\n")
    lines.append("tilefile:=tiles.txt\n")
    folders = []
    for lane in lanes.keys():
        folders.append(runfolder.get_lane_subfolder("", "$(analysis_subfolder)",
                                                    "", lane))
                                              
        lines.append("analysisdir%i:=%s\n" % (lane+1, folders[-1]))
    lines.append("analysisdirs:=%s\n" % " ".join(folders))
    ncolours = 4
    allmaxcycle = get_max_cycle(tiledic.values()[0])
    lines.append("maxcycle:=%02i\n" % allmaxcycle)
    lines.append("cycles:=%s\n" % " ".join(map(lambda x: "%02i" % (x+1),
                                           range(allmaxcycle))))
    lines.append("subfolders:=$(wildcard %s*)\n" %
                 bustard.basecaller_software)
    lines.append("logfile:=$(analysis_subfolder)/$(analysis_subfolder).log\n")
    offsetdir = runfolder.offsetdir
    lines.append("focusdir:=%s\n" % "Focus")
    lines.append("offsetdir:=%s\n" % runfolder.offsetdir)
    lines.append("offsetfile:=%s\n" % os.path.split(offsetsummaryfile)[1])
    lines.append("offsetstats:=$(offsetdir)/offset_summary.txt\n")
    #lines.append("offsetoutput:=%s\n" %
    #         offsetsummaryfile.replace(analysis_folder,
    #                                   "$(analysisdir)").replace(runfolder.offsetdir, "$(offsetdir)"))
    lines.append("offsetdependency?=true\n")
    if old_analysis_folder:
        lines.append("loadfile:= -loadsuffix=.old\n")
    else:
        lines.append("loadfile:=\n")
    
    lines.append("default_offsets:=%s\n" % "default_offsets.txt")
    if ((offsetfile != data_offsetfile)
        and (offsetfile != instrument_offsetfile)):
        lines.append("input_offsets:=%s\n" % offsetfile)
    lines.append("data_offsets:=../$(default_offsets)\n")
    if instrument_offsetfile:
        lines.append("instrument_offsets:=%s\n" % instrument_offsetfile)
    lines.append("offsets:=$(default_offsets)\n")

    lines.append("fwhm:=%f\n" %      image_params.fwhm);
    lines.append("dist:=%f\n" %      image_params.remapping_distance)
    lines.append("threshold:=%f\n" % image_params.threshold)

    # set up look up table for deblock string and maxcycle per tile
    # also find lowest common denominator for deblock length
    difflengthtiles = {}
    diffdeblocktiles = {}
    for tilestr in tiles:
        sample, lane, tile = runfolder.FileName.split_tile(tilestr)
        firstcycle = 0
        ncycles, ncolour, deblock = goat_pipeline.get_cycle_params(\
            tiledic[tilestr][ncolours*firstcycle:], FileNameClass)
        
        tilestr2 = get_var_tile(tilestr)
        maxcycle = get_max_cycle(tiledic[tilestr])
        diffdeblocktiles.setdefault(deblock, []).append(tilestr2)
        if maxcycle != allmaxcycle:
            difflengthtiles.setdefault(maxcycle, []).append(tilestr2)
    mindeblock = diffdeblocktiles.keys()[0]
    for deblock in diffdeblocktiles.keys():
        if (len(deblock) < len(mindeblock)):
            mindeblock = deblock

    lines.append("matrixdir:=%s\n" % runfolder.matrixdir)
    # identify cycles to use for matrix estimation (skip deblock cycles)
    matrix_maxcycle = 2
    matrix_cycles = []
    for i in range(len(mindeblock)):
        if mindeblock[i] == 'C':
            matrix_cycles.append(i)
        if len(matrix_cycles) == matrix_maxcycle:
            break
    if len(matrix_cycles) == 2: # skip first cycle
        matrix_cycles = [matrix_cycles[1]]

    for i in matrix_cycles:
        lines.append("matrixsummary%i:=s_%02i_summary.txt\n" %
                     (i+1, i+1))

    clean_targets = map(lambda x: "clean%i" % (x+1), lanes.keys()) 
    lines.append("\nvpath %%_jnl.txt %s\n" % " ".join(folders))
    lines.append("vpath %_qcm.xml $(analysisdirs)\n")
    lines.append("vpath %_mat.txt $(matrixdir)\n")
    lines.append("vpath %_matrix.txt $(matrixdir)\n")
    lines.append("vpath %_summary.txt $(matrixdir)\n")
    lines.append("vpath %offsets.txt $(offsetdir)\n")
    lines.append("\n"
                 ".PHONY: all allmsg $(lanes) update_offsets matrix focus recursive $(subfolders)\n"
                 ".PHONY: clean clean_intermediate clean_matrix clean_offsets cleanfocus %s\n" % (" ".join(clean_targets)))
    lines.append(".PRECIOUS: %_int.txt %_mat.txt %_matrix.txt %_offsets.txt %_summary.txt $(default_offsets)\n\n")
    lines.append("all: allmsg $(lanes) update_offsets matrix\n"
                 "\t@echo Image analysis has completed successfully.\n"
                 "\ttouch finished.txt\n"
                 "\t$(cmdf3)\n")
    lines.append("allmsg:\n\t@echo Analysing lanes $(lanes)\n")
    lines.append("\n")

    ##################################################################
    # generate rules for each lane
    ##################################################################
    for lane2 in lanes.keys():
        intfiles = []
        for tilestr in tiles:
            sample, lane, tile = runfolder.FileName.split_tile(tilestr)
            if lane2 == lane:
                intfiles.append("%s_int.txt" % get_var_tile(tilestr))
        lines.append("l%i_intfiles:=%s\n" % (lane2+1,
                                             " ".join(intfiles)))
    lines.append("#Alternatively, auto-generate target list...\n")
    for lane2 in lanes.keys():
        lines.append("#l%i_intfiles:=$(subst _jnl.txt,_int.txt,$(notdir $(wildcard $(analysisdir%i)/*_jnl.txt)))\n"
                     % (lane2+1, lane2+1))
    lines.append("intfiles:=%s\n\n" %
                 " ".join(map(lambda x: "$(l%i_intfiles)" % (x+1),
                              lanes.keys())))

    lines.append("ifeq ($(offsetdependency),true)\n"
                 "$(intfiles): $(default_offsets)\n"
                 "endif\n\n")
    lines.append("$(lanefile):\n")
    lines.append("\techo \"$(subst _int.txt,,$(intfiles))\" > $(tilefile)\n")
    lines.append("\techo \"$(lanes)\" > $@\n\n")

    ##################################################################
    # generate rules for int.txt files
    ##################################################################
    lines.append("# Generate intensity files\n")
    lines.append("%_int.txt: %_jnl.txt $(lanefile)\n")
    keys = difflengthtiles.keys()
    keys.sort()
    for cycle in keys:
        lines.append("%s: maxcycle:=%i\n" %
                     (" ".join(map(lambda x: "%s_int.txt" % x,
                                   difflengthtiles[cycle])), cycle))
    keys = diffdeblocktiles.keys()
    keys.sort()
    for deblock in keys:
        lines.append("%s: deblock:=%s\n" %
                     (" ".join(map(lambda x: "%s_int.txt" % x,
                                   diffdeblocktiles[deblock])), deblock))

    lines.append("%_int.txt: tile=$(subst _int.txt,,$@)\n")
    for lane in lanes.keys():
        lines.append("$(l%i_intfiles): adir:=$(analysisdir%i)\n" %
                     (lane+1, lane+1))
    lines.append("%_int.txt:\n"
                 "\t$(cmdprefix) $(image_exe) $(adir)/$(tile)_jnl.txt -fwhm=$(fwhm) -dist=$(dist) -threshold=$(threshold) -nc=0 -logfile=$(logfile) -prefix=$(adir)/ -prefix2=$(adir2)/ -offsets=$(offsets) -tile=$(tile) -deblocked=$(deblock) $(loadfile)\n")

    ##################################################################
    # generate default offsets
    ##################################################################
    lines.append("\n# Generate default offsets\n")
    lines.append("$(default_offsets): $(lanefile)\n")
    lines.append("\t@echo Auto-generating default offsets... ;\\\n")
    if (os.path.exists(offsetfile) and (offsetfile != data_offsetfile)
        and (offsetfile != instrument_offsetfile)):
        lines.append("\t$(CP) $(input_offsets) $(default_offsets);\\\n")
    elif os.path.exists(data_offsetfile):
        lines.append("\t$(CP) $(data_offsets) $(default_offsets);\\\n")
    elif os.path.exists(instrument_offsetfile):
        lines.append("\t$(CP) $(instrument_offsets) $(default_offsets);\\\n")
    if autooffsetflag:
        # Generate offsets by running the pipeline on 20 tiles
        ntiles = min(20, len(tiles))
        lines.append("\t$(MAKE) offsetdependency=false %s ; \\\n"
                     % " ".join(map(get_var_tile, tiles[:ntiles])))
        for tilestr in tiles[:ntiles]:
            sample, lane2, tile = runfolder.FileName.split_tile(tilestr)
            lines.append("\t$(cmdprefix) $(offsets_exe) $(foreach cycle,$(wordlist 1,%i,$(cycles)),$(analysisdir%i)/%s_$(cycle)_qcm.xml) >> $(offsetdir)/$(offsetfile) ; \\\n"
                         % (get_max_cycle(tiledic[tilestr]),
                            lane2+1, get_var_tile(tilestr)))
        lines.append("\t$(cmdprefix) $(offsets2_exe) $(offsetdir)/$(offsetfile) $(data_offsets) > $(offsetstats) ; \\\n"
                     "\tcat $(offsetstats); \\\n"
                     "\t$(CP) $(data_offsets) $(default_offsets) ; \\\n"
                     "\t$(MAKE) clean_lanes clean_offsets clean_matrix;\\\n")
    lines.append("\t$(cmdf1)\n")


    ##################################################################
    # update offsets
    ##################################################################
    lines.append("\n# Generate offsets\n")
    maxoffsetlines = 10
    # This bit is very awkward - to avoid command-line buffer
    # overruns while preserving the order of the commands,
    # this implements a dependency chain of commands.
    # Need to think of a better way of doing this...
    for lane in lanes.keys():
        target = "$(offsetdir)/$(sample%i)_%i_offsets.txt" %\
                 (lane+1, lane+1)
        subtargets = [ "clean_offsets%i" % (lane+1) ]
        lines.append("%s: $(l%i_intfiles)\n"
                     "\t$(RM) %s\n" % (subtargets[-1], lane+1, target))
        i = 0
        for tilestr in tiles:
            sample, lane2, tile = runfolder.FileName.split_tile(tilestr)
            if lane == lane2:
                if (i % maxoffsetlines) == 0:
                    subtargets.append("offsets%i_%i" % (lane+1, i))
                    lines.append("\n%s: %s\n" % (subtargets[-1],
                                                 subtargets[-2]))
                    lines.append("\t$(cmdprefix) $(offsets_exe) "
    "$(foreach stem,$(subst _int.txt,,$(wordlist %i,%i,$(l%i_intfiles))),"
          "$(foreach cycle,$(wordlist 1,%i,$(cycles)),"
                "$(analysisdir%i)/$(stem)_$(cycle)_qcm.xml)) "
                                 ">> %s\n"
                                 % (i+1, i+10, lane+1,
                                    get_max_cycle(tiledic[tilestr]),
                                    lane+1, target))
                                 
                #lines.append("\t$(cmdprefix) $(offsets_exe) $(foreach cycle,$(wordlist 1,%i,$(cycles)),$(analysisdir%i)/%s_$(cycle)_qcm.xml) >> %s ; \\\n"
                #             % (get_max_cycle(tiledic[tilestr]),
                #                lane+1, get_var_tile(tilestr), target))

                i += 1

        lines.append("\n%s: %s\n" % (target, subtargets[-1]))
        lines.append("\techo Generated %s.\n" % target)
        lines.append(".PHONY: %s\n" % " ".join(subtargets))

    lines.append("\n$(offsetdir)/$(offsetfile): %s\n" %
                 " ".join(map(lambda x:
                              "$(offsetdir)/$(sample%i)_%i_offsets.txt" %
                              (x+1, x+1), lanes.keys())))
    lines.append("\tcat $^ > $@\n")

    lines.append("\nupdate_offsets: $(offsetfile)\n")
    if instrument_offsetfile and not os.path.exists(data_offsetfile):
        lines.append("\t$(CP) $(instrument_offsets) $(data_offsets) ; \\\n")
    lines.append("\t$(cmdprefix) $(offsets2_exe) $(offsetdir)/$(offsetfile) $(data_offsets) > $(offsetstats) ; \\\n")
    lines.append("\tcat $(offsetstats) ; \\\n")
    if instrument_offsetfile:
        lines.append("\t$(CP) $(data_offsets) $(instrument_offsets)\n")
    else:
        lines.append("\techo Offsets updated\n")
        

    ##################################################################
    # matrix estimation
    ##################################################################
    # matrix estimation is performed on the first two cycles.
    # need to make sure that each tile has sequenced two cycles.
    lines.append("\n# Matrix estimation targets\n")
    # matrix targets by lane
    for lane2 in lanes.keys():
        for i in matrix_cycles:
            lines.append("l%i_%02i_matfiles:=$(addprefix $(matrixdir)/,$(subst _int.txt,_%02i_mat.txt,$(l%i_intfiles)))\n"
                         % (lane2+1, i+1, i+1, lane2+1))
        lines.append("l%i_matfiles:=%s\n"
                     % (lane2+1,
                        " ".join(map(lambda x: "$(l%i_%02i_matfiles)" %
                                     (lane2+1, x+1), matrix_cycles))))

    # main matrix target
    matrixsummary_targets = []
    for i in matrix_cycles:
        for lane2 in lanes.keys():
            target = "$(matrixdir)/$(sample%i)_%i_%02i_summary.txt" \
                     % (lane2+1, lane2+1, i+1)
            matrixsummary_targets.append(target)
            lines.append("%s: $(l%i_%02i_matfiles)\n"
                         % (target, lane2+1, i+1))
    lines.append("%s: \n" % " ".join(matrixsummary_targets))
    lines.append("\t$(RM) $@;\\\n"
                 "\tfor f in $^; do\\\n"
                 "\t    echo $$f;\\\n"
                 "\t    $(cmdprefix) $(matrix3_exe) $$f >> $@;\\\n"
                 "\tdone\n\n" )
    for i in matrix_cycles:
        lines.append("$(matrixdir)/$(matrixsummary%i): %s\n"
                     % (i+1,
                        " ".join(map(lambda x:
                                     "$(matrixdir)/$(sample%i)_%i_%02i_summary.txt"
                                     % (x+1, x+1, i+1), lanes.keys()))))
        lines.append("\tcat $^ > $@\n")

    for lane2 in lanes.keys():
        lines.append("$(matrixdir)/$(sample%i)_%i_matrix.txt: %s\n" %
                     (lane2+1, lane2+1,
                      " ".join(map(lambda x:
                                   "$(matrixdir)/$(sample%i)_%i_%02i_summary.txt" %
                                   (lane2+1, lane2+1, x+1),
                                   matrix_cycles))))
        lines.append("\t$(cmdprefix) $(matrix4_exe) "
                     "$(matrixdir)/$(sample%i)_%i_%02i_summary.txt > $@\n"
                     % (lane2+1, lane2+1, (matrix_cycles[-1]+1)))
    lines.append("$(matrixdir)/s_matrix.txt: %s\n" %
                 " ".join(map(lambda x:
                              "$(matrixdir)/$(matrixsummary%i)" % (x+1),
                              matrix_cycles)))
    lines.append("\t$(cmdprefix) $(matrix4_exe) $(matrixdir)/$(matrixsummary%i) > $@\n"
                 % (matrix_cycles[-1]+1))

    lines.append("\nmatrix: $(matrixdir)/s_matrix.txt %s\n\n" %
                 " ".join(map(lambda x:
                              "$(matrixdir)/$(sample%i)_%i_matrix.txt"
                              % (x+1, x+1), lanes.keys())))

    # matrix targets by tile
    lines.append("$(matrixdir)/%_mat.txt: "
                 "tilecycle=$(subst _mat.txt,,$(notdir $@))\n")
    for lane in lanes.keys():
        lines.append("$(l%i_matfiles): adir:=$(analysisdir%i)\n" % (lane+1,
                                                                lane+1))
    for i in matrix_cycles:
        lines.append("$(matrixdir)/%%_%02i_mat.txt: %%_int.txt\n"
                     "\t$(cmdprefix) $(matrix1_exe) "
                     "$(adir)/$(tilecycle)_int.txt > $@\n" % (i+1))
    
    ##################################################################
    # targets to make typing easier
    ##################################################################
    tiletargets = {}
    lines.append("\n# Some rules to make typing targets easier\n")
    for tilestr in tiles:
        sample, lane, tile = runfolder.FileName.split_tile(tilestr)
        tilestr2 = get_var_tile(tilestr)
        lines.append("%s: %s\n" %
                     (tilestr, " ".join(map(lambda x :"%s_%02i_mat.txt" %
                                            (tilestr2, x+1),
                                            matrix_cycles))))
        tiletargets.setdefault(tile, []).append(tilestr2)
    keys = tiletargets.keys()
    keys.sort()
    for key in keys:
        lines.append("_%04i: %s\n" % (key + 1, " ".join(tiletargets[key])))

    # lanes
    lines.append("\n# Lane targets\n")
    for lane in lanes.keys():
        lines.append("$(sample%i)_%i: $(sample%i)_%i_finished.txt l%i_msg\n" %

                     (lane+1, lane+1, lane+1, lane+1, lane+1))
    for lane in lanes.keys():
        lines.append("$(sample%i)_%i_finished.txt: $(default_offsets) %s\n" %
                     (lane+1, lane+1,
                       " ".join(map(lambda x:
                                   "$(matrixdir)/$(sample%i)_%i_%02i_summary.txt" %
                                   (lane+1, lane+1, x+1),
                                    matrix_cycles))))
        lines.append("\ttouch $@\n"
                     "  ifdef cmdf2\n"
                     "\ttouch $@; $(cmdf2) $(subst _finished.txt,,$@)\n"
                     "  endif\n")
        lines.append("l%i_msg:\n"
                     "\t@echo Analysing lane %i: $(l%i_intfiles)\n" %\
                     (lane+1, lane+1, lane+1))
    lines.append("\n")

    ##################################################################
    # clean targets
    ##################################################################
    lines.append("\n#Clean targets\n")
    lines.append("clean: clean_matrix clean_offsets clean_focus clean_tags clean_lanes\n")
    lines.append("clean_tags:\n"
                 "\t$(RM) *finished.txt $(lanefile) $(tilefile)\n")
    lines.append("clean_lanes: %s\n"
                 % " ".join(clean_targets))
    lines.append("\t$(RM) $(analysis_subfolder)/*.log\n")
    lines.append("\n")
    for lane in lanes.keys():
        lines.append("clean%i: clean%i_int clean%i_int_sub clean%i_all_sub\n"
                     % (lane+1, lane+1, lane+1, lane+1))
        
        lines.append("clean%i_int:\n"
                     "\tfind . -regex \".*$(sample%i)_%i_.*_\\(int\\|nse\\)\\.txt\" | xargs $(RM)\n" % (lane+1, lane+1, lane+1))
        # clean subdirectories
        lines.append("clean%i_int_sub:\n"
                     "\tfind $(analysisdir%i) -regex \".*_\\(int\\|nse\\)\\.txt\" | xargs $(RM)\n" % (lane+1, lane+1))
        lines.append("clean%i_all_sub:\n"
                     "\tfind $(analysisdir%i) -regex \".*_\\(idx\\|pos\\|clu\\)\\.txt\" | xargs $(RM)\n" % (lane+1, lane+1))
        lines.append("\tfind $(analysisdir%i) -name \"*_qcm.xml\" | xargs $(RM)\n" % (lane+1))
    lines.append("clean_intermediate: %s\n" %
                 " ".join(map(lambda x: "clean%i_int_sub" % (x+1),
                              lanes.keys())))
    lines.append("clean_offsets:\n"
                 "\tfind $(offsetdir) -name \"*.txt\" | xargs $(RM)\n")
    lines.append("clean_focus:\n"
                 "\tfind $(focusdir) -name \"*.txt\" | xargs $(RM)\n"
                 "\tfind $(focusdir) -name \"*.csv\" | xargs $(RM)\n")
    lines.append("clean_matrix:\n")
    lines.append("\t$(RM) %s\n" %
                 " ".join(map(lambda x: "$(matrixsummary%i)" % (x+1),
                              matrix_cycles)))
    lines.append("\tfind $(matrixdir) -name \"*_mat.txt\" | xargs $(RM)\n")
    lines.append("\tfind $(matrixdir) -name \"*_cmat.txt\" | xargs $(RM)\n")
    lines.append("\tfind $(matrixdir) -name \"*_matrix.txt\" | xargs $(RM)\n")
    lines.append("\tfind $(matrixdir) -name \"*_summary.txt\" | xargs $(RM)\n")

    ##################################################################
    # targets in other Makefiles
    ##################################################################
    lines.append("compress uncompress compress_intermediate uncompress_intermediate:\n"
                 "\t$(MAKE) -f Makefile.compress.firecrest $@\n")
    lines.append("compress_images uncompress_images:\n"
                 "\t$(MAKE) -f Makefile.compress.images $@\n")

    lines.append("\nvpath %_image.txt $(focusdir)\n")
    lines.append("focus: $(focusdir)/focus.csv\n")
    lines.append("focuscmd?=$(focusdir)Quality\n")
    lines.append("$(focusdir)/focus.csv: $(addprefix $(focusdir)/,$(subst _int.txt,_focus.csv,$(intfiles)))\n"
                 "\t$(RM) $@ ; \\\n"
                 "\tfor f in $^; do\\\n"
                 "\t  cat $$f >> $@;\\\n"
                 "\tdone\n"

                 "$(focusdir)/%_image.txt: %_jnl.txt\n"
                 "\tsed -n '{s/Load \"\\(.*\\)\"/\\1/p}' $< > $@\n"
                 "$(focusdir)/%_focus.csv: $(focusdir)/%_image.txt\n"
                 "\t$(focuscmd) -c $< $@\n")

    ##################################################################
    # recursive make
    ##################################################################
    lines.append("\nrecursive: $(subfolders) nonrecursive\n")
    lines.append("nonrecursive:\n")
    lines.append("\t$(MAKE) $(TARGET)\n")
    lines.append("$(subfolders): nonrecursive\n")
    lines.append("\t$(MAKE) --directory=$@ recursive TARGET=$(TARGET)\n")
    try:
        fd = open(filename, "w")
        if not fd:
            sys.stderr.write("Error: Could not open %s\n", filename)
            return
        fd.writelines(lines)
        fd.close()
    except:
        sys.stderr.write("Error: Failed to write %s\n", filename)

    makefilename = "Makefile.compress.images"
    makefilepath2 = runfolder.get_executable(makefilepath)
    try:
        shutil.copyfile(os.path.join(makefilepath2, makefilename),
                        os.path.join(analysis_folder, makefilename))
    except:
        sys.stderr.write("Error: Failed to copy %s\n" % makefilename)

    makefilename = "Makefile.compress.firecrest"
    try:
        shutil.copyfile(os.path.join(makefilepath2, makefilename),
                        os.path.join(analysis_folder, makefilename))
    except:
        sys.stderr.write("Error: Failed to copy %s\n" % makefilename)


def generate_basecaller_makefile(tiledic,
                                 analysis_folder, sequence_folder,
                                 matrixfile, phasing_rate, prephasing_rate,
                                 phasing_restarts,
                                 automatrix = 0, autophasing = 0,
                                 automatrixlane = 0, autophasinglane = 0):
    filename = os.path.join(sequence_folder, "Makefile")
    lines = []
    
    tiles = tiledic.keys()
    tiles.sort()
    lanes = runfolder.get_lane_samples(tiles)

    lines.append("# Makefile auto-generated by makefile.py (version %s)\n# %s\n\n"
                 % (bustard.basecaller_software_version,
                    runfolder.get_date_time()))
    installation_dir = os.path.split(runfolder.get_executable(bustard.basecaller_executable))[0]
    lines.append("installationdir:=%s\n" %  installation_dir)
    lines.append("basecall_exe:=%s\n" %
             runfolder.get_executable(bustard.basecaller_executable)\
                 .replace(installation_dir, "$(installationdir)"))
    lines.append("phasing1_exe:=%s\n" %
             runfolder.get_executable(bustard.phasing1_executable)\
                 .replace(installation_dir, "$(installationdir)"))
    lines.append("phasing2_exe:=%s\n" %
             runfolder.get_executable(bustard.phasing2_executable)\
                 .replace(installation_dir, "$(installationdir)"))
    lines.append("phasing3_exe:=%s\n" %
             runfolder.get_executable(bustard.phasing3_executable)\
                 .replace(installation_dir, "$(installationdir)"))
    lines.append("PASTE=paste\n")
    lines.append("CP=cp\n")
    lines.append("analysisdir:=%s\n" % analysis_folder)
    lines.append("sequencedir:=%s\n" % sequence_folder)
    for lane in lanes.keys():
        lines.append("sample%i:=%s\n" % (lane+1, lanes[lane]))
    lines.append("analysis_subfolder:=%s\n" % runfolder.analysissubdir)
    lines.append("sequence_subfolder:=%s\n" % runfolder.sequencesubdir)
    lines.append("lanes:=%s\n" %" ".join(map(lambda x: "$(sample%i)_%i" %
                                            (x+1, x+1),
                                            lanes.keys())))
    for lane in lanes.keys():
        folder = runfolder.get_lane_subfolder("", "$(analysis_subfolder)",
                                              "", lane)
        lines.append("analysisdir%i:=%s\n" % (lane+1, folder))
    for lane in lanes.keys():
        folder = runfolder.get_lane_subfolder("", "$(sequence_subfolder)",
                                              "", lane)
        lines.append("sequencedir%i:=%s\n" % (lane+1, folder))
    ncolours = 4
    allmaxcycle = get_max_cycle(tiledic.values()[0])
    lines.append("maxcycle:=%i\n" % allmaxcycle)
    lines.append("cycles:=%s\n\n" % " ".join(map(lambda x: "%02i" % (x+1),
                                                 range(allmaxcycle))))
             
    lines.append("subfolders:=$(wildcard %s*)\n" % "GERALD")
    
    lines.append("TOOLS_DIR:=%s\n" %
                 os.path.abspath(os.path.join(os.path.dirname(os.path.abspath(sys.argv[0])),
                                 gerald.toolspath)))
    lines.append("QH:=%s\n" %
                 os.path.join("$(TOOLS_DIR)",
                              runfolder.get_platform_executable("quahog")))
    pure_bases, use_bases = gerald.getQhgInfo()
    lines.append("USE_BASES:=-useBases %s\n" % use_bases)
    lines.append("PURE_BASES:=-pureBases %s\n" % pure_bases)
    lines.append("QH_PARAMS:=$(USE_BASES) $(PURE_BASES)\n\n")
    
    lines.append("logfile:=$(sequence_subfolder)/$(sequence_subfolder).log\n")
    lines.append("matrixdir:=%s\n" % runfolder.matrixdir)
    lines.append("matrix:=%s\n" % "matrix.txt")
    for lane in lanes.keys():
        lines.append("matrix%i:=matrix%i.txt\n" % (lane+1, lane+1))
    for lane in lanes.keys():
        lines.append("params%i:=params%i.xml\n" % (lane+1, lane+1))
    lines.append("phasingdir:=%s\n" % runfolder.phasingdir)
    lines.append("phasing_summary:=$(phasingdir)/phasing.txt\n")
    if automatrix:
        lines.append("automatrix?=true\n")
    else:
        lines.append("automatrix?=false\n")
    if autophasing:
        lines.append("autophasing?=true\n")
    else:
        lines.append("autophasing?=false\n")
    lines.append("restarts:=%s\n" % " ".join(map(lambda x: "-restart=%i" %x, phasing_restarts)))
        
    clean_targets = map(lambda x: "clean%i" % (x+1), lanes.keys()) 
    lines.append("\nvpath %_int.txt $(analysisdir)\n")
    lines.append("vpath %phasing.txt $(phasingdir)\n")
    lines.append("vpath %phasing.xml $(phasingdir)\n")
    lines.append("vpath %_matrix.txt $(analysisdir)/$(matrixdir)\n")
    lines.append("\n"
                 ".PHONY: allmsg $(lanes) clean clean_phasing "
                 "clean_intermediate %s "
                 "recursive nonrecursive $(subfolders)\n" %
                 (" ".join(clean_targets)))
    lines.append(".PRECIOUS: %_seq.txt\n"
                 ".PRECIOUS: %_phasing.txt %.xml\n")

    lines.append("all: allmsg $(lanes) phasing.xml\n"
                 "\ttouch finished.txt\n"
                 "\t@echo Base-calling has completed successfully.\n"
                 "\t$(cmdb4)\n")
    lines.append("allmsg:\n\t@echo Analysing lanes $(lanes)\n")
    lines.append("\n")

    for lane2 in lanes.keys():
        seqfiles = []
        for tilestr in tiles:
            sample, lane, tile = runfolder.FileName.split_tile(tilestr)
            if lane2 == lane:
                seqfiles.append("%s_seq.txt" % get_var_tile(tilestr))
        lines.append("l%i_seqfiles:=%s\n" %     (lane2+1, " ".join(seqfiles)))
        lines.append("l%i_qhgfiles:=$(subst seq,qhg,$(l%i_seqfiles))\n"
                     % (lane2+1, lane2+1))
        

    ##################################################################
    # generate rules for seq.txt files
    ##################################################################
    lines.append("# Auto-generate matrices\n")
    for lane in lanes.keys():
        lines.append("$(matrix%i):\n" % (lane+1))
        if automatrixlane:
            lines.append("\t$(CP) $(analysisdir)/$(matrixdir)/s_%i_matrix.txt"
                         " $@\n\n" % (automatrixlane))
        else:
            lines.append("\t$(CP) $(analysisdir)/$(matrixdir)/s_matrix.txt "
                         "$@\n\n")
        
    lines.append("# Generate sequence and probability files\n")
    lines.append("%_seq.txt: %_int.txt\n")
    difflengthtiles = {}
    for tilestr in tiles:
        maxcycle = get_max_cycle(tiledic[tilestr])
        if maxcycle != allmaxcycle:
            difflengthtiles.setdefault(maxcycle, []).append(tilestr)
    keys = difflengthtiles.keys()
    keys.sort()
    for cycle in keys:
            lines.append("%s: maxcycle:=%i\n" %
                         (" ".join(map(lambda x: "%s_seq.txt" % x,
                                       difflengthtiles[cycle])), cycle))
    lines.append("%_seq.txt: tile=$(subst _seq.txt,,$@)\n")
    for lane in lanes.keys():
        lines.append("$(l%i_seqfiles): $(matrix%i) $(params%i)\n"
                     % (lane+1, lane+1, lane+1))
        lines.append("$(l%i_seqfiles): adir:=$(analysisdir)/$(analysisdir%i)\n"
                     % (lane+1, lane+1))
        lines.append("$(l%i_seqfiles): sdir:=.\n"
                     % (lane+1))
        lines.append("$(l%i_seqfiles): matrix:=$(matrix%i)\n"
                     % (lane+1, lane+1))
        lines.append("$(l%i_seqfiles): paramsfile:=$(params%i)\n"
                     % (lane+1, lane+1))

    lines.append("%_seq.txt:\n"
                 "\t$(cmdprefix) $(basecall_exe) -nc=$(maxcycle) -logfile=$(logfile) -matrix=$(matrix) -params=$(paramsfile) -prefix=$(adir)/ -prefix2=$(sdir)/ -prefix3=../ -tile=$(tile) $(restarts)\n\n")

    #################################################################
    # qhg files
    ################################################################
    lines.append("%_qhg.txt: %_seq.txt\n"
                 "\t$(cmdprefix) $(QH) $(QH_PARAMS) $(subst seq,sig2,$<) > $@\n")
    

    ##################################################################
    # phasing estimation
    ##################################################################
    lines.append("\n# Phasing correction\n")

    phasinglanetargets = " ".join(map(lambda x: "$(phasingdir)/$(sample%i)_%i_phasing.txt" % (x+1, x+1),
                                      lanes.keys()))
    lines.append("$(phasingdir)/phasing.txt: %s\n" % phasinglanetargets)
    lines.append("\tcat $^ > $@\n")
    for lane in lanes.keys():
        lines.append("l%i_phasingfiles:=$(addprefix $(phasingdir)/,$(subst _seq.txt,_phasing.txt,$(l%i_seqfiles)))\n"
                     % (lane+1, lane+1))
    for lane in lanes.keys():
        lines.append("$(phasingdir)/$(sample%i)_%i_phasing.txt: $(l%i_phasingfiles)\n" % (lane+1, lane+1, lane+1))
    lines.append("%s:\n" % phasinglanetargets)
    lines.append("\tfor f in $^; do\\\n"
             "\t    echo $$f;\\\n"
             "\t    $(cmdprefix) $(phasing2_exe) $$f >> $@;\\\n"
             "\tdone\n" )
    
    for lane in lanes.keys():
        lines.append("$(phasingdir)/$(sample%i)_%i_%%_phasing.txt: "
                     "$(matrix%i) $(sample%i)_%i_%%_int.txt\n" %
                     (lane+1, lane+1, lane+1, lane+1, lane+1))
        lines.append("\t$(cmdprefix) $(phasing1_exe) "
                     "--matrix=$(subst _seq.txt,_sig.txt,$(word 1, $^)) "
                     "$(subst _seq.txt,_sig.txt,$(word 2, $^)) > $@\n")

    for lane in lanes.keys():
        lines.append("$(phasingdir)/$(sample%i)_%i_phasing.xml: $(phasingdir)/$(sample%i)_%i_phasing.txt\n"
                     % (lane+1, lane+1, lane+1, lane+1))
        lines.append("\t$(cmdprefix) $(phasing3_exe) $^ > $@\n"
                     "  ifdef cmdb1\n"
                     "\t$(cmdb1) $(sample%i)_%i\n"
                     "  endif\n" % (lane+1, lane+1))

    lines.append("$(phasingdir)/phasing.xml: $(phasingdir)/phasing.txt %s\n" %
                 " ".join(map(lambda x: "$(phasingdir)/$(sample%i)_%i_phasing.xml" %
                              (x+1, x+1), lanes.keys())))
    lines.append("\t$(cmdprefix) $(phasing3_exe) $< > $@\n\n"
                 "\t$(cmdb2)\n")

    lines.append("ifeq ($(autophasing),true)\n")
    for lane in lanes.keys():
        if autophasinglane:
            lines.append("$(params%i): $(phasingdir)/$(sample%i)_%i_phasing.xml $(phasingdir)/$(sample%i)_%i_phasing.xml\n"
                         % (lane+1, autophasinglane, autophasinglane,
                            lane+1, lane+1))
        else:
            lines.append("$(params%i): $(phasingdir)/phasing.xml $(phasingdir)/$(sample%i)_%i_phasing.xml\n"
                         % (lane+1, lane+1, lane+1))
        lines.append("\t$(CP) $< $@\n")
    lines.append("endif\n")

    ##################################################################
    # targets to make typing easier
    ##################################################################
    tiletargets = {}
    lines.append("\n# Some rules to make typing targets easier\n")
    for tilestr in tiles:
        lines.append("%s: %s_qhg.txt\n" % (tilestr, get_var_tile(tilestr)))
        sample, lane, tile = runfolder.FileName.split_tile(tilestr)
        tiletargets.setdefault(tile, []).append(get_var_tile(tilestr))
    keys = tiletargets.keys()
    keys.sort()
    for key in keys:
        lines.append("_%04i: %s\n" % (key + 1, " ".join(tiletargets[key])))

    ##################################################################
    # generate rules for each lane
    ##################################################################
    for lane in lanes.keys():
        lines.append("$(sample%i)_%i: l%i_msg $(sample%i)_%i_finished.txt\n" %
                     (lane+1, lane+1, lane+1, lane+1, lane+1))
        lines.append("$(sample%i)_%i_finished.txt: $(l%i_qhgfiles)\n" %
                     (lane+1, lane+1, lane+1))
        lines.append("\ttouch $@\n"
                     "  ifdef cmdb3\n"
                     "\ttouch $@; $(cmdb3) $(subst _finished.txt,,$@)\n"
                     "  endif\n")
        lines.append("l%i_msg:\n\t@echo Analysing lane %i: $(l%i_qhgfiles)\n" %\
                     (lane+1, lane+1, lane+1))
    lines.append("\n")

    ##################################################################
    # clean targets
    ##################################################################
    lines.append("\n# clean targets\n");
    lines.append("clean: %s clean_phasing\n" % " ".join(clean_targets))
    lines.append("\t$(RM) $(sequence_subfolder)/*.log\n")
    lines.append("\t$(RM) ztrarc.hash\n"
                 "\t$(RM) *finished.txt\n")
    lines.append("    ifeq ($(automatrix),true)\n"
                 "\t$(RM) matrix*.txt\n"
                 "    endif\n")
    lines.append("    ifeq ($(autophasing),true)\n"
                 "\t$(RM) params*.xml\n"
                 "    endif\n")
    for lane in lanes.keys():
        lines.append("clean%i: clean%i_seq clean%i_sig "
                     "clean%i_sig2 clean%i_qhg\n" %
                     (lane+1, lane+1, lane+1, lane+1, lane+1))
        lines.append("clean%i_seq:\n" % (lane+1))
        lines.append("\tfind . -regex \".*$(sample%i)_%i_.*_\\(seq\\|prb\\)\\.txt\" | xargs $(RM)\n" % (lane+1, lane+1))
        lines.append("clean%i_sig:\n" % (lane+1))
        lines.append("\tfind . -name \"$(sample%i)_%i_*_sig.txt\" | xargs $(RM)\n" % (lane+1, lane+1))
        lines.append("clean%i_qhg:\n" % (lane+1))
        lines.append("\tfind . -name \"$(sample%i)_%i_*_qhg.txt\" | xargs $(RM)\n" % (lane+1, lane+1))
        lines.append("clean%i_sig2:\n" % (lane+1))
        lines.append("\tfind . -name \"$(sample%i)_%i_*_sig2.txt\" | xargs $(RM)\n" % (lane+1, lane+1))

    lines.append("clean_intermediate:\n"
                 "\t@echo Target clean_intermediate is obsolete.\n")
    lines.append("clean_phasing:\n")
    lines.append("\tfind $(phasingdir) -regex \".*\\.\\(txt\\|xml\\)\" | xargs $(RM)\n\n")

    lines.append("compress uncompress:\n"
                 "\t$(MAKE) -f Makefile.compress.bustard $@\n")

    ##################################################################
    # recursive make
    ##################################################################
    lines.append("# Recursive analysis: do alignment\n")
    lines.append("recursive: $(subfolders) nonrecursive\n")
    lines.append("nonrecursive:\n"
                 "\t$(MAKE) $(TARGET)\n")
    lines.append("$(subfolders): nonrecursive\n"
                 "    ifeq ($(words $(filter-out clean%,$(filter-out all,$(TARGET)))), 0)\n"
                 "\t$(MAKE) --directory=$@ $(TARGET)\n"
                 "    else\n"
                 "\t$(MAKE) --directory=$@ TILE=$(TARGET) all\n"
                 "    endif\n")

    # ztr archive
    lines.append("\n# Export trace file for gap4\n")
    lines.append("ztrarc.hash:\n")
    lines.append("\tsolexa2ztr -p -o $@ *_seq.txt\n")
    try:
        fd = open(filename, "w")
        if not fd:
            sys.stderr.write("Error: Could not open %s\n", filename)
            return
        fd.writelines(lines)
        fd.close()
    except:
        sys.stderr.write("Error: Failed to write %s\n", filename)

    makefilename = "Makefile.compress.bustard"
    makefilepath2 = runfolder.get_executable(makefilepath)
    try:
        shutil.copyfile(os.path.join(makefilepath2, makefilename),
                        os.path.join(sequence_folder, makefilename))
    except:
        sys.stderr.write("Error: Failed to copy %s\n" % makefilename)
