/*
 * Statistical routines.
 *
 * Copyright (c) 2006
 * Author: Klaus Maisinger
 * This source file is covered by the "Solexa Public Source License"
 * agreement and bound by the terms therein.
 */

#if defined(__BORLANDC__)
#include <math.h>
#else
#include <cmath>
#endif
#include <cassert>
#include "Maths.h"

#define TINY 1.e-12
 
int round(float f)
{
    return (int)floor(f+0.5);
}

float max(float *a, int n)
{
    assert(n >= 1);
    int i;
    float m = a[0];
    for (i = 1; i < n; i++)
        if (a[i] > m)
            m = a[i];
    return m;
}

float min(float *a, int n)
{
    assert(n >= 1);
    int i;
    float m = a[0];
    for (i = 1; i < n; i++)
        if (a[i] < m)
            m = a[i];
    return m;
}

float average(float *x, int n)
{
    int i;
    double r = 0.;
    for (i = 0; i < n; i++)
        r += x[i];
    return (float)(r/n);
}


/* 
 * Linear regression.
 * Returns gradient and intercept of the <ndim> data points given by the
 * <x> and <y> arrays.
 */
int linregress(float *x, float *y, int ndim,
               float *gradient, float *intercept)
{
    assert(ndim > 1);
    *gradient = 0.f;
    *intercept = 0.f;
    if (fabs(max(x, ndim) - min(x, ndim)) < TINY)
        return 1;
    int i;
    double sum_x = 0.f, sum_y = 0.f, sum_xx = 0.f, sum_yx = 0.f;
    for (i = 0; i < ndim; i++) {
        sum_x += x[i];
        sum_xx += x[i] * x[i];
        sum_y += y[i];
        sum_yx += y[i] * x[i];
    }
    double denom = ((ndim * sum_xx) - (sum_x * sum_x));
    if (fabs(denom) < TINY)
        return 1;
    *gradient = (float)((ndim * sum_yx - sum_x * sum_y) / denom);
    *intercept = (float)((sum_y - (*gradient) * sum_x) / ndim);
    return 0;
}

float L1norm(float *x, float *y, int n)
{
    int i;
    double result = 0.;
    for (i = 0; i < n; i++) {
	result += fabs(x-y);
    }
    return (float)result;
}
