// QUAHOG.cpp
// Copyright (c) 2005, 2006 Solexa
// Author: A. J. Cox
// This source file is covered by the "Solexa Public Source License"
// agreement and bound by the terms therein.

#include<cstdio>
#include<cctype>
#include<algorithm>
#include<iostream>
#include<map>
#include<vector>
#include<cmath>
#include<cassert>

static char* quahog_cpp_id="@(#) $Id: QUAHOG.cpp,v 1.6 2007/02/28 15:10:15 km Exp $";


// Stands for _QUA_lity _H_euristic _O_utput _G_enerator
// Generates quality metrics from a cluster intensity file.
// Metrics are mainly to do with detecting the possibility of cluster mixing.

// Creates storage for this many cycles of sequencing - includes any deblocks
static  const uint maxNumCycles(200);

// max distance between clusters that will be guaranteed to be picked up
// neighbours more distant than this that cross bin boundaries may not
// be detected
static const uint maxDist(10);
static const uint maxDistSqrd(maxDist*maxDist);

// cluster positions are placed in bins of binSize x binSize pixels
static const uint binSize(128);

// NB: largest possible grid of bins is maxBinGridSize x maxBinGridSize
static const uint maxBinGridSize(512);

// no more than 8 lanes per chip in current design
static const uint maxNumLanes(8);


using namespace std;

typedef unsigned int uint;

struct ClusterPos
{
  ClusterPos( void ) {}
  ClusterPos(int lane,
             int tile,
             int clusX,
             int clusY,
             float iA, float iC, float iG, float iT
             ): lane_(lane), tile_(tile), clusX_(clusX), clusY_(clusY),
                 dist_(9999*9999) 
  {
    inten_[0]=iA;
    inten_[1]=iC;
    inten_[2]=iG;
    inten_[3]=iT;
  } // ~ctor

  ClusterPos( const ClusterPos& rhs ) : lane_(rhs.lane_), tile_(rhs.tile_), 
                                        clusX_(rhs.clusX_), 
                                        clusY_(rhs.clusY_), dist_(rhs.dist_) 
  {
    for (int i(0);i<4;i++) inten_[i]=rhs.inten_[i];
  } // ~copy ctor

  //  void operator=(const ClusterPos& rhs ) {} 
  int lane_;
  int tile_;
  int clusX_;
  int clusY_;
  uint dist_;
  float inten_[4];
};

// getDist: get distance between positions of two clusters
// Just store squared difference - can then deal with as integer, only
// take square root for output
uint getDist( const ClusterPos& c1, const ClusterPos& c2 )
{
  int dX(c1.clusX_-c2.clusX_), dY(c1.clusY_-c2.clusY_);
  dX*=dX; 
  dY*=dY;
  return ( dX+dY );
} // ~getDist

// getSimilarity: work out similarity between 1st cycle intensity vectors
// of two clusters. Worked out as cos of angle between them.
// This is 1 if they are in exactly same direction
// This is 0 if they are at right angles (e.g perfect A vs. perfect G)
// A negative value is a sign something weird is going on (e.g. with
// matrix correction), since it can't happen for positive intensity values
// Needs to be interpreted with care, as for single template experiment
// we expect all 1st cycle intensities to be similar. For a random mix
// of templates, any 2 templates have a 1 in 4 chance of being similar
float getSimilarity( const ClusterPos& c1, const ClusterPos& c2 )
{
  float dotProd(0.0), mag1(0.0), mag2(0.0);

  for (int j(0); j<4; j++)
  {
    mag1+=c1.inten_[j]*c1.inten_[j];
    mag2+=c2.inten_[j]*c2.inten_[j];
    dotProd+=c1.inten_[j]*c2.inten_[j];
  } // ~for
  mag1=sqrt(mag1);
  mag2=sqrt(mag2);
  mag1*=mag2;


  if (mag1!=0) dotProd/=mag1;

  // if mag1==0 at least one of the vectors must have had zero
  // magnitude, so will return zero
  return dotProd;

} // ~getSimilarity




// get bin number for a coordinate
uint getBin( uint coord )
{
  return (uint)(coord/binSize);
} // ~getBin

// get hash key for a cluster pos - unique key for each bin in each tile
uint getKey( const ClusterPos& c ) 
{
    return getBin(c.clusY_)
      + (getBin(c.clusX_)*maxBinGridSize) 
      + (c.lane_*maxBinGridSize*maxBinGridSize) 
      + (c.tile_*maxBinGridSize*maxBinGridSize*maxNumLanes);
} // ~getKey 


// class Sorter: use this as a sort criterion to sort list of clusters 
// by distance to a given cluster
class Sorter
{
public:
  Sorter( const ClusterPos& c ) : c_(c) {}
  bool operator()( const ClusterPos* p1, const ClusterPos* p2 )
  {
    return (getDist( c_, *p1 )<getDist( c_, *p2 ));
  } // ~operator()

private:
  const ClusterPos& c_;
}; // class ~Sorter


struct ClusterPurity
{
  ClusterPurity() {}
  ClusterPurity( const char* readBuf, 
                 const uint pureBases, 
                 const vector<bool>& basesToUse );

  // Mike's purity definition: largest intensity as a fraction of total
  float purity;
  // My purity definition: largest as fraction of (largest + next largest)
  float chastity;

  vector<vector<float> > intensity;

}; // ~struct ClusterPurity


int badLines = 0;

ClusterPurity::ClusterPurity
( const char* readBuf, 
  const uint pureBases, 
  const vector<bool>& basesToUse ) : purity(1000000), chastity(1000000)
{
  uint basesGot(0);
  //  uint cyclesFound=0;
  vector<float> thisInten(4);
  vector<bool>::const_iterator i(basesToUse.begin());
  float thisPure, thisChaste;

  const char* p(readBuf);
  for (uint j(0); j<3;j++)
  {
    p=strchr(p,'\t');
    if (p==NULL)
    {
      cerr << "Error: problem parsing line:\n" << readBuf << endl;
      exit (-1);
    } // ~if
    p++;
  } // ~for
  //  uint z(0);
  while ((p=strchr(p,'\t'))!=NULL)
  {
    p++;
    if (i==basesToUse.end())
    {
      cerr << "Error: need " << pureBases 
           << " bases, but -useBases only gives locations for "
           << basesGot << endl;
      exit (-1);
    }
    //    cout << z++;

    if ((*i)==true)
    {
      sscanf( p, "%f %f %f %f", 
              &thisInten[0], &thisInten[1],
              &thisInten[2], &thisInten[3] );
      intensity.push_back(thisInten);
      //    cout << ": " << thisInten[0] << " "  << thisInten[1] << " "  << thisInten[2] << " "  << thisInten[3] << endl; 
      sort( thisInten.begin(),thisInten.end() );
      if (thisInten[3]==0)
      {
        thisPure=0;
        thisChaste=0;
      } // ~if
      else
      {
        thisPure = thisInten[3]
          /(thisInten[0]+thisInten[1]+thisInten[2]+thisInten[3]);
        thisChaste = thisInten[3]
          /(thisInten[2]+thisInten[3]);
      } // ~else          

      if (thisPure<purity) purity=thisPure;
      if (thisChaste<chastity) chastity=thisChaste;

      if (++basesGot==pureBases) return;
    } // ~if
    i++;
  } // ~do
  if (badLines == 0) {
    cerr << "Warning: only found " << basesGot << " bases, needed " << pureBases 
       << ". Line =" << endl << readBuf << endl;
    cerr << "Skipping further warnings." << endl;    
    badLines++;
  }
} // ~ClusterPurity::ctor

float getSimilarity( const ClusterPurity& c1, const ClusterPurity& c2 )
{
  assert(c1.intensity.size()==c2.intensity.size());
  float similarity(999999);
  double dotProd(0.0), mag1(0.0), mag2(0.0);

  for (uint i(0);i<c1.intensity.size();i++)
  {
    dotProd=0.0;mag1=0.0;mag2=0.0;

    //  cout << c1.intensity[i][0] << " " << c1.intensity[i][1] << " " << c1.intensity[i][2] << " " << c1.intensity[i][3] << " - " << c2.intensity[i][0] << " " << c2.intensity[i][1] << " " << c2.intensity[i][2] << " " << c2.intensity[i][3] << endl;

    for (int j(0); j<4; j++)
    {
      mag1+=c1.intensity[i][j]*c1.intensity[i][j];
      mag2+=c2.intensity[i][j]*c2.intensity[i][j];
      dotProd+=c1.intensity[i][j]*c2.intensity[i][j];
    } // ~for
    mag1=sqrt(mag1);
    mag2=sqrt(mag2);
    mag1*=mag2;
    if (mag1!=0) dotProd/=mag1;
    if (dotProd<similarity) similarity=dotProd;
  }


  // if mag1==0 at least one of the vectors must have had zero
  // magnitude, so will return zero
  return similarity;

}



vector<ClusterPos> clusterPos;
vector<ClusterPurity> clusterPurity;
map< uint, vector<ClusterPos*> > bins;

vector<vector<ClusterPos*> > nbors;

// If the last two command args contain a '%' wild card, treat them as
// templates to convert input file names into output file names
// e.g. InputDir/%_in.txt -> OUTDIR/%_out.txt
// will take files in InputDir as input and put an output file for
// each in OutDir. Syntax is as for makefile targets
bool parseArgs( uint numArgs, const char** args, 
                vector<const char*>& inputFile,
                vector<char*>& outputFile,
                vector<bool>& basesToUse,
                uint& pureBases )
{
  const char wildCard('%');
  char c;

  if (numArgs==1)
    return false; // need at least one argument

  uint thisArg(1);

  if (strcmp( args[thisArg], "-useBases")==0)
  {
    ++thisArg;
    assert(thisArg<numArgs);

    if (strcmp( args[thisArg], "all")==0)
    {
      basesToUse=vector<bool>(maxNumCycles,true);
    } // ~if
    else if (strcmp( args[thisArg], "odd")!=0)
    {
      // Assume custom useBases argument, error check it
      for (uint i(0); i<strlen(args[thisArg]); i++)
      {
        c=toupper(args[thisArg][i]);
        basesToUse.push_back(c=='Y');
        if ((c!='Y')&&(c!='N'))
        {
          cerr << "Error: could not parse -useBases argument " << args[thisArg]
               << endl
               << "Must be 'all', 'odd' or some combination of y,Y,n,N"
               << endl;
          return false;
        } // ~if
      } // ~for
    } // ~else if
    ++thisArg;
    //    assert(thisArg<numArgs);
  } // ~if

  if (basesToUse.empty())
  { // add in default (same result as '-useBases odd')
      for (uint i(0); i<maxNumCycles; i++)
        basesToUse.push_back(((i&0x1)==0)?true:false);
  } // ~if

  if (strcmp( args[thisArg], "-pureBases")==0)
  {
    pureBases=atoi(args[++thisArg]);
    if (pureBases<=0)
    {
      cerr 
        << "Error: -pureBases argument must be numeric and greater than zero"
        << endl;
      return false;
    } // ~if

    ++thisArg;

  } // ~if


  for (uint i(thisArg);i<numArgs;i++)
    inputFile.push_back(args[i]);

  if (numArgs-thisArg+1>3)
  { // check for wild card
    char *suffIn, *suffOut;
    
    char prefIn[strlen(args[numArgs-2])+1];
    char prefOut[strlen(args[numArgs-1])+1];
    strcpy(prefIn,  args[numArgs-2] );
    strcpy(prefOut, args[numArgs-1] );

    if ((suffIn=strchr(prefIn,wildCard))!=NULL)
    { // found wildcard in 1st
      if ((suffOut=strchr(prefOut,wildCard))==NULL)
        return false; // no % in prev argument

      // OK, take wildcards off list
      inputFile.pop_back();
      inputFile.pop_back();

      // and try to parse them
      *suffIn++='\0';
      *suffOut++='\0';

      char buf[1024];

      strcpy(buf, prefOut);
      int prefInLen(strlen(prefIn));
      int prefOutLen(strlen(prefOut));
      int suffInLen(strlen(suffIn));

      for (vector<const char*>::iterator i(inputFile.begin());
           i!=inputFile.end(); ++i)
      {
        //        cout << *i << " P:" << prefIn << " S:" << suffIn << endl;
        if (strncmp( prefIn, *i, prefInLen )
             !=0)
        {
          cerr << "Error: could not match wildcard prefix " << prefIn
               << " to file name " << *i << endl;
          return false;
        } // ~if
        if (strncmp( suffIn, *i + strlen (*i) - suffInLen, suffInLen )
            !=0)
        {
          //          cerr << (*i+strlen(*i)-suffInLen) << endl;
          cerr << "Error: could not match wildcard suffix " << suffIn
               << " to file name " << *i << endl;
          return false;
        } // ~if


        buf[prefOutLen]='\0';

        strncat( buf, *i+prefInLen, strlen(*i)-prefInLen-suffInLen); 
        strcat( buf, suffOut );
        outputFile.push_back( new char[strlen(buf)+1] );
        strcpy( outputFile.back(), buf );
        cout << *i << " -> " << outputFile.back() << endl;
      } // ~for

    } // ~if suffIn...


  } // ~if numArgs...

  //  exit(0);
  return true;

} // ~parseArgs





int main( uint numArgs, const char** args )
{

  vector<const char*> inputFile;
  vector<char*> outputFile;
  vector<bool> basesToUse;
  uint pureBases=1;

  if (!parseArgs( numArgs, args, 
                  inputFile, outputFile, basesToUse, pureBases ))
  {
    cerr << "Usage: \n" 
         << args[0] 
         << " [-useBases basesToUse] [-pureBases pureBases] intensityFiles\n"
         << "> output.txt\n"
         << args[0] 
         << " [-useBases basesToUse] [-pureBases pureBases] intensityFiles\n"
         << "inPrefix%inSuffix outPrefix%outSuffix\n"
         << "e.g. " << args[0]
         << " exptDir/s_1_????_int.txt exptDir/%_int.txt %_qhg.txt\n\n"
         << "-useBases: specify which bases to align/ignore\n"
         << "-useBases odd: align odd-numbered bases only (default)\n"
         << "-useBases all: align all bases\n"
         << "-useBases YYNNYYY...: Y=align base, N=don't\n"
         << "-pureBases: bases to purity filter on (default=1)"
         << endl;

    exit (1);
  } // ~numArgs


  const int bufSize(4*65536);
  char readBuf[bufSize];

  uint lane, tile, clusX, clusY, clusXa, clusYa;
  vector<float> int1st(4);
  clusterPos.reserve(10000000);
  
  ClusterPos s;

  //  char outNameBuf[512];

  // send output to stdout unless redirected
  FILE* pOut=stdout;

  
  //  for (uint thisFile(1); thisFile<numArgs-2;thisFile++)
  for (uint thisFile(0); thisFile<inputFile.size();thisFile++)
  {

    FILE* pInten=fopen(inputFile[thisFile], "r");

    cerr << "About to analyse intensity file " << inputFile[thisFile] << endl;
    if (pInten==NULL)
    {
      cerr << "Error: Problem opening file " << inputFile[thisFile] << endl;
      exit (-1);
    } // ~if

    // make output file name
    //    strcpy (outNameBuf, args[thisFile]);
    //  char* pSuff(strstr(outNameBuf,args[numArgs-2]));
    //  if (pSuff==NULL)
    //   {
    //   cerr << "Error: did not find suffix " << args[numArgs-2]
    //        << " in filename " << args[thisFile] << endl;
    //   exit (-1);
    //   } // ~if
    //  else if (strlen(pSuff)!=strlen(args[numArgs-2]))
    //  {
    //   cerr << "Error: suffix string " << args[numArgs-2]
    //         << " not at end of filename " << args[thisFile] << endl;
    //   exit (-1);
    //  } // ~else if
    //  strcpy(pSuff, args[numArgs-1]);
    //  
    //  cerr << "Will send output to file " << outNameBuf << endl;
      
    //    FILE* pOut=fopen(outNameBuf, "w");

    if (!outputFile.empty())
    {
      cerr << "Will send output to file " << outputFile[thisFile] << endl;
      pOut=fopen(outputFile[thisFile], "w");
      if (pOut==NULL)
      {
        cerr << "Error: Problem opening output file " 
             << outputFile[thisFile] << endl;
        exit (-1);
      } // ~if
    } // ~if
    else
      cerr << "Will send output to standard output" << endl;

    clusterPos.resize(0);
    clusterPurity.resize(0);
    bins.clear();
    nbors.resize(0);

    while (fgets( readBuf, bufSize, pInten)!=NULL)
    {
      sscanf(readBuf, "%u\t%u\t%u\t%u\t%f\t%f\t%f\t%f", 
             &lane, &tile, &clusX, &clusY,
             &int1st[0], &int1st[1], &int1st[2], &int1st[3] );

      //      sscanf(readBuf, "%u\t%u\t%u\t%u", 
      //      &lane, &tile, &clusX, &clusY );


      clusterPurity.push_back
        ( ClusterPurity( readBuf, pureBases, basesToUse) );

      
      // add to cluster list
      clusterPos.push_back
        ( ClusterPos( lane, tile, clusX, clusY ,
                      int1st[0], int1st[1], int1st[2], int1st[3] ));
      s.lane_=lane;
      s.tile_=tile;
      s.clusX_=clusX;
      s.clusY_=clusY;

      // hash to bin for original cluster position
      bins[ getKey(s) ].push_back(&clusterPos.back());
      //    cout << "Hashing cluster " << clusterPos.size() << " to " << getBin(s.clusX_) << " " << getBin(s.clusY_) << endl; 

      // then check to see if needs to go in any adjacent bins
      clusXa=clusX+maxDist;
      clusYa=clusY+maxDist;

      if ((getBin(clusX)!=getBin(clusXa))&&(getBin(clusY)!=getBin(clusYa)))
      {
        s.clusX_=clusXa;
        s.clusY_=clusYa;
        bins[ getKey(s) ].push_back(&clusterPos.back());
        //  cout << "Hashing cluster +xy " << clusterPos.size() << " to " << getBin(s.clusX_) << " " << getBin(s.clusY_) << endl; 
      } // ~if
      if (getBin(clusX)!=getBin(clusXa))
      {
        s.clusX_=clusXa;
        s.clusY_=clusY;
        bins[ getKey(s) ].push_back(&clusterPos.back());
        //  cout << "Hashing cluster +x " << clusterPos.size() << " to " << getBin(s.clusX_) << " " << getBin(s.clusY_) << endl; 
      } // ~if
      if (getBin(clusY)!=getBin(clusYa))
      {
        s.clusX_=clusX;
        s.clusY_=clusYa;
        bins[ getKey(s) ].push_back(&clusterPos.back());
        //  cout << "Hashing cluster +y " << clusterPos.size() << " to " << getBin(s.clusX_) << " " << getBin(s.clusY_) << endl; 
      } // ~if

    } // ~while
    fclose(pInten);
    cerr << "Found " << clusterPos.size() << " clusters" << endl;
    nbors.resize(clusterPos.size());

    for (map< uint, vector<ClusterPos*> >::iterator i(bins.begin());
         i!=bins.end();++i)
    {
      if (i->second.size()>1)
      {
        // do an all against all comparison
        vector<ClusterPos*>::iterator jLast(i->second.end());
        --jLast;
        for (vector<ClusterPos*>::iterator j(i->second.begin());j!=jLast;j++)
        {
          for (vector<ClusterPos*>::iterator k(j+1);k!=i->second.end();k++)
          {
            //       cout << "comparing " << (&(**j)-&clusterPos[0]) << " and "
            //     << (&(**k)-&clusterPos[0]) << endl;

            uint d( getDist( **j, **k ) );
            if (d<(*j)->dist_) (*j)->dist_=d;
            if (d<(*k)->dist_) (*k)->dist_=d;
            if (d<maxDistSqrd)
            {
              //            cout << "they're neighbours" << endl;
              nbors[&(**j)-&clusterPos[0]].push_back(*k);
              nbors[&(**k)-&clusterPos[0]].push_back(*j);
            }
          } // ~for k
        } // ~for j
        
      } // ~if
    } // ~for i

    float thisDist, thisSim;
    char spacer;
    int lastNbor, thisNbor;
    for (uint i(0);i<clusterPos.size();i++)
    {

      fprintf ( pOut, "%u\t%u\t%i\t%i\t%7.2f\t%7.2f",
                clusterPos[i].lane_,
                clusterPos[i].tile_,
                clusterPos[i].clusX_,
                clusterPos[i].clusY_,
                clusterPurity[i].purity,
                clusterPurity[i].chastity );
      //                sqrt((double)clusterPos[i].dist_));

      Sorter sorter(clusterPos[i]);
      thisSim=-99;
      if (nbors[i].size()>0)
      {
        sort( nbors[i].begin(), nbors[i].end(), sorter );

        lastNbor=-1;
        spacer='\t';
        //  crowdedness=0.0; 


        fprintf(pOut, "\t%7.2f\t%7.2f",
                //               getSimilarity( clusterPos[i], *nbors[i][0] ),
               getSimilarity( clusterPurity[i], 
                              clusterPurity[nbors[i][0]-&clusterPos[0]] ),
               sqrt((double)clusterPos[i].dist_));

        for (uint j(0);j<nbors[i].size();j++)
        {
          // neighbour list will contain some repeats, as in some case
          // both neighbours end up in more than 1 bin. But these entries
          // will end up as adjacent in the sorted list, so filter them out
	  thisNbor=&*nbors[i][j]-&clusterPos[0];
          if (thisNbor!=lastNbor)
          {
            thisDist=sqrt((double)getDist( clusterPos[i], *nbors[i][j] ));
          
            fprintf(pOut, "%c%i", spacer, thisNbor);
            //                 thisDist );
            lastNbor=thisNbor;
            spacer=':';
          } // ~if
        } // ~for j
        
        fprintf(pOut, "\n");
      } // ~if
      else 
      { 
        // if no neighbours<maxDist, print out dummy similarity value
        // nearest neighbour found (may be an overestimate)
        // and dummy neighbourhood description
        fprintf(pOut, "\t%7.2f\t%7.2f\t-1\n",
                (float)-1,
                sqrt((double)clusterPos[i].dist_));

      } // ~else

    }  // ~for i
    if (!outputFile.empty())
    {
      fclose(pOut);
    }
  } // ~for thisFile

  // deallocate output filename strings
  for (vector<char*>::iterator i(outputFile.begin());
       i!=outputFile.end();i++)
  { 
    delete [] *i;
  } // ~for
  

} // ~main

// end of file QUAHOG.cpp
