/*
 * Copyright (c) Solexa 2006
 * Author: Klaus Maisinger
 * This source file is covered by the "Solexa Public Source License"
 * agreement and bound by the terms therein.
 */

/*
 * This class holds the integrated binned lookup table.
 * It has functions to integrate, discard bins, find the bin with
 * the lowest error rate.
 */
#ifndef ERRORTABLE_H_
#define ERRORTABLE_H_

#include <cmath>
#include <cstdio>
#include "DataTable.h"

class QualityCount {
public:
    long count;
    long error;
    int  q_phred;
    int  discarded;
    QualityCount() { count = 0; error = 0; q_phred = 0; discarded = 0; }
    virtual ~QualityCount() {};
};

/*
 * Implementation of the error lookup table described in the phred paper.
 */
class ErrorTable : public BinnedTable<QualityCount>
{
public:
    ErrorTable() {};
    virtual ~ErrorTable() {};
    virtual void create(DataTable *data)
    {
#if defined(DEBUG)
        printf("ErrorTable::create()\n");
#endif
        int i;
        binnings = data->binnings;
        int noparams = binnings.size();
        table.resize(data->table.size());
        if (table.size() == 0)
            return;
        computeCounts(data);
        computeScores();
    }

    /*
     * Calculate the cumulative counts for the error lookup table
     * by summing up the errors and counts for all bins with parameter indices 
     * less than the current bin.
     */
    virtual void computeCounts(DataTable *data)
    {
#if defined(DEBUG)
        printf("ErrorTable::computeCounts()\n");
#endif
        assert(table.size() > 0);
        int noparams = binnings.size();
        int i, i2, j;
        // copy counts/errors per bin from data        
        for (i = 0; i < table.size(); i++) {
            if (table[i].discarded) {
                table[i].count = 0;
                table[i].error = 0;
            }
            else {
                ErrorCount count = data->getErrorCount(i);
                table[i].count = count.count;
                table[i].error = count.error;
            }
        }

        // now sum over grid to accumulate all data for those points
        // for which each parameter is less than the corresponding parameter
        // values for the current point
        std::vector<int> indices;
        for (j = 0; j < noparams; j++) {
            for (i = 0; i < table.size(); i++) {
                indices = getBinIndices(i);
                if (indices[j] == 0) 
                    continue;
                indices[j] -= 1;
                i2 = getBinIndex(indices);
                table[i].count += table[i2].count;
                table[i].error += table[i2].error;
            }
        }
            
    }
    /*
     * Compute the phred scores for all bins.
     */
    virtual void computeScores()
    {
#if defined(DEBUG)
        printf("ErrorTable::computeScores()\n");
#endif
        int i;
        for (i = 0; i < size(); i++) {
            table[i].q_phred = 0;
            if (!table[i].discarded)
                table[i].q_phred = 
                    (int)floor(-10. * log10((1.0+double(table[i].error))
                                      / (1.+double(table[i].count)))+0.5);
        }
    }
    /*
     * Discards a whole set of bins from the analysis.
     * For an index i_k in the k-th parameter dimension, discard all bins
     * with values j_k <= i_k, k=1,..., no of dimensions by setting a flag.
     * The counts have to be recomputed afterwards - there is probably a more
     * efficient way of implementing this.
     */
    virtual void discardBins(int binidx)
    {
        std::vector<int> bin_indices = getBinIndices(binidx);
        int i, j;
        int noparams = bin_indices.size();
        std::vector<int> indices;
        for (i = 0; i <= binidx; i++) {
            if (table[i].discarded)
                continue;
            indices = getBinIndices(i);
            int flag = 0;
            for (j = 0; j < noparams; j++) 
                if (indices[j] > bin_indices[j]) {
                    flag = 1;
                    break;
                }
            if (!flag)
                table[i].discarded = 1;
        }
    }

    /*
     * Find the bin with the next maximum, following the criteria
     * described in the phred paper.
     */
    virtual int findMaximalBin()
    {
        // Find bin with maximum phred score.
        int i, j;
        int max = 0;
        std::vector<int> maxima;
        for (i = 0; i < table.size(); i++) {
            if (table[i].discarded)
                continue;
            if (table[i].q_phred > max) {
                max = table[i].q_phred;
                maxima.clear();
            }
            if (table[i].q_phred == max) {
                maxima.push_back(i);
            }
        }
        if (max < 1)
            return -1;
        if (maxima.size() == 0)
            return -1;
        if (maxima.size() == 1)
            return maxima[0];
        
        // Found more than one maximum if sorted by quality score.
        // Now look for maximum with largest number of data points.
        std::vector<int> maxima2;
        long maxcount = table[maxima[0]].count;
        for (i = 0; i < maxima.size(); i++) {
            long count = table[maxima[i]].count;
            if (count > maxcount) {
                maxcount = count;
                maxima2.clear();
            }
            if (count == maxcount)
                maxima2.push_back(maxima[i]);
        }
        if (maxima2.size() == 1)
            return maxima2[0];

        // Still more than one maximum; look at largest sum of indices
        // the phred paper assumes equal number of bins in all dimensions,
        // which this implementation doesn't. Presumably this criterion is still as
        // good as any other.
        maxima = maxima2;
        maxima2.clear();
        max = 0;
        for (i = 0; i < maxima.size(); i++) {
            std::vector<int> indices = getBinIndices(maxima[i]);
            int count = 0;
            for (j = 0; j < indices.size(); j++)
                count += indices[j];
            if (count > max) {
                maxima2.clear();
                max = count;
            }
            if (count == max)
                maxima2.push_back(maxima[i]);
        }
        
        // Now we've had enough...
        return maxima2[0];
    }

    virtual void print() 
    {
        int i, j, k;
        std::vector<int> indices;
        std::vector<float>params;
        for (i = 0; i < table.size(); i++) {
            indices = getBinIndices(i);
            params  = getBinParams(indices);
            printf("%6i:", i);
            for (j = 0; j < indices.size(); j++)
                printf(" %3i", indices[j]);
            printf(": %7li %7li %3i %1i (",
                   table[i].count, table[i].error,
                   table[i].q_phred, table[i].discarded);
                    
            for (j = 0; j < params.size(); j++)
                printf(" %6.1f", params[j]);
            printf(")\n");
        }
    }
};

#endif /*ERRORTABLE_H_*/
