/*
 * Copyright (c) Solexa 2006
 * Authors: Klaus Maisinger, A. J. Cox 
 * This source file is covered by the "Solexa Public Source License"
 * agreement and bound by the terms therein.
 */

#include <cstdio>
#include <string>
#include "QualityTable.h"

// Identifier to embed in source and binary files
static char* quality_apply_id="@(#) $Id: QualityApply.cpp,v 1.7 2006/12/11 14:07:23 km Exp $";

/*
 * Apply a quality calibration table as calculated by QualityCalibration.cpp
 * to a data set.
 */

const int bufferSize(128);

class QualityWriter
{
public:
  virtual void operator()(std::string& s, bool isFirst, int score)=0;
protected:
  char buffer[bufferSize];
}; // ~class QualityWriter


// write numerical output as per Klaus' original code
class QualityWriterNumeric : public QualityWriter
{
public:
  virtual void operator()(std::string& s, bool isFirst, int score)
  {
	snprintf(buffer, sizeof(buffer), " %i", score);
	s += (buffer+(isFirst?1:0));
  }
}; // ~class QualityWriterNumeric

// write compact ASCII-encoded output
class QualityWriterSymbolic : public QualityWriter
{
public:
  virtual void operator()(std::string& s, bool isFirst, int score)
  {
	snprintf(buffer, sizeof(buffer), "%c", 64+score);
	s+=buffer;
  }
}; // ~class QualityWriterSymbolic


int writeQualityOutput(FILE *fp,
                       DataSet *set, 
                       std::vector<int> *scores,
                       QualityWriter& writeQuality)
{
    int i;
    std::string s;
    std::vector<std::vector<float > > *values = set->getParameterValues();
    int cycle = 0;
    int prevcycle = 0;
    for (i = 0; i < scores->size(); i++) {
        cycle = int((*values)[0][i]);
        
        if (cycle < prevcycle) {
            fprintf(fp, "%s\n", s.c_str());
            s = "";
        }
        writeQuality(s,(cycle<prevcycle)||(i==0),scores->at(i));
        prevcycle = cycle;
        if (i == scores->size()-1)
           fprintf(fp, "%s\n", s.c_str());
    }
}

int writeQualityOutput(const char filename[],
                       DataSet *set, 
                       std::vector<int> *scores,
		       QualityWriter& writeQuality)
{
    FILE *fp = fopen(filename, "r");
    if (!fp)
        return -1;
    int result = writeQualityOutput(fp, set, scores, writeQuality);
    fclose(fp);
    return result;
}    
    
 
int main(int argc, char *argv[])
{
  bool wantNumeric(false);

  if (argc==4)
  {
    if ( (strcmp(argv[1],"--symbolic")!=0) 
	 && (strcmp(argv[1],"--numeric")!=0) )
    {
      printf
	( "Usage: %s [--symbolic|--numeric] <quality-table> <qval-file1>\n",
	  argv[0] );
	exit (0);
    } // ~if
    else
    {
      wantNumeric=(strcmp(argv[1],"--numeric")==0); 
      argc--; argv++;
    } // ~else
  } // ~if
  else if (argc!=3)
  {
    printf
      ( "Usage: %s [--symbolic|--numeric] <quality-table> <qval-file1>\n",
	argv[0] );
    exit (0);
  } // ~else if


  QualityTable *table = new QualityTable();
  QualityWriter* writer;
  if (wantNumeric)
  {
    writer= new QualityWriterNumeric();
  } // ~if
  else
  {
    writer= new QualityWriterSymbolic();
  } // ~else

  table->read(argv[1]);
#if 0
  table->print();
#endif
  DataSet *set = new DataSet(argv[2]);
  std::vector<int> scores = table->applySolexa(set);
  writeQualityOutput(stdout, set, &scores, *writer);
  delete set;
  delete table;
  delete writer;
  return 0;
} // ~main
